/*
 * JBoss, the OpenSource J2EE webOS
 *
 * Distributable under LGPL license.
 * See terms of license at gnu.org.
 */
package org.jboss.cache;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.jboss.cache.buddyreplication.BuddyGroup;
import org.jboss.cache.buddyreplication.BuddyManager;
import org.jboss.cache.buddyreplication.BuddyNotInitException;
import org.jboss.cache.buddyreplication.GravitateResult;
import org.jboss.cache.config.BuddyReplicationConfig;
import org.jboss.cache.config.Configuration;
import org.jboss.cache.config.Configuration.NodeLockingScheme;
import org.jboss.cache.config.Option;
import org.jboss.cache.config.RuntimeConfig;
import org.jboss.cache.factories.InterceptorChainFactory;
import org.jboss.cache.interceptors.Interceptor;
import org.jboss.cache.loader.CacheLoader;
import org.jboss.cache.loader.CacheLoaderManager;
import org.jboss.cache.lock.IsolationLevel;
import org.jboss.cache.lock.LockStrategyFactory;
import org.jboss.cache.lock.LockUtil;
import org.jboss.cache.lock.LockingException;
import org.jboss.cache.lock.NodeLock;
import org.jboss.cache.lock.TimeoutException;
import org.jboss.cache.marshall.InactiveRegionAwareRpcDispatcher;
import org.jboss.cache.marshall.Marshaller;
import org.jboss.cache.marshall.MethodCall;
import org.jboss.cache.marshall.MethodCallFactory;
import org.jboss.cache.marshall.MethodDeclarations;
import org.jboss.cache.marshall.NodeData;
import org.jboss.cache.marshall.VersionAwareMarshaller;
import org.jboss.cache.notifications.Notifier;
import org.jboss.cache.notifications.event.NodeModifiedEvent;
import org.jboss.cache.optimistic.DataVersion;
import org.jboss.cache.statetransfer.StateTransferManager;
import org.jboss.cache.transaction.GlobalTransaction;
import org.jboss.cache.transaction.OptimisticTransactionEntry;
import org.jboss.cache.transaction.TransactionEntry;
import org.jboss.cache.transaction.TransactionManagerLookup;
import org.jboss.cache.transaction.TransactionTable;
import org.jboss.cache.util.ExposedByteArrayOutputStream;
import org.jboss.cache.util.ThreadGate;
import org.jboss.cache.util.concurrent.ConcurrentHashSet;
import org.jboss.util.stream.MarshalledValueInputStream;
import org.jboss.util.stream.MarshalledValueOutputStream;
import org.jgroups.Address;
import org.jgroups.Channel;
import org.jgroups.ChannelClosedException;
import org.jgroups.ChannelException;
import org.jgroups.ChannelFactory;
import org.jgroups.ChannelNotConnectedException;
import org.jgroups.ExtendedMembershipListener;
import org.jgroups.ExtendedMessageListener;
import org.jgroups.JChannel;
import org.jgroups.Message;
import org.jgroups.StateTransferException;
import org.jgroups.View;
import org.jgroups.blocks.GroupRequest;
import org.jgroups.blocks.RpcDispatcher;
import org.jgroups.blocks.RspFilter;
import org.jgroups.util.Rsp;
import org.jgroups.util.RspList;
import org.jgroups.util.Util;

import javax.management.MBeanServerFactory;
import javax.transaction.Status;
import javax.transaction.SystemException;
import javax.transaction.Transaction;
import javax.transaction.TransactionManager;
import java.io.ByteArrayInputStream;
import java.io.InputStream;
import java.io.NotSerializableException;
import java.io.OutputStream;
import java.lang.reflect.Method;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.Vector;
import java.util.concurrent.ConcurrentHashMap;

/**
 * The default implementation class of {@link org.jboss.cache.Cache} and {@link org.jboss.cache.CacheSPI}.  This class
 * has its roots in the legacy (JBoss Cache 1.x.x) org.jboss.cache.TreeCache class.
 * <p/>
 * You should not use this class directly, or attempt to cast {@link org.jboss.cache.Cache} or {@link org.jboss.cache.CacheSPI}
 * interfaces directly to this class.
 *
 * @author Bela Ban
 * @author Ben Wang
 * @author <a href="mailto:manik@jboss.org">Manik Surtani (manik@jboss.org)</a>
 * @author Brian Stansberry
 * @author Daniel Huang (dhuang@jboss.org)
 * @see org.jboss.cache.Cache
 */
public class CacheImpl<K, V> implements CacheSPI<K, V>
{

   /**
    * CacheImpl log.
    */
   private Log log = LogFactory.getLog(CacheImpl.class);

   /**
    * Thread gate used to block Dispatcher during JGroups FLUSH protocol
    */
   private final ThreadGate flushBlockGate = new ThreadGate();

   /**
    * Root node.
    */
   private NodeSPI<K, V> root;

   /**
    * Cache's region manager.
    */
   private RegionManager regionManager = null;

   /**
    * The JGroups JChannel in use.
    */
   protected Channel channel = null;

   /**
    * True if this CacheImpl is the coordinator.
    */
   private volatile boolean coordinator = false;

   /**
    * List of cluster group members.
    */
   private final Vector<Address> members = new Vector<Address>();

   /**
    * JGroups RpcDispatcher in use.
    */
   private RpcDispatcher disp = null;

   /**
    * JGroups message listener.
    */
   private MessageListenerAdaptor ml = new MessageListenerAdaptor();

   /**
    * Maintains mapping of transactions (keys) and Modifications/Undo-Operations
    */
   private final TransactionTable tx_table = new TransactionTable();

   /**
    * HashMap<Thread, List<Lock>, maintains locks acquired by threads (used when no TXs are used)
    */
   private Map<Thread, List<NodeLock>> lock_table;

   /**
    * Set<Fqn> of Fqns of the topmost node of internal regions that should
    * not included in standard state transfers.
    */
   private Set<Fqn> internalFqns = new ConcurrentHashSet<Fqn>();

   /**
    * True if state was initialized during start-up.
    */
   private volatile boolean isStateSet = false;

   /**
    * Class name used to handle evictions.
    */
   private String evictionInterceptorClass = "org.jboss.cache.interceptors.EvictionInterceptor";

   /**
    * Marshaller if register to handle marshalling
    */
   private Marshaller marshaller_ = null;

   /**
    * {@link #invokeMethod(org.jboss.cache.marshall.MethodCall,boolean)} will dispatch to this chain of interceptors.
    * In the future, this will be replaced with JBossAop. This is a first step towards refactoring JBossCache.
    */
   private Interceptor interceptor_chain = null;

   /**
    * Method to acquire a TransactionManager. By default we use JBossTransactionManagerLookup. Has
    * to be set before calling {@link #start()}
    */
   private TransactionManagerLookup tm_lookup = null;

   /**
    * Used to get the Transaction associated with the current thread
    */
   private TransactionManager tm = null;

   /**
    * Cache loader manager.
    */
   private CacheLoaderManager cacheLoaderManager;

   /**
    * Queue used to replicate updates when mode is repl-async
    */
   private ReplicationQueue repl_queue = null;

   /**
    * The current lifecycle state.
    */
   CacheStatus cacheStatus;

   /**
    * Buddy Manager
    */
   private BuddyManager buddyManager;

   /**
    * State transfer manager. Do not access this field directly -- use the getter
    */
   private StateTransferManager stateTransferManager;

   /**
    * Cache notifier handler class.
    */
   private Notifier notifier;

   private ThreadLocal<InvocationContext> invocationContextContainer = new ThreadLocal<InvocationContext>()
   {
      @Override
      protected InvocationContext initialValue()
      {
         return new InvocationContext();
      }
   };

   private Configuration configuration;

   /**
    * Constructs an uninitialized CacheImpl.
    */
   protected CacheImpl() throws Exception
   {
      configuration = new Configuration(this);
      notifier = new Notifier(this);
      regionManager = new RegionManager(this);
      cacheStatus = CacheStatus.INSTANTIATED;
   }

   public StateTransferManager getStateTransferManager()
   {
      if (stateTransferManager == null)
      {
         stateTransferManager = new StateTransferManager(this);
      }
      return stateTransferManager;
   }

   public void setStateTransferManager(StateTransferManager manager)
   {
      this.stateTransferManager = manager;
   }

   public Configuration getConfiguration()
   {
      return configuration;
   }

   /**
    * Returns the CacheImpl implementation version.
    */
   public String getVersion()
   {
      return Version.printVersion();
   }

   /**
    * Returns the root node.
    */
   public NodeSPI<K, V> getRoot()
   {
      return root;
   }

   /**
    * Returns the local channel address.
    */
   public Address getLocalAddress()
   {
      return channel != null ? channel.getLocalAddress() : null;
   }

   /**
    * Returns the members as a List.
    * This list may be concurrently modified.
    */
   public List<Address> getMembers()
   {
      synchronized (members)
      {
         return new ArrayList<Address>(members);
      }
   }

   /**
    * Returns <code>true</code> if this node is the group coordinator.
    */
   public boolean isCoordinator()
   {
      return coordinator;
   }

   /**
    * Returns the transaction table.
    */
   public TransactionTable getTransactionTable()
   {
      return tx_table;
   }

   /**
    * Returns the lock table.
    */
   public Map<Thread, List<NodeLock>> getLockTable()
   {
      if (lock_table == null)
      {
         lock_table = new ConcurrentHashMap<Thread, List<NodeLock>>();
      }
      return lock_table;
   }

   /**
    * Returns the contents of the TransactionTable as a string.
    */
   public String dumpTransactionTable()
   {
      return tx_table.toString(true);
   }

   /**
    * Used for testing only - sets the interceptor chain.
    */
   public void setInterceptorChain(Interceptor i)
   {
      interceptor_chain = i;
   }

   /**
    * @return the list of interceptors.
    */
   public List<Interceptor> getInterceptors()
   {
      return InterceptorChainFactory.getInstance().asList(interceptor_chain);
   }

   /**
    * Returns the underlying cache loader in use.
    */
   public CacheLoader getCacheLoader()
   {
      if (cacheLoaderManager == null)
         return null;
      return cacheLoaderManager.getCacheLoader();
   }

   public String getEvictionInterceptorClass()
   {
      return this.evictionInterceptorClass;
   }

   private void setUseReplQueue(boolean flag)
   {
      if (flag)
      {
         if (repl_queue == null)
         {
            repl_queue = new ReplicationQueue(this, configuration.getReplQueueInterval(), configuration.getReplQueueMaxElements());
            if (configuration.getReplQueueInterval() >= 0)
            {
               repl_queue.start();
            }
         }
      }
      else
      {
         if (repl_queue != null)
         {
            repl_queue.stop();
            repl_queue = null;
         }
      }
   }


   /**
    * Returns the replication queue.
    */
   public ReplicationQueue getReplicationQueue()
   {
      return repl_queue;
   }

   /**
    * Sets the cache locking isolation level.
    */
   private void setIsolationLevel(IsolationLevel level)
   {
      LockStrategyFactory.setIsolationLevel(level);
   }

   /**
    * Sets the TransactionManagerLookup object
    */
   public void setTransactionManagerLookup(TransactionManagerLookup l)
   {
      this.tm_lookup = l;
   }

   /**
    * Returns the transaction manager in use.
    */
   public TransactionManager getTransactionManager()
   {
      return tm;
   }

   /**
    * Fetches the group state from the current coordinator. If successful, this
    * will trigger JChannel setState() call.
    */
   public void fetchState(long timeout) throws ChannelClosedException, ChannelNotConnectedException
   {
      if (channel == null)
      {
         throw new ChannelNotConnectedException();
      }
      boolean rc = channel.getState(null, timeout);
      if (rc)
      {
         log.debug("fetchState(): state was retrieved successfully");
      }
      else
      {
         log.debug("fetchState(): state could not be retrieved (first member)");
      }
   }

   void fetchPartialState(List<Address> sources, Fqn sourceTarget, Fqn integrationTarget) throws Exception
   {
      String encodedStateId = sourceTarget + StateTransferManager.PARTIAL_STATE_DELIMITER + integrationTarget;
      fetchPartialState(sources, encodedStateId);
   }

   void fetchPartialState(List<Address> sources, Fqn subtree) throws Exception
   {
      if (subtree == null)
      {
         throw new IllegalArgumentException("Cannot fetch partial state. Null subtree.");
      }
      fetchPartialState(sources, subtree.toString());
   }

   private void fetchPartialState(List<Address> sources, String stateId) throws Exception
   {
      if (sources == null || sources.isEmpty() || stateId == null)
      {
         // should this really be throwing an exception?  Are there valid use cases where partial state may not be available? - Manik
         // Yes -- cache is configured LOCAL but app doesn't know it -- Brian
         //throw new IllegalArgumentException("Cannot fetch partial state, targets are " + sources + " and stateId is " + stateId);
         if (log.isWarnEnabled())
         {
            log.warn("Cannot fetch partial state, targets are " + sources +
                    " and stateId is " + stateId);
         }
         return;
      }

      List<Address> targets = new LinkedList<Address>(sources);

      //skip *this* node as a target
      targets.remove(getLocalAddress());

      if (targets.isEmpty())
      {
         // Definitely no exception here -- this happens every time the 1st node in the
         // cluster activates a region!! -- Brian
         log.debug("Cannot fetch partial state. There are no target members specified");
         return;
      }

      log.debug("Node " + getLocalAddress() + " fetching partial state " + stateId + " from members " + targets);
      boolean successfulTransfer = false;
      for (Address target : targets)
      {
         log.debug("Node " + getLocalAddress() + " fetching partial state " + stateId + " from member " + target);
         isStateSet = false;
         successfulTransfer = channel.getState(target, stateId, configuration.getStateRetrievalTimeout());
         if (successfulTransfer)
         {
            try
            {
               ml.waitForState();
            }
            catch (Exception transferFailed)
            {
               successfulTransfer = false;
            }
         }
         log.debug("Node " + getLocalAddress() + " fetching partial state " + stateId + " from member " + target + (successfulTransfer ? " successful" : " failed"));
         if (successfulTransfer)
            break;
      }

      if (!successfulTransfer)
      {
         log.debug("Node " + getLocalAddress() + " could not fetch partial state " + stateId + " from any member " + targets);
      }
   }

   /**
    * Lifecycle method. This is like initialize.
    *
    * @throws Exception
    */
   public void create() throws CacheException
   {
      if (!cacheStatus.createAllowed())
      {
         if (cacheStatus.needToDestroyFailedCache())
            destroy();
         else
            return;
      }

      try
      {
         internalCreate();
      }
      catch (Throwable t)
      {
         handleLifecycleTransitionFailure(t);
      }
   }

   /**
    * Sets the cacheStatus to FAILED and rethrows the problem as one
    * of the declared types. Converts any non-RuntimeException Exception
    * to CacheException.
    *
    * @param t
    * @throws CacheException
    * @throws RuntimeException
    * @throws Error
    */
   private void handleLifecycleTransitionFailure(Throwable t)
           throws CacheException, RuntimeException, Error
   {
      cacheStatus = CacheStatus.FAILED;
      if (t instanceof CacheException)
         throw (CacheException) t;
      else if (t instanceof RuntimeException)
         throw (RuntimeException) t;
      else if (t instanceof Error)
         throw (Error) t;
      else
         throw new CacheException(t);
   }

   /**
    * The actual create implementation.
    *
    * @throws CacheException
    */
   private void internalCreate() throws CacheException
   {
      // Include our clusterName in our log category
      configureLogCategory();

      // initialise the node factory and set this in the runtime.
      NodeFactory<K, V> nf;
      if ((nf = configuration.getRuntimeConfig().getNodeFactory()) == null)
      {
         nf = new NodeFactory<K, V>(this);
         configuration.getRuntimeConfig().setNodeFactory(nf);
      }
      else
      {
         // don't create a new one each and every time.  After stopping and starting the cache, the old NodeFactory may still be valid.
         nf.init();
      }

      if (notifier == null)
         notifier = new Notifier(this);

      // create a new root temporarily.
      NodeSPI<K, V> tempRoot = nf.createRootDataNode();
      // if we don't already have a root or the new (temp) root is of a different class (optimistic vs pessimistic) to
      // the current root, then we use the new one.  Helps preserve data between cache restarts.
      if (root == null || !root.getClass().equals(tempRoot.getClass()))
         root = tempRoot;

      if (configuration.getCacheLoaderConfig() != null && cacheLoaderManager == null)
      {
         initialiseCacheLoaderManager();
      }
      // first set up the Buddy Manager and an RPCManager
      if (configuration.getCacheMode() != Configuration.CacheMode.LOCAL)
      {
         getConfiguration().getRuntimeConfig().setRPCManager(new RPCManagerImpl(this));
         setBuddyReplicationConfig(configuration.getBuddyReplicationConfig());
      }
      // build interceptor chain
      try
      {
         interceptor_chain = InterceptorChainFactory.getInstance().buildInterceptorChain(this);
      }
      catch (Exception e)
      {
         throw new CacheException("Unable to build interceptor chain", e);
      }


      setUseReplQueue(configuration.isUseReplQueue());
      setIsolationLevel(configuration.getIsolationLevel());

      getRegionManager();// make sure we create one
      createEvictionPolicy();

      getRegionManager().setDefaultInactive(configuration.isInactiveOnStartup());

      cacheStatus = CacheStatus.CREATED;
   }

   private void createTransactionManager()
   {
      // See if we had a TransactionManager injected into our config
      this.tm = configuration.getRuntimeConfig().getTransactionManager();
      if (tm == null)
      {
         // Nope. See if we can look it up from JNDI
         if (this.tm_lookup == null && configuration.getTransactionManagerLookupClass() != null)
         {
            try
            {
               Class clazz = Thread.currentThread().getContextClassLoader().loadClass(configuration.getTransactionManagerLookupClass());
               this.tm_lookup = (TransactionManagerLookup) clazz.newInstance();
            }
            catch (Exception e)
            {
               throw new CacheException("Problems creating the cache", e);
            }
         }

         try
         {
            if (tm_lookup != null)
            {
               tm = tm_lookup.getTransactionManager();
               configuration.getRuntimeConfig().setTransactionManager(tm);
            }
            else
            {
               if (configuration.getNodeLockingScheme() == NodeLockingScheme.OPTIMISTIC)
               {
                  log.fatal("No transaction manager lookup class has been defined. Transactions cannot be used and thus OPTIMISTIC locking cannot be used");
               }
               else
               {
                  log.info("No transaction manager lookup class has been defined. Transactions cannot be used");
               }
            }
         }
         catch (Exception e)
         {
            log.debug("failed looking up TransactionManager, will not use transactions", e);
         }
      }
   }

   protected boolean shouldFetchStateOnStartup()
   {
      boolean loaderFetch = cacheLoaderManager != null && cacheLoaderManager.isFetchPersistentState();
      return !configuration.isInactiveOnStartup() && buddyManager == null && (configuration.isFetchInMemoryState() || loaderFetch);
   }

   /**
    * Lifecyle method.
    *
    * @throws CacheException
    */
   public void start() throws CacheException
   {
      if (!cacheStatus.startAllowed())
      {
         if (cacheStatus.needToDestroyFailedCache())
            destroy(); // this will take us back to DESTROYED

         if (cacheStatus.needCreateBeforeStart())
            create();
         else
            return;
      }

      try
      {
         internalStart();
      }
      catch (Throwable t)
      {
         handleLifecycleTransitionFailure(t);
      }
   }

   /**
    * The actual start implementation.
    *
    * @throws CacheException
    * @throws IllegalArgumentException
    */
   private void internalStart() throws CacheException, IllegalArgumentException
   {
      cacheStatus = CacheStatus.STARTING;

      createTransactionManager();

      // cache loaders should be initialised *before* any state transfers take place to prevent
      // exceptions involving cache loaders not being started. - Manik
      // create cache loader
      if (cacheLoaderManager != null)
      {
         cacheLoaderManager.startCacheLoader();
      }
      // now that we have a TM we can init the interceptor chain
      InterceptorChainFactory.getInstance().initialiseInterceptors(interceptor_chain, this);

      switch (configuration.getCacheMode())
      {
         case LOCAL:
            log.debug("cache mode is local, will not create the channel");
            break;
         case REPL_SYNC:
         case REPL_ASYNC:
         case INVALIDATION_ASYNC:
         case INVALIDATION_SYNC:
            if (log.isDebugEnabled()) log.debug("cache mode is " + configuration.getCacheMode());
            initialiseChannelAndRpcDispatcher();

            //connect and transfer state
            if (shouldFetchStateOnStartup())
            {
               try
               {
                  long start = System.currentTimeMillis();
                  channel.connect(configuration.getClusterName(), null, null, configuration.getStateRetrievalTimeout());
                  //if I am not the only and the first member than wait for a state to arrive
                  if (getMembers().size()>1)
                  {
                     ml.waitForState();
                  }

                  if (log.isDebugEnabled())
                  {
                     log.debug("connected, state was retrieved successfully (in " + (System.currentTimeMillis() - start)
                             + " milliseconds)");
                  }
               }
               catch (StateTransferException ste)
               {
                  // make sure we disconnect from the channel before we throw this exception!
                  // JBCACHE-761
                  channel.disconnect();
                  channel.close();
                  throw new CacheException("Unable to fetch state on startup", ste);
               }
               catch (ChannelException e)
               {
                  throw new CacheException("Unable to connect to JGroups channel", e);
               }
               catch (Exception ex)
               {
                  throw new CacheException("Unable to fetch state on startup", ex);
               }
            }
            //otherwise just connect
            else
            {
               try
               {
                  channel.connect(configuration.getClusterName());
               }
               catch (ChannelException e)
               {
                  throw new CacheException("Unable to connect to JGroups channel", e);
               }
            }
            if (log.isInfoEnabled())
            {
               log.info("CacheImpl local address is " + channel.getLocalAddress());
            }
            if (buddyManager != null)
            {
               buddyManager.init(this);
               if (configuration.isUseReplQueue())
               {
                  log.warn("Replication queue not supported when using buddy replication.  Disabling repliction queue.");
                  configuration.setUseReplQueue(false);
                  repl_queue = null;
               }
            }
            break;
         default:
            throw new IllegalArgumentException("cache mode " + configuration.getCacheMode() + " is invalid");
      }

      //now attempt to preload the cache from the loader - Manik
      if (cacheLoaderManager != null)
      {
         cacheLoaderManager.preloadCache();
      }

      // start any eviction threads.
      if (regionManager.isUsingEvictions())
      {
         regionManager.startEvictionThread();
      }

      notifier.notifyCacheStarted(this, getInvocationContext());

      addShutdownHook();

      log.info("JBoss Cache version: " + getVersion());

      cacheStatus = CacheStatus.STARTED;
   }

   private void addShutdownHook()
   {
      ArrayList al = MBeanServerFactory.findMBeanServer(null);
      if (al.size() == 0)
      {
         // the only MBean server is the system (JDK) server.  So we need to register a shutdown hook.
         // install a VM shutdown hook
         Thread shutdownHook = new Thread()
         {
            public void run()
            {
               CacheImpl.this.stop();
            }
         };

      Runtime.getRuntime().addShutdownHook(shutdownHook);
      }
      else
      {
         log.trace("Running in an MBeanServer environment.  Not registering a shutdown hook with the VM as the MBeanServer will handle lifecycle.");
      }
   }

   /**
    * Lifecycle method.
    */
   public void destroy()
   {
      if (!cacheStatus.destroyAllowed())
      {
         if (cacheStatus.needStopBeforeDestroy())
         {
            try
            {
               stop();
            }
            catch (CacheException e)
            {
               log.warn("Needed to call stop() before destroying but stop() " +
                       "threw exception. Proceeding to destroy", e);
            }
         }
         else
            return;
      }

      try
      {
         internalDestroy();
      }
      finally
      {
         // We always progress to destroyed
         cacheStatus = CacheStatus.DESTROYED;
      }
   }

   /**
    * The actual destroy implementation.
    */
   private void internalDestroy()
   {
      cacheStatus = CacheStatus.DESTROYING;
      regionManager = null;
      cacheLoaderManager = null;
      notifier = null;

      // The rest of these should have already been taken care of in stop,
      // but we do it here as well in case stop failed.

      if (channel != null)
      {
         if (channel.isOpen())
         {
            try
            {
               channel.close();
               channel.disconnect();
            }
            catch (Exception toLog)
            {
               log.error("Problem closing channel; setting it to null", toLog);
            }
         }
         channel = null;
         configuration.getRuntimeConfig().setChannel(null);
      }
      disp = null;
      tm = null;
   }

   /**
    * Lifecycle method.
    */
   public void stop()
   {
      if (!cacheStatus.stopAllowed())
      {
         return;
      }

      // Trying to stop() from FAILED is valid, but may not work
      boolean failed = cacheStatus == CacheStatus.FAILED;

      try
      {
         internalStop();
      }
      catch (Throwable t)
      {
         if (failed)
         {
            log.warn("Attempted to stop() from FAILED state, " +
                    "but caught exception; try calling destroy()", t);
         }
         handleLifecycleTransitionFailure(t);
      }
   }

   /**
    * The actual stop implementation.
    */
   private void internalStop()
   {
      cacheStatus = CacheStatus.STOPPING;

      // before closing the channel stop the buddy manager
      if (buddyManager != null && buddyManager.isEnabled())
      {
         log.debug("stop(): stopping buddy manager");
         buddyManager.stop();
      }

      if (channel != null)
      {
         log.info("stop(): closing the channel");
         killChannel();
         channel = null;
         configuration.getRuntimeConfig().setChannel(null);
      }

      if (disp != null)
      {
         log.info("stop(): stopping the dispatcher");
         disp.stop();
         disp = null;
      }
      if (members != null)
      {
         synchronized (members)
         {
            members.clear();
         }
      }

      coordinator = false;

      if (repl_queue != null)
      {
         repl_queue.stop();
      }

      if (cacheLoaderManager != null)
      {
         log.debug("stop(): stopping cache loader manager");
         cacheLoaderManager.stopCacheLoader();
      }

      if (regionManager.isUsingEvictions()) regionManager.stopEvictionThread();

      if (notifier != null)
      {
         notifier.notifyCacheStopped(this, getInvocationContext());
         notifier.removeAllCacheListeners();
      }

      // unset transaction manager reference
      tm = null;

      cacheStatus = CacheStatus.STOPPED;

      // empty in-memory state
      root.clearDataDirect();
      root.removeChildrenDirect();
   }

   public CacheStatus getCacheStatus()
   {
      return cacheStatus;
   }

   /* ----------------------- End of MBeanSupport ----------------------- */

   /* ----------------------- Start of buddy replication specific methods ------------*/

   /**
    * Sets the buddy replication configuration element
    *
    * @param config
    */
   private void setBuddyReplicationConfig(BuddyReplicationConfig config)
   {
      if (config != null)
      {
         buddyManager = new BuddyManager(config);
         if (!buddyManager.isEnabled())
         {
            buddyManager = null;
         }
         else
         {
            internalFqns.add(BuddyManager.BUDDY_BACKUP_SUBTREE_FQN);
         }
      }
   }

   /**
    * Retrieves the Buddy Manager configured.
    *
    * @return null if buddy replication is not enabled.
    */
   public BuddyManager getBuddyManager()
   {
      return buddyManager;
   }

   /**
    * Returns a Set<Fqn> of Fqns of the topmost node of internal regions that
    * should not included in standard state transfers. Will include
    * {@link BuddyManager#BUDDY_BACKUP_SUBTREE} if buddy replication is
    * enabled.
    *
    * @return an unmodifiable Set<Fqn>.  Will not return <code>null</code>.
    */
   public Set<Fqn> getInternalFqns()
   {
      return Collections.unmodifiableSet(internalFqns);
   }

   /* ----------------------- End of buddy replication specific methods ------------*/

   protected void createEvictionPolicy()
   {
      if (configuration.getEvictionConfig() != null
              && configuration.getEvictionConfig().isValidConfig())
      {
         regionManager.setEvictionConfig(configuration.getEvictionConfig());
         regionManager.setUsingEvictions(true);
      }
      else
      {
         regionManager.setUsingEvictions(false);
         log.debug("Not using an EvictionPolicy");
      }
   }

   /**
    * Loads the indicated Fqn, plus all parents recursively from the
    * CacheLoader. If no CacheLoader is present, this is a no-op
    *
    * @param fqn
    * @throws Exception
    */
   public void load(String fqn) throws Exception
   {
      if (cacheLoaderManager != null)
      {
         cacheLoaderManager.preload(Fqn.fromString(fqn), true, true);
      }
   }

   /**
    * Returns the address of the coordinator or null if there is no
    * coordinator.
    * Waits until the membership view is updated.
    */
   public Address getCoordinator()
   {
      if (channel == null)
      {
         return null;
      }

      synchronized (members)
      {
         while (members.isEmpty())
         {
            log.debug("getCoordinator(): waiting on viewAccepted()");
            try
            {
               members.wait();
            }
            catch (InterruptedException e)
            {
               log.error("getCoordinator(): Interrupted while waiting for members to be set", e);
               break;
            }
         }
         return members.size() > 0 ? members.get(0) : null;
      }
   }

   /**
    * Evicts the node at <code>subtree</code> along with all descendant nodes.
    *
    * @param subtree Fqn indicating the uppermost node in the
    *                portion of the cache that should be evicted.
    * @throws CacheException
    */
   protected void _evictSubtree(Fqn subtree) throws CacheException
   {

      if (!exists(subtree))
      {
         return;// node does not exist. Maybe it has been recursively removed.
      }

      if (log.isTraceEnabled())
      {
         log.trace("_evictSubtree(" + subtree + ")");
      }

      // Recursively remove any children
      Set children = getChildrenNames(subtree);
      if (children != null)
      {
         for (Object s : children)
         {

            Fqn<Object> tmp = new Fqn<Object>(subtree, s);
            _remove(null, // no tx
                    tmp,
                    false, // no undo ops
                    false, // no nodeEvent
                    true);// is an eviction
         }
      }

      // Remove the root node of the subtree
      _remove(null, subtree, false, false, true);

   }

   private void removeLocksForDeadMembers(NodeSPI<K, V> node, List deadMembers)
   {
      Set<GlobalTransaction> deadOwners = new HashSet<GlobalTransaction>();
      NodeLock lock = node.getLock();
      Object owner = lock.getWriterOwner();

      if (isLockOwnerDead(owner, deadMembers))
      {
         deadOwners.add((GlobalTransaction) owner);
      }

      for (Object readOwner : lock.getReaderOwners())
      {
         if (isLockOwnerDead(readOwner, deadMembers))
         {
            deadOwners.add((GlobalTransaction) readOwner);
         }
      }

      for (GlobalTransaction deadOwner : deadOwners)
      {
         boolean localTx = deadOwner.getAddress().equals(getLocalAddress());
         boolean broken = LockUtil.breakTransactionLock(lock, deadOwner, localTx, this);

         if (broken && log.isTraceEnabled())
         {
            log.trace("Broke lock for node " + node.getFqn() +
                    " held by " + deadOwner);
         }
      }

      // Recursively unlock children
      for (NodeSPI<K, V> child : node.getChildrenDirect())
      {
         removeLocksForDeadMembers(child, deadMembers);
      }
   }

   private boolean isLockOwnerDead(Object owner, List deadMembers)
   {
      boolean result = false;
      if (owner != null && owner instanceof GlobalTransaction)
      {
         Object addr = ((GlobalTransaction) owner).getAddress();
         result = deadMembers.contains(addr);
      }
      return result;
   }

   // -----------  End Marshalling and State Transfer -----------------------

   /**
    * @param fqn fqn String name to retrieve from cache
    * @return DataNode corresponding to the fqn. Null if does not exist. No guarantees wrt replication,
    *         cache loading are given if the underlying node is modified
    */
   public Node get(String fqn) throws CacheException
   {
      return get(Fqn.fromString(fqn));
   }

   /**
    * Returns a DataNode corresponding to the fully qualified name or null if
    * does not exist.
    * No guarantees wrt replication, cache loading are given if the underlying node is modified
    *
    * @param fqn name of the DataNode to retreive
    */
   public Node<K, V> get(Fqn<?> fqn) throws CacheException
   {
      MethodCall m = MethodCallFactory.create(MethodDeclarations.getNodeMethodLocal, fqn);
      return (Node<K, V>) invokeMethod(m, true);
   }

   /**
    * Returns the raw data of the node; called externally internally.
    */
   public Node<K, V> _get(Fqn<?> fqn) throws CacheException
   {
      return findNode(fqn);
   }

   /**
    * Returns the raw data of the node; called externally internally.
    */
   public Map _getData(Fqn<?> fqn)
   {
      NodeSPI n = findNode(fqn);
      if (n == null) return null;
      return n.getDataDirect();
   }

   /**
    * Returns a set of attribute keys for the Fqn.
    * Returns null if the node is not found, otherwise a Set.
    * The set is a copy of the actual keys for this node.
    *
    * @param fqn name of the node
    */
   public Set getKeys(String fqn) throws CacheException
   {
      return getKeys(Fqn.fromString(fqn));
   }

   /**
    * Returns a set of attribute keys for the Fqn.
    * Returns null if the node is not found, otherwise a Set.
    * The set is a copy of the actual keys for this node.
    *
    * @param fqn name of the node
    */
   public Set<K> getKeys(Fqn<?> fqn) throws CacheException
   {
      MethodCall m = MethodCallFactory.create(MethodDeclarations.getKeysMethodLocal, fqn);
      return (Set<K>) invokeMethod(m, true);
   }


   /**
    * Retrieves a defensively copied data map of the underlying node.
    *
    * @param fqn
    * @return map of data, or an empty map
    * @throws CacheException
    */
   public Map<K, V> getData(Fqn<?> fqn) throws CacheException
   {
      MethodCall m = MethodCallFactory.create(MethodDeclarations.getDataMapMethodLocal, fqn);
      return (Map<K, V>) invokeMethod(m, true);
   }

   public Set _getKeys(Fqn<?> fqn) throws CacheException
   {
      NodeSPI<K, V> n = findNode(fqn);
      if (n == null)
      {
         return null;
      }
      Set<K> keys = n.getKeysDirect();
      return new HashSet<K>(keys);
   }

   /**
    * Finds a node given its name and returns the value associated with a given key in its <code>data</code>
    * map. Returns null if the node was not found in the cache or the key was not found in the hashmap.
    *
    * @param fqn The fully qualified name of the node.
    * @param key The key.
    */
   public V get(String fqn, K key) throws CacheException
   {
      return get(Fqn.fromString(fqn), key);
   }


   /**
    * Finds a node given its name and returns the value associated with a given key in its <code>data</code>
    * map. Returns null if the node was not found in the cache or the key was not found in the hashmap.
    *
    * @param fqn The fully qualified name of the node.
    * @param key The key.
    */
   public V get(Fqn<?> fqn, K key) throws CacheException
   {
      return get(fqn, key, true);
   }

   public V _get(Fqn<?> fqn, K key, boolean sendNodeEvent) throws CacheException
   {
      InvocationContext ctx = getInvocationContext();
      if (log.isTraceEnabled())
      {
         log.trace(new StringBuffer("_get(").append("\"").append(fqn).append("\", \"").append(key).append("\", \"").
                 append(sendNodeEvent).append("\")"));
      }
      if (sendNodeEvent) notifier.notifyNodeVisited(fqn, true, ctx);
      NodeSPI<K, V> n = findNode(fqn);
      if (n == null)
      {
         log.trace("node not found");
         return null;
      }
      if (sendNodeEvent) notifier.notifyNodeVisited(fqn, false, ctx);
      return n.getDirect(key);
   }


   protected V get(Fqn<?> fqn, K key, boolean sendNodeEvent) throws CacheException
   {
      MethodCall m = MethodCallFactory.create(MethodDeclarations.getKeyValueMethodLocal, fqn, key, sendNodeEvent);
      return (V) invokeMethod(m, true);
   }

   /**
    * Checks whether a given node exists in current in-memory state of the cache.
    * Does not acquire any locks in doing so (result may be dirty read). Does
    * not attempt to load nodes from a cache loader (may return false if a
    * node has been evicted).
    *
    * @param fqn The fully qualified name of the node
    * @return boolean Whether or not the node exists
    */
   public boolean exists(String fqn)
   {
      return exists(Fqn.fromString(fqn));
   }


   /**
    * Checks whether a given node exists in current in-memory state of the cache.
    * Does not acquire any locks in doing so (result may be dirty read). Does
    * not attempt to load nodes from a cache loader (may return false if a
    * node has been evicted).
    *
    * @param fqn The fully qualified name of the node
    * @return boolean Whether or not the node exists
    */
   public boolean exists(Fqn<?> fqn)
   {
      Node n = peek(fqn, false);
      return n != null;
   }

   public NodeSPI<K, V> peek(Fqn<?> fqn, boolean includeDeletedNodes)
   {
      return peek(fqn, includeDeletedNodes, false);
   }

   public NodeSPI<K, V> peek(Fqn<?> fqn, boolean includeDeletedNodes, boolean includeInvalidNodes)
   {
      if (fqn == null || fqn.size() == 0) return root;
      NodeSPI<K, V> n = root;
      int fqnSize = fqn.size();
      for (int i = 0; i < fqnSize; i++)
      {
         Object obj = fqn.get(i);
         n = n.getChildDirect(obj);
         if (n == null)
         {
            return null;
         }
         else if (!includeDeletedNodes && n.isDeleted())
         {
            return null;
         }
         else if (!includeInvalidNodes && !n.isValid())
         {
            return null;
         }
      }
      return n;
   }


   /**
    * @param fqn
    * @param key
    */
   public boolean exists(String fqn, Object key)
   {
      return exists(Fqn.fromString(fqn), key);
   }


   /**
    * Checks whether a given key exists in the given node. Does not interact with CacheLoader, so the behavior is
    * different from {@link #get(Fqn,Object)}
    *
    * @param fqn The fully qualified name of the node
    * @param key
    * @return boolean Whether or not the node exists
    */
   public boolean exists(Fqn<?> fqn, Object key)
   {
      NodeSPI n = peek(fqn, false);
      return n != null && n.getKeysDirect().contains(key);
   }


   /**
    * Adds a new node to the cache and sets its data. If the node doesn not yet exist, it will be created.
    * Also, parent nodes will be created if not existent. If the node already has data, then the new data
    * will override the old one. If the node already existed, a nodeModified() notification will be generated.
    * Otherwise a nodeCreated() motification will be emitted.
    *
    * @param fqn  The fully qualified name of the new node
    * @param data The new data. May be null if no data should be set in the node.
    */
   public void put(String fqn, Map data) throws CacheException
   {
      put(Fqn.fromString(fqn), data);
   }

   /**
    * Sets a node's data. If the node does not yet exist, it will be created.
    * Also, parent nodes will be created if not existent. If the node already has data, then the new data
    * will override the old one. If the node already existed, a nodeModified() notification will be generated.
    * Otherwise a nodeCreated() motification will be emitted.
    *
    * @param fqn  The fully qualified name of the new node
    * @param data The new data. May be null if no data should be set in the node.
    */
   public void put(Fqn<?> fqn, Map<K, V> data) throws CacheException
   {
      put(fqn, data, false);
   }

   public void put(Fqn<?> fqn, Map<K, V> data, boolean erase) throws CacheException
   {
      GlobalTransaction tx = getCurrentTransaction();
      MethodCall m;
      if (erase)
      {
         m = MethodCallFactory.create(MethodDeclarations.putDataEraseMethodLocal, tx, fqn, data, true, true);
      }
      else
      {
         m = MethodCallFactory.create(MethodDeclarations.putDataMethodLocal, tx, fqn, data, true);
      }
      invokeMethod(m, true);
   }

   /**
    * Adds a key and value to a given node. If the node doesn't exist, it will be created. If the node
    * already existed, a nodeModified() notification will be generated. Otherwise a
    * nodeCreated() motification will be emitted.
    *
    * @param fqn   The fully qualified name of the node
    * @param key   The key
    * @param value The value
    * @return Object The previous value (if any), if node was present
    */
   public V put(String fqn, K key, V value) throws CacheException
   {
      return put(Fqn.fromString(fqn), key, value);
   }

   /**
    * Adds a key and value to a given node. If the node doesn't exist, it will be created. If the node
    * already existed, a nodeModified() notification will be generated. Otherwise a
    * nodeCreated() motification will be emitted.
    *
    * @param fqn   The fully qualified name of the node
    * @param key   The key
    * @param value The value
    * @return Object The previous value (if any), if node was present
    */
   public V put(Fqn<?> fqn, K key, V value) throws CacheException
   {
      GlobalTransaction tx = getCurrentTransaction();
      MethodCall m = MethodCallFactory.create(MethodDeclarations.putKeyValMethodLocal, tx, fqn, key, value, true);
      return (V) invokeMethod(m, true);
   }

   /**
    * Removes the node from the cache.
    *
    * @param fqn The fully qualified name of the node.
    */
   public void remove(String fqn) throws CacheException
   {
      remove(Fqn.fromString(fqn));
   }

   /**
    * Removes the node from the cache.
    *
    * @param fqn The fully qualified name of the node.
    */
   public boolean remove(Fqn fqn) throws CacheException
   {
      GlobalTransaction tx = getCurrentTransaction();
      // special case if we are removing the root.  Remove all children instead.
      if (fqn.isRoot())
      {
         boolean result = true;
         // we need to preserve options
         InvocationContext ctx = getInvocationContext();
         Option o = ctx.getOptionOverrides();
         for (Object childName : _getChildrenNames(fqn))
         {
            ctx.setOptionOverrides(o);
            result = remove(new Fqn<Object>(fqn, childName)) && result;
         }

         return result;
      }
      else
      {
         MethodCall m = MethodCallFactory.create(MethodDeclarations.removeNodeMethodLocal, tx, fqn, true);
         Object retval = invokeMethod(m, true);
         return retval != null && (Boolean) retval;
      }
   }

   /**
    * Called by eviction policy provider. Note that eviction is done only in
    * local mode, that is, it doesn't replicate the node removal. This will
    * cause the replication nodes to not be synchronizing, which is fine since
    * the value will be fetched again when {@link #get} returns null. After
    * that, the contents will be in sync.
    *
    * @param fqn Will remove everythign assoicated with this fqn.
    * @throws CacheException
    */
   public void evict(Fqn fqn) throws CacheException
   {
      if (fqn.isRoot())
      {
         // special treatment for root eviction
         // we need to preserve options
         InvocationContext ctx = getInvocationContext();
         Option o = ctx.getOptionOverrides();
         for (Object childName : _getChildrenNames(fqn))
         {
            ctx.setOptionOverrides(o);
            evict(new Fqn<Object>(fqn, childName));
         }
      }
      else
      {
         MethodCall m = MethodCallFactory.create(MethodDeclarations.evictNodeMethodLocal, fqn);
         invokeMethod(m, true);
      }
   }

   /**
    * Removes <code>key</code> from the node's hashmap
    *
    * @param fqn The fullly qualified name of the node
    * @param key The key to be removed
    * @return The previous value, or null if none was associated with the given key
    */
   public V remove(String fqn, K key) throws CacheException
   {
      return remove(Fqn.fromString(fqn), key);
   }

   /**
    * Removes <code>key</code> from the node's hashmap
    *
    * @param fqn The fullly qualified name of the node
    * @param key The key to be removed
    * @return The previous value, or null if none was associated with the given key
    */
   public V remove(Fqn<?> fqn, K key) throws CacheException
   {
      GlobalTransaction tx = getCurrentTransaction();
      MethodCall m = MethodCallFactory.create(MethodDeclarations.removeKeyMethodLocal, tx, fqn, key, true);
      return (V) invokeMethod(m, true);
   }

   /**
    * Removes the keys and properties from a node.
    */
   public void removeData(String fqn) throws CacheException
   {
      removeData(Fqn.fromString(fqn));
   }

   /**
    * Removes the keys and properties from a named node.
    */
   public void removeData(Fqn fqn) throws CacheException
   {
      GlobalTransaction tx = getCurrentTransaction();
      MethodCall m = MethodCallFactory.create(MethodDeclarations.removeDataMethodLocal, tx, fqn, true);
      invokeMethod(m, true);
   }

   /**
    * Lock a given node (or the entire subtree starting at this node)
    * @param fqn The FQN of the node
    * @param owner The owner. This is simply a key into a hashtable, and can be anything, e.g.
    * a GlobalTransaction, the current thread, or a special object. If null, it is set to Thread.currentThread()
    * @param lock_type The type of lock (RO, RW). Needs to be of type DataNode.LOCK_TYPE_READ or DataNode.LOCK_TYPE_WRITE
    * @param lock_recursive If true, the entire subtree is locked, else only the given node
    * @throws CacheException If node doesn't exist, a NodeNotExistsException is throw. Other exceptions are
    * LockingException, TimeoutException and UpgradeException
    */
   //   public void lock(Fqn fqn, Object owner, int lock_type, boolean lock_recursive) throws CacheException {
   //
   //   }

   /**
    * Unlock a given node (or the entire subtree starting at this node)
    * @param fqn The FQN of the node
    * @param owner The owner. This is simply a key into a hashtable, and can be anything, e.g.
    * a GlobalTransaction, the current thread, or a special object. If null, it is set to Thread.currentThread()
    * @param unlock_recursive If true, the entire subtree is unlocked, else only the given node
    * @param force Release the lock even if we're not the owner
    */
   //   public void unlock(Fqn fqn, Object owner, boolean unlock_recursive, boolean force) {
   //
   //   }

   /**
    * Releases all locks for this node and the entire node subtree.
    */
   public void releaseAllLocks(String fqn)
   {
      releaseAllLocks(Fqn.fromString(fqn));
   }

   /**
    * Releases all locks for this node and the entire node subtree.
    */
   public void releaseAllLocks(Fqn fqn)
   {
      MethodCall m = MethodCallFactory.create(MethodDeclarations.releaseAllLocksMethodLocal, fqn);
      try
      {
         invokeMethod(m, true);
      }
      catch (CacheException e)
      {
         log.error("failed releasing all locks for " + fqn, e);
      }
   }

   /**
    * Prints a representation of the node defined by <code>fqn</code>.
    * Output includes name, fqn and data.
    */
   public String print(String fqn)
   {
      return print(Fqn.fromString(fqn));
   }

   /**
    * Prints a representation of the node defined by <code>fqn</code>.
    * Output includes name, fqn and data.
    */
   public String print(Fqn fqn)
   {
      MethodCall m = MethodCallFactory.create(MethodDeclarations.printMethodLocal, fqn);
      Object retval = null;
      try
      {
         retval = invokeMethod(m, true);
      }
      catch (Throwable e)
      {
         retval = e;
      }
      if (retval != null)
      {
         return retval.toString();
      }
      else
      {
         return "";
      }
   }


   /**
    * Returns all children of a given node.
    * Returns null of the parent node was not found, or if there are no
    * children.
    * The set is unmodifiable.
    *
    * @param fqn The fully qualified name of the node
    * @return Set A list of child names (as Strings)
    * @see #getChildrenNames(Fqn)
    */
   public Set getChildrenNames(String fqn) throws CacheException
   {
      return getChildrenNames(Fqn.fromString(fqn));
   }

   /**
    * Returns all children of a given node.  Returns an empty set if there are no children.
    * The set is unmodifiable.
    *
    * @param fqn The fully qualified name of the node
    * @return Set an unmodifiable set of children names, Object.
    */
   public <E> Set<E> getChildrenNames(Fqn<E> fqn) throws CacheException
   {
      MethodCall m = MethodCallFactory.create(MethodDeclarations.getChildrenNamesMethodLocal, fqn);
      Set<E> retval = null;
      retval = (Set<E>) invokeMethod(m, true);
      if (retval != null)
         retval = Collections.unmodifiableSet(new HashSet<E>(retval));
      else
         retval = Collections.emptySet();
      return retval;
   }

   @SuppressWarnings("unchecked")
   public <E> Set<E> _getChildrenNames(Fqn<E> fqn) throws CacheException
   {
      NodeSPI<K, V> n = findNode(fqn);
      if (n == null) return null;
      Set<E> childNames = new HashSet<E>();
      Map childrenMap = n.getChildrenMapDirect();
      if (childrenMap == null || childrenMap.isEmpty()) return Collections.emptySet();
      Collection s = childrenMap.values();
      // prune deleted children - JBCACHE-1136
      for (Object c : s)
      {
         NodeSPI child = (NodeSPI) c;
         if (!child.isDeleted())
         {
            E e = (E) child.getFqn().getLastElement();
            childNames.add(e);
         }
      }
      return childNames;
   }

   /**
    * Returns true if the FQN exists and the node has children.
    */
   public boolean hasChild(Fqn fqn)
   {
      if (fqn == null) return false;

      NodeSPI n = findNode(fqn);
      return n != null && n.hasChildrenDirect();
   }

   /**
    * Returns a debug string with few details.
    */
   public String toString()
   {
      return toString(false);
   }


   /**
    * Returns a debug string with optional details of contents.
    */
   public String toString(boolean details)
   {
      StringBuffer sb = new StringBuffer();
      int indent = 0;

      if (!details)
      {
         sb.append(getClass().getName()).append(" [").append(getNumberOfNodes()).append(" nodes, ");
         sb.append(getNumberOfLocksHeld()).append(" locks]");
      }
      else
      {
         if (root == null)
            return sb.toString();
         for (NodeSPI n : root.getChildrenDirect())
         {
            n.print(sb, indent);
            sb.append("\n");
         }
      }
      return sb.toString();
   }


   /**
    * Prints information about the contents of the nodes in the cache's current
    * in-memory state.  Does not load any previously evicted nodes from a
    * cache loader, so evicted nodes will not be included.
    */
   public String printDetails()
   {
      StringBuffer sb = new StringBuffer();
      root.printDetails(sb, 0);
      sb.append("\n");
      return sb.toString();
   }

   /**
    * Returns lock information.
    */
   public String printLockInfo()
   {
      StringBuffer sb = new StringBuffer("\n");
      int indent = 0;

      for (NodeSPI n : root.getChildrenDirect())
      {
         n.getLock().printLockInfo(sb, indent);
         sb.append("\n");
      }
      return sb.toString();
   }

   /**
    * Returns the number of read or write locks held across the entire cache.
    */
   public int getNumberOfLocksHeld()
   {
      return numLocks(root);
   }

   private int numLocks(NodeSPI<K, V> n)
   {
      int num = 0;
      if (n.getLock().isLocked())
      {
         num++;
      }
      for (NodeSPI<K, V> cn : n.getChildrenDirect(true))
      {
         num += numLocks(cn);
      }
      return num;
   }

   /**
    * Returns an <em>approximation</em> of the total number of nodes in the
    * cache. Since this method doesn't acquire any locks, the number might be
    * incorrect, or the method might even throw a
    * ConcurrentModificationException
    */
   public int getNumberOfNodes()
   {
      return numNodes(root) - 1;
   }

   private int numNodes(NodeSPI<K, V> n)
   {
      int count = 1;// for n
      for (NodeSPI<K, V> child : n.getChildrenDirect())
      {
         count += numNodes(child);
      }
      return count;
   }

   /**
    * Returns an <em>approximation</em> of the total number of attributes in
    * the cache. Since this method doesn't acquire any locks, the number might
    * be incorrect, or the method might even throw a
    * ConcurrentModificationException
    */
   public int getNumberOfAttributes()
   {
      return numAttributes(root);
   }

   /**
    * Returns an <em>approximation</em> of the total number of attributes in
    * this sub cache.
    *
    * @see #getNumberOfAttributes
    */
   public int getNumberOfAttributes(Fqn fqn)
   {
      return numAttributes(findNode(fqn));
   }

   private int numAttributes(NodeSPI<K, V> n)
   {
      int count = 0;
      for (NodeSPI<K, V> child : n.getChildrenDirect())
      {
         count += numAttributes(child);
      }
      count += n.getDataDirect().size();
      return count;
   }

   /* ---------------------- Remote method calls -------------------- */

   /**
    * @param mbrs
    * @param method_call
    * @param synchronous
    * @param exclude_self
    * @param timeout
    * @return
    * @throws Exception
    * @deprecated Note this is due to be moved to an interceptor.
    */
   @Deprecated
   public List callRemoteMethods(List<Address> mbrs, MethodCall method_call,
                                 boolean synchronous, boolean exclude_self, long timeout)
           throws Exception
   {
      return callRemoteMethods(mbrs, method_call, synchronous ? GroupRequest.GET_ALL : GroupRequest.GET_NONE, exclude_self, timeout);
   }

   @Deprecated
   public List callRemoteMethods(List<Address> mbrs, MethodCall method_call, int mode, boolean exclude_self, long timeout)
           throws Exception
   {
      return callRemoteMethods(mbrs, method_call, mode, exclude_self, timeout, null);
   }

   /**
    * Overloaded to allow a finer grained control over JGroups mode
    *
    * @param mbrs
    * @param method_call
    * @param mode
    * @param exclude_self
    * @param timeout
    * @return
    * @throws Exception
    * @deprecated Note this is due to be moved to an interceptor.
    */
   @Deprecated
   public List callRemoteMethods(List<Address> mbrs, MethodCall method_call, int mode, boolean exclude_self, long timeout, RspFilter rspFilter)
           throws Exception
   {
      int modeToUse = mode;
      int preferredMode;
      if ((preferredMode = getInvocationContext().getOptionOverrides().getGroupRequestMode()) > -1)
         modeToUse = preferredMode;

      RspList rsps = null;
      Rsp rsp;
      List retval;
      Vector<Address> validMembers;

      if (disp == null)
      {
         return null;
      }

      validMembers = null;
      if (mbrs != null)
         validMembers = new Vector<Address>(mbrs);
      else
      {
         synchronized (members)
         {
            validMembers = new Vector<Address>(this.members);
         }
      }

      if (exclude_self && !validMembers.isEmpty())
      {
         Object local_addr = getLocalAddress();
         if (local_addr != null)
         {
            validMembers.remove(local_addr);
         }
      }
      if (validMembers.isEmpty())
      {
         if (log.isTraceEnabled())
         {
            log.trace("destination list is empty, discarding call");
         }
         return null;
      }

      if (log.isTraceEnabled())
      {
         log.trace("callRemoteMethods(): valid members are " + validMembers + " methods: " + method_call.getArgs()[0]);
      }

      if (channel.flushSupported())
      {
         if (!flushBlockGate.await(configuration.getStateRetrievalTimeout()))
            throw new TimeoutException("State retrieval timed out waiting for flush unblock.");
      }
      rsps = rspFilter == null
              ? disp.callRemoteMethods(validMembers, method_call, modeToUse, timeout, buddyManager != null && buddyManager.isEnabled())
              : disp.callRemoteMethods(validMembers, method_call, modeToUse, timeout, buddyManager != null && buddyManager.isEnabled(), false, rspFilter);

      // a null response is 99% likely to be due to a marshalling problem - we throw a NSE, this needs to be changed when
      // JGroups supports http://jira.jboss.com/jira/browse/JGRP-193
      if (rsps == null)
      {
         // return null;
         throw new NotSerializableException("RpcDispatcher returned a null.  This is most often caused by args for " + method_call + " not being serializable.");
      }
      if (mode == GroupRequest.GET_NONE)
      {
         return Collections.EMPTY_LIST;// async case
      }

      if (log.isTraceEnabled())
      {
         log.trace("(" + getLocalAddress() + "): responses for method " + method_call.getName() + ":\n" + rsps);
      }

      retval = new ArrayList(rsps.size());
      for (int i = 0; i < rsps.size(); i++)
      {
         rsp = (Rsp) rsps.elementAt(i);
         if (rsp.wasSuspected() || !rsp.wasReceived())
         {
            CacheException ex;
            if (rsp.wasSuspected())
            {
               ex = new SuspectException("Suspected member: " + rsp.getSender());
            }
            else
            {
               ex = new TimeoutException("Replication timeout for " + rsp.getSender());
            }
            retval.add(new ReplicationException("rsp=" + rsp, ex));
         }
         else
         {
            retval.add(rsp.getValue());
         }
      }
      return retval;
   }

   /**
    * @param members
    * @param method
    * @param args
    * @param synchronous
    * @param exclude_self
    * @param timeout
    * @return
    * @throws Exception
    * @deprecated Note this is due to be moved to an interceptor.
    */
   @Deprecated
   public List callRemoteMethods(List<Address> members, Method method, Object[] args,
                                 boolean synchronous, boolean exclude_self, long timeout)
           throws Exception
   {
      return callRemoteMethods(members, MethodCallFactory.create(method, args), synchronous, exclude_self, timeout);
   }

   /**
    * @param members
    * @param method_name
    * @param types
    * @param args
    * @param synchronous
    * @param exclude_self
    * @param timeout
    * @return
    * @throws Exception
    * @deprecated Note this is due to be moved to an interceptor.
    */
   @Deprecated
   public List callRemoteMethods(Vector<Address> members, String method_name,
                                 Class[] types, Object[] args,
                                 boolean synchronous, boolean exclude_self, long timeout)
           throws Exception
   {
      Method method = getClass().getDeclaredMethod(method_name, types);
      return callRemoteMethods(members, method, args, synchronous, exclude_self, timeout);
   }

   /* -------------------- End Remote method calls ------------------ */

   /* --------------------- Callbacks -------------------------- */

   /* ----- These are VERSIONED callbacks to facilitate JBCACHE-843.  Also see docs/design/DataVersion.txt --- */

   public void _putForExternalRead(GlobalTransaction gtx, Fqn fqn, K key, V value, DataVersion dv) throws CacheException
   {
      _putForExternalRead(gtx, fqn, key, value);
   }

   public void _put(GlobalTransaction tx, Fqn fqn, Map<K, V> data, boolean create_undo_ops, DataVersion dv) throws CacheException
   {
      _put(tx, fqn, data, create_undo_ops, false, dv);
   }

   public void _put(GlobalTransaction tx, Fqn fqn, Map<K, V> data, boolean create_undo_ops, boolean erase_contents, DataVersion dv) throws CacheException
   {
      _put(tx, fqn, data, create_undo_ops, erase_contents);
   }

   public Object _put(GlobalTransaction tx, Fqn fqn, K key, V value, boolean create_undo_ops, DataVersion dv) throws CacheException
   {
      return _put(tx, fqn, key, value, create_undo_ops);
   }

   public boolean _remove(GlobalTransaction tx, Fqn fqn, boolean create_undo_ops, DataVersion dv) throws CacheException
   {
      return _remove(tx, fqn, create_undo_ops, true);
   }

   public Object _remove(GlobalTransaction tx, Fqn fqn, K key, boolean create_undo_ops, DataVersion dv) throws CacheException
   {
      return _remove(tx, fqn, key, create_undo_ops);
   }

   public void _removeData(GlobalTransaction tx, Fqn fqn, boolean create_undo_ops, DataVersion dv) throws CacheException
   {

      _removeData(tx, fqn, create_undo_ops, true);
   }

   /* ----- End VERSIONED callbacks - Now for the NORMAL callbacks. -------- */

   /**
    * Internal put method.
    * Does the real work. Needs to acquire locks if accessing nodes, depending on
    * the value of <tt>locking</tt>. If run inside a transaction, needs to (a) add
    * newly acquired locks to {@link TransactionEntry}'s lock list, (b) add nodes
    * that were created to {@link TransactionEntry}'s node list and (c) create
    * {@link Modification}s and add them to {@link TransactionEntry}'s modification
    * list and (d) create compensating modifications to undo the changes in case
    * of a rollback
    *
    * @param fqn
    * @param data
    * @param create_undo_ops If true, undo operations will be created (default is true).
    *                        Otherwise they will not be created (used by rollback()).
    */
   public void _put(GlobalTransaction tx, String fqn, Map<K, V> data, boolean create_undo_ops)
           throws CacheException
   {
      _put(tx, Fqn.fromString(fqn), data, create_undo_ops);
   }


   /**
    * Internal put method.
    * Does the real work. Needs to acquire locks if accessing nodes, depending on
    * the value of <tt>locking</tt>. If run inside a transaction, needs to (a) add
    * newly acquired locks to {@link TransactionEntry}'s lock list, (b) add nodes
    * that were created to {@link TransactionEntry}'s node list and (c) create
    * {@link Modification}s and add them to {@link TransactionEntry}'s modification
    * list and (d) create compensating modifications to undo the changes in case
    * of a rollback
    *
    * @param fqn
    * @param data
    * @param create_undo_ops If true, undo operations will be created (default is true).
    *                        Otherwise they will not be created (used by rollback()).
    */
   public void _put(GlobalTransaction tx, Fqn fqn, Map<K, V> data, boolean create_undo_ops)
           throws CacheException
   {
      _put(tx, fqn, data, create_undo_ops, false);
   }

   /**
    * Internal put method.
    * Does the real work. Needs to acquire locks if accessing nodes, depending on
    * the value of <tt>locking</tt>. If run inside a transaction, needs to (a) add
    * newly acquired locks to {@link TransactionEntry}'s lock list, (b) add nodes
    * that were created to {@link TransactionEntry}'s node list and (c) create
    * {@link Modification}s and add them to {@link TransactionEntry}'s modification
    * list and (d) create compensating modifications to undo the changes in case
    * of a rollback
    *
    * @param fqn
    * @param data
    * @param create_undo_ops If true, undo operations will be created (default is true).
    * @param erase_contents  Clear the existing hashmap before putting the new data into it
    *                        Otherwise they will not be created (used by rollback()).
    */
   public void _put(GlobalTransaction tx, Fqn fqn, Map<K, V> data, boolean create_undo_ops, boolean erase_contents)
           throws CacheException
   {
      if (log.isTraceEnabled())
      {
         log.trace("_put(" + tx + ", \"" + fqn + "\", " + data + " undo=" + create_undo_ops + " erase=" + erase_contents + ")");
      }
      InvocationContext ctx = getInvocationContext();
      boolean isRollback = checkIsRollingBack(ctx.getTransaction());
      NodeSPI<K, V> n = findNodeCheck(tx, fqn, isRollback);
      Map<K, V> rawData = n.getDataDirect();
      if (!isRollback) notifier.notifyNodeModified(fqn, true, NodeModifiedEvent.ModificationType.PUT_MAP, rawData, ctx);

      // create a compensating method call (reverting the effect of
      // this modification) and put it into the TX's undo list.
      if (tx != null && create_undo_ops)
      {
         // erase and set to previous hashmap contents
         MethodCall undo_op = MethodCallFactory.create(MethodDeclarations.putDataEraseMethodLocal, tx, fqn, new HashMap<K, V>(rawData), false, true);
         tx_table.addUndoOperation(tx, undo_op);
      }

      if (erase_contents)
         n.clearDataDirect();
      n.putAllDirect(data);

      if (!isRollback)
         notifier.notifyNodeModified(fqn, false, NodeModifiedEvent.ModificationType.PUT_MAP, n.getDataDirect(), ctx);
   }

   /**
    * Internal put method.
    *
    * @return Previous value (if any)
    */
   public Object _put(GlobalTransaction tx, String fqn, K key, V value, boolean create_undo_ops)
           throws CacheException
   {
      return _put(tx, Fqn.fromString(fqn), key, value, create_undo_ops);
   }

   private boolean checkIsRollingBack(Transaction tx)
   {
      try
      {
         return tx != null && (
                 tx.getStatus() == javax.transaction.Status.STATUS_ROLLEDBACK ||
                         tx.getStatus() == javax.transaction.Status.STATUS_ROLLING_BACK ||
                         tx.getStatus() == javax.transaction.Status.STATUS_MARKED_ROLLBACK);
      }
      catch (Exception e)
      {
         // can't get a hold of a transaction - probably no tx rolling back
         return false;
      }
   }

   /**
    * Internal put method.
    *
    * @return Previous value (if any)
    */
   public Object _put(GlobalTransaction tx, Fqn fqn, K key, V value, boolean create_undo_ops)
           throws CacheException
   {
      if (log.isTraceEnabled())
      {
         log.trace(new StringBuffer("_put(").append(tx).append(", \"").
                 append(fqn).append("\", k=").append(key).append(", v=").append(value).append(")"));
      }


      InvocationContext ctx = getInvocationContext();
      // if this is a rollback then don't fire notifications.
      boolean isRollback = checkIsRollingBack(ctx.getTransaction());

      NodeSPI<K, V> n = findNodeCheck(tx, fqn, isRollback);
      Map<K, V> rawData = n.getDataDirect();
      if (!isRollback)
         notifier.notifyNodeModified(fqn, true, NodeModifiedEvent.ModificationType.PUT_DATA, rawData, ctx);

      V old_value = n.putDirect(key, value);

      // create a compensating method call (reverting the effect of
      // this modification) and put it into the TX's undo list.
      if (tx != null && create_undo_ops)
      {
         MethodCall undo_op;
         if (old_value == null)
         {
            undo_op = MethodCallFactory.create(MethodDeclarations.removeKeyMethodLocal, tx, fqn, key, false);
         }
         else
         {
            undo_op = MethodCallFactory.create(MethodDeclarations.putKeyValMethodLocal, tx, fqn, key, old_value, false);
         }
         // 1. put undo-op in TX' undo-operations list (needed to rollback TX)
         tx_table.addUndoOperation(tx, undo_op);
      }

      Map<K, V> newData = Collections.singletonMap(key, value);
      if (!isRollback)
         notifier.notifyNodeModified(fqn, false, NodeModifiedEvent.ModificationType.PUT_DATA, newData, ctx);
      
      return old_value;
   }

   /**
    * Internal remove method.
    */
   public void _remove(GlobalTransaction tx, String fqn, boolean create_undo_ops) throws CacheException
   {
      _remove(tx, Fqn.fromString(fqn), create_undo_ops);
   }

   /**
    * Internal remove method.
    */
   public boolean _remove(GlobalTransaction tx, Fqn fqn, boolean create_undo_ops) throws CacheException
   {
      return _remove(tx, fqn, create_undo_ops, true);
   }

   public boolean _remove(GlobalTransaction tx, Fqn fqn, boolean create_undo_ops, boolean sendNodeEvent)
           throws CacheException
   {
      return _remove(tx, fqn, create_undo_ops, sendNodeEvent, false);
   }

   /**
    * Internal method to remove a node.
    *
    * @param tx
    * @param fqn
    * @param create_undo_ops
    * @param sendNodeEvent
    */
   public boolean _remove(GlobalTransaction tx, Fqn fqn, boolean create_undo_ops, boolean sendNodeEvent, boolean eviction)
           throws CacheException
   {
      return _remove(tx, fqn, create_undo_ops, sendNodeEvent, eviction, null);
   }

   /**
    * Internal method to remove a node.
    * Performs a remove on a node, passing in a {@link DataVersion} which is used with optimistically locked nodes.  Pass
    * in a null if optimistic locking is not used.
    *
    * @param tx
    * @param fqn
    * @param create_undo_ops
    * @param sendNodeEvent
    * @param eviction
    * @param version
    * @return true if the node was removed, false if not found
    * @throws CacheException
    */
   public boolean _remove(GlobalTransaction tx, Fqn fqn, boolean create_undo_ops, boolean sendNodeEvent, boolean eviction, DataVersion version)
           throws CacheException
   {

      NodeSPI<K, V> n;
      NodeSPI<K, V> parent_node;
      MethodCall undo_op = null;

      if (log.isTraceEnabled())
      {
         log.trace("_remove(" + tx + ", \"" + fqn + "\", undo=" + create_undo_ops + ")");
      }
      InvocationContext ctx = getInvocationContext();
      // check if this is triggered by a rollback operation ...
      boolean isRollback = checkIsRollingBack(ctx.getTransaction());
      if (tx != null)
      {
         try
         {
            if (isRollback)
            {
               log.trace("This remove call is triggered by a transaction rollback, as a compensation operation.  Do a realRemove() instead.");
               return realRemove(fqn, true);
            }
         }
         catch (Exception e)
         {
            // what do we do here?
            log.warn("Unable to get a hold of the transaction for a supposedly transactional call!  This *may* result in stale locks!", e);
         }
      }

      // Find the node. This will add the temporarily created parent nodes to the TX's node list if tx != null)
      n = findNode(fqn, version, true);
      if (n == null)
      {
         if (log.isTraceEnabled())
         {
            log.trace("node " + fqn + " not found");
         }
         return false;
      }

      if (!isRollback)
      {
         if (eviction)
         {
            notifier.notifyNodeEvicted(fqn, true, ctx);
         }
         else
         {
            notifier.notifyNodeRemoved(fqn, true, n.getDataDirect(), ctx);
         }
      }

      parent_node = n.getParent();
      boolean found;

      // remove subtree from parent
      if (eviction || configuration.isNodeLockingOptimistic())
      {
         // if there is no parent node and the fqn is root, found == true otherwise found == false.
         found = n.isValid() && parent_node == null ? fqn.isRoot() : parent_node.removeChildDirect(n.getFqn().getLastElement());
      }
      else
      {
         found = n.isValid() && !n.isDeleted();
         n.markAsDeleted(true);
      }

      if (eviction && parent_node != null)
      {
         parent_node.setChildrenLoaded(false);
      }

      // release all locks for the entire subtree
      // n.getNodeSPI().getLock().releaseAll(tx != null ? tx : (Object) Thread.currentThread());

      // create a compensating method call (reverting the effect of
      // this modification) and put it into the TX's undo list.
      if (tx != null && create_undo_ops && !eviction && found)
      {
         undo_op = MethodCallFactory.create(MethodDeclarations.addChildMethodLocal, tx, parent_node.getFqn(), n.getFqn().getLastElement(), n, false);

         // 1. put undo-op in TX' undo-operations list (needed to rollback TX)
         tx_table.addUndoOperation(tx, undo_op);
      }

      if (!isRollback)
      {
         if (eviction)
         {
            notifier.notifyNodeEvicted(fqn, false, ctx);
         }
         else
         {
            notifier.notifyNodeRemoved(fqn, false, null, ctx);
         }
      }

      return found;
   }

   /**
    * Internal method to remove a key.
    *
    * @param fqn
    * @param key
    * @return Object
    */
   public V _remove(GlobalTransaction tx, String fqn, K key, boolean create_undo_ops)
           throws CacheException
   {
      return _remove(tx, Fqn.fromString(fqn), key, create_undo_ops);
   }

   /**
    * Internal method to remove a key.
    *
    * @param fqn
    * @param key
    * @return Object
    */
   public V _remove(GlobalTransaction tx, Fqn fqn, K key, boolean create_undo_ops)
           throws CacheException
   {
      MethodCall undo_op = null;
      V old_value = null;

      if (log.isTraceEnabled())
      {
         log.trace("_remove(" + tx + ", \"" + fqn + "\", key=" + key + ")");
      }

      // Find the node. This will lock it (if <tt>locking</tt> is true) and
      // add the temporarily created parent nodes to the TX's node list if tx != null)
      NodeSPI<K, V> n = findNode(fqn);
      if (n == null)
      {
         log.warn("node " + fqn + " not found");
         return null;
      }
      InvocationContext ctx = getInvocationContext();
      boolean isRollback = checkIsRollingBack(ctx.getTransaction());
      if (!isRollback)
         notifier.notifyNodeModified(fqn, true, NodeModifiedEvent.ModificationType.REMOVE_DATA, n.getDataDirect(), ctx);

      old_value = n.removeDirect(key);

      // create a compensating method call (reverting the effect of
      // this modification) and put it into the TX's undo list.
      if (tx != null && create_undo_ops && old_value != null)
      {
         undo_op = MethodCallFactory.create(MethodDeclarations.putKeyValMethodLocal, tx, fqn, key, old_value, false);
         // 1. put undo-op in TX' undo-operations list (needed to rollback TX)
         tx_table.addUndoOperation(tx, undo_op);
      }

      Map<K, V> removedData = Collections.singletonMap(key, old_value);
      if (!isRollback)
         notifier.notifyNodeModified(fqn, false, NodeModifiedEvent.ModificationType.REMOVE_DATA, removedData, ctx);

      return old_value;
   }

   /**
    * Internal method to remove data from a node.
    */
   public void _removeData(GlobalTransaction tx, String fqn, boolean create_undo_ops)
           throws CacheException
   {
      _removeData(tx, Fqn.fromString(fqn), create_undo_ops);
   }

   /**
    * Internal method to remove data from a node.
    */
   public void _removeData(GlobalTransaction tx, Fqn fqn, boolean create_undo_ops)
           throws CacheException
   {
      _removeData(tx, fqn, create_undo_ops, true);
   }

   /**
    * Internal method to remove data from a node.
    */
   public void _removeData(GlobalTransaction tx, Fqn fqn, boolean create_undo_ops, boolean sendNodeEvent)
           throws CacheException
   {
      _removeData(tx, fqn, create_undo_ops, sendNodeEvent, false);
   }

   /**
    * Internal method to remove data from a node.
    */
   public void _removeData(GlobalTransaction tx, Fqn fqn, boolean create_undo_ops, boolean sendNodeEvent, boolean eviction)
           throws CacheException
   {
      _removeData(tx, fqn, create_undo_ops, sendNodeEvent, eviction, null);
   }

   /**
    * Internal method to remove data from a node.
    */
   public void _removeData(GlobalTransaction tx, Fqn fqn, boolean create_undo_ops, boolean sendNodeEvent, boolean eviction, DataVersion version)
           throws CacheException
   {
      MethodCall undo_op = null;

      if (log.isTraceEnabled())
      {
         log.trace("_removeData(" + tx + ", \"" + fqn + "\")");
      }

      // Find the node. This will lock it (if <tt>locking</tt> is true) and
      // add the temporarily created parent nodes to the TX's node list if tx != null)
      NodeSPI<K, V> n = findNode(fqn, version);
      if (n == null)
      {
         log.warn("node " + fqn + " not found");
         return;
      }

      Map<K, V> data = n.getDataDirect();
      InvocationContext ctx = getInvocationContext();
      boolean isRollback = checkIsRollingBack(ctx.getTransaction());
      // create a compensating method call (reverting the effect of
      // this modification) and put it into the TX's undo list.
      if (tx != null && create_undo_ops && !eviction)
      {
         if (!data.isEmpty())
         {
            undo_op = MethodCallFactory.create(MethodDeclarations.putDataMethodLocal,
                    tx, fqn, new HashMap<K, V>(data), false);
         }
      }

      if (!isRollback)
      {
         if (eviction)
         {
            notifier.notifyNodeEvicted(fqn, true, ctx);
         }
         else
         {
            notifier.notifyNodeModified(fqn, true, NodeModifiedEvent.ModificationType.REMOVE_DATA, data, ctx);
         }
      }

      n.clearDataDirect();
      if (eviction)
      {
         n.setDataLoaded(false);
      }

      if (!isRollback)
      {
         if (sendNodeEvent)
         {
            notifier.notifyNodeVisited(fqn, false, ctx);
         }
         else
         {// FIXME Bela did this so GUI view can refresh the view after node is evicted. But this breaks eviction policy, especially AOP!!!!
            if (eviction)
            {
               notifier.notifyNodeEvicted(fqn, false, ctx);
            }
            else
            {
               notifier.notifyNodeModified(fqn, false, NodeModifiedEvent.ModificationType.REMOVE_DATA, data, ctx);
            }
         }
      }

      // put undo-op in TX' undo-operations list (needed to rollback TX)
      if (tx != null && create_undo_ops)
      {
         tx_table.addUndoOperation(tx, undo_op);
      }
   }


   /**
    * Internal evict method called by eviction policy provider.
    *
    * @param fqn removes everything assoicated with this FQN
    * @return <code>true</code> if the node has been completely removed,
    *         <code>false</code> if only the data map was removed, due
    *         to the presence of children
    * @throws CacheException
    */
   public boolean _evict(Fqn fqn) throws CacheException
   {
      if (peek(fqn, false, true) == null) return true;
      // node does not exist. Maybe it has been recursively removed.

      // use remove method now if there is a child node. Otherwise, it is removed
      boolean create_undo_ops = false;
      boolean sendNodeEvent = false;
      boolean eviction = true;
      if (log.isTraceEnabled())
      {
         log.trace("_evict(" + fqn + ")");
      }
      if (hasChild(fqn))
      {
         _removeData(null, fqn, create_undo_ops, sendNodeEvent, eviction);
         return false;
      }
      else
      {
         _remove(null, fqn, create_undo_ops, sendNodeEvent, eviction);
         return true;
      }
   }

   /**
    * Internal evict method called by eviction policy provider.
    *
    * @param fqn
    * @param version
    * @return <code>true</code> if the node has been completely removed,
    *         <code>false</code> if only the data map was removed, due
    *         to the presence of children
    * @throws CacheException
    */
   public boolean _evict(Fqn fqn, DataVersion version) throws CacheException
   {
      if (!exists(fqn))
         return true;// node does not exist

      if (log.isTraceEnabled())
      {
         log.trace("_evict(" + fqn + ", " + version + ")");
      }

      if (hasChild(fqn))
      {
         _removeData(null, fqn, false, false, true, version);
         return false;
      }
      else
      {
         _remove(null, fqn, false, false, true, version);
         return true;
      }
   }

   /**
    * Very much like an evict(), except that regardless of whether there is a child present, this call will never
    * remove the node from memory - just remove its contents.
    * <p/>
    * Also, potentially throws a cache exception if data versioning is used and the node in memory has a newer data
    * version than what is passed in.
    * <p/>
    * Finally, the data version of the in-memory node is updated to the version being evicted to prevent versions
    * going out of sync.
    */
   public void invalidate(Fqn fqn, DataVersion versionToInvalidate)
   {
      Node<K, V> node = get(fqn); // force interceptor chain, load if necessary from cache loader.

      if (node == null)
      {
         // if pessimistic locking, just return.
         if (!configuration.isNodeLockingOptimistic()) return;

         // check if a tombstone already exists
         NodeSPI nodeSPI = peek(fqn, false, true);
         if (nodeSPI == null)
         {
            log.trace("Node doesn't exist; creating a tombstone");
            // create the node we need.
            Map<K, V> m = Collections.emptyMap();
            InvocationContext ic = getInvocationContext();
            boolean origCacheModeLocal = ic.getOptionOverrides().isCacheModeLocal();
            ic.getOptionOverrides().setCacheModeLocal(true);
            // if we are in a tx this call should happen outside of any tx
            try
            {
               Transaction suspended = null;
               if (getTransactionManager() != null)
               {
                  suspended = getTransactionManager().suspend();
               }
               put(fqn, m);
               if (suspended != null) getTransactionManager().resume(suspended);
               ic.getOptionOverrides().setCacheModeLocal(origCacheModeLocal);
            }
            catch (Exception e)
            {
               log.error("Unable to create tombstone!", e);
            }
            nodeSPI = (NodeSPI) root.getChild(fqn);
         }
         node = nodeSPI;
      }
      
      if (configuration.isNodeLockingOptimistic())
         _removeData(null, fqn, false, false, true, versionToInvalidate);
      else
         _evict(fqn);

      // mark the node to be removed (and all children) as invalid so anyone holding a direct reference to it will
      // be aware that it is no longer valid.
      ((NodeSPI) node).setValid(false, true);

      if (versionToInvalidate != null)
      {
         NodeSPI n = peek(fqn, false, true);
         n.setVersion(versionToInvalidate);
      }
   }

   /**
    * Evicts a key/value pair from a node's attributes. Note that this is <em>local</em>, will not be replicated.
    * @param fqn
    * @param key
    * @throws CacheException
    */
   //    public void _evict(Fqn fqn, Object key) throws CacheException {
   //       if(!exists(fqn)) return;
   //       boolean create_undo_ops = false;
   //       boolean sendNodeEvent = false;
   //       boolean eviction=true;
   //       _removeData(null, fqn, create_undo_ops, sendNodeEvent, eviction);
   //    }


   /**
    * Compensating method to {@link #_remove(GlobalTransaction,Fqn,boolean)}.
    */
   public void _addChild(GlobalTransaction gtx, Fqn parent_fqn, Object child_name, Node cn, boolean undoOps)
           throws CacheException
   {
      NodeSPI childNode = (NodeSPI) cn;
      if (log.isTraceEnabled())
      {
         log.trace("_addChild(\"" + parent_fqn + "\", \"" + child_name + "\", node=" + childNode + ")");
      }

      if (parent_fqn == null || child_name == null || childNode == null)
      {
         log.error("parent_fqn or child_name or childNode was null");
         return;
      }
      NodeSPI parentNode = findNode(parent_fqn);
      if (parentNode == null)
      {
         log.warn("node " + parent_fqn + " not found");
         return;
      }
      InvocationContext ctx = getInvocationContext();
      boolean isRollback = checkIsRollingBack(ctx.getTransaction());
      Fqn fqn = new Fqn(parent_fqn, child_name);
      if (!isRollback) notifier.notifyNodeCreated(fqn, true, ctx);
      parentNode.addChild(child_name, childNode);

      childNode.markAsDeleted(false, true);

      // tricky stuff here - this does look kludgy since we're recursively re-validating nodes
      // potentially mistakenly revalidating tombstones, but this method would only be called
      // when using pess locking and tombstones don't exist with PL, so this is OK.            
      childNode.setValid(true, true);

      if (gtx != null && undoOps)
      {
         // 1. put undo-op in TX' undo-operations list (needed to rollback TX)
         tx_table.addUndoOperation(gtx, MethodCallFactory.create(MethodDeclarations.removeNodeMethodLocal, gtx, fqn, false));
      }

      if (!isRollback) notifier.notifyNodeCreated(fqn, false, ctx);
   }


   /**
    * Replicates changes across to other nodes in the cluster.  Invoked by the
    * ReplicationInterceptor.  Calls need to be forwarded to the
    * ReplicationInterceptor in this interceptor chain. This method will later
    * be moved entirely into the ReplicationInterceptor.
    */
   public Object _replicate(MethodCall method_call) throws Throwable
   {
      try
      {
         Object retVal = invokeMethod(method_call, false);
         // we only need to return values for a set of remote calls; not every call.
         if (MethodDeclarations.returnValueForRemoteCall(method_call.getMethodId()))
         {
            return retVal;
         }
         else
         {
            return null;
         }
      }
      catch (Throwable ex)
      {
         if (method_call.getMethodId() != MethodDeclarations.putForExternalReadMethodLocal_id
             || method_call.getMethodId() != MethodDeclarations.putForExternalReadVersionedMethodLocal_id )
         {
            if (log.isWarnEnabled()) log.warn("replication failure with method_call " + method_call + " exception", ex);
            throw ex;
         }
         else return null;

      }
   }

   /**
    * Replicates a list of method calls.
    */
   public void _replicate(List<MethodCall> methodCalls) throws Throwable
   {
      for (MethodCall methodCall : methodCalls) _replicate(methodCall);
   }

   /**
    * A 'clustered get' call, called from a remote ClusteredCacheLoader.
    *
    * @return a List containing 2 elements: (true or false) and a value (Object).  If buddy replication
    *         is used one further element is added - an Fqn of the backup subtree in which this node may be found.
    */
   public List _clusteredGet(MethodCall methodCall, Boolean searchBackupSubtrees)
   {
      MethodCall call = methodCall;
      if (log.isTraceEnabled()) log.trace("Clustered Get called with params: " + call + ", " + searchBackupSubtrees);
      Method m = call.getMethod();
      Object[] args = call.getArgs();

      Object callResults = null;

      try
      {
         Fqn fqn = (Fqn) args[0];

         if (log.isTraceEnabled()) log.trace("Clustered get: invoking call " + m + " with Fqn " + fqn);
         callResults = m.invoke(this, args);
         boolean found = validResult(callResults, call, fqn);
         if (log.isTraceEnabled()) log.trace("Got result " + callResults + ", found=" + found);
         if (found && callResults == null) callResults = createEmptyResults(call);
      }
      catch (Exception e)
      {
         log.warn("Problems processing clusteredGet call", e);
      }

      List<Object> results = new ArrayList<Object>(2);
      if (callResults != null)
      {
         results.add(true);
         results.add(callResults);
      }
      else
      {
         results.add(false);
         results.add(null);
      }
      return results;
   }

   /**
    * Used with buddy replication's data gravitation interceptor.  If marshalling is necessary, ensure that the cache is
    * configured to use {@link org.jboss.cache.config.Configuration#useRegionBasedMarshalling} and the {@link org.jboss.cache.Region}
    * pertaining to the Fqn passed in is activated, and has an appropriate ClassLoader.
    *
    * @param fqn            the fqn to gravitate
    * @param searchSubtrees if true, buddy backup subtrees are searched and if false, they are not.
    * @return a GravitateResult which contains the data for the gravitation
    */
   public GravitateResult gravitateData(Fqn fqn, boolean searchSubtrees)
           throws CacheException
   {
      // we need to get the state for this Fqn and its sub-nodes.

      // for now, perform a very simple series of getData calls.
      InvocationContext ctx = getInvocationContext();
      if (log.isTraceEnabled()) log.trace("Caller is asking for " + fqn);

      try
      {
         ctx.setOriginLocal(false);

         // use a get() call into the cache to make sure cache loading takes place.
         // no need to cache the original skipDataGravitation setting here - it will always be false of we got here!!
         ctx.getOptionOverrides().setSkipDataGravitation(true);
         Node<K, V> actualNode = get(fqn);
         ctx.getOptionOverrides().setSkipDataGravitation(false);

         if (log.isTraceEnabled()) log.trace("In local tree, this is " + actualNode);

         Fqn backupNodeFqn = null;
         if (actualNode == null && searchSubtrees)
         {
            log.trace("Looking at backup trees.");
            NodeSPI backupSubtree = findNode(BuddyManager.BUDDY_BACKUP_SUBTREE_FQN);
            if (backupSubtree != null)
            {
               // need to loop through backupSubtree's children
               Set childNames = backupSubtree.getChildrenNamesDirect();
               if (childNames != null)
               {
                  for (Object childName : childNames)
                  {
                     // childName is the name of a buddy group since all child names in this
                     // collection are direct children of BUDDY_BACKUP_SUBTREE_FQN
                     backupNodeFqn = BuddyManager.getBackupFqn(childName.toString(), fqn);

                     // use a get() call into the cache to make sure cache loading takes place.
                     ctx.getOptionOverrides().setSkipDataGravitation(true);
                     actualNode = get(backupNodeFqn);
                     ctx.getOptionOverrides().setSkipDataGravitation(false);
                     if (log.isTraceEnabled()) log.trace("Looking for " + backupNodeFqn + ". Search result: " + actualNode);
                     if (actualNode != null) break;
                  }
               }
            }
         }

         if (actualNode == null)
         {
            return GravitateResult.noDataFound();
         }
         else
         {
            // make sure we LOAD data for this node!!
            getData(actualNode.getFqn());
         }

         if (backupNodeFqn == null && searchSubtrees)
         {
            backupNodeFqn = BuddyManager.getBackupFqn(BuddyManager.getGroupNameFromAddress(getLocalAddress()), fqn);
         }

         List<NodeData> list = getNodeData(new LinkedList<NodeData>(), (NodeSPI<K, V>) actualNode);

         return GravitateResult.subtreeResult(list, backupNodeFqn);
      }
      finally
      {
         ctx.setOriginLocal(true);
      }
   }

   private List<NodeData> getNodeData(List<NodeData> list, NodeSPI<K, V> node)
   {
      NodeData data = new NodeData(BuddyManager.getActualFqn(node.getFqn()), node.getDataDirect());
      list.add(data);
      for (NodeSPI<K, V> childNode : node.getChildrenDirect())
      {
         getNodeData(list, childNode);
      }
      return list;
   }

   // ------------- start: buddy replication specific 'lifecycle' method calls

   public void _remoteAssignToBuddyGroup(BuddyGroup group, Map<Fqn, byte[]> state) throws Exception
   {
      if (buddyManager != null)
         buddyManager.handleAssignToBuddyGroup(group, state);
      else if (log.isWarnEnabled())
         log.warn("Received assignToBuddyGroup call from group owner [" + group.getDataOwner() + "] but buddy replication is not enabled on this node!");
   }

   public void _remoteRemoveFromBuddyGroup(String groupName) throws BuddyNotInitException
   {
      if (buddyManager != null)
         buddyManager.handleRemoveFromBuddyGroup(groupName);
      else if (log.isWarnEnabled())
         log.warn("Received removeFromBuddyGroup call for group name [" + groupName + "] but buddy replication is not enabled on this node!");

   }

   public void _remoteAnnounceBuddyPoolName(Address address, String buddyPoolName)
   {
      if (buddyManager != null)
         buddyManager.handlePoolNameBroadcast(address, buddyPoolName);
      else if (log.isWarnEnabled())
         log.warn("Received annouceBuddyPoolName call from [" + address + "] but buddy replication is not enabled on this node!");
   }

   public void _dataGravitationCleanup(GlobalTransaction gtx, Fqn primary, Fqn backup) throws Exception
   {
      MethodCall primaryDataCleanup, backupDataCleanup;
      if (buddyManager.isDataGravitationRemoveOnFind())
      {
         if (log.isTraceEnabled())
            log.trace("DataGravitationCleanup: Removing primary (" + primary + ") and backup (" + backup + ")");
         primaryDataCleanup = MethodCallFactory.create(MethodDeclarations.removeNodeMethodLocal, null, primary, false);
         backupDataCleanup = MethodCallFactory.create(MethodDeclarations.removeNodeMethodLocal, null, backup, false);
      }
      else
      {
         if (log.isTraceEnabled())
            log.trace("DataGravitationCleanup: Evicting primary (" + primary + ") and backup (" + backup + ")");
         primaryDataCleanup = MethodCallFactory.create(MethodDeclarations.evictNodeMethodLocal, primary);
         backupDataCleanup = MethodCallFactory.create(MethodDeclarations.evictNodeMethodLocal, backup);
      }

      invokeMethod(primaryDataCleanup, true);
      invokeMethod(backupDataCleanup, true);
   }

   // ------------- end: buddy replication specific 'lifecycle' method calls


   /**
    * Returns true if the call results returned a valid result.
    */
   private boolean validResult(Object callResults, MethodCall mc, Fqn fqn)
   {
      switch (mc.getMethodId())
      {
         case MethodDeclarations.getDataMapMethodLocal_id:
         case MethodDeclarations.getChildrenNamesMethodLocal_id:
            return callResults != null || exists(fqn);
         case MethodDeclarations.existsMethod_id:
            return (Boolean) callResults;
         default:
            return false;
      }
   }

   /**
    * Creates an empty Collection class based on the return type of the method called.
    */
   private Object createEmptyResults(MethodCall mc)
   {
      switch (mc.getMethodId())
      {
         case MethodDeclarations.getDataMapMethodLocal_id:
         case MethodDeclarations.getChildrenNamesMethodLocal_id:
            return Collections.emptyMap();
         default:
            return null;
      }
   }

   /**
    * Releases all locks for a FQN.
    */
   public void _releaseAllLocks(Fqn fqn)
   {
      NodeSPI<K, V> n;

      try
      {
         n = findNode(fqn);
         if (n == null)
         {
            log.error("releaseAllLocks(): node " + fqn + " not found");
            return;
         }
         releaseAll(n);
      }
      catch (Throwable t)
      {
         log.error("releaseAllLocks(): failed", t);
      }
   }

   private void releaseAll(NodeSPI<K, V> n)
   {
      for (NodeSPI<K, V> child : n.getChildrenDirect())
      {
         releaseAll(child);
      }
      n.getLock().releaseAll();
   }


   /**
    * Finds and returns the string value for the Fqn.
    * Returns null if not found or upon error.
    */
   public String _print(Fqn fqn)
   {
      try
      {
         Node n = findNode(fqn);
         if (n == null) return null;
         return n.toString();
      }
      catch (Throwable t)
      {
         return null;
      }
   }

   /**
    * Should not be called.
    */
   public void _lock(Fqn fqn, NodeLock.LockType lock_type, boolean recursive)
           throws TimeoutException, LockingException
   {
      throw new UnsupportedOperationException("method _lock() should not be invoked on CacheImpl");
   }

   /**
    * Throws UnsupportedOperationException.
    */
   public void optimisticPrepare(GlobalTransaction gtx, List modifications, Map data, Address address, boolean onePhaseCommit)
   {
      throw new UnsupportedOperationException("optimisticPrepare() should not be called on CacheImpl directly");
   }

   /**
    * Throws UnsupportedOperationException.
    */
   public void prepare(GlobalTransaction global_tx, List modifications, Address coord, boolean onePhaseCommit)
   {
      throw new UnsupportedOperationException("prepare() should not be called on CacheImpl directly");
   }

   /**
    * Throws UnsupportedOperationException.
    */
   public void commit(GlobalTransaction tx)
   {
      throw new UnsupportedOperationException("commit() should not be called on CacheImpl directly");
   }

   /**
    * Throws UnsupportedOperationException.
    */
   public void rollback(GlobalTransaction tx)//, Boolean hasMods)
   {
      throw new UnsupportedOperationException("rollback() should not be called on CacheImpl directly");
   }

   /* ----------------- End of  Callbacks ---------------------- */

   /**
    * Adds an undo operatoin to the transaction table.
    */
   public void addUndoOperation(GlobalTransaction gtx, MethodCall undo_op)
   {
      tx_table.addUndoOperation(gtx, undo_op);
   }

   /**
    * Returns the CacheLoaderManager.
    */
   public CacheLoaderManager getCacheLoaderManager()
   {
      return cacheLoaderManager;
   }

   /**
    * Sets the CacheLoaderManager.
    */
   public void setCacheLoaderManager(CacheLoaderManager cacheLoaderManager)
   {
      this.cacheLoaderManager = cacheLoaderManager;
   }

   public void setConfiguration(Configuration configuration)
   {
      this.configuration = configuration;
      configuration.setCacheImpl(this);
   }

   /**
    * @return an instance of {@link Notifier} which has been configured with this instance of CacheImpl.
    */
   public Notifier getNotifier()
   {
      return notifier;
   }

   public InvocationContext getInvocationContext()
   {
      InvocationContext ctx = invocationContextContainer.get();
      if (ctx == null)
      {
         ctx = new InvocationContext();
         invocationContextContainer.set(ctx);
      }
      return ctx;
   }

   public void setInvocationContext(InvocationContext ctx)
   {
      invocationContextContainer.set(ctx);
   }

   /**
    * New API to efficiently relocate a node
    *
    * @since 2.0.0
    */
   public void move(Fqn<?> nodeToMove, Fqn<?> newParent)
   {
      // this needs to be passed up the interceptor chain
      MethodCall m = MethodCallFactory.create(MethodDeclarations.moveMethodLocal, nodeToMove, newParent);
      invokeMethod(m, true);
   }

   /**
    * Called by reflection
    *
    * @param newParentFqn
    * @param nodeToMoveFqn
    */
   public void _move(Fqn nodeToMoveFqn, Fqn newParentFqn)
   {
      // the actual move algorithm.
      NodeSPI<K, V> newParent = findNode(newParentFqn);

      if (newParent == null)
      {
         throw new NodeNotExistsException("New parent node " + newParentFqn + " does not exist when attempting to move node!!");
      }

      NodeSPI<K, V> node = findNode(nodeToMoveFqn);

      if (node == null)
      {
         throw new NodeNotExistsException("Node " + nodeToMoveFqn + " does not exist when attempting to move node!!");
      }

      NodeSPI oldParent = node.getParent();
      Object nodeName = nodeToMoveFqn.getLastElement();

      // now that we have the parent and target nodes:
      // first correct the pointers at the pruning point
      oldParent.removeChildDirect(nodeName);
      newParent.addChild(nodeName, node);
      InvocationContext ctx = getInvocationContext();
      // parent pointer is calculated on the fly using Fqns.
      boolean isRollback = checkIsRollingBack(ctx.getTransaction());

      // notify
      if (!isRollback)
         notifier.notifyNodeMoved(nodeToMoveFqn, new Fqn(newParentFqn, nodeToMoveFqn.getLastElement()), true, ctx);

      // now adjust Fqns of node and all children.
      moveFqns(node, newParent.getFqn());

      if (!isRollback)
         notifier.notifyNodeMoved(nodeToMoveFqn, new Fqn(newParentFqn, nodeToMoveFqn.getLastElement()), false, ctx);

      // now register an undo op
      if (ctx.getTransaction() != null)
      {
         MethodCall undo = MethodCallFactory.create(MethodDeclarations.moveMethodLocal, new Fqn(newParentFqn, nodeToMoveFqn.getLastElement()), oldParent.getFqn());
         tx_table.addUndoOperation(ctx.getGlobalTransaction(), undo);
      }
   }

   public void _block()
   {
      //intentionally empty, used only for reflection in MethodDeclarations.blockChannelLocal
   }

   public void _unblock()
   {
      //intentionally empty, used only for reflection in MethodDeclarations.unblockChannelLocal
   }

   private void moveFqns(NodeSPI node, Fqn newBase)
   {
      Fqn newFqn = new Fqn(newBase, node.getFqn().getLastElement());
      node.setFqn(newFqn);
   }

   /**
    * Set our log category to include our clusterName, if we have one.
    */
   private void configureLogCategory()
   {
      StringBuffer category = new StringBuffer(getClass().getName());
      if (configuration != null)
      {
         String clusterName = configuration.getClusterName();
         if (clusterName != null)
         {
            category.append('.');
            category.append(clusterName);
         }
      }

      log = LogFactory.getLog(category.toString());
   }

   /**
    * Kills the JGroups channel; an unclean channel disconnect
    */
   public void killChannel()
   {
      if (channel != null)
      {
         channel.disconnect();
         channel.close();
//         channel.disconnect();
      }
   }

   protected class MessageListenerAdaptor implements ExtendedMessageListener
   {
      /**
       * Reference to an exception that was raised during
       * state installation on this node.
       */
      protected volatile Exception setStateException;
      private final Object stateLock = new Object();

      protected MessageListenerAdaptor()
      {
      }

      public void waitForState() throws Exception
      {
         synchronized (stateLock)
         {
            while (!isStateSet)
            {
               if (setStateException != null)
               {
                  throw setStateException;
               }

               try
               {
                  stateLock.wait();
               }
               catch (InterruptedException iex)
               {
               }
            }
         }
      }

      protected void stateReceivedSuccess()
      {
         isStateSet = true;
         setStateException = null;
      }

      protected void stateReceivingFailed(Throwable t)
      {
         if (t instanceof CacheException)
         {
            log.debug(t);
         }
         else
         {
            log.error("failed setting state", t);
         }
         if (t instanceof Exception)
         {
            setStateException = (Exception) t;
         }
         else
         {
            setStateException = new Exception(t);
         }
      }

      protected void stateProducingFailed(Throwable t)
      {
         if (t instanceof CacheException)
         {
            log.debug(t);
         }
         else
         {
            log.error("Caught " + t.getClass().getName()
                    + " while responding to state transfer request", t);
         }
      }

      /**
       * Callback, does nothing.
       */
      public void receive(Message msg)
      {
      }

      public byte[] getState()
      {
         MarshalledValueOutputStream out = null;
         byte[] result = null;
         ExposedByteArrayOutputStream baos = new ExposedByteArrayOutputStream(16 * 1024);
         try
         {
            out = new MarshalledValueOutputStream(baos);

            getStateTransferManager().getState(out, Fqn.ROOT, configuration.getStateRetrievalTimeout(), true, true);
         }
         catch (Throwable t)
         {
            stateProducingFailed(t);
         }
         finally
         {
            result = baos.getRawBuffer();
            Util.close(out);
         }
         return result;
      }

      public void setState(byte[] new_state)
      {
         if (new_state == null)
         {
            log.debug("transferred state is null (may be first member in cluster)");
            return;
         }
         ByteArrayInputStream bais = new ByteArrayInputStream(new_state);
         MarshalledValueInputStream in = null;
         try
         {
            in = new MarshalledValueInputStream(bais);
            getStateTransferManager().setState(in, Fqn.ROOT);
            stateReceivedSuccess();
         }
         catch (Throwable t)
         {
            stateReceivingFailed(t);
         }
         finally
         {
            Util.close(in);
            synchronized (stateLock)
            {
               // Notify wait that state has been set.
               stateLock.notifyAll();
            }
         }
      }

      public byte[] getState(String state_id)
      {
         MarshalledValueOutputStream out = null;
         String sourceRoot = state_id;
         byte[] result = null;

         boolean hasDifferentSourceAndIntegrationRoots = state_id.indexOf(StateTransferManager.PARTIAL_STATE_DELIMITER) > 0;
         if (hasDifferentSourceAndIntegrationRoots)
         {
            sourceRoot = state_id.split(StateTransferManager.PARTIAL_STATE_DELIMITER)[0];
         }

         ExposedByteArrayOutputStream baos = new ExposedByteArrayOutputStream(16 * 1024);
         try
         {
            out = new MarshalledValueOutputStream(baos);

            getStateTransferManager().getState(out, Fqn.fromString(sourceRoot),
                    configuration.getStateRetrievalTimeout(), true, true);
         }
         catch (Throwable t)
         {
            stateProducingFailed(t);
         }
         finally
         {
            result = baos.getRawBuffer();
            Util.close(out);
         }
         return result;
      }

      public void getState(OutputStream ostream)
      {
         MarshalledValueOutputStream out = null;
         try
         {
            out = new MarshalledValueOutputStream(ostream);
            getStateTransferManager().getState(out, Fqn.ROOT, configuration.getStateRetrievalTimeout(), true, true);
         }
         catch (Throwable t)
         {
            stateProducingFailed(t);
         }
         finally
         {
            Util.close(out);
         }
      }

      public void getState(String state_id, OutputStream ostream)
      {
         String sourceRoot = state_id;
         MarshalledValueOutputStream out = null;
         boolean hasDifferentSourceAndIntegrationRoots = state_id.indexOf(StateTransferManager.PARTIAL_STATE_DELIMITER) > 0;
         if (hasDifferentSourceAndIntegrationRoots)
         {
            sourceRoot = state_id.split(StateTransferManager.PARTIAL_STATE_DELIMITER)[0];
         }
         try
         {
            out = new MarshalledValueOutputStream(ostream);
            getStateTransferManager().getState(out, Fqn.fromString(sourceRoot), configuration.getStateRetrievalTimeout(), true, true);
         }
         catch (Throwable t)
         {
            stateProducingFailed(t);
         }
         finally
         {
            Util.close(out);
         }
      }

      public void setState(InputStream istream)
      {
         if (istream == null)
         {
            log.debug("stream is null (may be first member in cluster)");
            return;
         }
         MarshalledValueInputStream in = null;
         try
         {
            in = new MarshalledValueInputStream(istream);
            getStateTransferManager().setState(in, Fqn.ROOT);
            stateReceivedSuccess();
         }
         catch (Throwable t)
         {
            stateReceivingFailed(t);
         }
         finally
         {
            Util.close(in);
            synchronized (stateLock)
            {
               // Notify wait that state has been set.
               stateLock.notifyAll();
            }
         }
      }

      public void setState(String state_id, byte[] state)
      {
         if (state == null)
         {
            log.debug("partial transferred state is null");
            return;
         }

         MarshalledValueInputStream in = null;
         String targetRoot = state_id;
         boolean hasDifferentSourceAndIntegrationRoots = state_id.indexOf(StateTransferManager.PARTIAL_STATE_DELIMITER) > 0;
         if (hasDifferentSourceAndIntegrationRoots)
         {
            targetRoot = state_id.split(StateTransferManager.PARTIAL_STATE_DELIMITER)[1];
         }
         try
         {
            log.debug("Setting received partial state for subroot " + state_id);
            Fqn subroot = Fqn.fromString(targetRoot);
//            Region region = regionManager.getRegion(subroot, false);
//            ClassLoader cl = null;
//            if (region != null)
//            {
//               // If a classloader is registered for the node's region, use it
//               cl = region.getClassLoader();
//            }
            ByteArrayInputStream bais = new ByteArrayInputStream(state);
            in = new MarshalledValueInputStream(bais);
            //getStateTransferManager().setState(in, subroot, cl);
            getStateTransferManager().setState(in, subroot);
            stateReceivedSuccess();
         }
         catch (Throwable t)
         {
            stateReceivingFailed(t);
         }
         finally
         {
            Util.close(in);
            synchronized (stateLock)
            {
               // Notify wait that state has been set.
               stateLock.notifyAll();
            }
         }
      }

      public void setState(String state_id, InputStream istream)
      {
         String targetRoot = state_id;
         MarshalledValueInputStream in = null;
         boolean hasDifferentSourceAndIntegrationRoots = state_id.indexOf(StateTransferManager.PARTIAL_STATE_DELIMITER) > 0;
         if (hasDifferentSourceAndIntegrationRoots)
         {
            targetRoot = state_id.split(StateTransferManager.PARTIAL_STATE_DELIMITER)[1];
         }
         if (istream == null)
         {
            log.debug("stream is null (may be first member in cluster). State is not set");
            return;
         }

         try
         {
            log.debug("Setting received partial state for subroot " + state_id);
            in = new MarshalledValueInputStream(istream);
            Fqn subroot = Fqn.fromString(targetRoot);
//            Region region = regionManager.getRegion(subroot, false);
//            ClassLoader cl = null;
//            if (region != null)
//            {
//               // If a classloader is registered for the node's region, use it
//               cl = region.getClassLoader();
//            }
            //getStateTransferManager().setState(in, subroot, cl);
            getStateTransferManager().setState(in, subroot);
            stateReceivedSuccess();
         }
         catch (Throwable t)
         {
            stateReceivingFailed(t);
         }
         finally
         {
            Util.close(in);
            synchronized (stateLock)
            {
               // Notify wait that state has been set.
               stateLock.notifyAll();
            }
         }
      }
   }

   /*-------------------- End of MessageListener ----------------------*/

   /*----------------------- MembershipListener ------------------------*/

   protected class MembershipListenerAdaptor implements ExtendedMembershipListener
   {

      public void viewAccepted(View new_view)
      {
         Vector<Address> new_mbrs = new_view.getMembers();
         if (log.isInfoEnabled()) log.info("viewAccepted(): " + new_view);
         synchronized (members)
         {
            boolean needNotification = false;
            if (new_mbrs != null)
            {
               // Determine what members have been removed
               // and roll back any tx and break any locks
               Vector<Address> removed = new Vector<Address>(members);
               removed.removeAll(new_mbrs);
               removeLocksForDeadMembers(root, removed);

               members.removeAllElements();
               members.addAll(new_mbrs);

               needNotification = true;
            }

            // Now that we have a view, figure out if we are the coordinator
            coordinator = (members.size() != 0 && members.get(0).equals(getLocalAddress()));

            // now notify listeners - *after* updating the coordinator. - JBCACHE-662
            if (needNotification && notifier != null)
            {
               InvocationContext ctx = getInvocationContext();
               notifier.notifyViewChange(new_view, ctx);
            }

            // Wake up any threads that are waiting to know who the members
            // are so they can figure out who the coordinator is
            members.notifyAll();
         }
      }

      /**
       * Called when a member is suspected.
       */
      public void suspect(Address suspected_mbr)
      {
      }

      /**
       * Indicates that a channel has received a BLOCK event from FLUSH protocol.
       */
      public void block()
      {
         flushBlockGate.close();
         if (log.isDebugEnabled())
         {
            log.debug("Block received at " + getLocalAddress());
         }
         MethodCall m = MethodCallFactory.create(MethodDeclarations.blockChannelLocal);
         getInvocationContext().getOptionOverrides().setSkipCacheStatusCheck(true);
         invokeMethod(m, true);
         if (log.isDebugEnabled())
         {
            log.debug("Block processed at " + getLocalAddress());
         }
      }

      /**
       * Indicates that a channel has received a UNBLOCK event from FLUSH protocol.
       */
      public void unblock()
      {
         if (log.isDebugEnabled())
         {
            log.debug("UnBlock received at " + getLocalAddress());
         }
         MethodCall m = MethodCallFactory.create(MethodDeclarations.unblockChannelLocal);
         getInvocationContext().getOptionOverrides().setSkipCacheStatusCheck(true);
         invokeMethod(m, true);
         if (log.isDebugEnabled())
         {
            log.debug("UnBlock processed at " + getLocalAddress());
         }
         flushBlockGate.open();
      }

   }

   /*------------------- End of MembershipListener ----------------------*/

   /* ------------------------------ Private methods --------------------------- */

   /**
    * Returns the transaction associated with the current thread. We get the
    * initial context and a reference to the TransactionManager to get the
    * transaction. This method is used by {@link #getCurrentTransaction()}
    */
   protected Transaction getLocalTransaction()
   {
      if (tm == null)
      {
         return null;
      }
      try
      {
         return tm.getTransaction();
      }
      catch (Throwable t)
      {
         return null;
      }
   }


   /**
    * Returns true if transaction is ACTIVE or PREPARING, false otherwise.
    */
   private boolean isValid(Transaction tx)
   {
      if (tx == null) return false;
      int status = -1;
      try
      {
         status = tx.getStatus();
         return status == Status.STATUS_ACTIVE || status == Status.STATUS_PREPARING;
      }
      catch (SystemException e)
      {
         log.error("failed getting transaction status", e);
         return false;
      }
   }


   /**
    * Returns the transaction associated with the current thread.
    * If a local transaction exists, but doesn't yet have a mapping to a
    * GlobalTransaction, a new GlobalTransaction will be created and mapped to
    * the local transaction.  Note that if a local transaction exists, but is
    * not ACTIVE or PREPARING, null is returned.
    *
    * @return A GlobalTransaction, or null if no (local) transaction was associated with the current thread
    */
   public GlobalTransaction getCurrentTransaction()
   {
      return getCurrentTransaction(true);
   }

   /**
    * Returns the transaction associated with the thread; optionally creating
    * it if is does not exist.
    */
   public GlobalTransaction getCurrentTransaction(boolean createIfNotExists)
   {
      Transaction tx;

      if ((tx = getLocalTransaction()) == null)
      {// no transaction is associated with the current thread
         return null;
      }

      if (!isValid(tx))
      {// we got a non-null transaction, but it is not active anymore
         int status = -1;
         try
         {
            status = tx.getStatus();
         }
         catch (SystemException e)
         {
         }

         // JBCACHE-982 -- don't complain if COMMITTED
         if (status != Status.STATUS_COMMITTED)
         {
            log.warn("status is " + status + " (not ACTIVE or PREPARING); returning null)", new Throwable());
         }
         else
         {
            log.trace("status is COMMITTED; returning null");
         }

         return null;
      }

      return getCurrentTransaction(tx, createIfNotExists);
   }

   /**
    * Returns the global transaction for this local transaction.
    */
   public GlobalTransaction getCurrentTransaction(Transaction tx)
   {
      return getCurrentTransaction(tx, true);
   }

   /**
    * Returns the global transaction for this local transaction.
    *
    * @param createIfNotExists if true, if a global transaction is not found; one is created
    */
   public GlobalTransaction getCurrentTransaction(Transaction tx, boolean createIfNotExists)
   {
      // removed synchronization on tx_table because underlying implementation is thread safe
      // and JTA spec (section 3.4.3 Thread of Control, par 2) says that only one thread may
      // operate on the transaction at one time so no concern about 2 threads trying to call
      // this method for the same Transaction instance at the same time
      //
      GlobalTransaction gtx = tx_table.get(tx);
      if (gtx == null && createIfNotExists)
      {
         Address addr = getLocalAddress();
         gtx = GlobalTransaction.create(addr);
         tx_table.put(tx, gtx);
         TransactionEntry ent = configuration.isNodeLockingOptimistic() ? new OptimisticTransactionEntry() : new TransactionEntry();
         ent.setTransaction(tx);
         tx_table.put(gtx, ent);
         if (log.isTraceEnabled())
         {
            log.trace("created new GTX: " + gtx + ", local TX=" + tx);
         }
      }
      return gtx;
   }


   private void blockUntilCacheStarts() throws InterruptedException
   {
      int pollFrequencyMS = 100;
      long startupWaitTime = configuration.getStateRetrievalTimeout();
      long giveUpTime = System.currentTimeMillis() + startupWaitTime;

      while (System.currentTimeMillis() < giveUpTime)
      {
         if (cacheStatus.allowInvocations()) break;
         Thread.sleep(pollFrequencyMS);
      }

   }

   /**
    * Invokes a method against this object. Contains the logger_ic for handling
    * the various use cases, e.g. mode (local, repl_async, repl_sync),
    * transaction (yes or no) and locking (yes or no).
    * <p/>
    * Only sets originLocal on the invocation context IF this is explicitly passed in as <tt>false</tt>
    * If passed in as <tt>true</tt> then it is not set; but whatever default exists in the invocation
    * context is used.  For example, it may be set prior to calling invokeMethod()
    */
   protected Object invokeMethod(MethodCall m, boolean originLocal) throws CacheException
   {
      // don't create a new one; get it from ThreadLocal just this once, in case a user has added any overrides.
      InvocationContext ctx = getInvocationContext();

      // BR methods should NOT block on the cache being started, since the cache depends on these completing to start.
      if (!MethodDeclarations.isBuddyGroupOrganisationMethod(m.getMethodId()) && !cacheStatus.allowInvocations() && !ctx.getOptionOverrides().isSkipCacheStatusCheck())
      {
         // only throw an exception if this is a locally originating call - JBCACHE-1179
         if (originLocal)
         {
            throw new IllegalStateException("Cache not in STARTED state!");
         }
         else
         {
            if (getCacheStatus() == CacheStatus.STARTING)
            {
               try
               {
                  blockUntilCacheStarts();
               }
               catch (InterruptedException e)
               {
                  Thread.currentThread().interrupt();
               }

               // if the cache STILL can't take invocations...
               if (!cacheStatus.allowInvocations()) throw new IllegalStateException("Cache not in STARTED state!");
            }
            else
            {
               log.warn("Received a remote call but the cache is not in STARTED state - ignoring call.");
               return null;
            }
         }
      }


      MethodCall oldCall = null;
      try
      {
         // check if we had a method call lurking around
         oldCall = ctx.getMethodCall();
         ctx.setMethodCall(m);
         // only set this if originLocal is EXPLICITLY passed in as FALSE.  Otherwise leave it as a default.
         if (!originLocal) ctx.setOriginLocal(false);
         return interceptor_chain.invoke(ctx);
      }
      catch (CacheException e)
      {
         throw e;
      }
      catch (RuntimeException e)
      {
         throw e;
      }
      catch (Throwable t)
      {
         throw new RuntimeException(t);
      }
      finally
      {
         if (!originLocal) ctx.setOriginLocal(true);
         // reset old method call
         ctx.setMethodCall(oldCall);
      }
   }

   /**
    * Returns an object suitable for use in node locking, either the current
    * transaction or the current thread if there is no transaction.
    */
   protected Object getOwnerForLock()
   {
      Object owner = getCurrentTransaction();
      if (owner == null)
      {
         owner = Thread.currentThread();
      }

      return owner;
   }

   /**
    * Finds a node given a fully qualified name.
    * Whenever nodes are created, and the global transaction is not null, the created
    * nodes have to be added to the transaction's {@link TransactionEntry}
    * field.<br>
    * When a lock is acquired on a node, a reference to the lock has to be
    * {@link TransactionEntry#addLock(org.jboss.cache.lock.NodeLock) added to the list of locked nodes}
    * in the {@link TransactionEntry}.
    * <p>This operation will also apply different locking to the cache nodes, depending on
    * <tt>operation_type</tt>. If it is <tt>read</tt> type, all nodes will be acquired with
    * read lock. Otherwise, the operation is <tt>write</tt> type, all parent nodes will be acquired
    * with read lock while the destination node acquires write lock.</p>
    *
    * @param fqn Fully qualified name for the corresponding node.
    * @return DataNode
    */
   public NodeSPI<K, V> findNode(Fqn fqn)
   {
      try
      {
         return findNode(fqn, null);
      }
      catch (CacheException e)
      {
         log.warn("Unexpected error", e);
         return null;
      }
   }

   private NodeSPI<K, V> findNodeCheck(GlobalTransaction tx, Fqn fqn, boolean includeInvalid)
   {
      NodeSPI<K, V> n = findNode(fqn, null, includeInvalid);
      if (n == null)
      {
         String errStr = "node " + fqn + " not found (gtx=" + tx + ", caller=" + Thread.currentThread() + ")";
         if (log.isTraceEnabled())
         {
            log.trace(errStr);
         }
         throw new NodeNotExistsException(errStr);
      }
      return n;
   }

   /**
    * Internal method; not to be used externally.
    * Returns true if the node was found, false if not.
    *
    * @param f
    */
   public boolean realRemove(Fqn f, boolean skipMarkerCheck)
   {
      NodeSPI n = peek(f, true);
      if (n == null)
      {
         return false;
      }


      if (log.isTraceEnabled()) log.trace("Performing a real remove for node " + f + ", marked for removal.");
      if (skipMarkerCheck || n.isDeleted())
      {
         if (n.getFqn().isRoot())
         {
            // do not actually delete; just remove deletion marker
            n.markAsDeleted(true);
            // but now remove all children, since the call has been to remove("/")
            n.removeChildrenDirect();

            // mark the node to be removed (and all children) as invalid so anyone holding a direct reference to it will
            // be aware that it is no longer valid.
            n.setValid(false, true);
            n.setValid(true, false);
            return true;
         }
         else
         {
            // mark the node to be removed (and all children) as invalid so anyone holding a direct reference to it will
            // be aware that it is no longer valid.            
            n.setValid(false, true);
            return n.getParent().removeChildDirect(n.getFqn().getLastElement());
         }
      }
      else
      {
         if (log.isDebugEnabled()) log.debug("Node " + f + " NOT marked for removal as expected, not removing!");
         return false;
      }
   }

   /**
    * Finds a node given a fully qualified name and DataVersion.  Does not include invalid nodes.
    */
   private NodeSPI<K, V> findNode(Fqn fqn, DataVersion version)
   {
      return findNode(fqn, version, false);
   }

   private NodeSPI<K, V> findNode(Fqn fqn, DataVersion version, boolean includeInvalidNodes)
   {
      if (fqn == null) return null;

      NodeSPI<K, V> toReturn = peek(fqn, false, includeInvalidNodes);

      if (toReturn != null && version != null && configuration.isNodeLockingOptimistic())
      {
         // we need to check the version of the data node...
         DataVersion nodeVersion = toReturn.getVersion();
         if (log.isTraceEnabled())
         {
            log.trace("looking for optimistic node [" + fqn + "] with version [" + version + "].  My version is [" + nodeVersion + "]");
         }
         if (nodeVersion.newerThan(version))
         {
            // we have a versioning problem; throw an exception!
            throw new CacheException("Unable to validate versions.");
         }
      }
      return toReturn;
   }

   public synchronized RegionManager getRegionManager()
   {
      if (regionManager == null)
      {
         regionManager = new RegionManager(this);
      }
      return regionManager;
   }


   public Marshaller getMarshaller()
   {
      if (marshaller_ == null)
      {
         synchronized (this)
         {
            if (marshaller_ == null)
            {
               if (configuration.getMarshallerClass() == null || configuration.getMarshallerClass().equals(VersionAwareMarshaller.class.getName()))
               {
                  marshaller_ = new VersionAwareMarshaller(getRegionManager(), configuration);
               }
               else
               {
                  try
                  {
                     marshaller_ = (Marshaller) org.jboss.cache.util.Util.loadClass(configuration.getMarshallerClass()).newInstance();
                  }
                  catch (Exception e)
                  {
                     log.error("Unable to load marshaller " + configuration.getMarshallerClass() + ".  Falling back to default (" + VersionAwareMarshaller.class.getName() + ")");
                     marshaller_ = new VersionAwareMarshaller(getRegionManager(), configuration);
                  }
               }
               if (log.isTraceEnabled()) log.trace("Using marshaller " + marshaller_.getClass().getName());
            }
         }
      }
      return marshaller_;
   }

   private void initialiseCacheLoaderManager() throws CacheException
   {
      if (cacheLoaderManager == null)
      {
         cacheLoaderManager = new CacheLoaderManager();
      }
      cacheLoaderManager.setConfig(configuration.getCacheLoaderConfig(), this);
   }

   /**
    * Sets the CacheLoader to use.
    * Provided for backwards compatibility.
    *
    * @param loader
    * @deprecated only provided for backward compat
    */
   @Deprecated
   public void setCacheLoader(CacheLoader loader)
   {
      log.warn("Using deprecated config method setCacheLoader.  This element will be removed in future, please use CacheLoaderConfiguration instead.");

      try
      {
         if (cacheLoaderManager == null) initialiseCacheLoaderManager();
      }
      catch (Exception e)
      {
         log.warn("Problem setting cache loader.  Perhaps your cache loader config has not been set yet?");
      }
      cacheLoaderManager.setCacheLoader(loader);
   }

   /**
    * Purges the contents of all configured {@link CacheLoader}s
    */
   public void purgeCacheLoaders() throws Exception
   {
      if (cacheLoaderManager != null) cacheLoaderManager.purgeLoaders(true);
   }

   // ---------------------------------------------------------------
   // END: Methods to provide backward compatibility with older cache loader config settings
   // ---------------------------------------------------------------

   private void initialiseChannelAndRpcDispatcher() throws CacheException
   {
      channel = configuration.getRuntimeConfig().getChannel();
      if (channel == null)
      {
         // Try to create a multiplexer channel
         channel = getMultiplexerChannel();

         if (channel != null)
         {
            configuration.setUsingMultiplexer(true);
            if (log.isDebugEnabled())
            {
               log.debug("Created Multiplexer Channel for cache cluster " + configuration.getClusterName() +
                         " using stack " + configuration.getMultiplexerStack());
            }
         }
         else
         {

            try
            {
               if (configuration.getClusterConfig() == null)
               {
                  log.debug("setting cluster properties to default value");
                  channel = new JChannel(configuration.getDefaultClusterConfig());
               }
               else
               {
                  if (log.isTraceEnabled())
                  {
                     log.trace("Cache cluster properties: " + configuration.getClusterConfig());
                  }
                  channel = new JChannel(configuration.getClusterConfig());
               }
            }
            catch (ChannelException el)
            {
               el.printStackTrace();
            }
         }

         configuration.getRuntimeConfig().setChannel(channel);
      }

      channel.setOpt(Channel.AUTO_RECONNECT, true);
      channel.setOpt(Channel.AUTO_GETSTATE, true);
      channel.setOpt(Channel.BLOCK, true);

      // always use the InactiveRegionAwareRpcDispatcher - exceptions due to regions not being active should not propagate to remote
      // nodes as errors. - Manik
      disp = new InactiveRegionAwareRpcDispatcher(channel, ml, new MembershipListenerAdaptor(), this);
      //            disp = new RpcDispatcher(channel, ml, this, this);

      disp.setRequestMarshaller(getMarshaller());
      disp.setResponseMarshaller(getMarshaller());
   }

   private JChannel getMultiplexerChannel() throws CacheException
   {
      String stackName = configuration.getMultiplexerStack();

      RuntimeConfig rtc = configuration.getRuntimeConfig();
      ChannelFactory channelFactory = rtc.getMuxChannelFactory();
      JChannel muxchannel = null;

      if (channelFactory != null)
      {
         try
         {
            muxchannel = (JChannel) channelFactory.createMultiplexerChannel(stackName, configuration.getClusterName());
         }
         catch (Exception e)
         {
            throw new CacheException("Failed to create multiplexed channel using stack " + stackName, e);
         }
      }

      return muxchannel;
   }

   // ================== methods to implement Cache and CacheSPI interfaces ============================

   public List<Interceptor> getInterceptorChain()
   {
      List<Interceptor> modifiable = getInterceptors();
      return modifiable == null ? null : Collections.unmodifiableList(modifiable);
   }

   public void addCacheListener(Object listener)
   {
      getNotifier().addCacheListener(listener);
   }

   public void addCacheListener(Fqn<?> region, Object listener)
   {
      throw new UnsupportedOperationException("Not implemented in this release");
   }

   public void removeCacheListener(Object listener)
   {
      // BES 2007/5/23 workaround to avoid NPE while we decide whether
      // to remove notifier in destroy()
      Notifier n = getNotifier();
      if (n != null)
         n.removeCacheListener(listener);
   }

   public void removeCacheListener(Fqn<?> region, Object listener)
   {
      throw new UnsupportedOperationException("Not implemented in this release");
   }

   public Set<Object> getCacheListeners()
   {
      return getNotifier().getCacheListeners();
   }

   public Set<Object> getCacheListeners(Fqn<?> region)
   {
      throw new UnsupportedOperationException("Not implemented in this release");
   }

   public synchronized void addInterceptor(Interceptor i, int position)
   {
      List<Interceptor> interceptors = getInterceptors();

      i.setCache(this);

      interceptors.add(position, i);

      // now correct the chaining of interceptors...
      Interceptor linkedChain = InterceptorChainFactory.getInstance().correctInterceptorChaining(interceptors);

      setInterceptorChain(linkedChain);
   }

   public synchronized void removeInterceptor(int position)
   {
      List<Interceptor> i = getInterceptors();
      i.remove(position);
      setInterceptorChain(InterceptorChainFactory.getInstance().correctInterceptorChaining(i));
   }

   public RPCManager getRPCManager()
   {
      return configuration.getRuntimeConfig().getRPCManager();
   }

   public String getClusterName()
   {
      return getConfiguration().getClusterName();
   }

   public void evict(Fqn<?> fqn, boolean recursive)
   {
      //this method should be called by eviction thread only, so no transaction - expected (sec param is false)
      Node<K, V> node = peek(fqn, false);
      if (node != null && node.isResident())
      {
         return;
      }
      if (recursive)
      {
         if (node != null)
         {
            evictChildren((NodeSPI<K, V>) node);
         }
      }
      else
      {
         evict(fqn);
      }
   }

   private void evictChildren(NodeSPI<K, V> n)
   {
      for (NodeSPI<K, V> child : n.getChildrenDirect())
      {
         evictChildren(child);
      }
      evict(n.getFqn());
   }

   public Region getRegion(Fqn<?> fqn, boolean createIfAbsent)
   {
      return getRegionManager().getRegion(fqn, createIfAbsent);
   }

   public boolean removeRegion(Fqn<?> fqn)
   {
      return getRegionManager().removeRegion(fqn);
   }

   public boolean removeNode(Fqn<?> fqn)
   {
      return remove(fqn);
   }

   public void putForExternalRead(Fqn<?> fqn, K key, V value)
   {
      // if the node exists then this should be a no-op.
      if (!exists(fqn))
      {
         getInvocationContext().getOptionOverrides().setFailSilently(true);
         GlobalTransaction tx = getCurrentTransaction();
         MethodCall m = MethodCallFactory.create(MethodDeclarations.putForExternalReadMethodLocal, tx, fqn, key, value);
         invokeMethod(m, true);
      }
      else
      {
         if (log.isDebugEnabled())
            log.debug("putForExternalRead() called with Fqn " + fqn + " and this node already exists.  This method is hence a no op.");
      }
   }

   public void _putForExternalRead(GlobalTransaction gtx, Fqn fqn, K key, V value)
   {
      _put(gtx, fqn, key, value, true);
   }

   public boolean isStarted()
   {
      return getCacheStatus() == CacheStatus.STARTED;
   }

   protected void setMessageListener(MessageListenerAdaptor ml)
   {
      this.ml = ml;
   }

}
