/*
 * JBoss, Home of Professional Open Source
 *
 * Distributable under LGPL license.
 * See terms of license at gnu.org.
 */
package org.jboss.cache;

import net.jcip.annotations.ThreadSafe;
import org.jboss.cache.buddyreplication.BuddyManager;
import org.jboss.cache.buddyreplication.GravitateResult;
import org.jboss.cache.interceptors.Interceptor;
import org.jboss.cache.loader.CacheLoader;
import org.jboss.cache.loader.CacheLoaderManager;
import org.jboss.cache.lock.NodeLock;
import org.jboss.cache.marshall.Marshaller;
import org.jboss.cache.notifications.Notifier;
import org.jboss.cache.statetransfer.StateTransferManager;
import org.jboss.cache.transaction.GlobalTransaction;
import org.jboss.cache.transaction.TransactionTable;

import javax.transaction.Transaction;
import javax.transaction.TransactionManager;
import java.util.List;
import java.util.Map;

/**
 * A more detailed interface to {@link Cache}, which is used when writing plugins for or extending JBoss Cache.  A reference
 * to this interface should only be obtained when it is passed in to your code, for example when you write an
 * {@link Interceptor} or {@link CacheLoader}.
 * <p/>
 * <B><I>You should NEVER attempt to directly cast a {@link Cache} instance to this interface.  In future, the implementation may not allow it.</I></B>
 * <p/>
 * This interface contains overridden method signatures of some methods from {@link Cache}, overridden to ensure return
 * types of {@link Node} are replaced with {@link NodeSPI}.
 * <p/>
 *
 * @author <a href="mailto:manik@jboss.org">Manik Surtani (manik@jboss.org)</a>
 * @see NodeSPI
 * @see Cache
 * @see org.jboss.cache.loader.CacheLoader
 * @see org.jboss.cache.interceptors.Interceptor
 * @since 2.0.0
 */
@ThreadSafe
public interface CacheSPI<K, V> extends Cache<K, V>
{
   /**
    * Overrides {@link org.jboss.cache.Cache#getRoot()} to return a {@link org.jboss.cache.NodeSPI} instead of a {@link org.jboss.cache.Node}.
    */
   NodeSPI<K, V> getRoot();

   /**
    * Retrieves a reference to a running {@link javax.transaction.TransactionManager}, if one is configured.
    *
    * @return a TransactionManager
    */
   TransactionManager getTransactionManager();

   /**
    * @return an immutable {@link List} of {@link Interceptor}s configured for this cache, or
    *         <code>null</code> if {@link Cache#create() create()} has not been invoked
    *         and the interceptors thus do not exist.
    */
   List<Interceptor> getInterceptorChain();

   /**
    * Adds a custom interceptor to the interceptor chain, at specified position, where the first interceptor in the chain
    * is at position 0 and the last one at getInterceptorChain().size() - 1.
    *
    * @param i        the interceptor to add
    * @param position the position to add the interceptor
    */
   void addInterceptor(Interceptor i, int position);

   /**
    * Removes the interceptor at a specified position, where the first interceptor in the chain
    * is at position 0 and the last one at getInterceptorChain().size() - 1.
    *
    * @param position the position at which to remove an interceptor
    */
   void removeInterceptor(int position);

   /**
    * @return Retrieves a reference to the currently configured {@link org.jboss.cache.loader.CacheLoaderManager} if one or more cache loaders are configured, null otherwise.
    */
   CacheLoaderManager getCacheLoaderManager();

   /**
    * @return an instance of {@link BuddyManager} if buddy replication is enabled, null otherwise.
    */
   BuddyManager getBuddyManager();

   /**
    * @return the current {@link TransactionTable}
    */
   TransactionTable getTransactionTable();

   /**
    * Gets a handle of the RPC manager.
    *
    * @return the {@link org.jboss.cache.RPCManager} configured.
    */
   RPCManager getRPCManager();

   /**
    * @return the current {@link org.jboss.cache.statetransfer.StateTransferManager}
    */
   StateTransferManager getStateTransferManager();

   /**
    * @return the name of the cluster.  Null if running in local mode.
    */
   String getClusterName();

   /**
    * @return the number of attributes in the cache.
    */
   int getNumberOfAttributes();

   /**
    * @return the number of nodes in the cache.
    */
   int getNumberOfNodes();

   /**
    * Retrieves the current table of locks.
    *
    * @return lock table.
    */
   Map<Thread, List<NodeLock>> getLockTable();

   /**
    * @return the {@link org.jboss.cache.RegionManager}
    */
   RegionManager getRegionManager();

   /**
    * Returns the global transaction for this local transaction.
    * Optionally creates a new global transaction if it does not exist.
    *
    * @param tx                the current transaction
    * @param createIfNotExists if true creates a new transaction if none exists
    * @return a GlobalTransaction
    */
   GlobalTransaction getCurrentTransaction(Transaction tx, boolean createIfNotExists);

   /**
    * @return the notifier attached with this instance of the cache.  See {@link Notifier}, a class
    *         that is responsible for emitting notifications to registered CacheListeners.
    */
   Notifier getNotifier();

   /**
    * Returns a node without accessing the interceptor chain.  Does not return any nodes marked as invalid.
    *
    * @param fqn                 the Fqn to look up.
    * @param includeDeletedNodes if you intend to see nodes marked as deleted within the current tx, set this to true
    * @return a node if one exists or null
    */
   NodeSPI<K, V> peek(Fqn<?> fqn, boolean includeDeletedNodes);

   /**
    * Returns a node without accessing the interceptor chain, optionally returning nodes that are marked as invalid ({@link org.jboss.cache.Node#isValid()} == false).
    *
    * @param fqn                 the Fqn to look up.
    * @param includeDeletedNodes if you intend to see nodes marked as deleted within the current tx, set this to true
    * @param includeInvalidNodes if true, nodes marked as being invalid are also returned.
    * @return a node if one exists or null
    */
   NodeSPI<K, V> peek(Fqn<?> fqn, boolean includeDeletedNodes, boolean includeInvalidNodes);

   /**
    * Used with buddy replication's data gravitation interceptor.  If marshalling is necessary, ensure that the cache is
    * configured to use {@link org.jboss.cache.config.Configuration#useRegionBasedMarshalling} and the {@link org.jboss.cache.Region}
    * pertaining to the Fqn passed in is activated, and has an appropriate ClassLoader.
    *
    * @param fqn                       the fqn to gravitate
    * @param searchBuddyBackupSubtrees if true, buddy backup subtrees are searched and if false, they are not.
    * @return a GravitateResult which contains the data for the gravitation
    */
   GravitateResult gravitateData(Fqn<?> fqn, boolean searchBuddyBackupSubtrees);

   /**
    * Retrieves an instance of a {@link Marshaller}, which is capable of
    * converting Java objects to bytestreams and back in an efficient manner, which is
    * also interoperable with bytestreams produced/consumed by other versions of JBoss
    * Cache.
    * <p/>
    * The use of this marshaller is the <b>recommended</b> way of creating efficient,
    * compatible, byte streams from objects.
    *
    * @return an instance of {@link Marshaller}
    */
   Marshaller getMarshaller();
}
