/*
 * JBoss, Home of Professional Open Source
 *
 * Distributable under LGPL license.
 * See terms of license at gnu.org.
 */
package org.jboss.cache;

import org.jboss.cache.config.Configuration;
import org.jboss.cache.config.ConfigurationException;
import org.jboss.cache.factories.XmlConfigurationParser;

/**
 * Default (singleton) implementation of the {@link org.jboss.cache.CacheFactory} interface.
 * Use {@link #getInstance()} to obtain an instance.
 *
 * @author <a href="mailto:manik@jboss.org">Manik Surtani (manik@jboss.org)</a>
 */
public class DefaultCacheFactory<K, V> implements CacheFactory<K, V>
{
   private static CacheFactory<?, ?> singleton = new DefaultCacheFactory();

   /**
    * @return a singleton instance of this class.
    */
   @SuppressWarnings("unchecked")
   public static <K, V> CacheFactory<K, V> getInstance()
   {
      return (CacheFactory<K, V>) singleton;
   }

   public Cache<K, V> createCache() throws ConfigurationException
   {
      return createCache(true);
   }

   public Cache<K, V> createCache(boolean start) throws ConfigurationException
   {
      return createCache(new Configuration(), start);
   }

   public Cache<K, V> createCache(String configFileName) throws ConfigurationException
   {
      return createCache(configFileName, true);
   }

   public Cache<K, V> createCache(String configFileName, boolean start) throws ConfigurationException
   {
      XmlConfigurationParser parser = new XmlConfigurationParser();
      Configuration c = parser.parseFile(configFileName);
      return createCache(c, start);
   }

   /**
    * This implementation clones the configuration passed in before using it.
    *
    * @param configuration to use
    * @return a cache
    * @throws ConfigurationException if there are problems with the cfg
    */
   public Cache<K, V> createCache(Configuration configuration) throws ConfigurationException
   {
      return createCache(configuration, true);
   }

   /**
    * This implementation clones the configuration passed in before using it.
    *
    * @param configuration to use
    * @param start         whether to start the cache
    * @return a cache
    * @throws ConfigurationException if there are problems with the cfg
    */
   public Cache<K, V> createCache(Configuration configuration, boolean start) throws ConfigurationException
   {
      try
      {
         CacheImpl<K, V> cache = new CacheImpl<K, V>();
         cache.setConfiguration(configuration);
         if (start) cache.start();
         return cache;
      }
      catch (ConfigurationException ce)
      {
         throw ce;
      }
      catch (RuntimeException re)
      {
         throw re;
      }
      catch (Exception e)
      {
         throw new RuntimeException(e);
      }
   }
}
