/*
 * JBoss, Home of Professional Open Source
 *
 * Distributable under LGPL license.
 * See terms of license at gnu.org.
 */
package org.jboss.cache;

import org.jboss.cache.optimistic.TransactionWorkspace;
import org.jboss.cache.optimistic.WorkspaceNode;
import org.jboss.cache.optimistic.WorkspaceNodeImpl;

import java.util.Map;

/**
 * Generates new nodes based on the {@link CacheSPI} configuration.
 *
 * @author <a href="mailto:manik@jboss.org">Manik Surtani (manik@jboss.org)</a>
 */
public class NodeFactory<K, V>
{
   public enum NodeType
   {
      UNVERSIONED_NODE, VERSIONED_NODE, WORKSPACE_NODE
   }

   private CacheSPI<K, V> cache;
   private boolean optimistic;

   /**
    * Constructs an instance of the factory
    */
   public NodeFactory(CacheSPI<K, V> cache)
   {
      this.cache = cache;
      init();
   }

   /**
    * Initialises the node factory with the configuration from the cache.
    */
   public void init()
   {
      optimistic = cache.getConfiguration().isNodeLockingOptimistic();
   }


   /**
    * Creates a new {@link Node} instance.
    *
    * @param childName the new node's name
    * @param fqn       the new node's Fqn
    * @param parent    the new node's parent
    * @param data      the new node's attribute map
    * @param mapSafe   <code>true</code> if param <code>data</code> can safely
    *                  be directly assigned to the new node's data field;
    *                  <code>false</code> if param <code>data</code>'s contents
    *                  should be copied into the new node's data field.
    * @return the new node
    */
   public NodeSPI<K, V> createDataNode(Object childName, Fqn fqn, NodeSPI<K, V> parent, Map<K, V> data, boolean mapSafe)
   {
      NodeSPI<K, V> n = optimistic ? new VersionedNode<K, V>(fqn, parent, data, cache) : new UnversionedNode<K, V>(childName, fqn, data, mapSafe, cache);
      // always assume that new nodes do not have data loaded
      n.setDataLoaded(false);
      return n;
   }

   public Node<K, V> createNode(Object childName, Node<K, V> parent, Map<K, V> data)
   {
      return createNodeOfType(parent, childName, parent, data);
   }

   public Node<K, V> createNodeOfType(Node<K, V> template, Object childName, Node<K, V> parent, Map<K, V> data)
   {
      if (template instanceof WorkspaceNode)
      {
         NodeSPI<K, V> dataNodeParent = ((WorkspaceNode<K, V>) parent).getNode();
         TransactionWorkspace workspace = ((WorkspaceNode) template).getTransactionWorkspace();
         return createWorkspaceNode(dataNodeParent, workspace);
      }

      // not a workspace node.
      return createDataNode(childName, new Fqn(parent.getFqn(), childName), (NodeSPI<K, V>) parent, data, false);
   }

   public WorkspaceNode<K, V> createWorkspaceNode(NodeSPI<K, V> dataNode, TransactionWorkspace workspace)
   {
      return new WorkspaceNodeImpl<K, V>(dataNode, workspace);
   }

   public NodeSPI<K, V> createRootDataNode()
   {
      return createDataNode(null, Fqn.ROOT, null, null, false);
   }

}
