/*
 * JBoss, Home of Professional Open Source
 *
 * Distributable under LGPL license.
 * See terms of license at gnu.org.
 */
package org.jboss.cache.factories;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.jboss.cache.CacheImpl;
import org.jboss.cache.CacheSPI;
import org.jboss.cache.interceptors.*;
import org.jboss.cache.loader.CacheLoaderManager;
import org.jboss.cache.util.Util;

import java.util.ArrayList;
import java.util.List;

/**
 * Factory class that builds an interceptor chain based on CacheImpl config.
 *
 * @author <a href="mailto:manik@jboss.org">Manik Surtani (manik@jboss.org)</a>
 */
public class InterceptorChainFactory
{
   private Log log = LogFactory.getLog(InterceptorChainFactory.class);
   private static InterceptorChainFactory singleton;

   private InterceptorChainFactory()
   {
      // private ctor
   }

   public static InterceptorChainFactory getInstance()
   {
      if (singleton == null)
      {
         synchronized (InterceptorChainFactory.class)
         {
            if (singleton == null) singleton = new InterceptorChainFactory();
         }
      }

      return singleton;
   }


   public Interceptor buildInterceptorChain(CacheImpl cache) throws IllegalAccessException, ClassNotFoundException, InstantiationException
   {
      if (cache.getConfiguration().isNodeLockingOptimistic())
      {
         return createOptimisticInterceptorChain(cache);
      }
      else
      {
         return createPessimisticInterceptorChain(cache);
      }
   }

   public Interceptor setLastInterceptorPointer(Interceptor first, Interceptor last)
   {
      Interceptor i = first;
      while (i != null)
      {
         i.setLast(last);
         i = i.getNext();
      }
      return first;
   }


   private Interceptor createInterceptor(String classname, CacheSPI cache) throws ClassNotFoundException, IllegalAccessException, InstantiationException
   {
      Class<Interceptor> clazz = Util.loadClass(classname);
      return createInterceptor(clazz, cache);
   }

   private Interceptor createInterceptor(Class<? extends Interceptor> clazz, CacheSPI cache) throws IllegalAccessException, InstantiationException
   {
      Interceptor i = clazz.newInstance();
      i.setCache(cache);
      i.setStatisticsEnabled(cache.getConfiguration().getExposeManagementStatistics());
      return i;
   }


   /**
    * Adds an interceptor at the end of the chain
    */
   private void addInterceptor(Interceptor first, Interceptor i)
   {
      if (first == null) return;
      while (first.getNext() != null) first = first.getNext();
      first.setNext(i);
   }

   private Interceptor createPessimisticInterceptorChain(CacheImpl cache) throws IllegalAccessException, InstantiationException, ClassNotFoundException
   {
      Interceptor call_interceptor;
      Interceptor lock_interceptor;
      Interceptor repl_interceptor = null;
      Interceptor cache_loader_interceptor = null;
      Interceptor cache_store_interceptor = null;
      Interceptor unlock_interceptor;
      Interceptor passivation_interceptor = null;
      Interceptor activation_interceptor = null;
      Interceptor cacheMgmtInterceptor;
      Interceptor txInterceptor;
      Interceptor eviction_interceptor;
      Interceptor dataGravitatorInterceptor = null;
      Interceptor invocationCtxInterceptor = createInterceptor(InvocationContextInterceptor.class, cache);
      Interceptor notificationInterceptor = createInterceptor(NotificationInterceptor.class, cache);
      Interceptor first = null;


      call_interceptor = createInterceptor(CallInterceptor.class, cache);
      ((CallInterceptor) call_interceptor).setTreeCacheInstance(cache);

      if (cache.getBuddyManager() != null)
      {
         dataGravitatorInterceptor = createInterceptor(DataGravitatorInterceptor.class, cache);
      }

      lock_interceptor = createInterceptor(PessimisticLockInterceptor.class, cache);

      unlock_interceptor = createInterceptor(UnlockInterceptor.class, cache);

      cacheMgmtInterceptor = createInterceptor(CacheMgmtInterceptor.class, cache);

      txInterceptor = createInterceptor(TxInterceptor.class, cache);

      switch (cache.getConfiguration().getCacheMode())
      {
         case REPL_SYNC:
         case REPL_ASYNC:
            repl_interceptor = createInterceptor(ReplicationInterceptor.class, cache);
            break;
         case INVALIDATION_SYNC:
         case INVALIDATION_ASYNC:
            repl_interceptor = createInterceptor(InvalidationInterceptor.class, cache);
            break;
         case LOCAL:
            //Nothing...
      }

      CacheLoaderManager cacheLoaderMgr = cache.getCacheLoaderManager();

      if (cacheLoaderMgr != null && cacheLoaderMgr.getCacheLoader() != null)
      {
         if (cacheLoaderMgr.isPassivation())
         {
            activation_interceptor = createInterceptor(ActivationInterceptor.class, cache);
            passivation_interceptor = createInterceptor(PassivationInterceptor.class, cache);
         }
         else
         {
            cache_loader_interceptor = createInterceptor(CacheLoaderInterceptor.class, cache);
            cache_store_interceptor = createInterceptor(CacheStoreInterceptor.class, cache);
         }
      }

      // load the icInterceptor first
      if (first == null) first = invocationCtxInterceptor;

      // load the cache management interceptor next
      if (cache.getConfiguration().getExposeManagementStatistics())
      {
         if (first == null)
         {
            first = cacheMgmtInterceptor;
         }
         else
         {
            addInterceptor(first, cacheMgmtInterceptor);
         }
      }

      // load the tx interceptor
      if (first == null)
      {
         first = txInterceptor;
      }
      else
      {
         addInterceptor(first, txInterceptor);
      }

      if (first == null)
         first = notificationInterceptor;
      else
         addInterceptor(first, notificationInterceptor);

      // create the stack from the bottom up
//      if (activation_interceptor != null)
//      {
//         if (!cacheLoaderMgr.isFetchPersistentState())
//         {
//            if (first == null)
//            {
//               first = passivation_interceptor;
//            }
//            else
//            {
//               addInterceptor(first, passivation_interceptor);
//            }
//         }
//      }
//
//      if (cache_loader_interceptor != null)
//      {
//         if (!cacheLoaderMgr.isFetchPersistentState())
//         {
//            if (first == null)
//            {
//               first = cache_store_interceptor;
//            }
//            else
//            {
//               addInterceptor(first, cache_store_interceptor);
//            }
//         }
//      }

      if (repl_interceptor != null)
      {
         if (first == null)
         {
            first = repl_interceptor;
         }
         else
         {
            addInterceptor(first, repl_interceptor);
         }
      }

      if (first == null)
      {
         first = lock_interceptor;
      }
      else
      {
         addInterceptor(first, lock_interceptor);
      }


      if (unlock_interceptor != null)
      {
         if (first == null)
         {
            first = unlock_interceptor;
         }
         else
         {
            addInterceptor(first, unlock_interceptor);
         }
      }

      if (activation_interceptor != null)
      {
//         if (!cacheLoaderMgr.isFetchPersistentState())
         if (false)
         {
            if (first == null)
            {
               first = activation_interceptor;
            }
            else
            {
               addInterceptor(first, activation_interceptor);
            }
         }
         else
         {
            if (first == null)
            {
               first = activation_interceptor;
            }
            else
            {
               addInterceptor(first, activation_interceptor);
            }
            if (first == null)
            {
               first = passivation_interceptor;
            }
            else
            {
               addInterceptor(first, passivation_interceptor);
            }
         }
      }

      if (cache_loader_interceptor != null)
      {
//         if (!cacheLoaderMgr.isFetchPersistentState())
         if (false)
         {
            if (first == null)
            {
               first = cache_loader_interceptor;
            }
            else
            {
               addInterceptor(first, cache_loader_interceptor);
            }
         }
         else
         {
            if (first == null)
            {
               first = cache_loader_interceptor;
            }
            else
            {
               addInterceptor(first, cache_loader_interceptor);
            }
            if (first == null)
            {
               first = cache_store_interceptor;
            }
            else
            {
               addInterceptor(first, cache_store_interceptor);
            }
         }
      }

      if (dataGravitatorInterceptor != null)
      {
         if (first == null)
         {
            first = dataGravitatorInterceptor;
         }
         else
         {
            addInterceptor(first, dataGravitatorInterceptor);
         }
      }
//
//      if (first == null)
//      {
//         first = lock_interceptor;
//      }
//      else
//      {
//         addInterceptor(first, lock_interceptor);
//      }

      if (cache.getConfiguration().getEvictionConfig() != null && cache.getConfiguration().getEvictionConfig().isValidConfig())
      {
         eviction_interceptor = createInterceptor(cache.getEvictionInterceptorClass(), cache);
         if (first == null)
         {
            first = eviction_interceptor;
         }
         else
         {
            addInterceptor(first, eviction_interceptor);
         }
      }

      if (first == null)
      {
         first = call_interceptor;
      }
      else
      {
         addInterceptor(first, call_interceptor);
      }

      if (log.isDebugEnabled())
      {
         log.debug("interceptor chain is:\n" + printInterceptorChain(first));
      }

      return setLastInterceptorPointer(first, call_interceptor);
   }

   private Interceptor createOptimisticInterceptorChain(CacheImpl cache) throws IllegalAccessException, InstantiationException, ClassNotFoundException
   {
      Interceptor txInterceptor, replicationInterceptor = null, lockInterceptor, validationInterceptor;
      Interceptor createIfNotExistsInterceptor, nodeInterceptor, invokerInterceptor, activationInterceptor = null;
      Interceptor passivationInterceptor = null, cacheLoaderInterceptor = null, cacheStoreInterceptor = null, first = null;
      Interceptor cacheMgmtInterceptor, evictionInterceptor = null, dataGravitatorInterceptor = null;
      Interceptor invocationCtxInterceptor = createInterceptor(InvocationContextInterceptor.class, cache);
      Interceptor notificationInterceptor = createInterceptor(NotificationInterceptor.class, cache);

      CacheLoaderManager cacheLoaderManager = cache.getCacheLoaderManager();
      if (cacheLoaderManager != null && cacheLoaderManager.getCacheLoader() != null)
      {
         if (cacheLoaderManager.isPassivation())
         {
            activationInterceptor = createInterceptor(ActivationInterceptor.class, cache);
            passivationInterceptor = createInterceptor(PassivationInterceptor.class, cache);
         }
         else
         {
            cacheLoaderInterceptor = createInterceptor(CacheLoaderInterceptor.class, cache);
            cacheStoreInterceptor = createInterceptor(CacheStoreInterceptor.class, cache);
         }
      }

      txInterceptor = createInterceptor(TxInterceptor.class, cache);

      if (cache.getBuddyManager() != null)
      {
         dataGravitatorInterceptor = createInterceptor(DataGravitatorInterceptor.class, cache);
      }

      switch (cache.getConfiguration().getCacheMode())
      {
         case REPL_SYNC:
         case REPL_ASYNC:
            replicationInterceptor = createInterceptor(OptimisticReplicationInterceptor.class, cache);
            break;
         case INVALIDATION_SYNC:
         case INVALIDATION_ASYNC:
            replicationInterceptor = createInterceptor(InvalidationInterceptor.class, cache);
            break;
         case LOCAL:
            //Nothing...
      }

      lockInterceptor = createInterceptor(OptimisticLockingInterceptor.class, cache);

      validationInterceptor = createInterceptor(OptimisticValidatorInterceptor.class, cache);

      createIfNotExistsInterceptor = createInterceptor(OptimisticCreateIfNotExistsInterceptor.class, cache);

      nodeInterceptor = createInterceptor(OptimisticNodeInterceptor.class, cache);

      invokerInterceptor = createInterceptor(CallInterceptor.class, cache);
      ((CallInterceptor) invokerInterceptor).setTreeCacheInstance(cache);

      if (cache.getConfiguration().getEvictionConfig() != null && cache.getConfiguration().getEvictionConfig().isValidConfig())
      {
         evictionInterceptor = createInterceptor(cache.getEvictionInterceptorClass(), cache);
      }

      if (first == null) first = invocationCtxInterceptor;

      if (cache.getConfiguration().getExposeManagementStatistics())
      {
         cacheMgmtInterceptor = createInterceptor(CacheMgmtInterceptor.class, cache);
         if (first == null)
         {
            first = cacheMgmtInterceptor;
         }
         else
         {
            addInterceptor(first, cacheMgmtInterceptor);
         }
      }

      if (txInterceptor != null)
      {
         if (first == null)
         {
            first = txInterceptor;
         }
         else
         {
            addInterceptor(first, txInterceptor);
         }
      }

      if (first == null)
         first = notificationInterceptor;
      else
         addInterceptor(first, notificationInterceptor);

      if (first == null)
      {
         first = replicationInterceptor;
      }
      else
      {
         addInterceptor(first, replicationInterceptor);
      }

      if (passivationInterceptor != null && !cacheLoaderManager.isFetchPersistentState())
      {
         if (first == null)
         {
            first = passivationInterceptor;
         }
         else
         {
            addInterceptor(first, passivationInterceptor);
         }
      }

      // add the cache store interceptor here
      if (cacheStoreInterceptor != null && !cacheLoaderManager.isFetchPersistentState())
      {
         if (first == null)
         {
            first = cacheStoreInterceptor;
         }
         else
         {
            addInterceptor(first, cacheStoreInterceptor);
         }
      }

      // cache loader interceptor is only invoked if we are ready to write to the actual tree cache
      if (activationInterceptor != null)
      {
         if (first == null)
         {
            first = activationInterceptor;
         }
         else
         {
            addInterceptor(first, activationInterceptor);
         }

         if (cacheLoaderManager.isFetchPersistentState())
         {
            if (first == null)
            {
               first = passivationInterceptor;
            }
            else
            {
               addInterceptor(first, passivationInterceptor);
            }
         }
      }

      if (cacheLoaderInterceptor != null)
      {
         if (first == null)
         {
            first = cacheLoaderInterceptor;
         }
         else
         {
            addInterceptor(first, cacheLoaderInterceptor);
         }

         if (cacheLoaderManager.isFetchPersistentState())
         {
            if (first == null)
            {
               first = cacheStoreInterceptor;
            }
            else
            {
               addInterceptor(first, cacheStoreInterceptor);
            }
         }
      }

      if (dataGravitatorInterceptor != null)
      {
         if (first == null)
         {
            first = dataGravitatorInterceptor;
         }
         else
         {
            addInterceptor(first, dataGravitatorInterceptor);
         }
      }


      if (first == null)
      {
         first = lockInterceptor;
      }
      else
      {
         addInterceptor(first, lockInterceptor);
      }

      if (first == null)
      {
         first = validationInterceptor;
      }
      else
      {
         addInterceptor(first, validationInterceptor);
      }

      if (first == null)
      {
         first = createIfNotExistsInterceptor;
      }
      else
      {
         addInterceptor(first, createIfNotExistsInterceptor);
      }

      // eviction interceptor to come before the optimistic node interceptor
      if (first == null)
      {
         first = evictionInterceptor;
      }
      else
      {
         addInterceptor(first, evictionInterceptor);
      }

      if (first == null)
      {
         first = nodeInterceptor;
      }
      else
      {
         addInterceptor(first, nodeInterceptor);
      }


      if (first == null)
      {
         first = invokerInterceptor;
      }
      else
      {
         addInterceptor(first, invokerInterceptor);
      }

      if (log.isInfoEnabled())
      {
         log.info("interceptor chain is:\n" + printInterceptorChain(first));
      }

      return setLastInterceptorPointer(first, invokerInterceptor);
   }

   public String printInterceptorChain(Interceptor i)
   {
      StringBuffer sb = new StringBuffer();
      if (i != null)
      {
         if (i.getNext() != null)
         {
            sb.append(printInterceptorChain(i.getNext())).append("\n");
         }
         sb.append(i.getClass());
      }
      return sb.toString();
   }

   public List<Interceptor> asList(Interceptor interceptor)
   {
      if (interceptor == null)
      {
         return null;
      }
      int num = 1;
      Interceptor tmp = interceptor;
      while ((tmp = tmp.getNext()) != null)
      {
         num++;
      }
      List<Interceptor> retval = new ArrayList<Interceptor>(num);
      tmp = interceptor;
      num = 0;
      do
      {
         retval.add(tmp);
         tmp = tmp.getNext();
      }
      while (tmp != null);
      return retval;
   }

   /**
    * "Fixes" the next() and last() pointers for each interceptor, based on the order presented in the list passed in
    *
    * @param interceptors
    * @return the first interceptor in the chain.
    */
   public Interceptor correctInterceptorChaining(List<Interceptor> interceptors)
   {
      Interceptor first = null, last = null;

      for (Interceptor next : interceptors)
      {
         if (first == null)
         {
            first = last = next;
            continue;
         }
         last.setNext(next);
         last = next;
      }

      if (last != null) last.setNext(null);

      // now set the 'last' pointer.
      return setLastInterceptorPointer(first, last);
   }

   /**
    * Initialises the interceptors with a CacheSPI instance.  Essentially calls setCache().  This is done AFTER the interceptor chain
    * is created, and when Cache.start() is called, so the interceptors get a fully initialised cache.
    *
    * @param firstInterceptor interceptor chain
    * @param cache            cache instance
    */
   public void initialiseInterceptors(Interceptor firstInterceptor, CacheSPI cache)
   {
      for (Interceptor i : asList(firstInterceptor)) i.setCache(cache);
   }
}
