package org.jboss.cache.interceptors;

import org.jboss.cache.Fqn;
import org.jboss.cache.InvocationContext;
import org.jboss.cache.Modification;
import org.jboss.cache.NodeSPI;
import org.jboss.cache.marshall.MethodCall;
import org.jboss.cache.marshall.MethodDeclarations;
import org.jboss.cache.transaction.GlobalTransaction;
import org.jboss.cache.transaction.TransactionEntry;
import org.jboss.cache.transaction.TransactionTable;

import javax.transaction.TransactionManager;
import java.lang.reflect.Method;
import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.concurrent.ConcurrentHashMap;

/**
 * Loads nodes that don't exist at the time of the call into memory from the CacheLoader.
 * If the nodes were evicted earlier then we remove them from the cache loader after
 * their attributes have been initialized and their children have been loaded in memory.
 *
 * @author <a href="mailto:{hmesha@novell.com}">{Hany Mesha}</a>
 * @version $Id: ActivationInterceptor.java 4149 2007-07-17 02:03:26Z msurtani $
 */
public class ActivationInterceptor extends CacheLoaderInterceptor implements ActivationInterceptorMBean
{

   protected TransactionManager tx_mgr = null;
   protected TransactionTable tx_table = null;
   private HashMap m_txActivations = new HashMap();
   private long m_activations = 0;

   /**
    * List<Transaction> that we have registered for
    */
   protected ConcurrentHashMap transactions = new ConcurrentHashMap(16);
   protected static final Object NULL = new Object();

   public ActivationInterceptor()
   {
      this.useCacheStore = false;
      isActivation = true;
   }

   /**
    * Makes sure a node is loaded into memory before a call executes. If node is
    * already loaded and its attributes already initialized, then remove it from
    * the cache loader and notify the cache listeners that the node has been activated.
    *
    * @return
    * @throws Throwable
    */
   public Object invoke(InvocationContext ctx) throws Throwable
   {
      MethodCall m = ctx.getMethodCall();
      Fqn fqn = null;
      Object[] args = m.getArgs();
      Object retval;

      // First call the parent class to load the node
      retval = super.invoke(ctx);

      // is this a node removal operation?
      boolean removeData = false, nodeRemoved = false;

      // Could be TRANSACTIONAL. If so, we register for TX completion (if we haven't done so yet)
      if (tx_mgr != null && tx_mgr.getTransaction() != null)
      {
         GlobalTransaction gtx = ctx.getGlobalTransaction();
         switch (m.getMethodId())
         {
            case MethodDeclarations.commitMethod_id:
               if (hasModifications(args))
               {
                  loader.commit(gtx);
                  if (configuration.getExposeManagementStatistics() && getStatisticsEnabled())
                  {
                     Integer acts = (Integer) m_txActivations.get(gtx);
                     if (acts != null)
                     {
                        m_activations = m_activations + acts;
                     }
                     m_txActivations.remove(gtx);
                  }
               }
               break;
            case MethodDeclarations.rollbackMethod_id:
               if (hasModifications(args))
               {
                  loader.rollback(gtx);
                  if (configuration.getExposeManagementStatistics() && getStatisticsEnabled())
                  {
                     m_txActivations.remove(gtx);
                  }
               }
               break;
            case MethodDeclarations.optimisticPrepareMethod_id:
            case MethodDeclarations.prepareMethod_id:
               prepareCacheLoader(ctx);
               break;
         }
      }

      // CacheLoaderInterceptor normally doesn't load the node
      // since CacheStoreInterceptor.put() returns the old value
      switch (m.getMethodId())
      {
         case MethodDeclarations.putForExternalReadMethodLocal_id:
         case MethodDeclarations.putDataMethodLocal_id:
         case MethodDeclarations.putDataEraseMethodLocal_id:
         case MethodDeclarations.putKeyValMethodLocal_id:
         case MethodDeclarations.removeKeyMethodLocal_id:
         case MethodDeclarations.addChildMethodLocal_id:
            fqn = (Fqn) args[1];
            break;
         case MethodDeclarations.getKeyValueMethodLocal_id:
         case MethodDeclarations.getNodeMethodLocal_id:
         case MethodDeclarations.getKeysMethodLocal_id:
         case MethodDeclarations.getChildrenNamesMethodLocal_id:
         case MethodDeclarations.releaseAllLocksMethodLocal_id:
         case MethodDeclarations.printMethodLocal_id:
            fqn = (Fqn) args[0];
            break;
         case MethodDeclarations.removeNodeMethodLocal_id:
            nodeRemoved = true;
            fqn = (Fqn) args[1];
            break;
         case MethodDeclarations.removeDataMethodLocal_id:
            removeData = true;
            fqn = (Fqn) args[1];
            break;
      }

      NodeSPI n;
      if (fqn != null)
      {
         if (nodeRemoved)
         {
            log.trace("This is a remove operation; removing the node from the loader, no activation processing needed.");
            loader.remove(fqn);
         }
         else if (removeData)
         {
            log.trace("This is a remove data operation; removing the data from the loader, no activation processing needed.");
            loader.removeData(fqn);
         }
         else if (((n = getNode(fqn)) != null) && n.isDataLoaded() && loader.exists(fqn))
         {
            // Remove the node from the cache loader if it exists in memory,
            // its attributes have been initialized, its children have been loaded,
            // AND it was found in the cache loader (nodeLoaded = true).
            // Then notify the listeners that the node has been activated.

            // node not null and attributes have been loaded?
            if (!n.getChildrenDirect().isEmpty())
            {
               if (allInitialized(n))
               {
                  log.debug("children all initialized");
                  remove(ctx, fqn);
               }
            }
            else if (loaderNoChildren(fqn))
            {
               if (log.isDebugEnabled()) log.debug("no children " + n);
               remove(ctx, fqn);
            }
         }
      }
      return retval;
   }

   private void remove(InvocationContext ctx, Fqn fqn) throws Exception
   {
      cache.getNotifier().notifyNodeActivated(fqn, true, Collections.emptyMap(), ctx);
      loader.remove(fqn);
      if (configuration.getExposeManagementStatistics() && getStatisticsEnabled())
      {
         m_activations++;
      }
   }

   /**
    * Returns true if a node has all children loaded and initialized.
    */
   private boolean allInitialized(NodeSPI<?, ?> n)
   {
      if (!n.isChildrenLoaded())
      {
         return false;
      }

      for (NodeSPI child : n.getChildrenDirect())
      {
         if (!child.isDataLoaded())
         {
            return false;
         }
      }
      return true;
   }

   /**
    * Returns true if the loader indicates no children for this node.
    * Return false on error.
    */
   private boolean loaderNoChildren(Fqn fqn)
   {
      try
      {
         Set children_names = loader.getChildrenNames(fqn);
         return (children_names == null);
      }
      catch (Exception e)
      {
         log.error("failed getting the children names for " + fqn + " from the cache loader", e);
         return false;
      }
   }

   public long getActivations()
   {
      return m_activations;
   }

   public void resetStatistics()
   {
      super.resetStatistics();
      m_activations = 0;
   }

   public Map<String, Object> dumpStatistics()
   {
      Map<String, Object> retval = super.dumpStatistics();
      if (retval == null)
      {
         retval = new HashMap<String, Object>();
      }
      retval.put("Activations", m_activations);
      return retval;
   }

   protected boolean hasModifications(Object[] args)
   {
      int hint = 1;
      if (args[hint] instanceof Boolean) return (Boolean) args[hint];
      for (Object arg : args)
      {
         if (arg instanceof Boolean) return (Boolean) arg;
      }
      return false;
   }

   private void prepareCacheLoader(InvocationContext ctx) throws Exception
   {
      List<MethodCall> modifications;
      TransactionEntry entry;
      int txActs = 0;
      GlobalTransaction gtx = ctx.getGlobalTransaction();

      entry = tx_table.get(gtx);
      if (entry == null)
      {
         throw new Exception("entry for transaction " + gtx + " not found in transaction table");
      }
      modifications = entry.getCacheLoaderModifications();
      if (modifications.size() == 0)
      {
         return;
      }
      List cache_loader_modifications = new ArrayList();
      for (MethodCall methodCall : modifications)
      {
         Method method = methodCall.getMethod();
         Object[] args;
         if (method == null)
         {
            throw new Exception("method call has no method: " + methodCall);
         }
         args = methodCall.getArgs();
         switch (methodCall.getMethodId())
         {
            case MethodDeclarations.removeNodeMethodLocal_id:
               // just remove it from loader, don't trigger activation processing
               Modification mod = new Modification(Modification.ModificationType.REMOVE_NODE, (Fqn) args[1]);
               cache_loader_modifications.add(mod);
               break;
            case MethodDeclarations.putDataMethodLocal_id:
            case MethodDeclarations.putDataEraseMethodLocal_id:
            case MethodDeclarations.putKeyValMethodLocal_id:
               // On the way out, remove the node from the cache loader.
               // Only remove the node if it exists in memory, its attributes have
               // been initialized, its children have been loaded
               // AND it was found in the cache loader (nodeLoaded = true).
               // Then notify the listeners that the node has been activated.
               Fqn fqn = (Fqn) args[1];
               if (fqn != null && cache.peek(fqn, false) != null && loader.exists(fqn))
               {
                  NodeSPI n = getNode(fqn);// don't load
                  // node not null and attributes have been loaded?
                  if (n != null && n.isDataLoaded())
                  {
                     // has children?
                     if (!n.getChildrenDirect().isEmpty() && allInitialized(n))
                     {
                        // children have been loaded, remove the node
                        addRemoveMod(ctx, cache_loader_modifications, fqn, n.getDataDirect());
                        txActs++;
                     }
                     // doesn't have children, check the cache loader
                     else if (loaderNoChildren(fqn))
                     {
                        addRemoveMod(ctx, cache_loader_modifications, fqn, n.getDataDirect());
                        txActs++;
                     }
                  }
               }
               break;
         }
      }
      if (cache_loader_modifications.size() > 0)
      {
         loader.prepare(gtx, cache_loader_modifications, false);
         if (configuration.getExposeManagementStatistics() && getStatisticsEnabled() && txActs > 0)
         {
            m_txActivations.put(gtx, txActs);
         }
      }
   }

   private void addRemoveMod(InvocationContext ctx, List l, Fqn fqn, Map data)
   {
      Modification mod = new Modification(Modification.ModificationType.REMOVE_NODE, fqn);
      l.add(mod);
      cache.getNotifier().notifyNodeActivated(fqn, false, data, ctx);
   }

}
