/**
 *
 */
package org.jboss.cache.interceptors;

import org.jboss.cache.CacheSPI;
import org.jboss.cache.InvocationContext;
import org.jboss.cache.buddyreplication.BuddyManager;
import org.jboss.cache.config.Configuration.CacheMode;
import org.jboss.cache.config.Option;
import org.jboss.cache.marshall.MethodCall;
import org.jboss.cache.marshall.MethodCallFactory;
import org.jboss.cache.marshall.MethodDeclarations;
import org.jboss.cache.transaction.GlobalTransaction;
import org.jboss.cache.transaction.TransactionEntry;
import org.jgroups.Address;

import javax.transaction.Transaction;
import java.util.List;

/**
 * Acts as a base for all RPC calls - subclassed by
 * {@link ReplicationInterceptor} and {@link OptimisticReplicationInterceptor}.
 *
 * @author <a href="mailto:manik@jboss.org">Manik Surtani (manik@jboss.org)</a>
 */
public abstract class BaseRpcInterceptor extends Interceptor
{

   private BuddyManager buddyManager;
   private boolean usingBuddyReplication;
   protected boolean defaultSynchronous;

   public void setCache(CacheSPI cache)
   {
      super.setCache(cache);
      buddyManager = cache.getBuddyManager();
      usingBuddyReplication = buddyManager != null;
      CacheMode mode = cache.getConfiguration().getCacheMode();
      defaultSynchronous = (mode == CacheMode.REPL_SYNC || mode == CacheMode.INVALIDATION_SYNC);
   }

   /**
    * Checks whether any of the responses are exceptions. If yes, re-throws
    * them (as exceptions or runtime exceptions).
    *
    * @param rsps
    * @throws Throwable
    */
   protected void checkResponses(List rsps) throws Throwable
   {
      if (rsps != null)
      {
         for (Object rsp : rsps)
         {
            if (rsp != null && rsp instanceof Throwable)
            {
               // lets print a stack trace first.
               if (log.isDebugEnabled())
                  log.debug("Received Throwable from remote node", (Throwable) rsp);
               throw (Throwable) rsp;
            }
         }
      }
   }

   protected void replicateCall(MethodCall call, boolean sync, Option o) throws Throwable
   {
      replicateCall(null, call, sync, o);
   }

   protected void replicateCall(List<Address> recipients, MethodCall call, boolean sync, Option o) throws Throwable
   {

      if (log.isTraceEnabled()) log.trace("Broadcasting call " + call + " to recipient list " + recipients);
      Transaction tx = null;
      if (cache.getTransactionManager() != null && (tx = cache.getTransactionManager().getTransaction()) != null)
      {
         GlobalTransaction gtx = cache.getTransactionTable().get(tx);
         TransactionEntry te = cache.getTransactionTable().get(gtx);
         if (te != null)
         {
            if (te.isForceAsyncReplication()) sync = false;
            else if (te.isForceSyncReplication()) sync = true;
         }
      }
      if (!sync && cache.getRPCManager().getReplicationQueue() != null && !usingBuddyReplication)
      {
         putCallOnAsyncReplicationQueue(call);
      }
      else
      {
         if (usingBuddyReplication) call = buddyManager.transformFqns(call);

         List<Address> callRecipients = recipients;
         if (callRecipients == null)
         {
            callRecipients = usingBuddyReplication ? buddyManager.getBuddyAddresses() : cache.getMembers();
            if (log.isTraceEnabled())
               log.trace("Setting call recipients to " + callRecipients + " since the original list of recipients passed in is null.");
         }

         long syncReplTimeout = o.getSyncReplTimeout();
         if (syncReplTimeout < 0) syncReplTimeout = configuration.getSyncReplTimeout();

         List rsps = cache.getRPCManager().callRemoteMethods(callRecipients,
                 MethodDeclarations.replicateMethod,
                 new Object[]{call},
                 sync, // is synchronised?
                 true, // ignore self?
                 syncReplTimeout);
         if (log.isTraceEnabled())
         {
            log.trace("responses=" + rsps);
         }
         if (sync) checkResponses(rsps);
      }

   }

   protected void putCallOnAsyncReplicationQueue(MethodCall call)
   {
      if (log.isDebugEnabled()) log.debug("Putting call " + call + " on the replication queue.");
      cache.getRPCManager().getReplicationQueue().add(MethodCallFactory.create(MethodDeclarations.replicateMethod, call));
   }

   protected boolean containsModifications(InvocationContext ctx)
   {
      switch (ctx.getMethodCall().getMethodId())
      {
         case MethodDeclarations.prepareMethod_id:
         case MethodDeclarations.optimisticPrepareMethod_id:
            List mods = (List) ctx.getMethodCall().getArgs()[1];
            return mods.size() > 0;
         case MethodDeclarations.commitMethod_id:
         case MethodDeclarations.rollbackMethod_id:
            return ctx.isTxHasMods();
         default:
            return false;
      }
   }

   protected boolean isSynchronous(Option option)
   {
      if (option != null)
      {
         if (option.isForceSynchronous())
            return true;
         else if (option.isForceAsynchronous())
            return false;
      }
      return defaultSynchronous;
   }
}