/*
 * JBoss, Home of Professional Open Source
 * Copyright 2005, JBoss Inc., and individual contributors as indicated
 * by the @authors tag. See the copyright.txt in the distribution for a
 * full listing of individual contributors.
 *
 * This is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as
 * published by the Free Software Foundation; either version 2.1 of
 * the License, or (at your option) any later version.
 *
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this software; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA, or see the FSF site: http://www.fsf.org.
 */
package org.jboss.cache.interceptors;

import org.jboss.cache.CacheSPI;
import org.jboss.cache.InvocationContext;
import org.jboss.cache.marshall.MethodCall;
import org.jboss.cache.marshall.MethodDeclarations;

import java.util.HashMap;
import java.util.Map;

/**
 * Captures cache management statistics
 *
 * @author Jerry Gauthier
 * @version $Id: CacheMgmtInterceptor.java 3973 2007-06-06 12:43:50Z msurtani $
 */
public class CacheMgmtInterceptor
      extends Interceptor
      implements CacheMgmtInterceptorMBean
{
   private long m_hit_times = 0;
   private long m_miss_times = 0;
   private long m_store_times = 0;
   private long m_hits = 0;
   private long m_misses = 0;
   private long m_stores = 0;
   private long m_evictions = 0;
   private long m_start = System.currentTimeMillis();
   private long m_reset = m_start;

   public void setCache(CacheSPI cache)
   {
      super.setCache(cache);
   }

   /**
    * Pass the method on and capture cache statistics
    *
    * @return
    * @throws Throwable
    */
   public Object invoke(InvocationContext ctx) throws Throwable
   {
      MethodCall m = ctx.getMethodCall();
      Map attributes;
      Object[] args = m.getArgs();
      Object retval;

      // if statistics not enabled, execute the method and return
      if (!getStatisticsEnabled())
         return super.invoke(ctx);

      long t1, t2;
      switch (m.getMethodId())
      {
         case MethodDeclarations.getKeyValueMethodLocal_id:
            //fqn = (Fqn) args[0];
            //key = args[1];
            t1 = System.currentTimeMillis();
            retval = super.invoke(ctx);
            t2 = System.currentTimeMillis();
            if (retval == null)
            {
               m_miss_times = m_miss_times + (t2 - t1);
               m_misses++;
            }
            else
            {
               m_hit_times = m_hit_times + (t2 - t1);
               m_hits++;
            }
            break;
         case MethodDeclarations.putForExternalReadMethodLocal_id:
         case MethodDeclarations.putKeyValMethodLocal_id:
            t1 = System.currentTimeMillis();
            retval = super.invoke(ctx);
            t2 = System.currentTimeMillis();
            m_store_times = m_store_times + (t2 - t1);
            m_stores++;
            break;
         case MethodDeclarations.putDataMethodLocal_id:
         case MethodDeclarations.putDataEraseMethodLocal_id:
            //fqn = (Fqn) args[1];
            attributes = (Map) args[2];
            t1 = System.currentTimeMillis();
            retval = super.invoke(ctx);
            t2 = System.currentTimeMillis();

            if (attributes != null && attributes.size() > 0)
            {
               m_store_times = m_store_times + (t2 - t1);
               m_stores = m_stores + attributes.size();
            }
            break;
         case MethodDeclarations.evictNodeMethodLocal_id:
         case MethodDeclarations.evictVersionedNodeMethodLocal_id:
            //fqn = (Fqn) args[0];
            retval = super.invoke(ctx);
            m_evictions++;
            break;
         default:
            retval = super.invoke(ctx);
            break;
      }

      return retval;
   }

   public long getHits()
   {
      return m_hits;
   }

   public long getMisses()
   {
      return m_misses;
   }

   public long getStores()
   {
      return m_stores;
   }

   public long getEvictions()
   {
      return m_evictions;
   }

   public double getHitMissRatio()
   {
      double total = m_hits + m_misses;
      if (total == 0)
         return 0;
      return (m_hits / total);
   }

   public double getReadWriteRatio()
   {
      if (m_stores == 0)
         return 0;
      return (((double) (m_hits + m_misses) / (double) m_stores));
   }

   public long getAverageReadTime()
   {
      long total = m_hits + m_misses;
      if (total == 0)
         return 0;
      return (m_hit_times + m_miss_times) / total;
   }

   public long getAverageWriteTime()
   {
      if (m_stores == 0)
         return 0;
      return (m_store_times) / m_stores;
   }

   public int getNumberOfAttributes()
   {
      return cache.getNumberOfAttributes();
   }

   public int getNumberOfNodes()
   {
      return cache.getNumberOfNodes();
   }

   public long getElapsedTime()
   {
      return (System.currentTimeMillis() - m_start) / 1000;
   }

   public long getTimeSinceReset()
   {
      return (System.currentTimeMillis() - m_reset) / 1000;
   }

   public Map<String, Object> dumpStatistics()
   {
      Map<String, Object> retval = new HashMap<String, Object>();
      retval.put("Hits", m_hits);
      retval.put("Misses", m_misses);
      retval.put("Stores", m_stores);
      retval.put("Evictions", m_evictions);
      retval.put("NumberOfAttributes", cache.getNumberOfAttributes());
      retval.put("NumberOfNodes", cache.getNumberOfNodes());
      retval.put("ElapsedTime", getElapsedTime());
      retval.put("TimeSinceReset", getTimeSinceReset());
      retval.put("AverageReadTime", getAverageReadTime());
      retval.put("AverageWriteTime", getAverageWriteTime());
      retval.put("HitMissRatio", getHitMissRatio());
      retval.put("ReadWriteRatio", getReadWriteRatio());
      return retval;
   }

   public void resetStatistics()
   {
      m_hits = 0;
      m_misses = 0;
      m_stores = 0;
      m_evictions = 0;
      m_hit_times = 0;
      m_miss_times = 0;
      m_store_times = 0;
      m_reset = System.currentTimeMillis();
   }
}

