package org.jboss.cache.interceptors;

import org.jboss.cache.CacheException;
import org.jboss.cache.CacheSPI;
import org.jboss.cache.Fqn;
import org.jboss.cache.InvocationContext;
import org.jboss.cache.Modification;
import org.jboss.cache.NodeSPI;
import org.jboss.cache.config.CacheLoaderConfig;
import org.jboss.cache.loader.CacheLoader;
import org.jboss.cache.marshall.MethodCall;
import org.jboss.cache.marshall.MethodDeclarations;
import org.jboss.cache.transaction.GlobalTransaction;
import org.jboss.cache.transaction.TransactionEntry;
import org.jboss.cache.transaction.TransactionTable;

import javax.transaction.TransactionManager;
import java.lang.reflect.Method;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.concurrent.ConcurrentHashMap;

/**
 * Writes modifications back to the store on the way out: stores modifications back
 * through the CacheLoader, either after each method call (no TXs), or at TX commit.
 *
 * @author Bela Ban
 * @version $Id: CacheStoreInterceptor.java 4663 2007-10-22 23:08:18Z manik.surtani@jboss.com $
 */
public class CacheStoreInterceptor extends Interceptor implements CacheStoreInterceptorMBean
{

   protected CacheLoaderConfig loaderConfig = null;
   protected TransactionManager tx_mgr = null;
   protected TransactionTable tx_table = null;
   private HashMap m_txStores = new HashMap();
   private Map<GlobalTransaction, Set<Fqn>> preparingTxs = new ConcurrentHashMap<GlobalTransaction, Set<Fqn>>();
   private long m_cacheStores = 0;
   protected CacheLoader loader;

   public void setCache(CacheSPI cache)
   {
      super.setCache(cache);
      this.loaderConfig = cache.getCacheLoaderManager().getCacheLoaderConfig();
      tx_mgr = cache.getTransactionManager();
      tx_table = cache.getTransactionTable();
      this.loader = cache.getCacheLoaderManager().getCacheLoader();
   }

   /**
    * Pass the method on. When it returns, store the modification back to the store using the CacheLoader.
    * In case of a transaction, register for TX completion (2PC) and at TX commit, write modifications made
    * under the given TX to the CacheLoader
    *
    * @return
    * @throws Throwable
    */
   public Object invoke(InvocationContext ctx) throws Throwable
   {

      MethodCall m = ctx.getMethodCall();
      // if this is a shared cache loader and the call is of remote origin, pass up the chain. - Manik
      // see http://www.jboss.com/index.html?module=bb&op=viewtopic&t=76090

      if (!ctx.isOriginLocal() && loaderConfig.isShared())
      {
         log.trace("Passing up method call and bypassing this interceptor since the cache loader is shared and this call originated remotely.");
         return super.invoke(ctx);
      }

      Fqn fqn;
      Object key, value;
      Object[] args = m.getArgs();
      Object retval, tmp_retval = null;
      boolean use_tmp_retval = false;


      if (log.isTraceEnabled())
      {
         log.trace("invoke " + m);
      }

      if (tx_mgr != null && tx_mgr.getTransaction() != null)
      {
         // we have a tx running.
         log.trace("transactional so don't put stuff in the cloader yet.");
         GlobalTransaction gtx = ctx.getGlobalTransaction();
         switch (m.getMethodId())
         {
            case MethodDeclarations.commitMethod_id:
               if (ctx.isTxHasMods())
               {
                  // this is a commit call.
                  if (log.isTraceEnabled()) log.trace("Calling loader.commit() for gtx " + gtx);
                  // sync call (a write) on the loader
                  // ignore modified FQNs
                  // List fqnsModified = getFqnsFromModificationList(tx_table.get(gtx).getCacheLoaderModifications());
                  try
                  {
                     loader.commit(gtx);
                  }
                  catch (Throwable t)
                  {
                     preparingTxs.remove(gtx);
                     throw t;
                  }
                  if (configuration.getExposeManagementStatistics() && getStatisticsEnabled())
                  {
                     Integer puts = (Integer) m_txStores.get(gtx);
                     if (puts != null)
                     {
                        m_cacheStores = m_cacheStores + puts;
                     }
                     m_txStores.remove(gtx);
                  }
               }
               else
               {
                  log.trace("Commit called with no modifications; ignoring.");
               }
               break;
            case MethodDeclarations.rollbackMethod_id:
               if (ctx.isTxHasMods())
               {
                  // this is a rollback method
                  if (preparingTxs.containsKey(gtx))
                  {
                     preparingTxs.remove(gtx);
                     loader.rollback(gtx);
                  }
                  if (configuration.getExposeManagementStatistics() && getStatisticsEnabled())
                  {
                     m_txStores.remove(gtx);
                  }
               }
               else
               {
                  log.trace("Rollback called with no modifications; ignoring.");
               }
               break;
            case MethodDeclarations.optimisticPrepareMethod_id:
            case MethodDeclarations.prepareMethod_id:
               prepareCacheLoader(gtx, isOnePhaseCommitPrepareMehod(m));
               break;
         }

         // pass up the chain
         retval = super.invoke(ctx);

         if (m.getMethodId() == MethodDeclarations.commitMethod_id && ctx.isTxHasMods())
         {
            // persist additional internal state, if any, and then clean up internal resources
            Set<Fqn> affectedFqns = preparingTxs.remove(gtx);
            if (affectedFqns != null) storeInternalState(affectedFqns);
         }
         return retval;
      }

      // if we're here we don't run in a transaction

      // remove() methods need to be applied to the CacheLoader before passing up the call: a listener might
      // access an element just removed, causing the CacheLoader to *load* the element before *removing* it.
      //      synchronized(this) {
      switch (m.getMethodId())
      {
         case MethodDeclarations.removeNodeMethodLocal_id:
            fqn = (Fqn) args[1];
            loader.remove(fqn);
            break;
         case MethodDeclarations.removeKeyMethodLocal_id:
            fqn = (Fqn) args[1];
            key = args[2];
            tmp_retval = loader.remove(fqn, key);
            use_tmp_retval = true;
            break;
         case MethodDeclarations.removeDataMethodLocal_id:
            fqn = (Fqn) args[1];
            loader.removeData(fqn);
            // we need to mark this node as data loaded
            NodeSPI n = cache.peek(fqn, false);
            if (n != null) n.setDataLoaded(true);
            break;
      }
      //      }

      retval = super.invoke(ctx);

      // put() methods need to be applied *after* the call
      //      synchronized(this) {
      switch (m.getMethodId())
      {
         case MethodDeclarations.moveMethodLocal_id:
            doMove((Fqn) args[0], (Fqn) args[1]);
            break;
         case MethodDeclarations.putDataEraseMethodLocal_id:
            // aside from this removeData call, this is the same as a putData call.  Needed since CacheLoader.put(Map)
            // does not overwrite but append.
            fqn = (Fqn) args[1];
            loader.removeData(fqn);
            // if we are erasing all the data then consider this node loaded
            NodeSPI n = cache.peek(fqn, false);
            n.setDataLoaded(true);
         case MethodDeclarations.putDataMethodLocal_id:
            loader.put((Fqn) args[1], (Map) args[2]);
            if (configuration.getExposeManagementStatistics() && getStatisticsEnabled())
            {
               m_cacheStores++;
            }
            break;
         case MethodDeclarations.putForExternalReadMethodLocal_id:
         case MethodDeclarations.putKeyValMethodLocal_id:
            fqn = (Fqn) args[1];
            key = args[2];
            value = args[3];
            tmp_retval = loader.put(fqn, key, value);
            use_tmp_retval = true;
            if (configuration.getExposeManagementStatistics() && getStatisticsEnabled())
            {
               m_cacheStores++;
            }
            break;
      }
      //      }

      if (use_tmp_retval)
      {
         return tmp_retval;
      }
      else
      {
         return retval;
      }
   }

   private void storeInternalState(Set<Fqn> affectedFqns) throws Exception
   {
      if (cache.getConfiguration().isNodeLockingOptimistic())
      {
         for (Fqn f : affectedFqns)
         {
            // NOT going to store tombstones!!
            NodeSPI n = cache.peek(f, false, false);
            if (n != null)
            {
               Map internalState = n.getInternalState(true);
               loader.put(f, internalState);
            }
         }
      }
   }

   private void doMove(Fqn node, Fqn parent) throws Exception
   {
      Fqn newNodeFqn = new Fqn(parent, node.getLastElement());

      recursiveMove(node, newNodeFqn);
      loader.remove(node);
   }

   private void recursiveMove(Fqn fqn, Fqn newFqn) throws Exception
   {
      List fqns = new ArrayList();
      fqns.add(fqn);
      fqns.add(newFqn);
      loader.put(newFqn, loader.get(fqn));
      //recurse
      Set childrenNames = loader.getChildrenNames(fqn);
      if (childrenNames != null)
      {
         for (Object child : childrenNames)
         {
            recursiveMove(new Fqn(fqn, child), new Fqn(newFqn, child));
         }
      }
   }

   public long getCacheLoaderStores()
   {
      return m_cacheStores;
   }

   public void resetStatistics()
   {
      m_cacheStores = 0;
   }

   public Map<String, Object> dumpStatistics()
   {
      Map<String, Object> retval = new HashMap<String, Object>();
      retval.put("CacheLoaderStores", m_cacheStores);
      return retval;
   }

   private void prepareCacheLoader(GlobalTransaction gtx, boolean onePhase) throws Exception
   {
      List<MethodCall> modifications;
      TransactionEntry entry;
      int txPuts = 0;

      entry = tx_table.get(gtx);
      if (entry == null)
      {
         throw new Exception("entry for transaction " + gtx + " not found in transaction table");
      }
      Set<Fqn> affectedFqns = new HashSet<Fqn>();
      modifications = entry.getCacheLoaderModifications();
      if (modifications.size() == 0)
      {
         log.trace("Transaction has not logged any modifications!");
         return;
      }
      if (log.isTraceEnabled()) log.trace("Cache loader modification list: " + modifications);
      List cache_loader_modifications = new ArrayList();
      for (MethodCall methodCall : modifications)
      {
         Modification mod = convertMethodCallToModification(methodCall, affectedFqns);
         cache_loader_modifications.add(mod);

         if (configuration.getExposeManagementStatistics() && getStatisticsEnabled())
         {
            if ((mod.getType() == Modification.ModificationType.PUT_DATA) ||
                    (mod.getType() == Modification.ModificationType.PUT_DATA_ERASE) ||
                    (mod.getType() == Modification.ModificationType.PUT_KEY_VALUE))
            {
               txPuts++;
            }
         }
      }
      if (log.isTraceEnabled())
      {
         log.trace("Converted method calls to cache loader modifications.  List size: " + cache_loader_modifications.size());
      }
      if (cache_loader_modifications.size() > 0)
      {
         loader.prepare(gtx, cache_loader_modifications, onePhase);

         preparingTxs.put(gtx, affectedFqns);
         if (configuration.getExposeManagementStatistics() && getStatisticsEnabled() && txPuts > 0)
         {
            m_txStores.put(gtx, txPuts);
         }
      }
   }

   /**
    * Converts a method call to a Modification, to be sent to a cache loader.  In addition, adds any affected Fqns to the
    * affectedFqns collection, which is then used by the calling code.
    *
    */
   private Modification convertMethodCallToModification(MethodCall methodCall, Set<Fqn> affectedFqns) throws Exception
   {
      if (log.isTraceEnabled()) log.trace("Converting method call " + methodCall + " to modification.");
      Method method = methodCall.getMethod();
      Object[] args;
      if (method == null)
      {
         throw new Exception("method call has no method: " + methodCall);
      }

      args = methodCall.getArgs();
      Modification mod;
      Fqn fqn = (Fqn) args[1];

      switch (methodCall.getMethodId())
      {
         case MethodDeclarations.putDataMethodLocal_id:
            mod = new Modification(Modification.ModificationType.PUT_DATA,
                    fqn, (Map) args[2]);// data
            break;
         case MethodDeclarations.putDataEraseMethodLocal_id:
            mod = new Modification(Modification.ModificationType.PUT_DATA_ERASE,
                    fqn,  (Map) args[2]);// data
            break;
         case MethodDeclarations.putKeyValMethodLocal_id:
            mod = new Modification(Modification.ModificationType.PUT_KEY_VALUE,
                    fqn,  args[2], args[3]);// key + value
            break;
         case MethodDeclarations.removeNodeMethodLocal_id:
            mod = new Modification(Modification.ModificationType.REMOVE_NODE,
                    fqn);
            break;
         case MethodDeclarations.removeKeyMethodLocal_id:
            mod = new Modification(Modification.ModificationType.REMOVE_KEY_VALUE,
                    fqn, args[2]);// key
            break;
         case MethodDeclarations.removeDataMethodLocal_id:
            mod = new Modification(Modification.ModificationType.REMOVE_DATA,
                    fqn);
            break;
         case MethodDeclarations.moveMethodLocal_id:
            Fqn moveFrom = (Fqn) args[0];
            affectedFqns.add(moveFrom);
            mod = new Modification(Modification.ModificationType.MOVE, moveFrom, fqn);
            break;
         default:
            throw new CacheException("method call " + method.getName() + " cannot be converted to a modification");
      }
      affectedFqns.add(fqn);
      if (log.isTraceEnabled()) log.trace("Converted " + methodCall + " to Modification of type " + mod.getType());
      return mod;
   }
}
