/*
 * JBoss, Home of Professional Open Source
 *
 * Distributable under LGPL license.
 * See terms of license at gnu.org.
 */
package org.jboss.cache.interceptors;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.jboss.cache.CacheException;
import org.jboss.cache.CacheSPI;
import org.jboss.cache.Fqn;
import org.jboss.cache.InvocationContext;
import org.jboss.cache.Node;
import org.jboss.cache.NodeSPI;
import org.jboss.cache.buddyreplication.BuddyManager;
import org.jboss.cache.buddyreplication.GravitateResult;
import org.jboss.cache.config.Configuration;
import org.jboss.cache.marshall.MethodCall;
import org.jboss.cache.marshall.MethodCallFactory;
import org.jboss.cache.marshall.MethodDeclarations;
import org.jboss.cache.marshall.NodeData;
import org.jboss.cache.transaction.GlobalTransaction;
import org.jboss.cache.transaction.TransactionEntry;
import org.jgroups.Address;
import org.jgroups.blocks.GroupRequest;
import org.jgroups.blocks.RspFilter;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.List;
import java.util.Map;
import java.util.concurrent.ConcurrentHashMap;

/**
 * The Data Gravitator interceptor intercepts cache misses and attempts t
 * gravitate data from other parts of the cluster.
 * <p/>
 * Only used if Buddy Replication is enabled.  Also, the interceptor only kicks
 * in if an {@link org.jboss.cache.config.Option} is passed in to force Data
 * Gravitation for a specific invocation or if <b>autoDataGravitation</b> is
 * set to <b>true</b> when configuring Buddy Replication.
 * <p/>
 * See the JBoss Cache User Guide for more details on configuration options.
 * There is a section dedicated to Buddy Replication in the Replication
 * chapter.
 *
 * @author <a href="mailto:manik@jboss.org">Manik Surtani (manik@jboss.org)</a>
 */
public class DataGravitatorInterceptor extends BaseRpcInterceptor
{
   private BuddyManager buddyManager;
   private boolean syncCommunications = false;
   private Log log = LogFactory.getLog(DataGravitatorInterceptor.class);
   private Map<GlobalTransaction, MethodCall> transactionMods = new ConcurrentHashMap<GlobalTransaction, MethodCall>();

   public void setCache(CacheSPI cache)
   {
      super.setCache(cache);
      this.buddyManager = cache.getBuddyManager();
      syncCommunications = configuration.getCacheMode() == Configuration.CacheMode.REPL_SYNC || configuration.getCacheMode() == Configuration.CacheMode.INVALIDATION_SYNC;
   }

   public Object invoke(InvocationContext ctx) throws Throwable
   {
      MethodCall m = ctx.getMethodCall();
      if (log.isTraceEnabled()) log.trace("Invoked with method call " + m);

      if (MethodDeclarations.isBlockUnblockMethod(m.getMethodId()) || ctx.getOptionOverrides().isSkipDataGravitation())
      {
         return super.invoke(ctx);
      }

      // Transactional lifecycle methods should be handled regardless of whether data gravitation is enabled or not.
      if (!MethodDeclarations.isTransactionLifecycleMethod(m.getMethodId()))
      {
         if (isGravitationEnabled(ctx) && MethodDeclarations.isGetMethod(m.getMethodId()))
         {
            // test that the Fqn being requested exists locally in the cache.
            Fqn fqn = extractFqn(m.getMethodId(), m.getArgs());
            if (log.isTraceEnabled()) log.trace("Checking local existence of fqn " + fqn);
            if (BuddyManager.isBackupFqn(fqn))
            {
               log.info("Is call for a backup Fqn, not performing any gravitation.  Direct calls on internal backup nodes are *not* supported.");
            }
            else
            {
               if (cache.peek(fqn, false) == null)
               {
                  log.trace("Gravitating from local backup tree");
                  BackupData data = localBackupGet(fqn, ctx);

                  if (data == null)
                  {
                     log.trace("Gravitating from remote backup tree");
                     // gravitate remotely.
                     data = remoteBackupGet(fqn);
                  }

                  if (data != null)
                  {
                     // create node locally so I don't gravitate again
                     // when I do the put() call to the cluster!
                     //createNode(data.backupData, true);
                     // Make sure I replicate to my buddies.
                     log.trace("Passing the put call locally to make sure state is persisted and ownership is correctly established.");
                     createNode(data.backupData, false);

                     // very strange, the invocation contexts get twisted up here, and will need preservation.
                     // a bit crappy and hacky, all will be solved when we move to JBoss AOP in 2.1.0
                     //ctx.setMethodCall(m);

                     // Clean up the other nodes
                     cleanBackupData(data, ctx.getGlobalTransaction(), ctx);
                  }
               }
               else
               {
                  log.trace("No need to gravitate; have this already.");
               }
            }
         }
         else
         {
            if (log.isTraceEnabled())
            {
               log.trace("Suppressing data gravitation for this call.");
            }
         }
      }
      else
      {

         try
         {
            switch (m.getMethodId())
            {
               case MethodDeclarations.prepareMethod_id:
               case MethodDeclarations.optimisticPrepareMethod_id:
                  Object o = super.invoke(ctx);
                  doPrepare(ctx.getGlobalTransaction(), ctx);
                  return o;
               case MethodDeclarations.rollbackMethod_id:
                  transactionMods.remove(ctx.getGlobalTransaction());
                  return super.invoke(ctx);
               case MethodDeclarations.commitMethod_id:
                  doCommit(ctx.getGlobalTransaction(), ctx);
                  transactionMods.remove(ctx.getGlobalTransaction());
                  return super.invoke(ctx);
            }
         }
         catch (Throwable throwable)
         {
            transactionMods.remove(ctx.getGlobalTransaction());
            throw throwable;
         }
      }
      //            }
      //        }
      //        else
      //        {
      //           if (log.isTraceEnabled())
      //              log.trace("Suppressing data gravitation for this call.");
      //        }
      return super.invoke(ctx);
   }

   private boolean isGravitationEnabled(InvocationContext ctx)
   {
      boolean enabled = ctx.isOriginLocal();
      if (enabled)
      {
         if (!buddyManager.isAutoDataGravitation())
         {
            enabled = ctx.getOptionOverrides().getForceDataGravitation();
         }
      }
      return enabled;
   }

   private void doPrepare(GlobalTransaction gtx, InvocationContext ctx) throws Throwable
   {
      MethodCall cleanup = transactionMods.get(gtx);
      if (log.isTraceEnabled()) log.trace("Broadcasting prepare for cleanup ops " + cleanup);
      if (cleanup != null)
      {
         MethodCall prepare;
         List<MethodCall> mods = new ArrayList<MethodCall>(1);
         mods.add(cleanup);
         if (configuration.isNodeLockingOptimistic())
         {
            prepare = MethodCallFactory.create(MethodDeclarations.optimisticPrepareMethod, gtx, mods, null, cache.getLocalAddress(), false);
         }
         else
         {
            prepare = MethodCallFactory.create(MethodDeclarations.prepareMethod, gtx, mods, cache.getLocalAddress(), syncCommunications);
         }

         replicateCall(getMembersOutsideBuddyGroup(), prepare, syncCommunications, ctx.getOptionOverrides());
      }
      else
      {
         if (log.isTraceEnabled()) log.trace("Nothing to broadcast in prepare phase for gtx " + gtx);
      }
   }

   private void doCommit(GlobalTransaction gtx, InvocationContext ctx) throws Throwable
   {
      if (transactionMods.containsKey(gtx))
      {
         if (log.isTraceEnabled()) log.trace("Broadcasting commit for gtx " + gtx);
         replicateCall(getMembersOutsideBuddyGroup(), MethodCallFactory.create(MethodDeclarations.commitMethod, gtx), syncCommunications, ctx.getOptionOverrides());
      }
      else
      {
         if (log.isTraceEnabled()) log.trace("Nothing to broadcast in commit phase for gtx " + gtx);
      }
   }

   private List<Address> getMembersOutsideBuddyGroup()
   {
      List<Address> members = new ArrayList<Address>(cache.getMembers());
      members.remove(cache.getLocalAddress());
      members.removeAll(buddyManager.getBuddyAddresses());
      return members;
   }

   private BackupData remoteBackupGet(Fqn name) throws Exception
   {

      BackupData result = null;

      GravitateResult gr = gravitateData(name);

      if (gr.isDataFound())
      {
         if (log.isTraceEnabled())
         {
            log.trace("Got response " + gr);
         }

//         if (configuration.isUseRegionBasedMarshalling())
//         {
//            ClassLoader cl = Thread.currentThread().getContextClassLoader();
//            try
//            {
//               cache.getRegionManager().setContextClassLoaderAsCurrent(name);
//
//               byte[] nodeData = (byte[]) resp[0];
//               ByteArrayInputStream bais = new ByteArrayInputStream(nodeData);
//               MarshalledValueInputStream mais = new MarshalledValueInputStream(bais);
//               nodes = (List) mais.readObject();
//               mais.close();
//            }
//            finally
//            {
//               Thread.currentThread().setContextClassLoader(cl);
//            }
//         }
//         else
//         {
//         }

         result = new BackupData(name, gr);
      }

      return result;
   }

   private void cleanBackupData(BackupData backup, GlobalTransaction gtx, InvocationContext ctx) throws Throwable
   {
      //       MethodCall primaryDataCleanup, backupDataCleanup;
      //       if (buddyManager.isDataGravitationRemoveOnFind())
      //       {
      //           primaryDataCleanup = MethodCallFactory.create(MethodDeclarations.removeNodeMethodLocal, new Object[]{null, backup.primaryFqn, Boolean.FALSE});
      //           backupDataCleanup = MethodCallFactory.create(MethodDeclarations.removeNodeMethodLocal, new Object[]{null, backup.backupFqn, Boolean.FALSE});
      //       }
      //       else
      //       {
      //           primaryDataCleanup = MethodCallFactory.create(MethodDeclarations.evictNodeMethodLocal, new Object[]{backup.primaryFqn});
      //           backupDataCleanup = MethodCallFactory.create(MethodDeclarations.evictNodeMethodLocal, new Object[]{backup.backupFqn});
      //       }

      MethodCall cleanup = MethodCallFactory.create(MethodDeclarations.dataGravitationCleanupMethod, gtx, backup.primaryFqn, backup.backupFqn);


      if (log.isTraceEnabled()) log.trace("Performing cleanup on [" + backup.primaryFqn + "]");
      if (gtx == null)
      {
         // broadcast removes
         // remove main Fqn
         //replicateCall(cache.getMembers(), primaryDataCleanup, syncCommunications);

         if (log.isTraceEnabled()) log.trace("Performing cleanup on [" + backup.backupFqn + "]");
         // remove backup Fqn
         //replicateCall(cache.getMembers(), backupDataCleanup, syncCommunications);
         replicateCall(cache.getMembers(), cleanup, syncCommunications, ctx.getOptionOverrides());
      }
      else
      {
         if (log.isTraceEnabled())
         {
            log.trace("Data gravitation performed under global transaction " + gtx + ".  Not broadcasting cleanups until the tx commits.  Adding to tx mod list instead.");
         }
         transactionMods.put(gtx, cleanup);
         TransactionEntry te = getTransactionEntry(gtx);
         te.addModification(cleanup);
      }
   }

   private GravitateResult gravitateData(Fqn fqn) throws Exception
   {
      if (log.isTraceEnabled())
      {
         log.trace("cache=" + cache.getLocalAddress() + "; requesting data gravitation for Fqn " + fqn);
      }
      List<Address> mbrs = cache.getMembers();
      Boolean searchSubtrees = (buddyManager.isDataGravitationSearchBackupTrees() ? Boolean.TRUE : Boolean.FALSE);
      MethodCall dGrav = MethodCallFactory.create(MethodDeclarations.dataGravitationMethod, fqn, searchSubtrees);
      // doing a GET_ALL is crappy but necessary since JGroups' GET_FIRST could return null results from nodes that do
      // not have either the primary OR backup, and stop polling other valid nodes.
      List resps = cache.getRPCManager().callRemoteMethods(mbrs, dGrav, GroupRequest.GET_FIRST, true, buddyManager.getBuddyCommunicationTimeout(), new ResponseValidityFilter());
      if (log.isTraceEnabled())
      {
         log.trace("got responses " + resps);
      }
      if (resps == null)
      {
         log.error("No replies to call " + dGrav + ".  Perhaps we're alone in the cluster?");
         return GravitateResult.noDataFound();
      }

      // test for and remove exceptions
      GravitateResult result = GravitateResult.noDataFound();
      for (Object o : resps)
      {
         if (o instanceof Throwable)
         {
            if (log.isDebugEnabled())
            {
               log.debug("Found remote Throwable among responses - removing from responses list", (Exception) o);
            }
         }
         else if (o != null)
         {
            result = (GravitateResult) o;
            if (result.isDataFound())
            {
               break;
            }
         }
         else if (!configuration.isUseRegionBasedMarshalling())
         {
            // Null is OK if we are using region based marshalling; it
            // is what is returned if a region is inactive. Otherwise
            // getting a null is an error condition
            log.error("Unexpected null response to call " + dGrav + ".");
         }

      }

      return result;
   }

   @SuppressWarnings("unchecked")
   private void createNode(List<NodeData> nodeData, boolean localOnly) throws CacheException
   {
      for (NodeData data : nodeData)
      {
         if (localOnly)
         {
            if (cache.peek(data.getFqn(), false) == null)
            {
               createNodesLocally(data.getFqn(), data.getAttributes());
            }
         }
         else
         {
            cache.put(data.getFqn(), data.getAttributes());
         }
      }
   }

   @SuppressWarnings("unchecked")
   private void createNodesLocally(Fqn<?> fqn, Map<?, ?> data) throws CacheException
   {
      int treeNodeSize;
      if ((treeNodeSize = fqn.size()) == 0) return;
      NodeSPI n = cache.getRoot();
      for (int i = 0; i < treeNodeSize; i++)
      {
         Object child_name = fqn.get(i);
         NodeSPI child_node = n.addChildDirect(new Fqn<Object>(child_name));
         if (child_node == null)
         {
            if (log.isTraceEnabled())
            {
               log.trace("failed to find or create child " + child_name + " of node " + n.getFqn());
            }
            return;
         }
         if (i == treeNodeSize - 1)
         {
            // set data
            child_node.putAllDirect(data);
         }
         n = child_node;
      }
   }

   private TransactionEntry getTransactionEntry(GlobalTransaction gtx)
   {
      return cache.getTransactionTable().get(gtx);
   }

   private Fqn extractFqn(int methodId, Object[] args)
   {
      return (Fqn) args[MethodDeclarations.isCrudMethod(methodId) ? 1 : 0];
   }

   @SuppressWarnings("unchecked")
   private boolean localBackupExists(Fqn<?> fqn)
   {
      boolean exists = false;

      for (Node node : getBackupRootCollection())
      {
         Fqn<?> newSearchFqn = new Fqn(node.getFqn(), fqn);
         exists = cache.peek(newSearchFqn, false) != null;
         if (exists) break;
      }

      return exists;
   }

   private BackupData localBackupGet(Fqn fqn, InvocationContext ctx) throws CacheException
   {
      GravitateResult result = cache.gravitateData(fqn, true);// a "local" gravitation
      boolean found = result.isDataFound();
      BackupData data = null;

      if (found)
      {
         Fqn backupFqn = result.getBuddyBackupFqn();
         data = new BackupData(fqn, result);
         // now the cleanup
         if (buddyManager.isDataGravitationRemoveOnFind())
         {
            // Remove locally only; the remote call will
            // be broadcast later
            ctx.getOptionOverrides().setCacheModeLocal(true);
            cache.removeNode(backupFqn);
         }
         else
         {
            cache.evict(backupFqn, true);
         }
      }

      return data;
   }

   @SuppressWarnings("unchecked")
   private Collection<Node> getBackupRootCollection()
   {
      NodeSPI backupRoot = cache.peek(BuddyManager.BUDDY_BACKUP_SUBTREE_FQN, true);
      return backupRoot == null ? Collections.emptySet() : backupRoot.getChildrenDirect();
   }

   private static class BackupData
   {
      Fqn primaryFqn;
      Fqn backupFqn;
      List<NodeData> backupData;

      public BackupData(Fqn primaryFqn, GravitateResult gr)
      {
         this.primaryFqn = primaryFqn;
         this.backupFqn = gr.getBuddyBackupFqn();
         this.backupData = gr.getNodeData();
      }
   }

   public static class ResponseValidityFilter implements RspFilter
   {
      private int numValidResponses = 0;

      public boolean isAcceptable(Object object, Address address)
      {
         if (!(object instanceof GravitateResult)) return false;

         GravitateResult response = (GravitateResult) object;
         if (response.isDataFound()) numValidResponses++;

         return response.isDataFound();
      }

      public boolean needMoreResponses()
      {
         return numValidResponses < 1;
      }
   }
}
