/*
 * JBoss, the OpenSource J2EE webOS
 *
 * Distributable under LGPL license.
 * See terms of license at gnu.org.
 * Created on March 25 2003
 */
package org.jboss.cache.interceptors;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.jboss.cache.*;
import org.jboss.cache.eviction.EvictedEventNode;
import org.jboss.cache.eviction.NodeEventType;
import org.jboss.cache.marshall.MethodCall;
import org.jboss.cache.marshall.MethodDeclarations;

import java.util.HashMap;
import java.util.Map;

/**
 * Eviction Interceptor.
 * <p/>
 * This interceptor is used to handle eviction events.
 *
 * @author Daniel Huang
 * @version $Revision: 4527 $
 */
public class EvictionInterceptor extends Interceptor
{
   private static final Log log = LogFactory.getLog(EvictionInterceptor.class);

   protected RegionManager regionManager;
   protected Map<Integer, EvictionMethodHandler> evictionMethodHandlers = new HashMap<Integer, EvictionMethodHandler>();

   public EvictionInterceptor()
   {
      EvictionMethodHandler handler = new GetNodeEvictionMethodHandler();
      evictionMethodHandlers.put(MethodDeclarations.getNodeMethodLocal_id, handler);
      evictionMethodHandlers.put(MethodDeclarations.getDataMapMethodLocal_id, handler);

      handler = new GetKeyEvictionMethodHandler();
      evictionMethodHandlers.put(MethodDeclarations.getKeyValueMethodLocal_id, handler);

      handler = new RemoveNodeEvictionMethodHandler();
      evictionMethodHandlers.put(MethodDeclarations.removeNodeMethodLocal_id, handler);
      evictionMethodHandlers.put(MethodDeclarations.removeDataMethodLocal_id, handler);

      handler = new RemoveKeyEvictionMethodHandler();
      evictionMethodHandlers.put(MethodDeclarations.removeKeyMethodLocal_id, handler);

      handler = new PutDataEvictionMethodHandler();
      evictionMethodHandlers.put(MethodDeclarations.putDataMethodLocal_id, handler);

      handler = new PutDataEraseEvictionMethodHandler();
      evictionMethodHandlers.put(MethodDeclarations.putDataEraseMethodLocal_id, handler);

      handler = new PutKeyEvictionMethodHandler();
      evictionMethodHandlers.put(MethodDeclarations.putKeyValMethodLocal_id, handler);
      evictionMethodHandlers.put(MethodDeclarations.putForExternalReadMethodLocal_id, handler);

      handler = new PartialEvictionEvictionMethodHandler();
      evictionMethodHandlers.put(MethodDeclarations.evictNodeMethodLocal_id, handler);
      evictionMethodHandlers.put(MethodDeclarations.evictVersionedNodeMethodLocal_id, handler);
   }

   /**
    * this method is for ease of unit testing. thus package access.
    * <p/>
    * Not to be attempted to be used anywhere else.
    */
   void setRegionManager(RegionManager regionManager)
   {
      this.regionManager = regionManager;
   }

   public void setCache(CacheSPI cache)
   {
      super.setCache(cache);
      this.regionManager = cache.getRegionManager();
   }

   public Object invoke(InvocationContext ctx) throws Throwable
   {
      MethodCall m = ctx.getMethodCall();
      Object ret = super.invoke(ctx);

      if (log.isTraceEnabled())
      {
         log.trace("Invoking EvictionInterceptor");
      }

      // skip the TX. this interceptor will invoke around/after the call and lock. if the ret == null or if an exception
      // is thrown, this interceptor is terminated. there is no need for explicit rollback logic.
      this.updateNode(m, ret);



      if (log.isTraceEnabled())
      {
         log.trace("Finished invoking EvictionInterceptor");
      }

      return ret;
   }

   protected void updateNode(MethodCall m, Object retVal)
   {
      if (log.isTraceEnabled())
      {
         log.trace("Updating node/element events with no tx");
      }

      EvictedEventNode event = this.extractEvent(m, retVal);
      if (event == null)
      {
         // no node modifications.
         return;
      }

      NodeSPI<?,?> nodeSPI = cache.peek(event.getFqn(), false);
      //we do not trigger eviction events for resident nodes
      if (nodeSPI != null && nodeSPI.isResident())
      {
         return;
      }

      this.doEventUpdatesOnRegionManager(event);

      if (log.isTraceEnabled())
      {
         log.trace("Finished updating node");
      }
   }

   protected EvictedEventNode extractEvent(MethodCall m, Object retVal)
   {
      EvictionMethodHandler handler = this.evictionMethodHandlers.get(m.getMethodId());
      if (handler == null)
      {
         return null;
      }

      return handler.extractEvictedEventNode(m, retVal);
   }

   protected boolean canIgnoreEvent(Fqn fqn, NodeEventType type)
   {
      Region r = regionManager.getRegion(fqn, Region.Type.EVICTION, false);
      if (r == null) return true;// should never happen, we should at least get the default region.
      return r.getEvictionPolicy().canIgnoreEvent(fqn, type);
   }

   protected void doEventUpdatesOnRegionManager(EvictedEventNode event)
   {
      Region region = regionManager.getRegion(event.getFqn(), false);
      region.putNodeEvent(event);

      if (log.isTraceEnabled())
      {
         log.trace("Adding event " + event + " to region at " + region.getFqn());
      }
   }

   protected class GetNodeEvictionMethodHandler implements EvictionMethodHandler
   {
      public EvictedEventNode extractEvictedEventNode(MethodCall mc, Object retVal)
      {
         if (retVal == null)
         {
            if (log.isTraceEnabled())
            {
               log.trace("No event added. Node does not exist");
            }

            return null;
         }

         Object args[] = mc.getArgs();
         Fqn fqn = (Fqn) args[0];

         if (fqn != null && !EvictionInterceptor.this.canIgnoreEvent(fqn, NodeEventType.VISIT_NODE_EVENT))
         {
            return new EvictedEventNode(fqn, NodeEventType.VISIT_NODE_EVENT);
         }

         return null;
      }
   }

   protected class GetKeyEvictionMethodHandler implements EvictionMethodHandler
   {
      public EvictedEventNode extractEvictedEventNode(MethodCall mc, Object retVal)
      {
         if (retVal == null)
         {
            if (log.isTraceEnabled())
            {
               log.trace("No event added. Element does not exist");
            }

            return null;
         }

         Object args[] = mc.getArgs();
         Fqn fqn = (Fqn) args[0];
         Object key = args[1];
         if (fqn != null && key != null
                 && !EvictionInterceptor.this.canIgnoreEvent(fqn, NodeEventType.VISIT_NODE_EVENT))
         {
            return new EvictedEventNode(fqn, NodeEventType.VISIT_NODE_EVENT);
         }

         return null;
      }

   }

   protected class RemoveNodeEvictionMethodHandler implements EvictionMethodHandler
   {
      public EvictedEventNode extractEvictedEventNode(MethodCall mc, Object retVal)
      {
         Object args[] = mc.getArgs();
         Fqn fqn = (Fqn) args[1];

         if (fqn != null
                 && !EvictionInterceptor.this.canIgnoreEvent(fqn, NodeEventType.REMOVE_NODE_EVENT))
         {
            return new EvictedEventNode(fqn, NodeEventType.REMOVE_NODE_EVENT);
         }

         return null;
      }

   }

   protected class RemoveKeyEvictionMethodHandler implements EvictionMethodHandler
   {
      public EvictedEventNode extractEvictedEventNode(MethodCall mc, Object retVal)
      {
         if (retVal == null)
         {
            if (log.isTraceEnabled())
            {
               log.trace("No event added. Element does not exist");
            }

            return null;
         }

         Object args[] = mc.getArgs();
         Fqn fqn = (Fqn) args[1];
         Object key = args[2];
         if (fqn != null && key != null
                 && !EvictionInterceptor.this.canIgnoreEvent(fqn, NodeEventType.REMOVE_ELEMENT_EVENT))
         {
            return new EvictedEventNode(fqn, NodeEventType.REMOVE_ELEMENT_EVENT, 1);
         }
         return null;
      }

   }

   protected class PutDataEvictionMethodHandler implements EvictionMethodHandler
   {
      public EvictedEventNode extractEvictedEventNode(MethodCall mc, Object retVal)
      {
         Object[] args = mc.getArgs();
         Fqn fqn = (Fqn) args[1];
         Map putData = (Map) args[2];
         if (fqn != null
                 && !EvictionInterceptor.this.canIgnoreEvent(fqn, NodeEventType.ADD_NODE_EVENT))
         {
            if (putData == null)
            {
               if (log.isTraceEnabled())
               {
                  log.trace("Putting null data under fqn " + fqn + ".");
               }

               return null;
            }

            int size;
            synchronized (putData)
            {
               size = putData.size();
            }

            return new EvictedEventNode(fqn, NodeEventType.ADD_NODE_EVENT, size);
         }

         return null;
      }

   }

   protected class PutDataEraseEvictionMethodHandler implements EvictionMethodHandler
   {
      public EvictedEventNode extractEvictedEventNode(MethodCall mc, Object retVal)
      {
         Object[] args = mc.getArgs();
         Fqn fqn = (Fqn) args[1];
         Map putData = (Map) args[2];
         Boolean resetElementCount = (Boolean) args[4];
         if (fqn != null
                 && !EvictionInterceptor.this.canIgnoreEvent(fqn, NodeEventType.ADD_NODE_EVENT))
         {
            if (putData == null)
            {
               if (log.isTraceEnabled())
               {
                  log.trace("Putting null data under fqn " + fqn + ".");
               }

               return null;
            }

            int size;
            synchronized (putData)
            {
               size = putData.size();
            }

            EvictedEventNode event = new EvictedEventNode(fqn, NodeEventType.ADD_NODE_EVENT, size);
            event.setResetElementCount(resetElementCount);
            return event;
         }

         return null;
      }
   }

   protected class PutKeyEvictionMethodHandler implements EvictionMethodHandler
   {

      public EvictedEventNode extractEvictedEventNode(MethodCall mc, Object retVal)
      {
         Object[] args = mc.getArgs();
         Fqn fqn = (Fqn) args[1];
         Object key = args[2];
         if (fqn != null && key != null
                 && !EvictionInterceptor.this.canIgnoreEvent(fqn, NodeEventType.ADD_ELEMENT_EVENT))
         {
            return new EvictedEventNode(fqn, NodeEventType.ADD_ELEMENT_EVENT, 1);
         }

         return null;
      }

   }

   protected class PartialEvictionEvictionMethodHandler implements EvictionMethodHandler
   {
      public EvictedEventNode extractEvictedEventNode(MethodCall mc, Object retVal)
      {
         // See if the node still exists; i.e. was only data removed
         // because it still has children.
         // If yes, put an ADD event in the queue so the node gets revisited

         boolean complete = (retVal != null && (Boolean) retVal);
         if (!complete)
         {
            Object[] args = mc.getArgs();
            Fqn fqn = (Fqn) args[0];
            if (fqn != null
                    && !EvictionInterceptor.this.canIgnoreEvent(fqn, NodeEventType.ADD_NODE_EVENT))
            {
               return new EvictedEventNode(fqn, NodeEventType.ADD_NODE_EVENT, 0);
            }
         }
         return null;
      }
   }

   protected interface EvictionMethodHandler
   {
      EvictedEventNode extractEvictedEventNode(MethodCall mc, Object retVal);
   }

}
