/*
 * JBoss, Home of Professional Open Source
 * Copyright 2005, JBoss Inc., and individual contributors as indicated
 * by the @authors tag. See the copyright.txt in the distribution for a
 * full listing of individual contributors.
 *
 * This is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as
 * published by the Free Software Foundation; either version 2.1 of
 * the License, or (at your option) any later version.
 *
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this software; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA, or see the FSF site: http://www.fsf.org.
 */
package org.jboss.cache.interceptors;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.jboss.cache.CacheSPI;
import org.jboss.cache.InvocationContext;
import org.jboss.cache.config.Configuration;
import org.jboss.cache.marshall.MethodCall;
import org.jboss.cache.marshall.MethodDeclarations;

import javax.transaction.Status;
import javax.transaction.SystemException;
import javax.transaction.Transaction;
import java.util.Collections;
import java.util.Map;

/**
 * Class representing an interceptor.
 * <em>Note that this will be replaced by {@link org.jboss.aop.advice.Interceptor} in one of the next releases</em>
 *
 * @author Bela Ban
 * @version $Id: Interceptor.java 4183 2007-07-19 12:53:35Z msurtani $
 */
public abstract class Interceptor implements InterceptorMBean
{
   protected Interceptor next = null, last = null;
   protected CacheSPI<?, ?> cache;
   protected Log log = null;
   protected Configuration configuration;
   private boolean statsEnabled = false;

   public Interceptor()
   {
      log = LogFactory.getLog(getClass());
   }

   public void setNext(Interceptor i)
   {
      next = i;
   }

   public Interceptor getNext()
   {
      return next;
   }

   public void setCache(CacheSPI cache)
   {
      this.cache = cache;
      this.configuration = cache.getConfiguration();
   }

   public Object invoke(InvocationContext ctx) throws Throwable
   {
      return next.invoke(ctx);
   }

   public boolean getStatisticsEnabled()
   {
      return statsEnabled;
   }

   public void setStatisticsEnabled(boolean enabled)
   {
      statsEnabled = enabled;
   }

   public Interceptor getLast()
   {
      return last;
   }

   public void setLast(Interceptor last)
   {
      this.last = last;
   }

   /**
    * This implementation returns an empty Map.  If individual Interceptors wish to expose statistics, they can override this
    * method.
    */
   public Map<String, Object> dumpStatistics()
   {
      return Collections.emptyMap();
   }

   /**
    * This implementation is a no-op.  If individual Interceptors wish to expose statistics, they can override this
    * method.
    */
   public void resetStatistics()
   {
   }

   /**
    * Returns true if transaction is ACTIVE, false otherwise
    */
   protected boolean isActive(Transaction tx)
   {
      if (tx == null) return false;
      int status = -1;
      try
      {
         status = tx.getStatus();
         return status == Status.STATUS_ACTIVE;
      }
      catch (SystemException e)
      {
         log.error("failed getting transaction status", e);
         return false;
      }
   }

   /**
    * Returns true if transaction is PREPARING, false otherwise
    */
   protected boolean isPreparing(Transaction tx)
   {
      if (tx == null) return false;
      int status = -1;
      try
      {
         status = tx.getStatus();
         return status == Status.STATUS_PREPARING;
      }
      catch (SystemException e)
      {
         log.error("failed getting transaction status", e);
         return false;
      }
   }

   /**
    * Return s true of tx's status is ACTIVE or PREPARING
    *
    * @param tx
    * @return true if the tx is active or preparing
    */
   protected boolean isValid(Transaction tx)
   {
      return isActive(tx) || isPreparing(tx);
   }

   /**
    * This only works for prepare() and optimisticPrepare() method calls.
    *
    * @param m
    */
   protected boolean isOnePhaseCommitPrepareMehod(MethodCall m)
   {
      switch (m.getMethodId())
      {
         case MethodDeclarations.prepareMethod_id:
            return (Boolean) m.getArgs()[3];
         case MethodDeclarations.optimisticPrepareMethod_id:
            return (Boolean) m.getArgs()[4];
         default:
            return false;
      }
   }

   public String toString()
   {
      return getClass().getName()
              + "{next: "
              + (getNext() == null ? null : getNext().getClass())
              + "; last: "
              + (getLast() == null ? null : getLast().getClass())
              + "}";
   }
}
