/*
 * JBoss, Home of Professional Open Source
 *
 * Distributable under LGPL license.
 * See terms of license at gnu.org.
 */
package org.jboss.cache.interceptors;

import org.jboss.cache.CacheSPI;
import org.jboss.cache.Fqn;
import org.jboss.cache.InvocationContext;
import org.jboss.cache.config.Option;
import org.jboss.cache.marshall.MethodCall;
import org.jboss.cache.marshall.MethodCallFactory;
import org.jboss.cache.marshall.MethodDeclarations;
import org.jboss.cache.optimistic.DataVersion;
import org.jboss.cache.optimistic.DefaultDataVersion;
import org.jboss.cache.optimistic.TransactionWorkspace;
import org.jboss.cache.optimistic.WorkspaceNode;
import org.jboss.cache.transaction.GlobalTransaction;
import org.jboss.cache.transaction.OptimisticTransactionEntry;
import org.jboss.cache.transaction.TransactionEntry;
import org.jboss.cache.transaction.TransactionTable;

import javax.transaction.SystemException;
import javax.transaction.Transaction;
import java.util.HashMap;
import java.util.HashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.concurrent.ConcurrentHashMap;

/**
 * This interceptor acts as a replacement to the replication interceptor when
 * the CacheImpl is configured with ClusteredSyncMode as INVALIDATE.
 * <p/>
 * The idea is that rather than replicating changes to all caches in a cluster
 * when CRUD (Create, Remove, Update, Delete) methods are called, simply call
 * evict(Fqn) on the remote caches for each changed node.  This allows the
 * remote node to look up the value in a shared cache loader which would have
 * been updated with the changes.
 *
 * @author <a href="mailto:manik@jboss.org">Manik Surtani (manik@jboss.org)</a>
 */
public class InvalidationInterceptor extends BaseRpcInterceptor implements InvalidationInterceptorMBean
{
   private long m_invalidations = 0;
   protected TransactionTable txTable;
   protected Map<GlobalTransaction, List<MethodCall>> txMods;
   protected boolean optimistic;

   public void setCache(CacheSPI cache)
   {
      super.setCache(cache);
      txTable = cache.getTransactionTable();
      optimistic=cache.getConfiguration().isNodeLockingOptimistic();
      if (optimistic) txMods = new ConcurrentHashMap<GlobalTransaction, List<MethodCall>>();
   }

   public Object invoke(InvocationContext ctx) throws Throwable
   {
      MethodCall m = ctx.getMethodCall();
      Option optionOverride = ctx.getOptionOverrides();
      if (optionOverride != null && optionOverride.isCacheModeLocal() && (ctx.getTransaction() == null || MethodDeclarations.isTransactionLifecycleMethod(m.getMethodId())))
      {
         // skip replication!!
         return super.invoke(ctx);
      }

      Transaction tx = ctx.getTransaction();
      Object retval = super.invoke(ctx);

      if (log.isTraceEnabled()) log.trace("(" + cache.getLocalAddress() + ") method call " + m);

      // now see if this is a CRUD method:
      if (MethodDeclarations.isCrudMethod(m.getMethodId()))
      {
         if (m.getMethodId() != MethodDeclarations.putForExternalReadMethodLocal_id && m.getMethodId() != MethodDeclarations.putForExternalReadVersionedMethodLocal_id)
         {
            if (log.isDebugEnabled()) log.debug("Is a CRUD method");
            Set<Fqn> fqns = new HashSet<Fqn>();
            findAndAddFqns(m.getArgs(), fqns, m.getMethodId() == MethodDeclarations.moveMethodLocal_id);
            if (!fqns.isEmpty())
            {
               // could be potentially TRANSACTIONAL.  Ignore if it is, until we see a prepare().
               if (tx == null || !isValid(tx))
               {
                  // the no-tx case:
                  //replicate an evict call.
                  for (Fqn fqn : fqns) invalidateAcrossCluster(fqn, null, isSynchronous(optionOverride), ctx);
               }
            }
         }
         else
         {
            log.debug("Encountered a putForExternalRead() - is a no op.");
         }
      }
      else
      {
         // not a CRUD method - lets see if it is a tx lifecycle method.
         if (tx != null)
         {

            GlobalTransaction gtx;
            TransactionEntry entry;
            List<MethodCall> modifications;
            // lets see if we are in the prepare phase (as this is the only time we actually do anything)
            switch (m.getMethodId())
            {

               case MethodDeclarations.prepareMethod_id:
                  if (!optimistic)
                  {
                     log.debug("Entering InvalidationInterceptor's prepare phase");
                     // fetch the modifications before the transaction is committed (and thus removed from the txTable)
                     gtx = ctx.getGlobalTransaction();
                     entry = txTable.get(gtx);
                     if (entry == null) throw new IllegalStateException("cannot find transaction entry for " + gtx);
                     modifications = new LinkedList<MethodCall>(entry.getModifications());

                     if (modifications.size() > 0)
                     {
                        broadcastInvalidate(modifications, gtx, tx, ctx);
                     }
                     else
                     {
                        log.debug("Nothing to invalidate - no modifications in the transaction.");
                     }
                  }
                  break;
               case MethodDeclarations.optimisticPrepareMethod_id:
                  // here we just record the modifications but actually do the invalidate in commit.
                  gtx = ctx.getGlobalTransaction();
                  entry = txTable.get(gtx);
                  if (entry == null) throw new IllegalStateException("cannot find transaction entry for " + gtx);
                  modifications = new LinkedList<MethodCall>(entry.getModifications());

                  if (modifications.size() > 0)
                  {
                     txMods.put(gtx, modifications);
                  }
                  break;
               case MethodDeclarations.commitMethod_id:
                  if (optimistic)
                  {
                     gtx = ctx.getGlobalTransaction();
                     modifications = txMods.remove(gtx);
                     broadcastInvalidate(modifications, gtx, tx, ctx);
                     log.debug("Committing.  Broadcasting invalidations.");
                  }
                  break;
               case MethodDeclarations.rollbackMethod_id:
                  if (optimistic)
                  {
                     gtx = ctx.getGlobalTransaction();
                     txMods.remove(gtx);
                     log.debug("Caught a rollback.  Clearing modification in txMods");
                  }
                  break;
            }
         }

      }
      return retval;
   }

   private void broadcastInvalidate(List<MethodCall> modifications, GlobalTransaction gtx, Transaction tx, InvocationContext ctx)
   {
      if (containsPutForExternalRead(modifications))
      {
         log.debug("Modification list contains a putForExternalRead operation.  Not invalidating.");
      }
      else
      {
         try
         {
            invalidateModifications(modifications, configuration.isNodeLockingOptimistic() ? getWorkspace(gtx) : null, defaultSynchronous, ctx);
         }
         catch (Throwable t)
         {
            log.warn("Unable to broadcast evicts as a part of the prepare phase.  Rolling back.", t);
            try
            {
               tx.setRollbackOnly();
            }
            catch (SystemException se)
            {
               throw new RuntimeException("setting tx rollback failed ", se);
            }
            if (t instanceof RuntimeException)
               throw (RuntimeException) t;
            else
               throw new RuntimeException("Unable to broadcast invalidation messages", t);
         }
      }
   }

   private boolean containsPutForExternalRead(List<MethodCall> l)
   {
      if (l == null) return false;
      
      for (MethodCall m : l)
         if (m.getMethodId() == MethodDeclarations.putForExternalReadMethodLocal_id || m.getMethodId() == MethodDeclarations.putForExternalReadVersionedMethodLocal_id)
            return true;
      return false;
   }

   public long getInvalidations()
   {
      return m_invalidations;
   }

   public void resetStatistics()
   {
      m_invalidations = 0;
   }

   public Map<String, Object> dumpStatistics()
   {
      Map<String, Object> retval = new HashMap<String, Object>();
      retval.put("Invalidations", m_invalidations);
      return retval;
   }

   /**
    * Same as <code>invalidateAcrossCluster(fqn, workspace, defaultSynchronous)</code>
    *
    * @deprecated use {@link #invalidateAcrossCluster(Fqn,TransactionWorkspace,boolean,InvocationContext)
    */
   protected void invalidateAcrossCluster(Fqn fqn, TransactionWorkspace workspace, InvocationContext ctx) throws Throwable
   {
      invalidateAcrossCluster(fqn, workspace, defaultSynchronous, ctx);
   }

   protected void invalidateAcrossCluster(Fqn fqn, TransactionWorkspace workspace, boolean synchronous, InvocationContext ctx) throws Throwable
   {
      // increment invalidations counter if statistics maintained
      if (configuration.getExposeManagementStatistics() && getStatisticsEnabled())
         m_invalidations++;

      // only propagate version details if we're using explicit versioning.
      /*
      MethodCall call = workspace != null && !workspace.isVersioningImplicit() ?
              MethodCallFactory.create(MethodDeclarations.evictVersionedNodeMethodLocal, fqn, workspace.getNode(fqn).getVersion()) :
              MethodCallFactory.create(MethodDeclarations.evictNodeMethodLocal, fqn);
              */
      MethodCall call = MethodCallFactory.create(MethodDeclarations.invalidateMethodLocal, fqn, getNodeVersion(workspace, fqn));

      if (log.isDebugEnabled()) log.debug("Cache [" + cache.getLocalAddress() + "] replicating " + call);
      // voila, invalidated!
      replicateCall(call, synchronous, ctx.getOptionOverrides());
   }

   protected DataVersion getNodeVersion(TransactionWorkspace w, Fqn f)
   {
      if (w == null) return null;
      WorkspaceNode wn = w.getNode(f);
      DataVersion v = wn.getVersion();

      if (wn.isVersioningImplicit())
      {
         // then send back an incremented version
         v = ((DefaultDataVersion) v).increment();
      }

      return v;
   }

   /**
    * Same as <code>invalidateModifications(modifications, workspace, defaultSynchronous)</code>
    *
    * @deprecated use {@link #invalidateAcrossCluster(org.jboss.cache.Fqn,org.jboss.cache.optimistic.TransactionWorkspace,boolean,InvocationContext)}
    */
   protected void invalidateModifications(List<MethodCall> modifications, TransactionWorkspace workspace, InvocationContext ctx) throws Throwable
   {
      invalidateModifications(modifications, workspace, defaultSynchronous, ctx);
   }

   protected void invalidateModifications(List<MethodCall> modifications, TransactionWorkspace workspace, boolean synchronous, InvocationContext ctx) throws Throwable
   {
      // optimise the calls list here.
      Set<Fqn> modifiedFqns = optimisedIterator(modifications);
      for (Fqn fqn : modifiedFqns) invalidateAcrossCluster(fqn, workspace, synchronous, ctx);
   }

   protected TransactionWorkspace getWorkspace(GlobalTransaction gtx)
   {
      OptimisticTransactionEntry entry = (OptimisticTransactionEntry) txTable.get(gtx);
      return entry.getTransactionWorkSpace();
   }

   @SuppressWarnings("unchecked")
   protected void findAndAddFqns(Object[] objects, Set<Fqn> fqns, boolean isMove)
   {
      if (isMove)
      {
         Fqn f = (Fqn) objects[0];
         fqns.add(f);
         // now if this is a "move" operation, then we also have another Fqn -
         Object le = f.getLastElement();
         Fqn parent = (Fqn) objects[1];
         fqns.add(new Fqn(parent, le));
      }
      else
      {
         // it *should* be the 2nd param...
         fqns.add((Fqn) objects[1]);
      }
   }

   /**
    * Removes non-crud methods, plus clobs together common calls to Fqn's.
    * E.g, if we have put("/a/b", "1", "2") followed by a put("/a/b", "3",
    * "4") we should only evict "/a/b" once.
    *
    * @param list
    * @return Iterator containing a unique set of Fqns of crud methods in this tx
    */
   protected Set<Fqn> optimisedIterator(List<MethodCall> list)
   {
      Set<Fqn> fqns = new HashSet<Fqn>();
      if (list != null)
      {
         for (MethodCall mc : list)
         {
            if (MethodDeclarations.isCrudMethod(mc.getMethodId()))
            {
               findAndAddFqns(mc.getArgs(), fqns, MethodDeclarations.moveMethodLocal_id == mc.getMethodId());
            }
         }
      }
      return fqns;
   }
}
