/*
 * JBoss, Home of Professional Open Source
 *
 * Distributable under LGPL license.
 * See terms of license at gnu.org.
 */
package org.jboss.cache.interceptors;

import org.jboss.cache.CacheException;
import org.jboss.cache.CacheSPI;
import org.jboss.cache.Fqn;
import org.jboss.cache.InvocationContext;
import org.jboss.cache.NodeSPI;
import org.jboss.cache.optimistic.TransactionWorkspace;
import org.jboss.cache.transaction.GlobalTransaction;
import org.jboss.cache.transaction.OptimisticTransactionEntry;
import org.jboss.cache.transaction.TransactionTable;

import javax.transaction.Transaction;
import javax.transaction.TransactionManager;
import java.util.List;

/**
 * Abstract interceptor for optimistic locking
 *
 * @author <a href="mailto:manik@jboss.org">Manik Surtani (manik@jboss.org)</a>
 */
public class OptimisticInterceptor extends Interceptor
{
   protected TransactionManager txManager = null;
   protected TransactionTable txTable = null;
   protected boolean trace;

   public void setCache(CacheSPI cache)
   {
      super.setCache(cache);
      txManager = cache.getTransactionManager();
      txTable = cache.getTransactionTable();
      trace = log != null && log.isTraceEnabled();
   }

   protected TransactionWorkspace getTransactionWorkspace(GlobalTransaction gtx) throws CacheException
   {
      OptimisticTransactionEntry transactionEntry = (OptimisticTransactionEntry) txTable.get(gtx);

      if (transactionEntry == null)
      {
         throw new CacheException("Unable to map global transaction " + gtx + " to transaction entry when trying to retrieve transaction workspace.");
      }

      // try and get the workspace from the transaction
      return transactionEntry.getTransactionWorkSpace();
   }

   /**
    * Adds the Fqn of the node as well as all children and childrens children to the list.
    *
    * @param list
    * @param n
    */
   protected void greedyGetFqns(List<Fqn> list, NodeSPI<?, ?> n, Fqn newBase)
   {
      list.add(n.getFqn());
      Fqn newFqn = new Fqn(newBase, n.getFqn().getLastElement());
      list.add(newFqn);

      for (NodeSPI child : n.getChildrenDirect())
      {
         greedyGetFqns(list, child, newFqn);
      }
   }

   /**
    * @return the {@link org.jboss.cache.transaction.GlobalTransaction}, extracted from the current {@link org.jboss.cache.InvocationContext}.
    * @throws CacheException if the {@link org.jboss.cache.transaction.GlobalTransaction} or {@link javax.transaction.Transaction} associated with the
    *                        {@link org.jboss.cache.InvocationContext} is null.
    */
   protected GlobalTransaction getGlobalTransaction(InvocationContext ctx) throws CacheException
   {
      Transaction tx = ctx.getTransaction();
      if (tx == null) throw new CacheException("Transaction associated with the current invocation is null!");
      GlobalTransaction gtx = ctx.getGlobalTransaction();
      if (gtx == null) throw new CacheException("GlobalTransaction associated with the current invocation is null!");
      return gtx;
   }


}
