/*
 * JBoss, Home of Professional Open Source
 *
 * Distributable under LGPL license.
 * See terms of license at gnu.org.
 */
package org.jboss.cache.interceptors;

import org.jboss.cache.CacheException;
import org.jboss.cache.CacheSPI;
import org.jboss.cache.InvocationContext;
import org.jboss.cache.NodeSPI;
import org.jboss.cache.lock.NodeLock;
import org.jboss.cache.marshall.MethodCall;
import org.jboss.cache.marshall.MethodDeclarations;
import org.jboss.cache.optimistic.TransactionWorkspace;
import org.jboss.cache.optimistic.WorkspaceNode;
import org.jboss.cache.transaction.GlobalTransaction;
import org.jboss.cache.transaction.TransactionEntry;

/**
 * Locks nodes during transaction boundaries.  Only affects prepare/commit/rollback method calls; other method calls
 * are simply passed up the interceptor stack.
 *
 * @author <a href="mailto:manik@jboss.org">Manik Surtani (manik@jboss.org)</a>
 * @author <a href="mailto:stevew@jofti.com">Steve Woodcock (stevew@jofti.com)</a>
 */
public class OptimisticLockingInterceptor extends OptimisticInterceptor
{
   private long lockAcquisitionTimeout;

   public void setCache(CacheSPI cache)
   {
      super.setCache(cache);
      lockAcquisitionTimeout = cache.getConfiguration().getLockAcquisitionTimeout();
   }

   public Object invoke(InvocationContext ctx) throws Throwable
   {
      MethodCall m = ctx.getMethodCall();
      Object retval = null;

      //we are interested in the prepare/commit/rollback
      //this is irrespective of whether we are local or remote
      switch (m.getMethodId())
      {
         case MethodDeclarations.optimisticPrepareMethod_id:
            //try and acquire the locks - before passing on
            GlobalTransaction gtx = getGlobalTransaction(ctx);
            long timeout = lockAcquisitionTimeout;
            if (ctx.getOptionOverrides() != null 
                  && ctx.getOptionOverrides().getLockAcquisitionTimeout() >= 0)
            {
               timeout = ctx.getOptionOverrides().getLockAcquisitionTimeout();
            }
            try
            {
               lockNodes(gtx, timeout);
            }
            catch (Throwable e)
            {
               log.debug("Caught exception attempting to lock nodes ", e);
               //we have failed - set to rollback and throw exception
               try
               {
                  unlock(gtx);
               }
               catch (Throwable t)
               {
                  // we have failed to unlock - now what?
                  log.error("Failed to unlock nodes, after failing to lock nodes during a prepare!  Locks are possibly in a very inconsistent state now!", t);
               }
               throw e;
            }

            // locks have acquired so lets pass on up
            retval = super.invoke(ctx);
            break;
         case MethodDeclarations.commitMethod_id:
         case MethodDeclarations.rollbackMethod_id:
            // we need to let the stack run its commits or rollbacks first -
            // we unlock last - even if an exception occurs
            try
            {
               retval = super.invoke(ctx);
            }
            finally
            {
               try
               {
                  unlock(getGlobalTransaction(ctx));
               }
               catch (Exception e)
               {
                  // we have failed to unlock - now what?
                  log.error("Failed to unlock nodes after a commit or rollback!  Locks are possibly in a very inconsistent state now!", e);
               }
            }
            break;
         case MethodDeclarations.lockMethodLocal_id:
            // bail out if _lock() is being called on the tree cache... this should never be called with o/l enabled.
            throw new CacheException("_lock() passed up the interceptor stack when Optimistic Locking is used.  This is NOT supported.");
         default:
            //we do not care, just pass up the chain.
            retval = super.invoke(ctx);
            break;
      }

      return retval;
   }

   /**
    * Locks all nodes held in the transaction workspace registered with the given global transaction.
    *
    * @param gtx global transaction which contains a workspace
    */
   private void lockNodes(GlobalTransaction gtx, long timeout) throws InterruptedException
   {
      TransactionWorkspace<?, ?> workspace = getTransactionWorkspace(gtx);
      if (log.isDebugEnabled()) log.debug("Locking nodes in transaction workspace for GlobalTransaction " + gtx);

      for (WorkspaceNode workspaceNode : workspace.getNodes().values())
      {
         NodeSPI node = workspaceNode.getNode();
         boolean acquired = node.getLock().acquire(gtx, lockAcquisitionTimeout, NodeLock.LockType.WRITE);
         if (acquired)
         {
            if (trace) log.trace("Acquired lock on node " + node.getFqn());
            cache.getTransactionTable().addLock(gtx, node.getLock());
         }
         else
         {
            throw new CacheException("Unable to acquire lock on node " + node.getFqn());
         }

      }
   }

   /**
    * Releases all locks held by the specified global transaction.
    *
    * @param gtx which holds locks
    */
   private void unlock(GlobalTransaction gtx)
   {
      TransactionEntry entry = txTable.get(gtx);
      entry.releaseAllLocksFIFO(gtx);
   }

}
