package org.jboss.cache.interceptors;

import org.jboss.cache.CacheSPI;
import org.jboss.cache.Fqn;
import org.jboss.cache.InvocationContext;
import org.jboss.cache.NodeSPI;
import org.jboss.cache.loader.CacheLoader;
import org.jboss.cache.marshall.MethodCall;
import org.jboss.cache.marshall.MethodDeclarations;

import java.util.Collections;
import java.util.HashMap;
import java.util.Map;
import java.util.concurrent.atomic.AtomicLong;

/**
 * Writes evicted nodes back to the store on the way in through the
 * CacheLoader, either before each method call (no TXs), or at TX commit.
 *
 * @author <a href="mailto:{hmesha@novell.com}">{Hany Mesha}</a>
 * @version $Id: PassivationInterceptor.java 4138 2007-07-13 14:31:42Z msurtani $
 */
public class PassivationInterceptor extends Interceptor implements PassivationInterceptorMBean
{

   protected CacheLoader loader = null;
   private AtomicLong m_passivations = new AtomicLong(0);

   public void setCache(CacheSPI cache)
   {
      super.setCache(cache);
      this.loader = cache.getCacheLoaderManager().getCacheLoader();
   }

   /**
    * Notifies the cache instance listeners that the evicted node is about to
    * be passivated and stores the evicted node and its attributes back to the
    * store using the CacheLoader.
    *
    * @return
    * @throws Throwable
    */
   public Object invoke(InvocationContext ctx) throws Throwable
   {
      MethodCall m = ctx.getMethodCall();
      // hmesha- We don't need to handle transaction during passivation since
      // passivation happens local to a node and never replicated

      // evict() method need to be applied to the CacheLoader before passing the call on
      if (m.getMethodId() == MethodDeclarations.evictNodeMethodLocal_id)
      {
         Object[] args = m.getArgs();
         Fqn fqn = (Fqn) args[0];
         try
         {
            // synchronizations now handled in the cache loader implementation
//            synchronized (this)
//            {
            // evict method local doesn't hold attributes therefore we have
            // to get them manually
            Map attributes = getNodeAttributes(fqn);
            // notify listeners that this node is about to be passivated
            cache.getNotifier().notifyNodePassivated(fqn, true, attributes, ctx);

            loader.put(fqn, attributes);

            cache.getNotifier().notifyNodePassivated(fqn, false, Collections.emptyMap(), ctx);
//            }

            if (getStatisticsEnabled() && configuration.getExposeManagementStatistics())
            {
               m_passivations.getAndIncrement();
            }
         }
         catch (NodeNotLoadedException e)
         {
            if (log.isTraceEnabled())
            {
               log.trace("Node " + fqn + " not loaded in memory; passivation skipped");
            }
         }
      }

      return super.invoke(ctx);
   }

   public long getPassivations()
   {
      return m_passivations.get();
   }

   public void resetStatistics()
   {
      m_passivations.set(0);
   }

   public Map<String, Object> dumpStatistics()
   {
      Map<String, Object> retval = new HashMap<String, Object>();
      retval.put("Passivations", m_passivations.get());
      return retval;
   }

   /**
    * Returns attributes for a node.
    */
   private Map getNodeAttributes(Fqn fqn) throws NodeNotLoadedException
   {
      if (fqn == null)
      {
         throw new NodeNotLoadedException();
      }
      NodeSPI n = cache.peek(fqn, true);

      if (n != null)
      {
         return n.getDataDirect();
      }
      else
      {
         throw new NodeNotLoadedException();
      }
   }

   private static class NodeNotLoadedException extends Exception
   {
      /**
       * The serialVersionUID
       */
      private static final long serialVersionUID = -4078972305344328905L;
   }
}
