/*
 * JBoss, Home of Professional Open Source
 *
 * Distributable under LGPL license.
 * See terms of license at gnu.org.
 */
package org.jboss.cache.interceptors;

import org.jboss.cache.CacheImpl;
import org.jboss.cache.CacheSPI;
import org.jboss.cache.Fqn;
import org.jboss.cache.InvocationContext;
import org.jboss.cache.NodeSPI;
import org.jboss.cache.lock.IsolationLevel;
import org.jboss.cache.lock.LockingException;
import org.jboss.cache.lock.NodeLock;
import org.jboss.cache.lock.TimeoutException;
import org.jboss.cache.marshall.MethodCall;
import org.jboss.cache.marshall.MethodDeclarations;
import org.jboss.cache.transaction.GlobalTransaction;
import org.jboss.cache.transaction.TransactionEntry;
import org.jboss.cache.transaction.TransactionTable;

import java.util.Collections;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Set;

/**
 * An interceptor that handles locking. When a TX is associated, we register
 * for TX completion and unlock the locks acquired within the scope of the TX.
 * When no TX is present, we keep track of the locks acquired during the
 * current method and unlock when the method returns.
 *
 * @author Bela Ban
 * @version $Id: PessimisticLockInterceptor.java 4659 2007-10-22 10:45:50Z mircea.markus $
 */
public class PessimisticLockInterceptor extends Interceptor
{
   private TransactionTable tx_table = null;

   /**
    * Map<Object, java.util.List>. Keys = threads, values = lists of locks held by that thread
    */
   private Map<Thread, List<NodeLock>> lock_table;
   private long lock_acquisition_timeout;
   private LockManager lockManager = new LockManager();


   public void setCache(CacheSPI cache)
   {
      super.setCache(cache);
      tx_table = cache.getTransactionTable();
      lock_table = cache.getLockTable();
      lock_acquisition_timeout = cache.getConfiguration().getLockAcquisitionTimeout();
   }


   public Object invoke(InvocationContext ctx) throws Throwable
   {
      MethodCall m = ctx.getMethodCall();
      Fqn fqn = null;
      NodeLock.LockType lock_type = NodeLock.LockType.NONE;
      Object[] args = m.getArgs();
      boolean lockNecessary = false;
      boolean locksAlreadyObtained = false;

      if (log.isTraceEnabled()) log.trace("PessimisticLockInterceptor invoked for method " + m);
      if (ctx.getOptionOverrides() != null && ctx.getOptionOverrides().isSuppressLocking())
      {
         log.trace("Suppressing locking");
         switch (m.getMethodId())
         {
            case MethodDeclarations.putDataMethodLocal_id:
            case MethodDeclarations.putDataEraseMethodLocal_id:
            case MethodDeclarations.putKeyValMethodLocal_id:
               log.trace("Creating nodes if necessary");
               createNodes((Fqn) args[1], ctx.getGlobalTransaction());
               break;
         }

         return super.invoke(ctx);
      }

      /** List<IdentityLock> locks. Locks acquired during the current method; will be released later by UnlockInterceptor.
       *  This list is only populated when there is no TX, otherwise the TransactionTable maintains the locks
       * (keyed by TX) */
      // List locks=null;

      boolean recursive = false;
      boolean createIfNotExists = false;
      boolean zeroLockTimeout = false;// only used if the call is an evict() call.  See JBCACHE-794
      boolean isDeleteOperation = false;// needed for JBCACHE-871
      boolean isEvictOperation = false;
      boolean isRemoveDataOperation = false;

      // 1. Determine the type of lock (read, write, or none) depending on the method. If no lock is required, invoke
      //    the method, then return immediately
      //    Set the Fqn
      switch (m.getMethodId())
      {
         case MethodDeclarations.moveMethodLocal_id:
            fqn = (Fqn) args[0];
            obtainLocksForMove(ctx, fqn, (Fqn) args[1]);
            locksAlreadyObtained = true;
            lockNecessary = true;
            isDeleteOperation = true;
            break;
         case MethodDeclarations.putForExternalReadMethodLocal_id:
            createIfNotExists = true;
            fqn = (Fqn) args[1];
            lock_type = NodeLock.LockType.WRITE;
            zeroLockTimeout = true;
            break;
         case MethodDeclarations.removeNodeMethodLocal_id:
            isDeleteOperation = true;
            fqn = (Fqn) args[1];
            lock_type = NodeLock.LockType.WRITE;
            recursive = true;// remove node and *all* child nodes
            break;
         case MethodDeclarations.putDataMethodLocal_id:
         case MethodDeclarations.putDataEraseMethodLocal_id:
         case MethodDeclarations.putKeyValMethodLocal_id:
            createIfNotExists = true;
            fqn = (Fqn) args[1];
            lock_type = NodeLock.LockType.WRITE;
            break;
         case MethodDeclarations.removeKeyMethodLocal_id:
         case MethodDeclarations.removeDataMethodLocal_id:
            isRemoveDataOperation = true;
         case MethodDeclarations.addChildMethodLocal_id:
            fqn = (Fqn) args[1];
            lock_type = NodeLock.LockType.WRITE;
            break;
         case MethodDeclarations.evictNodeMethodLocal_id:
            zeroLockTimeout = true;
            fqn = (Fqn) args[0];
            lock_type = NodeLock.LockType.WRITE;
            isEvictOperation = true;
            break;
         case MethodDeclarations.getKeyValueMethodLocal_id:
         case MethodDeclarations.getNodeMethodLocal_id:
         case MethodDeclarations.getKeysMethodLocal_id:
         case MethodDeclarations.getChildrenNamesMethodLocal_id:
         case MethodDeclarations.releaseAllLocksMethodLocal_id:
         case MethodDeclarations.printMethodLocal_id:
            fqn = (Fqn) args[0];
            lock_type = NodeLock.LockType.READ;
            break;
         case MethodDeclarations.lockMethodLocal_id:
            fqn = (Fqn) args[0];
            lock_type = (NodeLock.LockType) args[1];
            recursive = (Boolean) args[2];
            break;
         case MethodDeclarations.commitMethod_id:
            // commit propagated up from the tx interceptor
            commit(ctx.getGlobalTransaction());
            break;
         case MethodDeclarations.rollbackMethod_id:
            // rollback propagated up from the tx interceptor
            rollback(ctx.getGlobalTransaction());
            break;
         default:
            if (isOnePhaseCommitPrepareMehod(m))
            {
               // commit propagated up from the tx interceptor
               commit(ctx.getGlobalTransaction());
            }
            break;
      }

      // Lock the node (must be either read or write if we get here)
      // If no TX: add each acquired lock to the list of locks for this method (locks)
      // If TX: [merge code from TransactionInterceptor]: register with TxManager, on commit/rollback,
      // release the locks for the given TX
      if (fqn != null)
      {
         if (!locksAlreadyObtained)
         {
            long timeout = zeroLockTimeout ? 0 : getLockAcquisitionTimeout(ctx);
            do
            {
               lock(ctx, fqn, lock_type, recursive, createIfNotExists, timeout, isDeleteOperation, isEvictOperation, isRemoveDataOperation);
            }
            while (createIfNotExists && cache.peek(fqn, false) == null);// keep trying until we have the lock (fixes concurrent remove())
         }
      }
      else if (!lockNecessary)
      {
         if (log.isTraceEnabled())
         {
            log.trace("bypassed locking as method " + m.getName() + "() doesn't require locking");
         }
      }
      if (m.getMethodId() == MethodDeclarations.lockMethodLocal_id)
      {
         return null;
      }
      Object o = super.invoke(ctx);
      // FIXME this should be done in UnlockInterceptor, but I didn't want
      // to add the removedNodes map to CacheImpl
      if (isDeleteOperation && ctx.getGlobalTransaction() == null)
      {
         //cache.getRemovedNodesMap().remove(fqn);
         //cache.peek(fqn);
         // do a REAL remove here.
         NodeSPI n = cache.peek(fqn, true);
         if (n != null)
         {
            lockManager.getLock(n).releaseAll(Thread.currentThread());
         }
         //CacheIml._move internally removes the source node. 'if clause' solve the scenario when a node is
         // moved in the same place, i.e. it will be removed from the source - which is also destination - so it ends
         // up being removed for good
         if (!(MethodDeclarations.moveMethodLocal_id == m.getMethodId()))
         {
            ((CacheImpl) cache).realRemove(fqn, true);
         }
      }
      else
      if (m.getMethodId() == MethodDeclarations.commitMethod_id || isOnePhaseCommitPrepareMehod(m) || m.getMethodId() == MethodDeclarations.rollbackMethod_id)
      {
         cleanup(ctx.getGlobalTransaction());
      }
      return o;
   }

   private long getLockAcquisitionTimeout(InvocationContext ctx)
   {
      long timeout = lock_acquisition_timeout;
      if (ctx.getOptionOverrides() != null
              && ctx.getOptionOverrides().getLockAcquisitionTimeout() >= 0)
      {
         timeout = ctx.getOptionOverrides().getLockAcquisitionTimeout();
      }
      return timeout;
   }


   private void obtainLocksForMove(InvocationContext ctx, Fqn node, Fqn parent) throws InterruptedException
   {
      // parent node (new parent) and current node's existing parent should both get RLs.
      // node should have a WL.

      long timeout = getLockAcquisitionTimeout(ctx);

      // this call will ensure the node gets a WL and it's current parent gets RL.
      if (log.isTraceEnabled()) log.trace("Attempting to get WL on node to be moved [" + node + "]");
      lock(ctx, node, NodeLock.LockType.WRITE, true, false, timeout, true, false, false);

      //now for an RL for the new parent.
      if (log.isTraceEnabled()) log.trace("Attempting to get RL on new parent [" + parent + "]");
      lock(ctx, parent, NodeLock.LockType.READ, true, false, timeout, false, false, false);
   }


   /**
    * Locks a given node.
    *
    * @param fqn
    * @param lock_type DataNode.LOCK_TYPE_READ, DataNode.LOCK_TYPE_WRITE or DataNode.LOCK_TYPE_NONE
    * @param recursive Lock children recursively
    */
   private void lock(InvocationContext ctx, Fqn fqn, NodeLock.LockType lock_type, boolean recursive, boolean createIfNotExists, long timeout, boolean isDeleteOperation, boolean isEvictionOperation, boolean isRemoveDataOperation)
           throws TimeoutException, LockingException, InterruptedException
   {
      NodeSPI n;
      NodeSPI child_node;
      Object child_name;
      Thread currentThread = Thread.currentThread();
      GlobalTransaction gtx = ctx.getGlobalTransaction();
      Object owner = (gtx != null) ? gtx : currentThread;
      int treeNodeSize;

      if (log.isTraceEnabled()) log.trace("Attempting to lock node " + fqn + " for owner " + owner);

      if (fqn == null)
      {
         log.error("fqn is null - this should not be the case");
         return;
      }

      if (configuration.getIsolationLevel() == IsolationLevel.NONE)
      {
         lock_type = NodeLock.LockType.NONE;
      }

      n = cache.getRoot();
      treeNodeSize = fqn.size();
      for (int i = -1; i < treeNodeSize; i++)
      {
         if (i == -1)
         {
            // this is the root node
            child_name = Fqn.ROOT.getLastElement();
            child_node = n;
         }
         else
         {
            child_name = fqn.get(i);
            child_node = n.getChildDirect(child_name);
         }

         boolean created = false;
         if (child_node == null && createIfNotExists)
         {
            child_node = n.addChildDirect(new Fqn(child_name));
            created = true;
         }

         if (child_node == null)
         {
            if (log.isTraceEnabled())
            {
               log.trace("failed to find or create child " + child_name + " of node " + n);
            }
            return;
         }

         NodeLock.LockType lockTypeRequired;
         if (lock_type == NodeLock.LockType.NONE)
         {
            n = child_node;
            continue;
         }
         else
         {
            if (created || writeLockNeeded(ctx, lock_type, i, treeNodeSize, isEvictionOperation, isDeleteOperation, createIfNotExists, isRemoveDataOperation, fqn, child_node))
            {
               lockTypeRequired = NodeLock.LockType.WRITE;

            }
            else
            {
               lockTypeRequired = NodeLock.LockType.READ;
            }
         }

         // reverse the "remove" if the node has been previously removed in the same tx, if this operation is a put()
         if (gtx != null && needToReverseRemove(child_node, tx_table.get(gtx), lock_type, isDeleteOperation, createIfNotExists))
         {
            reverseRemove(child_node);
         }

         // actually acquire the lock we need.
         acquireNodeLock(child_node, owner, gtx, lockTypeRequired, timeout);

         if (recursive && isTargetNode(i, treeNodeSize))
         {
            Set<NodeLock> acquired_locks = lockManager.acquireAll(child_node, owner, lock_type, timeout);
            if (acquired_locks.size() > 0)
            {
               if (gtx != null)
               {
                  cache.getTransactionTable().addLocks(gtx, acquired_locks);
               }
               else
               {
                  List<NodeLock> locks = getLocks(currentThread);
                  locks.addAll(acquired_locks);
               }
            }
         }
         n = child_node;
      }

      // Add the Fqn to be removed to the transaction entry so we can clean up after ourselves during commit/rollback
      if (isDeleteOperation && gtx != null) cache.getTransactionTable().get(gtx).addRemovedNode(fqn);
   }

   private boolean needToReverseRemove(NodeSPI n, TransactionEntry te, NodeLock.LockType lockTypeRequested, boolean isRemoveOperation, boolean createIfNotExists)
   {
      return !isRemoveOperation && createIfNotExists && lockTypeRequested == NodeLock.LockType.WRITE && n.isDeleted() && te.getRemovedNodes().contains(n.getFqn());
   }

   private void reverseRemove(NodeSPI n)
   {
      n.markAsDeleted(false);
   }

   private boolean writeLockNeeded(InvocationContext ctx, NodeLock.LockType lock_type, int currentNodeIndex, int treeNodeSize, boolean isEvictOperation, boolean isRemoveOperation, boolean isPutOperation, boolean isRemoveDataOperation, Fqn targetFqn, NodeSPI currentNode)
   {
      // write lock forced!!
      boolean isTargetNode = isTargetNode(currentNodeIndex, treeNodeSize);

      if (ctx.getOptionOverrides().isForceWriteLock() && isTargetNode) return true;

      if (currentNode.isLockForChildInsertRemove())
      {
         if (isRemoveOperation && currentNodeIndex == treeNodeSize - 2)
         {
            return true;// we're doing a remove and we've reached the PARENT node of the target to be removed.
         }

         if (!isTargetNode && cache.peek(new Fqn(currentNode.getFqn(), targetFqn.get(currentNodeIndex + 1)), false) == null)
         {
            return isPutOperation;// we're at a node in the tree, not yet at the target node, and we need to create the next node.  So we need a WL here.
         }
      }
      return lock_type == NodeLock.LockType.WRITE && isTargetNode && (isPutOperation || isRemoveOperation || isEvictOperation || isRemoveDataOperation);//normal operation, write lock explicitly requested and this is the target to be written to.
   }

   private boolean isTargetNode(int nodePosition, int treeNodeSize)
   {
      return nodePosition == (treeNodeSize - 1);
   }

   private void acquireNodeLock(NodeSPI node, Object owner, GlobalTransaction gtx, NodeLock.LockType lock_type, long lock_timeout) throws LockingException, TimeoutException, InterruptedException
   {
      boolean acquired = lockManager.acquire(node, owner, lock_type, lock_timeout);
      if (acquired)
      {
         // Record the lock for release on method return or tx commit/rollback
         recordNodeLock(gtx, lockManager.getLock(node));
      }
      else
      {
         //if (log.isDebugEnabled()) log.debug("Unable to acquire lock on node " + node.getFqn());
         //throw new LockingException("Could not acquite lock for " + node.getFqn());
      }
   }

   private void recordNodeLock(GlobalTransaction gtx, NodeLock lock)
   {
      if (gtx != null)
      {
         // add the lock to the list of locks maintained for this transaction
         // (needed for release of locks on commit or rollback)
         cache.getTransactionTable().addLock(gtx, lock);
      }
      else
      {
         Thread currentThread = Thread.currentThread();
         List<NodeLock> locks = getLocks(currentThread);
         if (!locks.contains(lock))
         {
            locks.add(lock);
            lock_table.put(currentThread, locks);
         }
      }
   }

   private List<NodeLock> getLocks(Thread currentThread)
   {
      // This sort of looks like a get/put race condition, but
      // since we key off the Thread, it's not
      List<NodeLock> locks = lock_table.get(currentThread);
      if (locks == null)
      {
         locks = Collections.synchronizedList(new LinkedList<NodeLock>());
         lock_table.put(currentThread, locks);
      }
      return locks;
   }


   private void createNodes(Fqn fqn, GlobalTransaction gtx)
   {
      int treeNodeSize;
      if ((treeNodeSize = fqn.size()) == 0) return;
      NodeSPI n = cache.getRoot();
      for (int i = 0; i < treeNodeSize; i++)
      {
         Object child_name = fqn.get(i);
         Fqn childFqn = new Fqn(child_name);

         NodeSPI child_node = n.getChildDirect(childFqn);
         if (child_node == null) child_node = n.addChildDirect(childFqn);
         // test if this node needs to be 'undeleted'
         // reverse the "remove" if the node has been previously removed in the same tx, if this operation is a put()
         if (gtx != null && needToReverseRemove(child_node, tx_table.get(gtx), NodeLock.LockType.WRITE, false, true))
         {
            reverseRemove(child_node);
         }

         if (child_node == null)
         {
            if (log.isTraceEnabled())
            {
               log.trace("failed to find or create child " + child_name + " of node " + n.getFqn());
            }
            return;
         }
         n = child_node;
      }
   }

   /**
    * Remove all locks held by <tt>tx</tt>, remove the transaction from the transaction table
    *
    * @param gtx
    */
   private void commit(GlobalTransaction gtx)
   {
      if (log.isTraceEnabled())
      {
         log.trace("committing cache with gtx " + gtx);
      }

      TransactionEntry entry = tx_table.get(gtx);
      if (entry == null)
      {
         log.error("entry for transaction " + gtx + " not found (maybe already committed)");
         return;
      }

      // first remove nodes that should be deleted.
      Iterator removedNodes = entry.getRemovedNodes().iterator();
      CacheImpl cacheImpl = (CacheImpl) cache;
      while (removedNodes.hasNext())
      {
         Fqn f = (Fqn) removedNodes.next();
         cacheImpl.realRemove(f, false);
      }
   }

   private void cleanup(GlobalTransaction gtx)
   {
      TransactionEntry entry = tx_table.get(gtx);
      // Let's do it in stack style, LIFO
      entry.releaseAllLocksLIFO(gtx);

/*
      Transaction ltx = entry.getTransaction();
      if (log.isTraceEnabled())
      {
         log.trace("removing local transaction " + ltx + " and global transaction " + gtx);
      }
      tx_table.remove(ltx);
      tx_table.remove(gtx);
*/
   }

   /**
    * Revert all changes made inside this TX: invoke all method calls of the undo-ops
    * list. Then release all locks and remove the TX from the transaction table.
    * <ol>
    * <li>Revert all modifications done in the current TX<li/>
    * <li>Release all locks held by the current TX</li>
    * <li>Remove all temporary nodes created by the current TX</li>
    * </ol>
    *
    * @param tx
    */
   private void rollback(GlobalTransaction tx)
   {
      TransactionEntry entry = tx_table.get(tx);

      if (log.isTraceEnabled())
      {
         log.trace("called to rollback cache with GlobalTransaction=" + tx);
      }

      if (entry == null)
      {
         log.error("entry for transaction " + tx + " not found (transaction has possibly already been rolled back)");
         return;
      }

      Iterator removedNodes = entry.getRemovedNodes().iterator();
      CacheImpl cacheImpl = (CacheImpl) cache;
      while (removedNodes.hasNext())
      {
         Fqn f = (Fqn) removedNodes.next();
         cacheImpl.realRemove(f, false);

      }

      // 1. Revert the modifications by running the undo-op list in reverse. This *cannot* throw any exceptions !
      entry.undoOperations(cache);

      // This was removed as we don't use temporary nodes anymore; we now create undo-operations on put(), e.g.
      // put(/a/b/c) on /a, create b and c, plus undo operations _remove(a/b/c) and _remove(/a/b)

      // 2. Remove all temporary nodes. Need to do it backwards since node is LIFO.
      //      for(ListIterator it=new LinkedList(entry.getNodes()).listIterator(entry.getNodes().size());
      //          it.hasPrevious();) {
      //         node_name=(Fqn)it.previous();
      //         try {
      //            cache._remove(tx, node_name, false);
      //         }
      //         catch(Throwable t) {
      //            log.error("failed removing node \"" + node_name + "\"", t);
      //         }
      //      }

      // 3. Finally, release all locks held by this TX
      // Let's do it in stack style, LIFO
      // Note that the lock could have been released already so don't panic.
   }

   private static class LockManager
   {
      boolean acquire(NodeSPI node, Object owner, NodeLock.LockType lockType, long timeout) throws InterruptedException
      {
         return getLock(node).acquire(owner, timeout, lockType);
      }

      NodeLock getLock(NodeSPI node)
      {
         return node.getLock();
      }

      public Set<NodeLock> acquireAll(NodeSPI node, Object owner, NodeLock.LockType lockType, long timeout) throws InterruptedException
      {
         return getLock(node).acquireAll(owner, timeout, lockType);
      }
   }

}
