/*
 * JBoss, Home of Professional Open Source
 *
 * Distributable under LGPL license.
 * See terms of license at gnu.org.
 */
package org.jboss.cache.interceptors;

import org.jboss.cache.CacheException;
import org.jboss.cache.CacheSPI;
import org.jboss.cache.InvocationContext;
import org.jboss.cache.ReplicationException;
import org.jboss.cache.config.Configuration;
import org.jboss.cache.config.Option;
import org.jboss.cache.marshall.MethodCall;
import org.jboss.cache.marshall.MethodCallFactory;
import org.jboss.cache.marshall.MethodDeclarations;
import org.jboss.cache.optimistic.DataVersion;
import org.jboss.cache.transaction.GlobalTransaction;
import org.jboss.cache.transaction.OptimisticTransactionEntry;
import org.jboss.cache.transaction.TransactionEntry;

import javax.transaction.Status;
import javax.transaction.Synchronization;
import javax.transaction.SystemException;
import javax.transaction.Transaction;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.concurrent.ConcurrentHashMap;

/**
 * This interceptor is the new default at the head of all interceptor chains,
 * and makes transactional attributes available to all interceptors in the chain.
 * This interceptor is also responsible for registering for synchronisation on
 * transaction completion.
 *
 * @author <a href="mailto:manik@jboss.org">Manik Surtani (manik@jboss.org)</a>
 * @author <a href="mailto:stevew@jofti.com">Steve Woodcock (stevew@jofti.com)</a>
 */
public class TxInterceptor extends BaseTransactionalContextInterceptor implements TxInterceptorMBean
{
   private final static Object NULL = new Object();

   /**
    * List <Transaction>that we have registered for
    */
   private Map transactions = new ConcurrentHashMap(16);
   private Map rollbackTransactions = new ConcurrentHashMap(16);
   private long m_prepares = 0;
   private long m_commits = 0;
   private long m_rollbacks = 0;


   /**
    * Set<GlobalTransaction> of GlobalTransactions that originated somewhere else (we didn't create them).
    * This is a result of a PREPARE phase. GlobalTransactions in this list should be ignored by this
    * interceptor when registering for TX completion
    */
   private Map remoteTransactions = new ConcurrentHashMap();

   public Object invoke(InvocationContext ctx) throws Throwable
   {
      MethodCall m = ctx.getMethodCall();

      // skip block/unblock calls
      if (MethodDeclarations.isBlockUnblockMethod(m.getMethodId())) return super.invoke(ctx);

      if (log.isTraceEnabled())
      {
         log.trace("(" + cache.getLocalAddress() + ") call on method [" + m + "]");
      }
      // bypass for buddy group org metod calls.
      if (MethodDeclarations.isBuddyGroupOrganisationMethod(m.getMethodId())) return super.invoke(ctx);

      boolean scrubTxsOnExit = false;
      Option optionOverride = ctx.getOptionOverrides();

      Object result = null;

      try
      {
         // first of all deal with tx methods - these are only going to be
         // prepare/commit/rollback called by a remote cache, since calling
         // such methods on CacheImpl directly would fail.

         if (MethodDeclarations.isTransactionLifecycleMethod(m.getMethodId()))
         {
            // this is a prepare, commit, or rollback.
            if (log.isDebugEnabled()) log.debug("Got gtx from invocation context " + ctx.getGlobalTransaction());

            if (ctx.getGlobalTransaction().isRemote()) remoteTransactions.put(ctx.getGlobalTransaction(), NULL);

            switch (m.getMethodId())
            {
               case MethodDeclarations.optimisticPrepareMethod_id:
               case MethodDeclarations.prepareMethod_id:
                  if (ctx.getGlobalTransaction().isRemote())
                  {
                     result = handleRemotePrepare(ctx, m);
                     scrubTxsOnExit = true;
                     if (configuration.getExposeManagementStatistics() && getStatisticsEnabled())
                     {
                        m_prepares++;
                     }
                  }
                  else
                  {
                     if (log.isTraceEnabled()) log.trace("received my own message (discarding it)");
                     result = null;
                  }
                  break;
               case MethodDeclarations.commitMethod_id:
               case MethodDeclarations.rollbackMethod_id:
                  if (ctx.getGlobalTransaction().isRemote())
                  {
                     result = handleRemoteCommitRollback(ctx);
                     scrubTxsOnExit = true;
                  }
                  else
                  {
                     if (log.isTraceEnabled()) log.trace("received my own message (discarding it)");
                     result = null;
                  }
                  break;
            }
         }
         else
         {
            // non-transaction lifecycle method.
            result = handleNonTxMethod(ctx);
         }
      }
      catch (Exception e)
      {
         if (optionOverride == null || !optionOverride.isFailSilently()) throw e;
         log.trace("There was a problem handling this request, but " +
                 "failSilently was set, so suppressing exception", e);
      }
      finally
      {
         // we should scrub txs after every call to prevent race conditions
         // basically any other call coming in on the same thread and hijacking any running tx's
         // was highlighted in JBCACHE-606

         if (scrubTxsOnExit)
         {
            setTransactionalContext(null, null, ctx);
         }
      }
      return result;
   }

   public long getPrepares()
   {
      return m_prepares;
   }

   public long getCommits()
   {
      return m_commits;
   }

   public long getRollbacks()
   {
      return m_rollbacks;
   }

   public void resetStatistics()
   {
      m_prepares = 0;
      m_commits = 0;
      m_rollbacks = 0;
   }

   public Map<String, Object> dumpStatistics()
   {
      Map<String, Object> retval = new HashMap<String, Object>(3);
      retval.put("Prepares", m_prepares);
      retval.put("Commits", m_commits);
      retval.put("Rollbacks", m_rollbacks);
      return retval;
   }

   private Object handleRemotePrepare(InvocationContext ctx, MethodCall m) throws Throwable
   {
      GlobalTransaction gtx = ctx.getGlobalTransaction();
      List<MethodCall> modifications = (List<MethodCall>) m.getArgs()[1];
      boolean onePhase = (Boolean) m.getArgs()[configuration.isNodeLockingOptimistic() ? 4 : 3];

      // Is there a local transaction associated with GTX ?
      Transaction ltx = txTable.getLocalTransaction(gtx);

      Transaction currentTx = txManager.getTransaction();
      Object retval = null;

      try
      {
         if (ltx == null)
         {
            if (currentTx != null) txManager.suspend();
            ltx = createLocalTxForGlobalTx(gtx, ctx);// creates new LTX and associates it with a GTX
            if (log.isDebugEnabled())
            {
               log.debug("Started new local TX as result of remote PREPARE: local TX=" + ltx + " (Status=" + ltx.getStatus() + "), global TX=" + gtx);
            }
         }
         else
         {
            //this should be valid
            if (!isValid(ltx)) throw new CacheException("Transaction " + ltx + " not in correct state to be prepared");

            //associate this thread with this ltx if this ltx is NOT the current tx.
            if (currentTx == null || !ltx.equals(currentTx))
            {
               txManager.suspend();
               txManager.resume(ltx);
            }
         }


         if (log.isTraceEnabled())
         {
            log.trace("Resuming existing transaction " + ltx + ", global TX=" + gtx);
         }

         // at this point we have a non-null ltx

         // Asssociate the local TX with the global TX. Create new
         // entry for TX in txTable, the modifications
         // below will need this entry to add their modifications
         // under the GlobalTx key
         if (txTable.get(gtx) == null)
         {
            // create a new transaction entry

            TransactionEntry entry = configuration.isNodeLockingOptimistic() ? new OptimisticTransactionEntry() : new TransactionEntry();
            entry.setTransaction(ltx);
            log.debug("creating new tx entry");
            txTable.put(gtx, entry);
            if (log.isTraceEnabled()) log.trace("TxTable contents: " + txTable);
         }

         setTransactionalContext(ltx, gtx, ctx);
         // register a sync handler for this tx.
         registerHandler(ltx, new RemoteSynchronizationHandler(gtx, ltx, cache), ctx);

         if (configuration.isNodeLockingOptimistic())
         {
            retval = handleOptimisticPrepare(ctx, gtx, modifications, onePhase, ltx);
         }
         else
         {
            retval = handlePessimisticPrepare(ctx, m, gtx, modifications, onePhase, ltx);
         }
      }
      finally
      {
         txManager.suspend();// suspends ltx - could be null
         // resume whatever else we had going.
         if (currentTx != null) txManager.resume(currentTx);
         if (log.isDebugEnabled()) log.debug("Finished remote prepare " + gtx);
      }

      return retval;
   }

   // --------------------------------------------------------------
   //   handler methods.
   // --------------------------------------------------------------

   /**
    * Tests if we already have a tx running.  If so, register a sync handler for this method invocation.
    * if not, create a local tx if we're using opt locking.
    *
    * @return
    * @throws Throwable
    */
   private Object handleNonTxMethod(InvocationContext ctx) throws Throwable
   {
      MethodCall m = ctx.getMethodCall();
      Transaction tx = ctx.getTransaction();
      Object result;
      // if there is no current tx and we're using opt locking, we need to use an implicit tx.
      boolean implicitTransaction = configuration.isNodeLockingOptimistic() && tx == null;
      if (implicitTransaction)
      {
         tx = createLocalTx();
         // we need to attach this tx to the InvocationContext.
         ctx.setTransaction(tx);
      }
      if (tx != null)
         attachGlobalTransaction(ctx, tx, m);

      GlobalTransaction gtx = ctx.getGlobalTransaction();

      try
      {
         result = super.invoke(ctx);
         if (implicitTransaction)
         {
            copyInvocationScopeOptionsToTxScope(ctx);
            copyForcedCacheModeToTxScope(ctx);
            txManager.commit();
         }
      }
      catch (Throwable t)
      {
         if (implicitTransaction)
         {
            log.warn("Rolling back, exception encountered", t);
            result = t;
            try
            {
               setTransactionalContext(tx, gtx, ctx);
               txManager.rollback();
            }
            catch (Throwable th)
            {
               log.warn("Roll back failed encountered", th);
            }
         }
         else
         {
            throw t;
         }
      }
      return result;
   }

   /**
    * @param ctx
    */
   private void copyForcedCacheModeToTxScope(InvocationContext ctx)
   {
      Option optionOverride = ctx.getOptionOverrides();
      if (optionOverride != null 
            && (optionOverride.isForceAsynchronous() || optionOverride.isForceSynchronous()))
      {
         TransactionEntry entry = txTable.get(ctx.getGlobalTransaction());
         if (entry != null)
         {
            if (optionOverride.isForceAsynchronous())
               entry.setForceAsyncReplication(true);
            else
               entry.setForceSyncReplication(true);
         }
      }
   }

   private MethodCall attachGlobalTransaction(InvocationContext ctx, Transaction tx, MethodCall m) throws Exception
   {
      if (log.isDebugEnabled())
      {
         log.debug(" local transaction exists - registering global tx if not present for " + Thread.currentThread());
      }
      if (log.isTraceEnabled())
      {
         GlobalTransaction tempGtx = txTable.get(tx);
         log.trace("Associated gtx in txTable is " + tempGtx);
      }

      // register a sync handler for this tx - only if the gtx is not remotely initiated.
      GlobalTransaction gtx = registerTransaction(tx, ctx);
      if (gtx != null)
      {
         m = replaceGtx(m, gtx);
      }
      else
      {
         // get the current gtx from the txTable.
         gtx = txTable.get(tx);
      }

      // make sure we attach this gtx to the invocation context.
      ctx.setGlobalTransaction(gtx);

      return m;
   }

   /**
    * This is called by invoke() if we are in a remote gtx's prepare() phase.
    * Finds the appropriate tx, suspends any existing txs, registers a sync handler
    * and passes up the chain.
    * <p/>
    * Resumes any existing txs before returning.
    *
    * @return
    * @throws Throwable
    */
   private Object handleOptimisticPrepare(InvocationContext ctx, GlobalTransaction gtx, List<MethodCall> modifications, boolean onePhase, Transaction ltx) throws Throwable
   {
      Object retval;
      if (log.isDebugEnabled()) log.debug("Handling optimistic remote prepare " + gtx);
      replayModifications(modifications, ctx, true);
      retval = super.invoke(ctx);
      // JBCACHE-361 Confirm that the transaction is ACTIVE
      if (!isActive(ltx))
      {
         throw new ReplicationException("prepare() failed -- " +
                 "local transaction status is not STATUS_ACTIVE;" +
                 " is " + ltx.getStatus());
      }
      return retval;
   }

   private Object handlePessimisticPrepare(InvocationContext ctx, MethodCall m, GlobalTransaction gtx, List<MethodCall> modifications, boolean commit, Transaction ltx) throws Exception
   {
      boolean success = true;
      Object retval;
      try
      {
         // now pass up the prepare method itself.
         try
         {
            replayModifications(modifications, ctx, false);
            if (isOnePhaseCommitPrepareMehod(m))
            {
               log.trace("Using one-phase prepare.  Not propagating the prepare call up the stack until called to do so by the sync handler.");
            }
            else
            {
               super.invoke(ctx);
            }

            // JBCACHE-361 Confirm that the transaction is ACTIVE
            if (!isActive(ltx))
            {
               throw new ReplicationException("prepare() failed -- " +
                       "local transaction status is not STATUS_ACTIVE;" +
                       " is " + ltx.getStatus());
            }
         }
         catch (Throwable th)
         {
            log.error("prepare method invocation failed", th);
            retval = th;
            success = false;
            if (retval instanceof Exception)
            {
               throw (Exception) retval;
            }
         }
      }
      finally
      {

         if (log.isTraceEnabled())
         {
            log.trace("Are we running a 1-phase commit? " + commit);
         }
         // 4. If commit == true (one-phase-commit): commit (or rollback) the TX; this will cause
         //    {before/after}Completion() to be called in all registered interceptors: the TransactionInterceptor
         //    will then commit/rollback against the cache

         if (commit)
         {
            try
            {
               //                    invokeOnePhaseCommitMethod(gtx, modifications.size() > 0, success);
               if (success)
               {
                  ltx.commit();
               }
               else
               {
                  ltx.rollback();
               }
            }
            catch (Throwable t)
            {
               log.error("Commit/rollback failed.", t);
               if (success)
               {
                  // try another rollback...
                  try
                  {
                     log.info("Attempting anotehr rollback");
                     //invokeOnePhaseCommitMethod(gtx, modifications.size() > 0, false);
                     ltx.rollback();
                  }
                  catch (Throwable t2)
                  {
                     log.error("Unable to rollback", t2);
                  }
               }
            }
            finally
            {
               transactions.remove(ltx);// JBAS-298
               remoteTransactions.remove(gtx);// JBAS-308
            }
         }
      }
      return null;
   }

   private Object replayModifications(List<MethodCall> modifications, InvocationContext ctx, boolean injectDataVersions)
   {
      Object retval = null;
      MethodCall originalMethodCall = ctx.getMethodCall();
      Option originalOption = ctx.getOptionOverrides();

      if (modifications != null)
      {
         for (MethodCall modification : modifications)
         {
            try
            {
               if (injectDataVersions && !MethodDeclarations.isDataGravitationMethod(modification.getMethodId()))
               {
                  Object[] origArgs = modification.getArgs();
                  // there may be instances (e.g., data gravitation calls) where a data version is not passed in or not even relevant.
                  // make sure we are aware of this.
                  Option o = null;
                  if (origArgs[origArgs.length - 1] instanceof DataVersion)
                  {
                     o = new Option();
                     o.setDataVersion((DataVersion) origArgs[origArgs.length - 1]);
                  }
                  // modify the call to the non-dataversioned counterpart since we've popped out the data version
                  Object[] args = new Object[origArgs.length - 1];
                  System.arraycopy(origArgs, 0, args, 0, args.length);

                  ctx.setMethodCall(MethodCallFactory.create(MethodDeclarations.getUnversionedMethod(modification.getMethodId()), args));
                  if (o != null) ctx.setOptionOverrides(o);
               }
               else
               {
                  ctx.setMethodCall(modification);
               }

               retval = super.invoke(ctx);

               if (!isActive(ctx.getTransaction()))
               {
                  throw new ReplicationException("prepare() failed -- " + "local transaction status is not STATUS_ACTIVE; is " + ctx.getTransaction().getStatus());
               }
            }
            catch (Throwable t)
            {
               log.error("method invocation failed", t);
               retval = t;
            }
            finally
            {
               // reset any options
               if (injectDataVersions) ctx.setOptionOverrides(originalOption);
               ctx.setMethodCall(originalMethodCall);
            }
            if (retval != null && retval instanceof Exception)
            {
               if (retval instanceof RuntimeException)
                  throw (RuntimeException) retval;
               else
                  throw new RuntimeException((Exception) retval);
            }
         }
      }
      // need to pass up the prepare as well and return value from that
      return retval;
   }

   /**
    * Handles a commit or a rollback for a remote gtx.  Called by invoke().
    *
    * @return
    * @throws Throwable
    */
   private Object handleRemoteCommitRollback(InvocationContext ctx) throws Throwable
   {
      Transaction ltx;
      GlobalTransaction gtx = ctx.getGlobalTransaction();
      MethodCall m = ctx.getMethodCall();

      try
      {
         ltx = getLocalTxForGlobalTx(gtx);
      }
      catch (IllegalStateException e)
      {
         if (m.getMethodId() == MethodDeclarations.rollbackMethod_id)
         {
            log.warn("No local transaction for this remotely originating rollback.  Possibly rolling back before a prepare call was broadcast?");
            return null;
         }
         else
         {
            throw e;
         }
      }

      // disconnect if we have a current tx associated
      Transaction currentTx = txManager.getTransaction();
      boolean resumeCurrentTxOnCompletion = false;
      try
      {
         if (!ltx.equals(currentTx))
         {
            currentTx = txManager.suspend();
            resumeCurrentTxOnCompletion = true;
            txManager.resume(ltx);
            // make sure we set this in the ctx
            ctx.setTransaction(ltx);
         }
         if (log.isDebugEnabled()) log.debug(" executing " + m + "() with local TX " + ltx + " under global tx " + gtx);

         // pass commit up the chain
         // super.invoke(ctx);
         // commit or rollback the tx.
         if (m.getMethodId() == MethodDeclarations.commitMethod_id)
         {
            txManager.commit();
            if (configuration.getExposeManagementStatistics() && getStatisticsEnabled())
            {
               m_commits++;
            }
         }
         else
         {
            txManager.rollback();
            if (configuration.getExposeManagementStatistics() && getStatisticsEnabled())
            {
               m_rollbacks++;
            }
         }
      }
      finally
      {
         //resume the old transaction if we suspended it
         if (resumeCurrentTxOnCompletion)
         {
            if (log.isTraceEnabled()) log.trace("Resuming suspended transaction " + currentTx);
            txManager.suspend();
            if (currentTx != null)
            {
               txManager.resume(currentTx);
               ctx.setTransaction(currentTx);
            }
         }

         // remove from local lists.
         remoteTransactions.remove(gtx);
         transactions.remove(ltx);

         // this tx has completed.  Clean up in the tx table.
         txTable.remove(gtx);
         txTable.remove(ltx);
      }

      if (log.isDebugEnabled()) log.debug("Finished remote commit/rollback method for " + gtx);

      return null;
   }

   private Transaction getLocalTxForGlobalTx(GlobalTransaction gtx) throws IllegalStateException
   {
      Transaction ltx = txTable.getLocalTransaction(gtx);
      if (ltx != null)
      {
         if (log.isDebugEnabled()) log.debug("Found local TX=" + ltx + ", global TX=" + gtx);
      }
      else
      {
         throw new IllegalStateException(" found no local TX for global TX " + gtx);
      }
      return ltx;
   }

   /**
    * Handles a commit or a rollback.  Called by the synch handler.  Simply tests that we are in the correct tx and
    * passes the meth call up the interceptor chain.
    *
    * @return
    * @throws Throwable
    */
   private Object handleCommitRollback(InvocationContext ctx) throws Throwable
   {
      //GlobalTransaction gtx = findGlobalTransaction(m.getArgs());
      GlobalTransaction gtx = ctx.getGlobalTransaction();
      Object result;

      // this must have a local transaction associated if a prepare has been
      // callled before
      //Transaction ltx = getLocalTxForGlobalTx(gtx);

      //        Transaction currentTx = txManager.getTransaction();

      //if (!ltx.equals(currentTx)) throw new IllegalStateException(" local transaction " + ltx + " transaction does not match running tx " + currentTx);

      result = super.invoke(ctx);

      if (log.isDebugEnabled()) log.debug("Finished local commit/rollback method for " + gtx);
      return result;
   }

   // --------------------------------------------------------------
   //   Transaction phase runners
   // --------------------------------------------------------------

   /**
    * creates a commit() MethodCall and feeds it to handleCommitRollback();
    *
    * @param gtx
    */
   protected void runCommitPhase(InvocationContext ctx, GlobalTransaction gtx, Transaction tx, List modifications, boolean onePhaseCommit)
   {
      // set the hasMods flag in the invocation ctx.  This should not be replicated, just used locally by the interceptors.
      ctx.setTxHasMods(modifications != null && modifications.size() > 0);
      try
      {
         MethodCall commitMethod;
         if (onePhaseCommit)
         {
            // running a 1-phase commit.
            if (configuration.isNodeLockingOptimistic())
            {
               commitMethod = MethodCallFactory.create(MethodDeclarations.optimisticPrepareMethod,
                       gtx, modifications, null, cache.getLocalAddress(), true);
            }
            else
            {
               commitMethod = MethodCallFactory.create(MethodDeclarations.prepareMethod,
                       gtx, modifications, cache.getLocalAddress(),
                       true);
            }
         }
         else
         {
            commitMethod = MethodCallFactory.create(MethodDeclarations.commitMethod, gtx);
         }

         if (log.isTraceEnabled())
         {
            log.trace(" running commit for " + gtx);
         }
         ctx.setMethodCall(commitMethod);
         handleCommitRollback(ctx);
      }
      catch (Throwable e)
      {
         log.warn("Commit failed.  Clearing stale locks.");
         try
         {
            cleanupStaleLocks(gtx);
         }
         catch (RuntimeException re)
         {
            log.error("Unable to clear stale locks", re);
            throw re;
         }
         catch (Throwable e2)
         {
            log.error("Unable to clear stale locks", e2);
            throw new RuntimeException(e2);
         }
         if (e instanceof RuntimeException)
            throw (RuntimeException) e;
         else
            throw new RuntimeException("Commit failed.", e);
      }
   }


   private void cleanupStaleLocks(GlobalTransaction gtx) throws Throwable
   {
      TransactionEntry entry = txTable.get(gtx);
      if (entry != null)
      {
         entry.releaseAllLocksLIFO(gtx);
      }
   }

   /**
    * creates a rollback() MethodCall and feeds it to handleCommitRollback();
    *
    * @param gtx
    */
   protected void runRollbackPhase(InvocationContext ctx, GlobalTransaction gtx, Transaction tx, List modifications)
   {
      //Transaction ltx = null;
      try
      {
         ctx.setTxHasMods(modifications != null && modifications.size() > 0);
         // JBCACHE-457
         //            MethodCall rollbackMethod = MethodCall(CacheImpl.rollbackMethod, new Object[]{gtx, hasMods ? true : false});
         MethodCall rollbackMethod = MethodCallFactory.create(MethodDeclarations.rollbackMethod, gtx);
         if (log.isTraceEnabled())
         {
            log.trace(" running rollback for " + gtx);
         }

         //JBCACHE-359 Store a lookup for the gtx so a listener
         // callback can find it
         //ltx = getLocalTxForGlobalTx(gtx);
         rollbackTransactions.put(tx, gtx);

         ctx.setMethodCall(rollbackMethod);
         handleCommitRollback(ctx);
      }
      catch (Throwable e)
      {
         log.warn("Rollback had a problem", e);
      }
      finally
      {
         if (tx != null) rollbackTransactions.remove(tx);
      }
   }

   /**
    * Handles a local prepare - invoked by the sync handler.  Tests if the current tx matches the gtx passed in to the
    * method call and passes the prepare() call up the chain.
    *
    * @return
    * @throws Throwable
    */
   protected Object runPreparePhase(InvocationContext ctx, GlobalTransaction gtx, List modifications) throws Throwable
   {
      // build the method call
      MethodCall prepareMethod;
      //        if (cache.getCacheModeInternal() != CacheImpl.REPL_ASYNC)
      //        {
      // running a 2-phase commit.
      if (configuration.isNodeLockingOptimistic())
      {
         prepareMethod = MethodCallFactory.create(MethodDeclarations.optimisticPrepareMethod, gtx, modifications, null, cache.getLocalAddress(), false);
      }
      else if (configuration.getCacheMode() != Configuration.CacheMode.REPL_ASYNC)
      {
         prepareMethod = MethodCallFactory.create(MethodDeclarations.prepareMethod,
                 gtx, modifications, cache.getLocalAddress(),
                 false);// don't commit or rollback - wait for call
      }
      //}
      else
      {
         // this is a REPL_ASYNC call - do 1-phase commit.  break!
         log.trace("This is a REPL_ASYNC call (1 phase commit) - do nothing for beforeCompletion()");
         return null;
      }

      // passes a prepare call up the local interceptor chain.  The replication interceptor
      // will do the broadcasting if needed.  This is so all requests (local/remote) are
      // treated the same
      Object result;

      // Is there a local transaction associated with GTX ?
      Transaction ltx = ctx.getTransaction();

      //if ltx is not null and it is already running
      if (txManager.getTransaction() != null && ltx != null && txManager.getTransaction().equals(ltx))
      {
         ctx.setMethodCall(prepareMethod);
         result = super.invoke(ctx);
      }
      else
      {
         log.warn("Local transaction does not exist or does not match expected transaction " + gtx);
         throw new CacheException(" local transaction " + ltx + " does not exist or does not match expected transaction " + gtx);
      }
      return result;
   }

   // --------------------------------------------------------------
   //   Private helper methods
   // --------------------------------------------------------------


   /**
    * Creates a gtx (if one doesnt exist), a sync handler, and registers the tx.
    *
    * @param tx
    * @return
    * @throws Exception
    */
   private GlobalTransaction registerTransaction(Transaction tx, InvocationContext ctx) throws Exception
   {
      GlobalTransaction gtx;
      if (isValid(tx) && transactions.put(tx, NULL) == null)
      {
         gtx = cache.getCurrentTransaction(tx, true);
         if (gtx.isRemote())
         {
            // should be no need to register a handler since this a remotely initiated gtx
            if (log.isTraceEnabled())
            {
               log.trace("is a remotely initiated gtx so no need to register a tx for it");
            }
         }
         else
         {
            if (log.isTraceEnabled())
            {
               log.trace("Registering sync handler for tx " + tx + ", gtx " + gtx);
            }
            // see the comment in the LocalSyncHandler for the last isOriginLocal param.
            LocalSynchronizationHandler myHandler = new LocalSynchronizationHandler(gtx, tx, cache, !ctx.isOriginLocal());
            registerHandler(tx, myHandler, ctx);
         }
      }
      else if ((gtx = (GlobalTransaction) rollbackTransactions.get(tx)) != null)
      {
         if (log.isDebugEnabled()) log.debug("Transaction " + tx + " is already registered and is rolling back.");
      }
      else
      {
         if (log.isDebugEnabled()) log.debug("Transaction " + tx + " is already registered.");

      }
      return gtx;
   }

   /**
    * Registers a sync hander against a tx.
    *
    * @param tx
    * @param handler
    * @throws Exception
    */
   private void registerHandler(Transaction tx, Synchronization handler, InvocationContext ctx) throws Exception
   {
      OrderedSynchronizationHandler orderedHandler = OrderedSynchronizationHandler.getInstance(tx);

      if (log.isTraceEnabled()) log.trace("registering for TX completion: SynchronizationHandler(" + handler + ")");

      orderedHandler.registerAtHead(handler);// needs to be invoked first on TX commit

      cache.getNotifier().notifyTransactionRegistered(tx, ctx);
   }

   /**
    * Replaces the global transaction in a method call with a new global transaction passed in.
    */
   private MethodCall replaceGtx(MethodCall m, GlobalTransaction gtx)
   {
      Class[] argClasses = m.getMethod().getParameterTypes();
      Object[] args = m.getArgs();

      for (int i = 0; i < argClasses.length; i++)
      {
         if (argClasses[i].equals(GlobalTransaction.class))
         {
            if (!gtx.equals(args[i]))
            {
               args[i] = gtx;
               m.setArgs(args);
            }
            break;
         }
      }
      return m;
   }

   /**
    * Creates and starts a local tx
    *
    * @return
    * @throws Exception
    */
   private Transaction createLocalTx() throws Exception
   {
      if (log.isTraceEnabled())
      {
         log.trace("Creating transaction for thread " + Thread.currentThread());
      }
      Transaction localTx;
      if (txManager == null) throw new Exception("Failed to create local transaction; TransactionManager is null");
      txManager.begin();
      localTx = txManager.getTransaction();
      return localTx;
   }

   /**
    * Creates a new local transaction for a given global transaction.
    *
    * @param gtx
    * @return
    * @throws Exception
    */
   private Transaction createLocalTxForGlobalTx(GlobalTransaction gtx, InvocationContext ctx) throws Exception
   {
      Transaction localTx = createLocalTx();
      txTable.put(localTx, gtx);
      // attach this to the context
      ctx.setTransaction(localTx);
      if (log.isTraceEnabled()) log.trace("Created new tx for gtx " + gtx);
      return localTx;
   }

   // ------------------------------------------------------------------------
   // Synchronization classes
   // ------------------------------------------------------------------------

   // this controls the whole transaction

   private class RemoteSynchronizationHandler implements Synchronization
   {
      Transaction tx = null;
      GlobalTransaction gtx = null;
      CacheSPI cache = null;
      List modifications = null;
      TransactionEntry entry = null;
      protected InvocationContext ctx; // the context for this call.


      RemoteSynchronizationHandler(GlobalTransaction gtx, Transaction tx, CacheSPI cache)
      {
         this.gtx = gtx;
         this.tx = tx;
         this.cache = cache;
      }

      public void beforeCompletion()
      {
         if (!cache.getCacheStatus().allowInvocations()) throw new IllegalStateException("Cache not in STARTED state!");

         if (log.isTraceEnabled()) log.trace("Running beforeCompletion on gtx " + gtx);
         entry = txTable.get(gtx);
         if (entry == null)
         {
            log.error("Transaction has a null transaction entry - beforeCompletion() will fail.");
            log.error("TxTable contents: " + txTable);
            throw new IllegalStateException("cannot find transaction entry for " + gtx);
         }

         modifications = entry.getModifications();
         ctx = cache.getInvocationContext();
         ctx.setOriginLocal(false);
      }

      // this should really not be done here -
      // it is supposed to be post commit not actually run the commit
      public void afterCompletion(int status)
      {
         if (!cache.getCacheStatus().allowInvocations()) throw new IllegalStateException("Cache not in STARTED state!");
         try
         {
            // could happen if a rollback is called and beforeCompletion() doesn't get called.
            if (ctx == null) ctx = cache.getInvocationContext();
            setTransactionalContext(tx, gtx, ctx);

            try
            {
               if (txManager.getTransaction() != null && !txManager.getTransaction().equals(tx)) txManager.resume(tx);
            }
            catch (Exception e)
            {
               log.error("afterCompletion error: " + status, e);
            }


            if (log.isTraceEnabled()) log.trace("calling aftercompletion for " + gtx);
            // set any transaction wide options as current for this thread.
            if ((entry = txTable.get(gtx)) != null)
            {
               modifications = entry.getModifications();
               ctx.setOptionOverrides(entry.getOption());
            }
            transactions.remove(tx);

            switch (status)
            {
               case Status.STATUS_COMMITTED:

                  // if this is optimistic or sync repl
                  boolean onePhaseCommit = !configuration.isNodeLockingOptimistic() && configuration.getCacheMode() == Configuration.CacheMode.REPL_ASYNC;
                  if (log.isDebugEnabled()) log.debug("Running commit phase.  One phase? " + onePhaseCommit);
                  runCommitPhase(ctx, gtx, tx, modifications, onePhaseCommit);
                  log.debug("Finished commit phase");
                  break;
               case Status.STATUS_UNKNOWN:
                  log.warn("Received JTA STATUS_UNKNOWN in afterCompletion()!  XA resources may not be in sync.  The app should manually clean up resources at this point.");
               case Status.STATUS_MARKED_ROLLBACK:
               case Status.STATUS_ROLLEDBACK:
                  log.debug("Running rollback phase");
                  runRollbackPhase(ctx, gtx, tx, modifications);
                  log.debug("Finished rollback phase");
                  break;

               default:
                  throw new IllegalStateException("illegal status: " + status);
            }
         }
         finally
         {
            // clean up the tx table
            txTable.remove(gtx);
            txTable.remove(tx);
            setTransactionalContext(null, null, ctx);
            cleanupInternalState();
         }
      }

      /**
       * Cleans out (nullifies) member variables held by the sync object for easier gc.  Could be (falsely) seen as a mem
       * leak if the TM implementation hangs on to the synchronizations for an unnecessarily long time even after the tx
       * completes.  See JBCACHE-1007.
       */
      private void cleanupInternalState()
      {
         this.tx = null;
         this.gtx = null;
         this.cache = null;
         this.modifications = null;
         this.entry = null;
      }

      public String toString()
      {
         return "TxInterceptor.RemoteSynchronizationHandler(gtx=" + gtx + ", tx=" + getTxAsString() + ")";
      }

      protected String getTxAsString()
      {
         // JBCACHE-1114 -- don't call toString() on tx or it can lead to stack overflow
         if (tx == null)
            return null;

         return tx.getClass().getName() + "@" + System.identityHashCode(tx);
      }
   }

   private class LocalSynchronizationHandler extends RemoteSynchronizationHandler
   {
      private boolean localRollbackOnly = true;
      // a VERY strange situation where a tx has remote origins, but since certain buddy group org methods perform local
      // cleanups even when remotely triggered, and optimistic locking is used, you end up with an implicit local tx.
      // This is STILL remotely originating though and this needs to be made explicit here.
      // this can be checked by inspecting the InvocationContext.isOriginLocal() at the time of registering the sync.
      private boolean remoteLocal = false;
      private Option originalOptions, transactionalOptions;

      /**
       * A Synchronization for locally originating txs.
       * <p/>
       * a VERY strange situation where a tx has remote origins, but since certain buddy group org methods perform local
       * cleanups even when remotely triggered, and optimistic locking is used, you end up with an implicit local tx.
       * This is STILL remotely originating though and this needs to be made explicit here.
       * this can be checked by inspecting the InvocationContext.isOriginLocal() at the time of registering the sync.
       *
       * @param gtx
       * @param tx
       * @param cache
       * @param remoteLocal
       */
      LocalSynchronizationHandler(GlobalTransaction gtx, Transaction tx, CacheSPI cache, boolean remoteLocal)
      {
         super(gtx, tx, cache);
         this.remoteLocal = remoteLocal;
      }

      @Override
      public void beforeCompletion()
      {
         super.beforeCompletion();
         ctx.setOriginLocal(!remoteLocal); // this is the LOCAL sync handler after all!
         // fetch the modifications before the transaction is committed
         // (and thus removed from the txTable)
         setTransactionalContext(tx, gtx, ctx);
         if (modifications.size() == 0)
         {
            if (log.isTraceEnabled()) log.trace("No modifications in this tx.  Skipping beforeCompletion()");
            return;
         }

         // set any transaction wide options as current for this thread.
         originalOptions = ctx.getOptionOverrides();
         transactionalOptions = entry.getOption();
         ctx.setOptionOverrides(transactionalOptions);

         try
         {
            switch (tx.getStatus())
            {
               // if we are active or preparing then we can go ahead
               case Status.STATUS_ACTIVE:
               case Status.STATUS_PREPARING:
                  // run a prepare call.
                  Object result = runPreparePhase(ctx, gtx, modifications);

                  if (result instanceof Throwable)
                  {
                     if (log.isDebugEnabled())
                        log.debug("Transaction needs to be rolled back - the cache returned an instance of Throwable for this prepare call (tx=" + tx + " and gtx=" + gtx + ")", (Throwable) result);
                     tx.setRollbackOnly();
                     throw (Throwable) result;
                  }
                  break;
               default:
                  throw new CacheException("transaction " + tx + " in status " + tx.getStatus() + " unbale to start transaction");
            }
         }
         catch (Throwable t)
         {
            try
            {
               tx.setRollbackOnly();
            }
            catch (SystemException se)
            {
               throw new RuntimeException("setting tx rollback failed ", se);
            }
            if (t instanceof RuntimeException)
               throw (RuntimeException) t;
            else
               throw new RuntimeException("", t);
         }
         finally
         {
            localRollbackOnly = false;
            setTransactionalContext(null, null, ctx);
            ctx.setOptionOverrides(originalOptions);
         }
      }

      @Override
      public void afterCompletion(int status)
      {
         // could happen if a rollback is called and beforeCompletion() doesn't get called.
         if (ctx == null) ctx = cache.getInvocationContext();
         ctx.setLocalRollbackOnly(localRollbackOnly);
         ctx.setOptionOverrides(transactionalOptions);
         try
         {
            super.afterCompletion(status);
         }
         finally
         {
            ctx.setOptionOverrides(originalOptions);
         }
      }

      public String toString()
      {
         return "TxInterceptor.LocalSynchronizationHandler(gtx=" + gtx + ", tx=" + getTxAsString() + ")";
      }
   }
}