/*
 * JBoss, the OpenSource J2EE webOS
 *
 * Distributable under LGPL license.
 * See terms of license at gnu.org.
 */
package org.jboss.cache.loader;

import org.jboss.cache.Fqn;
import org.jboss.cache.Modification;
import org.jboss.cache.config.CacheLoaderConfig.IndividualCacheLoaderConfig;
import org.jboss.cache.loader.tcp.TcpCacheOperations;

import java.io.BufferedInputStream;
import java.io.BufferedOutputStream;
import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.net.Socket;
import java.util.List;
import java.util.Map;
import java.util.Set;

/**
 * DelegatingCacheLoader implementation which delegates to a remote (not in the same VM)
 * CacheImpl using TCP/IP for communication. Example configuration for connecting to a TcpCacheServer
 * running at myHost:12345:<pre>
 * <attribute name="CacheLoaderClass">org.jboss.cache.loader.TcpDelegatingCacheLoader</attribute>
 * <attribute name="CacheLoaderConfig">
 * host=localhost
 * port=2099
 * </attribute>
 * </pre>
 *
 * @author Bela Ban
 * @version $Id: TcpDelegatingCacheLoader.java 4458 2007-09-13 05:47:28Z genman $
 */
public class TcpDelegatingCacheLoader extends AbstractCacheLoader
{
   private Socket sock;
   private TcpDelegatingCacheLoaderConfig config;
   ObjectInputStream in;
   ObjectOutputStream out;


   /**
    * Default constructor.
    */
   public TcpDelegatingCacheLoader()
   {
      // Empty.
   }

   /**
    * Allows programmatic configuration.
    *
    * @param host The host on which to look up the remote object.
    * @param port The port on which to look up the remote object.
    */
   public TcpDelegatingCacheLoader(String host, int port)
   {
      this.config = new TcpDelegatingCacheLoaderConfig(host, port);
   }

   /**
    * Allows configuration via XML config file.
    */
   public void setConfig(IndividualCacheLoaderConfig base)
   {
      if (base instanceof TcpDelegatingCacheLoaderConfig)
      {
         this.config = (TcpDelegatingCacheLoaderConfig) base;
      }
      else
      {
         config = new TcpDelegatingCacheLoaderConfig(base);
      }
   }

   public IndividualCacheLoaderConfig getConfig()
   {
      return config;
   }

   public Set<?> getChildrenNames(Fqn fqn) throws Exception
   {
      Set cn = null;
      synchronized (out)
      {
         out.reset();
         out.writeByte(TcpCacheOperations.GET_CHILDREN_NAMES);
         out.writeObject(fqn);
         out.flush();
         Object retval = in.readObject();
         if (retval instanceof Exception)
         {
            throw (Exception) retval;
         }
         cn = (Set) retval;
      }

      // the cache loader contract is a bit different from the cache when it comes to dealing with childrenNames
      if (cn.isEmpty()) return null;
      return cn;
   }

   public Map<Object, Object> get(Fqn name) throws Exception
   {
      synchronized (out)
      {
         out.reset();

         out.writeByte(TcpCacheOperations.GET);
         out.writeObject(name);
         out.flush();
         Object retval = in.readObject();
         if (retval instanceof Exception)
         {
            throw (Exception) retval;
         }
         return (Map) retval;
      }
   }

   public boolean exists(Fqn name) throws Exception
   {
      synchronized (out)
      {
         out.reset();

         out.writeByte(TcpCacheOperations.EXISTS);
         out.writeObject(name);
         out.flush();
         Object retval = in.readObject();
         if (retval instanceof Exception)
         {
            throw (Exception) retval;
         }
         return (Boolean) retval;
      }
   }

   public Object put(Fqn name, Object key, Object value) throws Exception
   {
      synchronized (out)
      {
         out.reset();

         out.writeByte(TcpCacheOperations.PUT_KEY_VAL);
         out.writeObject(name);
         out.writeObject(key);
         out.writeObject(value);
         out.flush();
         Object retval = in.readObject();
         if (retval instanceof Exception)
         {
            throw (Exception) retval;
         }
         return retval;
      }
   }

   public void put(Fqn name, Map<Object, Object> attributes) throws Exception
   {
      synchronized (out)
      {
         out.reset();

         out.writeByte(TcpCacheOperations.PUT);
         out.writeObject(name);
         out.writeObject(attributes);
         out.flush();
         Object retval = in.readObject();
         if (retval instanceof Exception)
         {
            throw (Exception) retval;
         }
      }
   }

   @Override
   public void start() throws Exception
   {
      init();
   }

   @Override
   public void stop()
   {
      try
      {
         if (in != null) in.close();
      }
      catch (IOException e)
      {
      }
      try
      {
         if (out != null) out.close();
      }
      catch (IOException e)
      {
      }
      try
      {
         if (sock != null) sock.close();
      }
      catch (IOException e)
      {
      }
   }


   private void init() throws IOException
   {
      sock = new Socket(config.getHost(), config.getPort());
      out = new ObjectOutputStream(new BufferedOutputStream(sock.getOutputStream()));
      out.flush();
      in = new ObjectInputStream(new BufferedInputStream(sock.getInputStream()));
   }


   @Override
   public void put(List<Modification> modifications) throws Exception
   {
      synchronized (out)
      {
         out.reset();

         out.writeByte(TcpCacheOperations.PUT_LIST);
         int length = modifications.size();
         out.writeInt(length);
         for (Modification m : modifications) 
         {
            m.writeExternal(out);
         }
         out.flush();
         Object retval = in.readObject();
         if (retval instanceof Exception)
         {
            throw (Exception) retval;
         }
      }
   }

   public Object remove(Fqn fqn, Object key) throws Exception
   {
      synchronized (out)
      {
         out.reset();

         out.writeByte(TcpCacheOperations.REMOVE_KEY);
         out.writeObject(fqn);
         out.writeObject(key);
         out.flush();
         Object retval = in.readObject();
         if (retval instanceof Exception)
         {
            throw (Exception) retval;
         }
         return retval;
      }
   }

   public void remove(Fqn fqn) throws Exception
   {
      synchronized (out)
      {
         out.reset();

         out.writeByte(TcpCacheOperations.REMOVE);
         out.writeObject(fqn);
         out.flush();
         Object retval = in.readObject();
         if (retval instanceof Exception)
         {
            throw (Exception) retval;
         }
      }
   }

   public void removeData(Fqn fqn) throws Exception
   {
      synchronized (out)
      {
         out.reset();

         out.writeByte(TcpCacheOperations.REMOVE_DATA);
         out.writeObject(fqn);
         out.flush();
         Object retval = in.readObject();
         if (retval instanceof Exception)
         {
            throw (Exception) retval;
         }
      }
   }

   @Override
   public void loadEntireState(ObjectOutputStream os) throws Exception
   {
      throw new UnsupportedOperationException("operation is not currently supported - need to define semantics first");
   }

   @Override
   public void loadState(Fqn subtree, ObjectOutputStream os) throws Exception
   {
      throw new UnsupportedOperationException("operation is not currently supported - need to define semantics first");
   }

   @Override
   public void storeEntireState(ObjectInputStream is) throws Exception
   {
      throw new UnsupportedOperationException("operation is not currently supported - need to define semantics first");
   }

   @Override
   public void storeState(Fqn subtree, ObjectInputStream is) throws Exception
   {
      throw new UnsupportedOperationException("operation is not currently supported - need to define semantics first");
   }

}
