/*
 * JBoss, Home of Professional Open Source
 *
 * Distributable under LGPL license.
 * See terms of license at gnu.org.
 */
package org.jboss.cache.marshall;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.jboss.cache.Fqn;
import org.jboss.cache.Region;
import org.jboss.cache.RegionManager;
import org.jboss.cache.buddyreplication.BuddyManager;
import org.jboss.cache.transaction.GlobalTransaction;

import java.io.ByteArrayOutputStream;
import java.io.InputStream;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.util.List;
import java.util.Map;
import java.util.concurrent.ConcurrentHashMap;

/**
 * Abstract AbstractMarshaller for JBoss Cache.
 *
 * @author <a href="mailto:manik@jboss.org">Manik Surtani (manik@jboss.org)</a>
 */
public abstract class AbstractMarshaller implements Marshaller
{
   protected boolean useRegionBasedMarshalling;
   protected RegionManager regionManager;
   protected boolean defaultInactive;
   private static Log log = LogFactory.getLog(AbstractMarshaller.class);

   /**
    * Map<GlobalTransaction, Fqn> for prepared tx that have not committed
    */
   private Map<GlobalTransaction, Fqn> transactions = new ConcurrentHashMap<GlobalTransaction, Fqn>(16);

   protected void init(RegionManager manager, boolean defaultInactive, boolean useRegionBasedMarshalling)
   {
      this.useRegionBasedMarshalling = useRegionBasedMarshalling;
      this.defaultInactive = defaultInactive;
      this.regionManager = manager;
   }

   // implement the basic contract set in RPcDispatcher.AbstractMarshaller
   public byte[] objectToByteBuffer(Object obj) throws Exception
   {
      ByteArrayOutputStream baos = new ByteArrayOutputStream();
      ObjectOutputStream out = ObjectSerializationFactory.createObjectOutputStream(baos);
      objectToObjectStream(obj, out);
      out.close();
      return baos.toByteArray();
   }

   public Object objectFromByteBuffer(byte[] bytes) throws Exception
   {
      ObjectInputStream in = ObjectSerializationFactory.createObjectInputStream(bytes);
      return objectFromObjectStream(in);
   }

   public Object objectFromStream(InputStream in) throws Exception
   {
      // by default just create an OIS from this IS and pass in to the relevant method
      if (in instanceof ObjectInputStream)
      {
         return objectFromObjectStream((ObjectInputStream) in);
      }
      else
      {
         return objectFromObjectStream(ObjectSerializationFactory.createObjectInputStream(in));
      }
   }

   /**
    * This is "replicate" call with a single MethodCall argument.
    *
    * @param call
    */
   protected Fqn extractFqnFromMethodCall(MethodCall call)
   {
      MethodCall c0 = (MethodCall) call.getArgs()[0];
      return extractFqn(c0);
   }

   /**
    * This is "replicate" call with a list of MethodCall argument.
    *
    * @param call
    */
   protected Fqn extractFqnFromListOfMethodCall(MethodCall call)
   {
      Object[] args = call.getArgs();
      // We simply pick the first one and assume everyone will need to operate under the same region!
      MethodCall c0 = (MethodCall) ((List) args[0]).get(0);
      return extractFqn(c0);
   }

   protected Fqn extractFqn(MethodCall methodCall)
   {
      if (methodCall == null)
      {
         throw new NullPointerException("method call is null");
      }

      Fqn fqn = null;
      Object[] args = methodCall.getArgs();
      switch (methodCall.getMethodId())
      {
         case MethodDeclarations.optimisticPrepareMethod_id:
         case MethodDeclarations.prepareMethod_id:
            // Prepare method has a list of modifications. We will just take the first one and extract.
            List modifications = (List) args[1];
            fqn = extractFqn((MethodCall) modifications.get(0));

            // the last arg of a prepare call is the one-phase flag
            boolean one_phase_commit = (Boolean) args[args.length - 1];

            // If this is two phase commit, map the FQN to the GTX so
            // we can find it when the commit/rollback comes through
            if (!one_phase_commit)
            {
               transactions.put((GlobalTransaction) args[0], fqn);
            }
            break;
         case MethodDeclarations.rollbackMethod_id:
         case MethodDeclarations.commitMethod_id:
            // We stored the fqn in the transactions map during the prepare phase
            fqn = transactions.remove(args[0]);
            break;
         case MethodDeclarations.getPartialStateMethod_id:
         case MethodDeclarations.dataGravitationMethod_id:
         case MethodDeclarations.evictNodeMethodLocal_id:
         case MethodDeclarations.evictVersionedNodeMethodLocal_id:
         case MethodDeclarations.invalidateMethodLocal_id:   
         case MethodDeclarations.getChildrenNamesMethodLocal_id:
         case MethodDeclarations.getDataMapMethodLocal_id:
         case MethodDeclarations.getKeysMethodLocal_id:
         case MethodDeclarations.getKeyValueMethodLocal_id:
         case MethodDeclarations.existsMethod_id:
            fqn = (Fqn) args[0];
            break;
         case MethodDeclarations.dataGravitationCleanupMethod_id:
            fqn = (Fqn) args[1];
            break;
         case MethodDeclarations.remoteAnnounceBuddyPoolNameMethod_id:
         case MethodDeclarations.remoteAssignToBuddyGroupMethod_id:
         case MethodDeclarations.remoteRemoveFromBuddyGroupMethod_id:
            break;
         case MethodDeclarations.replicateMethod_id:
         case MethodDeclarations.clusteredGetMethod_id:
            // possible when we have a replication queue.
            fqn = extractFqn((MethodCall) args[0]);
            break;
         default:
            if (MethodDeclarations.isCrudMethod(methodCall.getMethodId()))
            {
               fqn = (Fqn) args[1];
            }
            else
            {
               throw new IllegalArgumentException("AbstractMarshaller.extractFqn(): Unknown id in method call: " + methodCall);
            }
            break;

      }

      if (log.isTraceEnabled())
      {
         log.trace("extract(): received " + methodCall + "extracted fqn: " + fqn);
      }

      return fqn;
   }

   /**
    * Retrieves the {@link Region} from the {@link RegionManager} after taking into account that this may be a Buddy Backup Fqn.
    * If the fqn passed in is null, the region has been deactivated or if a region cannot be found, this method returns a null.
    *
    * @param fqn of the region to locate
    * @return a region
    */
   protected Region getRegion(Fqn fqn)
   {
      if (fqn == null) return null;
      if (BuddyManager.isBackupFqn(fqn))
      {
         // Strip out the buddy group portion
         fqn = BuddyManager.getActualFqn(fqn);
      }
      Region r = regionManager.getRegion(fqn, Region.Type.MARSHALLING, false);
      return r;
      //return r != null && r.isActive() ? r : null;
   }
}
