/*
 * JBoss, Home of Professional Open Source
 *
 * Distributable under LGPL license.
 * See terms of license at gnu.org.
 */
package org.jboss.cache.marshall;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.jboss.cache.Fqn;
import org.jboss.cache.Region;
import static org.jboss.cache.Region.Status;
import org.jboss.cache.RegionManager;
import org.jboss.cache.buddyreplication.GravitateResult;
import org.jboss.cache.optimistic.DefaultDataVersion;
import org.jboss.cache.transaction.GlobalTransaction;
import org.jboss.cache.util.MapCopy;
import org.jgroups.Address;
import org.jgroups.stack.IpAddress;

import java.io.Externalizable;
import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.io.Serializable;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.TreeMap;
import java.util.TreeSet;

/**
 * An enhanced marshaller for RPC calls between CacheImpl instances.
 *
 * @author <a href="mailto:manik@jboss.org">Manik Surtani (manik@jboss.org)</a>
 */
public class CacheMarshaller200 extends AbstractMarshaller
{
   // logger
   private Log log = LogFactory.getLog(CacheMarshaller200.class);

   // magic numbers
   protected static final int MAGICNUMBER_METHODCALL = 1;
   protected static final int MAGICNUMBER_FQN = 2;
   protected static final int MAGICNUMBER_GTX = 3;
   protected static final int MAGICNUMBER_IPADDRESS = 4;
   protected static final int MAGICNUMBER_ARRAY_LIST = 5;
   protected static final int MAGICNUMBER_INTEGER = 6;
   protected static final int MAGICNUMBER_LONG = 7;
   protected static final int MAGICNUMBER_BOOLEAN = 8;
   protected static final int MAGICNUMBER_STRING = 9;
   protected static final int MAGICNUMBER_DEFAULT_DATA_VERSION = 10;
   protected static final int MAGICNUMBER_LINKED_LIST = 11;
   protected static final int MAGICNUMBER_HASH_MAP = 12;
   protected static final int MAGICNUMBER_TREE_MAP = 13;
   protected static final int MAGICNUMBER_HASH_SET = 14;
   protected static final int MAGICNUMBER_TREE_SET = 15;
   protected static final int MAGICNUMBER_NODEDATA_MARKER = 16;
   protected static final int MAGICNUMBER_NODEDATA_EXCEPTION_MARKER = 17;
   protected static final int MAGICNUMBER_NODEDATA = 18;
   protected static final int MAGICNUMBER_GRAVITATERESULT = 19;
   protected static final int MAGICNUMBER_SHORT = 20;
   protected static final int MAGICNUMBER_MAPCOPY = 21;
   protected static final int MAGICNUMBER_NULL = 99;
   protected static final int MAGICNUMBER_SERIALIZABLE = 100;
   protected static final int MAGICNUMBER_REF = 101;

   protected static final InactiveRegionException IRE = new InactiveRegionException("Cannot unmarshall to an inactive region");

   // this is pretty nasty, and may need more thought.
   protected final ThreadLocal<Fqn> regionForCall = new ThreadLocal<Fqn>();

   public CacheMarshaller200(RegionManager manager, boolean defaultInactive, boolean useRegionBasedMarshalling)
   {
      init(manager, defaultInactive, useRegionBasedMarshalling);
      if (useRegionBasedMarshalling)
      {
         log.debug("Using region based marshalling logic : marshalling Fqn first for every call.");
      }
   }

   /**
    * Tests if the type of object being marshalled is a method call or a return value
    *
    * @param o object to marshall
    * @return true if the object is a return value to a method call; false otherwise
    */
   protected boolean isReturnValue(Object o)
   {
      return !(o instanceof MethodCall);
   }

   // -------- AbstractMarshaller interface

   public void objectToObjectStream(Object o, ObjectOutputStream out) throws Exception
   {
      if (useRegionBasedMarshalling)
      {
         Fqn region;
         if (o == null)
         {
            // if the return value we're trying to marshall is null we're easy ...
            region = null;
            // we still need to clear the thread local though.
            regionForCall.remove();
         }
         else if (isReturnValue(o))
         {
            // we are marshalling a return value from a method call.
            // let's see if an incoming unmarshalling call for this exists, in the same thread stack and had registered
            // a Fqn region.
            region = regionForCall.get();
            regionForCall.remove();
            if (log.isTraceEnabled()) log.trace("Suspect this is a return value.  Extract region from ThreadLocal as " + region);

            // otherwise, we need to marshall the retval.
         }
         else
         {
            // this is an outgoing method call.
            // we first marshall the Fqn as a String
            MethodCall call = (MethodCall) o;
            region = extractFqnRegion(call);
         }
         if (log.isTraceEnabled()) log.trace("Region based call.  Using region " + region);
         objectToObjectStream(o, out, region);
      }
      else
      {
         // not region based!
         if (log.isTraceEnabled()) log.trace("Marshalling object " + o);
         Map<Object, Integer> refMap = new HashMap<Object, Integer>();
         marshallObject(o, out, refMap);
      }
   }

   public Object objectFromObjectStream(ObjectInputStream in) throws Exception
   {
      if (useRegionBasedMarshalling)
      {
         return objectFromObjectStreamRegionBased(in);
      }
      else
      {
         Map<Integer, Object> refMap = new HashMap<Integer, Object>();
         Object retValue = unmarshallObject(in, refMap);
         if (log.isTraceEnabled()) log.trace("Unmarshalled object " + retValue);
         return retValue;
      }
   }

   public void objectToObjectStream(Object o, ObjectOutputStream out, Fqn region) throws Exception
   {
      if (log.isTraceEnabled()) log.trace("Marshalling object " + o);
      Map<Object, Integer> refMap = new HashMap<Object, Integer>();
      if (useRegionBasedMarshalling) // got to check again in case this meth is called directly
      {
         log.trace("Writing region " + region + " to stream");
         marshallObject(region, out, refMap);
      }
      marshallObject(o, out, refMap);
   }

   protected Object objectFromObjectStreamRegionBased(ObjectInputStream in) throws Exception
   {
      Map<Integer, Object> refMap = new HashMap<Integer, Object>();
      Object o = unmarshallObject(in, refMap);
      Fqn regionFqn = null;
      if (o == null)
      {
         // a null region.  Could happen.  Use std marshalling.
         log.trace("Unmarshalled region as null.  Not using a context class loader to unmarshall.");
      }
      else
      {
         regionFqn = (Fqn) o;
      }

      if (log.isTraceEnabled()) log.trace("Unmarshalled regionFqn " + regionFqn + " from stream");

      Region region = null;
      Object retValue;

      if (regionFqn != null)
      {
         region = findRegion(regionFqn);
      }
      if (region == null)
      {
         if (log.isDebugEnabled()) log.debug("Region does not exist for Fqn " + regionFqn + " - not using a context classloader.");
         retValue = unmarshallObject(in, refMap);
      }
      else
      {
         retValue = unmarshallObject(in, region.getClassLoader(), refMap);

         // only set this if this is an incoming method call and not a return value.
         if (!isReturnValue(retValue)) regionForCall.set(regionFqn);
      }
      if (log.isTraceEnabled()) log.trace("Unmarshalled object " + retValue);
      return retValue;
   }

   private Region findRegion(Fqn fqn) throws InactiveRegionException
   {
      Region region = getRegion(fqn);

      if (region != null)
      {
         Status status = region.getStatus();
         if (status == Status.INACTIVATING || status == Status.INACTIVE)
         {
            if (log.isDebugEnabled())
            {
               throw new InactiveRegionException("Cannot unmarshall message for region " + fqn + ". This region is inactive.");
            }
            else
            {
               throw IRE;
            }
         }
      }
      else if (defaultInactive)
      {
         if (log.isDebugEnabled())
         {
            // No region but default inactive means region is inactive
            throw new InactiveRegionException("Cannot unmarshall message for region " + fqn + ". By default region " + fqn
                    + " is inactive.");
         }
         else
         {
            throw IRE;
         }
      }

      return region;
   }

   private Fqn extractFqnRegion(MethodCall call) throws Exception
   {
      Fqn fqn;
      if (call.getMethod().equals(MethodDeclarations.replicateMethod))
      {
         fqn = extractFqnFromMethodCall(call);
      }
      else if (call.getMethod().equals(MethodDeclarations.replicateAllMethod))
      {
         fqn = extractFqnFromListOfMethodCall(call);
      }
      else
      {
         fqn = extractFqn(call);
      }

      Region r = getRegion(fqn);
      return r == null ? null : r.getFqn();
   }

   // --------- Marshalling methods

   private void marshallObject(Object o, ObjectOutputStream out, Map<Object, Integer> refMap) throws Exception
   {
      if (o == null)
      {
         out.writeByte(MAGICNUMBER_NULL);
      }
      else if (refMap.containsKey(o))// see if this object has been marshalled before.
      {
         out.writeByte(MAGICNUMBER_REF);
         out.writeShort(refMap.get(o));
      }
      else if (o instanceof MethodCall)
      {
         // first see if this is a 'known' method call.
         MethodCall call = (MethodCall) o;

         if (call.getMethodId() > -1)
         {
            out.writeByte(MAGICNUMBER_METHODCALL);
            marshallMethodCall(call, out, refMap);
         }
         else
         {
            throw new IllegalArgumentException("MethodCall does not have a valid method id.  Was this method call created with MethodCallFactory?");
         }
      }
      else if (o instanceof Fqn)
      {
         int refId = createReference(o, refMap);
         out.writeByte(MAGICNUMBER_FQN);
         out.writeShort(refId);
         marshallFqn((Fqn) o, out, refMap);
      }
      else if (o instanceof GlobalTransaction)
      {
         int refId = createReference(o, refMap);
         out.writeByte(MAGICNUMBER_GTX);
         out.writeShort(refId);
         marshallGlobalTransaction((GlobalTransaction) o, out, refMap);
      }
      else if (o instanceof IpAddress)
      {
         out.writeByte(MAGICNUMBER_IPADDRESS);
         marshallIpAddress((IpAddress) o, out);
      }
      else if (o instanceof DefaultDataVersion)
      {
         out.writeByte(MAGICNUMBER_DEFAULT_DATA_VERSION);
         out.writeLong(((DefaultDataVersion) o).getRawVersion());
      }
      else if (o.getClass().equals(ArrayList.class))
      {
         out.writeByte(MAGICNUMBER_ARRAY_LIST);
         marshallCollection((Collection) o, out, refMap);
      }
      else if (o.getClass().equals(LinkedList.class))
      {
         out.writeByte(MAGICNUMBER_LINKED_LIST);
         marshallCollection((Collection) o, out, refMap);
      }
      else if (o.getClass().equals(HashMap.class))
      {
         out.writeByte(MAGICNUMBER_HASH_MAP);
         marshallMap((Map) o, out, refMap);
      }
      else if (o.getClass().equals(TreeMap.class))
      {
         out.writeByte(MAGICNUMBER_TREE_MAP);
         marshallMap((Map) o, out, refMap);
      }
      else if (o.getClass().equals(MapCopy.class))
      {
         out.writeByte(MAGICNUMBER_MAPCOPY);
         marshallMap((Map) o, out, refMap);
      }
      else if (o.getClass().equals(HashSet.class))
      {
         out.writeByte(MAGICNUMBER_HASH_SET);
         marshallCollection((Collection) o, out, refMap);
      }
      else if (o.getClass().equals(TreeSet.class))
      {
         out.writeByte(MAGICNUMBER_TREE_SET);
         marshallCollection((Collection) o, out, refMap);
      }
      else if (o instanceof Boolean)
      {
         out.writeByte(MAGICNUMBER_BOOLEAN);
         out.writeBoolean(((Boolean) o).booleanValue());
      }
      else if (o instanceof Integer)
      {
         out.writeByte(MAGICNUMBER_INTEGER);
         out.writeInt(((Integer) o).intValue());
      }
      else if (o instanceof Long)
      {
         out.writeByte(MAGICNUMBER_LONG);
         out.writeLong(((Long) o).longValue());
      }
      else if (o instanceof Short)
      {
         out.writeByte(MAGICNUMBER_SHORT);
         out.writeShort(((Short) o).shortValue());
      }
      else if (o instanceof String)
      {
         int refId = createReference(o, refMap);
         out.writeByte(MAGICNUMBER_STRING);
         out.writeShort(refId);
         marshallString((String) o, out);
      }
      else if (o instanceof NodeDataMarker)
      {
         out.writeByte(MAGICNUMBER_NODEDATA_MARKER);
         ((Externalizable) o).writeExternal(out);
      }
      else if (o instanceof NodeDataExceptionMarker)
      {
         out.writeByte(MAGICNUMBER_NODEDATA_EXCEPTION_MARKER);
         ((Externalizable) o).writeExternal(out);
      }
      else if (o instanceof NodeData)
      {
         out.writeByte(MAGICNUMBER_NODEDATA);
         ((Externalizable) o).writeExternal(out);
      }
      else if (o instanceof GravitateResult)
      {
         out.writeByte(MAGICNUMBER_GRAVITATERESULT);
         marshallGravitateResult((GravitateResult) o, out, refMap);
      }
      else if (o instanceof Serializable)
      {
         int refId = createReference(o, refMap);
         if (log.isTraceEnabled())
         {
            log.trace("Warning: using object serialization for " + o.getClass());
         }
         out.writeByte(MAGICNUMBER_SERIALIZABLE);
         out.writeShort(refId);
         out.writeObject(o);
      }
      else
      {
         throw new Exception("Don't know how to marshall object of type " + o.getClass());
      }
   }

   private void marshallGravitateResult(GravitateResult gravitateResult, ObjectOutputStream out, Map<Object, Integer> refMap) throws Exception
   {
      marshallObject(gravitateResult.isDataFound(), out, refMap);
      if (gravitateResult.isDataFound())
      {
         marshallObject(gravitateResult.getNodeData(), out, refMap);
         marshallObject(gravitateResult.getBuddyBackupFqn(), out, refMap);
      }

   }

   private int createReference(Object o, Map<Object, Integer> refMap)
   {
      int reference = refMap.size();
      refMap.put(o, reference);
      return reference;
   }

   private void marshallString(String s, ObjectOutputStream out) throws Exception
   {
      //StringUtil.saveString(out, s);
      out.writeObject(s);
   }

   private void marshallMethodCall(MethodCall methodCall, ObjectOutputStream out, Map<Object, Integer> refMap) throws Exception
   {
      out.writeShort(methodCall.getMethodId());
      Object[] args = methodCall.getArgs();
      byte numArgs = (byte) (args == null ? 0 : args.length);
      out.writeByte(numArgs);

      for (int i = 0; i < numArgs; i++)
      {
         marshallObject(args[i], out, refMap);
      }
   }

   private void marshallGlobalTransaction(GlobalTransaction globalTransaction, ObjectOutputStream out, Map<Object, Integer> refMap) throws Exception
   {
      out.writeLong(globalTransaction.getId());
      marshallObject(globalTransaction.getAddress(), out, refMap);
   }


   private void marshallFqn(Fqn fqn, ObjectOutputStream out, Map<Object, Integer> refMap) throws Exception
   {
      boolean isRoot = fqn.isRoot();
      out.writeBoolean(isRoot);
      if (!isRoot)
      {
         out.writeShort(fqn.size());
         for (int i = 0; i < fqn.size(); i++)
         {
            marshallObject(fqn.get(i), out, refMap);
         }
      }
   }

   private void marshallIpAddress(IpAddress ipAddress, ObjectOutputStream out) throws Exception
   {
      ipAddress.writeExternal(out);
   }

   private void marshallCollection(Collection c, ObjectOutputStream out, Map refMap) throws Exception
   {
      out.writeInt(c.size());
      for (Object o : c)
      {
         marshallObject(o, out, refMap);
      }
   }

   private void marshallMap(Map map, ObjectOutputStream out, Map<Object, Integer> refMap) throws Exception
   {
      int mapSize = map.size();
      out.writeInt(mapSize);
      if (mapSize == 0) return;

      for (Map.Entry me : (Set<Map.Entry>) map.entrySet())
      {
         marshallObject(me.getKey(), out, refMap);
         marshallObject(me.getValue(), out, refMap);
      }
   }

   // --------- Unmarshalling methods

   private Object unmarshallObject(ObjectInputStream in, ClassLoader loader, Map<Integer, Object> refMap) throws Exception
   {
      if (loader == null)
      {
         return unmarshallObject(in, refMap);
      }
      else
      {
         Thread currentThread = Thread.currentThread();
         ClassLoader old = currentThread.getContextClassLoader();
         try
         {
            currentThread.setContextClassLoader(loader);
            return unmarshallObject(in, refMap);
         }
         finally
         {
            currentThread.setContextClassLoader(old);
         }
      }
   }

   private Object unmarshallObject(ObjectInputStream in, Map<Integer, Object> refMap) throws Exception
   {
      byte magicNumber = in.readByte();
      Integer reference;
      Object retVal;
      switch (magicNumber)
      {
         case MAGICNUMBER_NULL:
            return null;
         case MAGICNUMBER_REF:
            reference = (int) in.readShort();
            if (!refMap.containsKey(reference))
            {
               throw new IOException("Unable to locate object reference " + reference + " in byte stream!");
            }
            return refMap.get(reference);
         case MAGICNUMBER_SERIALIZABLE:
            reference = (int) in.readShort();
            retVal = in.readObject();
            refMap.put(reference, retVal);
            return retVal;
         case MAGICNUMBER_METHODCALL:
            retVal = unmarshallMethodCall(in, refMap);
            return retVal;
         case MAGICNUMBER_FQN:
            reference = (int) in.readShort();
            retVal = unmarshallFqn(in, refMap);
            refMap.put(reference, retVal);
            return retVal;
         case MAGICNUMBER_GTX:
            reference = (int) in.readShort();
            retVal = unmarshallGlobalTransaction(in, refMap);
            refMap.put(reference, retVal);
            return retVal;
         case MAGICNUMBER_IPADDRESS:
            retVal = unmarshallIpAddress(in);
            return retVal;
         case MAGICNUMBER_DEFAULT_DATA_VERSION:
            retVal = new DefaultDataVersion(in.readLong());
            return retVal;
         case MAGICNUMBER_ARRAY_LIST:
            return unmarshallArrayList(in, refMap);
         case MAGICNUMBER_LINKED_LIST:
            return unmarshallLinkedList(in, refMap);
         case MAGICNUMBER_HASH_MAP:
            return unmarshallHashMap(in, refMap);
         case MAGICNUMBER_TREE_MAP:
            return unmarshallTreeMap(in, refMap);
         case MAGICNUMBER_HASH_SET:
            return unmarshallHashSet(in, refMap);
         case MAGICNUMBER_TREE_SET:
            return unmarshallTreeSet(in, refMap);
         case MAGICNUMBER_MAPCOPY:
            return unmarshallMapCopy(in, refMap);
         case MAGICNUMBER_BOOLEAN:
            return in.readBoolean() ? Boolean.TRUE : Boolean.FALSE;
         case MAGICNUMBER_INTEGER:
            return in.readInt();
         case MAGICNUMBER_LONG:
            return in.readLong();
         case MAGICNUMBER_SHORT:
            return in.readShort();
         case MAGICNUMBER_STRING:
            reference = (int) in.readShort();
            retVal = unmarshallString(in);
            refMap.put(reference, retVal);
            return retVal;
         case MAGICNUMBER_NODEDATA_MARKER:
            retVal = new NodeDataMarker();
            ((NodeDataMarker) retVal).readExternal(in);
            return retVal;
         case MAGICNUMBER_NODEDATA_EXCEPTION_MARKER:
            retVal = new NodeDataExceptionMarker();
            ((NodeDataExceptionMarker) retVal).readExternal(in);
            return retVal;
         case MAGICNUMBER_NODEDATA:
            retVal = new NodeData();
            ((NodeData) retVal).readExternal(in);
            return retVal;
         case MAGICNUMBER_GRAVITATERESULT:
            return unmarshallGravitateResult(in, refMap);
         default:
            if (log.isErrorEnabled())
            {
               log.error("Unknown Magic Number " + magicNumber);
            }
            throw new Exception("Unknown magic number " + magicNumber);
      }
   }

   private GravitateResult unmarshallGravitateResult(ObjectInputStream in, Map<Integer, Object> refMap) throws Exception
   {
      Boolean found = (Boolean) unmarshallObject(in, refMap);
      if (found)
      {
         List<NodeData> stuff = (List<NodeData>) unmarshallObject(in, refMap);
         Fqn fqn = (Fqn) unmarshallObject(in, refMap);
         return GravitateResult.subtreeResult(stuff, fqn);
      }
      else
      {
         return GravitateResult.noDataFound();
      }
   }

   private String unmarshallString(ObjectInputStream in) throws Exception
   {
      //return StringUtil.readString(in, null);
      return (String) in.readObject();
   }

   private MethodCall unmarshallMethodCall(ObjectInputStream in, Map<Integer, Object> refMap) throws Exception
   {
      short methodId = in.readShort();
      byte numArgs = in.readByte();
      Object[] args = null;

      if (numArgs > 0)
      {
         args = new Object[numArgs];

         for (int i = 0; i < numArgs; i++)
         {
            args[i] = unmarshallObject(in, refMap);
         }
      }
      return MethodCallFactory.create(MethodDeclarations.lookupMethod(methodId), args);
   }

   private GlobalTransaction unmarshallGlobalTransaction(ObjectInputStream in, Map<Integer, Object> refMap) throws Exception
   {
      GlobalTransaction gtx = new GlobalTransaction();
      long id = in.readLong();
      Object address = unmarshallObject(in, refMap);
      gtx.setId(id);
      gtx.setAddress((Address) address);
      return gtx;
   }

   Fqn unmarshallFqn(ObjectInputStream in, Map<Integer, Object> refMap) throws Exception
   {

      boolean isRoot = in.readBoolean();
      Fqn fqn;
      if (!isRoot)
      {
         int numElements = in.readShort();
         List<Object> elements = new ArrayList<Object>(numElements);
         for (int i = 0; i < numElements; i++)
         {
            elements.add(unmarshallObject(in, refMap));
         }
         fqn = new Fqn(elements);
      }
      else
      {
         fqn = Fqn.ROOT;
      }
      return fqn;
   }

   private IpAddress unmarshallIpAddress(ObjectInputStream in) throws Exception
   {
      IpAddress ipAddress = new IpAddress();
      ipAddress.readExternal(in);
      return ipAddress;
   }

   private List unmarshallArrayList(ObjectInputStream in, Map refMap) throws Exception
   {
      int listSize = in.readInt();
      List list = new ArrayList(listSize);
      populateFromStream(in, refMap, list, listSize);
      return list;
   }

   private List unmarshallLinkedList(ObjectInputStream in, Map refMap) throws Exception
   {
      List list = new LinkedList();
      populateFromStream(in, refMap, list, in.readInt());
      return list;
   }

   private Map unmarshallHashMap(ObjectInputStream in, Map refMap) throws Exception
   {
      Map map = new HashMap();
      populateFromStream(in, refMap, map);
      return map;
   }

   private Map unmarshallMapCopy(ObjectInputStream in, Map<Integer, Object> refMap) throws Exception
   {
      // read in as a HashMap first
      Map m = unmarshallHashMap(in, refMap);
      MapCopy mc = new MapCopy(m);
      return mc;
   }

   private Map unmarshallTreeMap(ObjectInputStream in, Map refMap) throws Exception
   {
      Map map = new TreeMap();
      populateFromStream(in, refMap, map);
      return map;
   }

   private Set unmarshallHashSet(ObjectInputStream in, Map refMap) throws Exception
   {
      Set set = new HashSet();
      populateFromStream(in, refMap, set);
      return set;
   }

   private Set unmarshallTreeSet(ObjectInputStream in, Map refMap) throws Exception
   {
      Set set = new TreeSet();
      populateFromStream(in, refMap, set);
      return set;
   }

   private void populateFromStream(ObjectInputStream in, Map refMap, Map mapToPopulate) throws Exception
   {
      int size = in.readInt();
      for (int i = 0; i < size; i++)
      {
         mapToPopulate.put(unmarshallObject(in, refMap), unmarshallObject(in, refMap));
      }
   }

   private void populateFromStream(ObjectInputStream in, Map refMap, Set setToPopulate) throws Exception
   {
      int size = in.readInt();
      for (int i = 0; i < size; i++)
      {
         setToPopulate.add(unmarshallObject(in, refMap));
      }
   }

   private void populateFromStream(ObjectInputStream in, Map refMap, List listToPopulate, int listSize) throws Exception
   {
      for (int i = 0; i < listSize; i++)
      {
         listToPopulate.add(unmarshallObject(in, refMap));
      }
   }
}
