/*
 * JBoss, the OpenSource J2EE webOS
 * 
 * Distributable under LGPL license.
 * See terms of license at gnu.org.
 */
package org.jboss.cache.marshall;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.jboss.cache.CacheException;
import org.jboss.cache.CacheImpl;
import org.jboss.cache.Fqn;
import org.jboss.cache.Node;
import org.jboss.cache.buddyreplication.BuddyGroup;
import org.jboss.cache.lock.NodeLock;
import org.jboss.cache.optimistic.DataVersion;
import org.jboss.cache.transaction.GlobalTransaction;
import org.jgroups.Address;

import java.lang.reflect.Method;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

/**
 * Class containing Method and Method id definitions as well methods
 * allowing lookup operations both ways.
 *
 * @author <a href="galder.zamarreno@jboss.com">Galder Zamarreno</a>
 * @version $Revision: 4484 $
 */
public class MethodDeclarations
{
   private static Log log = LogFactory.getLog(MethodDeclarations.class);

   static final Set<Integer> crudMethodIds = new HashSet<Integer>();
   static final Set<Integer> transactionLifecycleMethodIds = new HashSet<Integer>();
   static final Set<Integer> buddyGroupOrganisationMethodIds = new HashSet<Integer>();
   static final Set<Integer> blockUnblockMethodIds = new HashSet<Integer>();
   static final Set<Integer> putMethodIds = new HashSet<Integer>();
   static final Set<Integer> methodsThatNeedToReturnValuesToRemoteCallers = new HashSet<Integer>();

   // maintain a list of method IDs that correspond to Methods in CacheImpl
   static final Map<Integer, Method> methods = new HashMap<Integer, Method>();

   // and for reverse lookup
   static final Map<Method, Integer> methodIds = new HashMap<Method, Integer>();

   public static final Method putDataMethodLocal;

   public static final Method putDataEraseMethodLocal;

   public static final Method putKeyValMethodLocal;

   public static final Method removeNodeMethodLocal;

   public static final Method removeKeyMethodLocal;

   public static final Method removeDataMethodLocal;

   public static final Method evictNodeMethodLocal;

   public static final Method evictVersionedNodeMethodLocal;

   public static final Method prepareMethod;

   public static final Method commitMethod;

   public static final Method rollbackMethod;

   public static final Method replicateMethod;

   public static final Method replicateAllMethod;

   public static final Method addChildMethodLocal;

   public static final Method getKeyValueMethodLocal;

   public static final Method getNodeMethodLocal;

   public static final Method getKeysMethodLocal;

   public static final Method getChildrenNamesMethodLocal;

   public static final Method getDataMapMethodLocal;

   public static final Method existsMethod;

   public static final Method releaseAllLocksMethodLocal;

   public static final Method printMethodLocal;

   public static final Method lockMethodLocal;

   public static final Method optimisticPrepareMethod;

   public static final Method clusteredGetMethod;

   public static final Method remoteAssignToBuddyGroupMethod;

   public static final Method remoteRemoveFromBuddyGroupMethod;

   public static final Method remoteAnnounceBuddyPoolNameMethod;

   public static final Method dataGravitationCleanupMethod;

   public static final Method dataGravitationMethod;

   public static final Method moveMethodLocal;

   public static final Method blockChannelLocal;

   public static final Method unblockChannelLocal;

   public static final Method putForExternalReadMethodLocal;

   // these are basic crud methods that are version-aware - JBCACHE-843.

   public static final Method putDataVersionedMethodLocal;

   public static final Method putDataEraseVersionedMethodLocal;

   public static final Method putKeyValVersionedMethodLocal;

   public static final Method putForExternalReadVersionedMethodLocal;

   public static final Method removeNodeVersionedMethodLocal;

   public static final Method removeKeyVersionedMethodLocal;

   public static final Method removeDataVersionedMethodLocal;

   public static final Method invalidateMethodLocal;

   //not all of these are used for RPC - trim accordingly.
   public static final int putDataMethodLocal_id = 1;

   public static final int putDataEraseMethodLocal_id = 2;

   public static final int putKeyValMethodLocal_id = 3;

   //    public static final int putFailFastKeyValueMethodLocal_id = 4;  // this method is no longer needed.

   public static final int removeNodeMethodLocal_id = 5;

   public static final int removeKeyMethodLocal_id = 6;

   public static final int removeDataMethodLocal_id = 7;

   public static final int evictNodeMethodLocal_id = 8;

   public static final int evictVersionedNodeMethodLocal_id = 9;

   public static final int prepareMethod_id = 10;

   public static final int commitMethod_id = 11;

   public static final int rollbackMethod_id = 12;

   public static final int replicateMethod_id = 13;

   public static final int replicateAllMethod_id = 14;

   public static final int addChildMethodLocal_id = 15;

   public static final int existsMethod_id = 16;

   public static final int releaseAllLocksMethodLocal_id = 17;

   public static final int optimisticPrepareMethod_id = 18;

   public static final int getPartialStateMethod_id = 19;

   public static final int clusteredGetMethod_id = 22;

   public static final int getChildrenNamesMethodLocal_id = 23;

   public static final int getDataMapMethodLocal_id = 24;

   public static final int getKeysMethodLocal_id = 25;

   public static final int getKeyValueMethodLocal_id = 26;

   public static final int remoteAnnounceBuddyPoolNameMethod_id = 28;

   public static final int remoteAssignToBuddyGroupMethod_id = 29;

   public static final int remoteRemoveFromBuddyGroupMethod_id = 30;

   /* Method id added as they did not exist before refactoring */
   public static final int getNodeMethodLocal_id = 31;

   public static final int printMethodLocal_id = 32;

   public static final int lockMethodLocal_id = 33;

   public static final int dataGravitationCleanupMethod_id = 34;

   public static final int dataGravitationMethod_id = 35;

   public static final int moveMethodLocal_id = 36;

   // these are basic crud methods that are version-aware - JBCACHE-843.

   public static final int putDataVersionedMethodLocal_id = 37;

   public static final int putDataEraseVersionedMethodLocal_id = 38;

   public static final int putKeyValVersionedMethodLocal_id = 39;

   public static final int removeNodeVersionedMethodLocal_id = 40;

   public static final int removeKeyVersionedMethodLocal_id = 41;

   public static final int removeDataVersionedMethodLocal_id = 42;

   public static final int blockChannelMethodLocal_id = 43;

   public static final int unblockChannelMethodLocal_id = 44;

   public static final int putForExternalReadMethodLocal_id = 45;

   public static final int putForExternalReadVersionedMethodLocal_id = 46;

   public static final int invalidateMethodLocal_id = 47;


   static
   {
      try
      {
         // these methods need to return values even to remote callers.
         methodsThatNeedToReturnValuesToRemoteCallers.add(dataGravitationMethod_id);
         methodsThatNeedToReturnValuesToRemoteCallers.add(clusteredGetMethod_id);


         getDataMapMethodLocal = CacheImpl.class.getDeclaredMethod("_getData", Fqn.class);
         existsMethod = CacheImpl.class.getDeclaredMethod("exists", Fqn.class);
         putDataMethodLocal = CacheImpl.class.getDeclaredMethod("_put", GlobalTransaction.class, Fqn.class, Map.class, boolean.class);
         putDataEraseMethodLocal = CacheImpl.class.getDeclaredMethod("_put", GlobalTransaction.class, Fqn.class, Map.class, boolean.class, boolean.class);
         putKeyValMethodLocal = CacheImpl.class.getDeclaredMethod("_put", GlobalTransaction.class, Fqn.class, Object.class, Object.class, boolean.class);
         putForExternalReadMethodLocal = CacheImpl.class.getDeclaredMethod("_putForExternalRead", GlobalTransaction.class, Fqn.class, Object.class, Object.class);
         removeNodeMethodLocal = CacheImpl.class.getDeclaredMethod("_remove", GlobalTransaction.class, Fqn.class, boolean.class);
         removeKeyMethodLocal = CacheImpl.class.getDeclaredMethod("_remove", GlobalTransaction.class, Fqn.class, Object.class, boolean.class);
         removeDataMethodLocal = CacheImpl.class.getDeclaredMethod("_removeData", GlobalTransaction.class, Fqn.class, boolean.class);
         evictNodeMethodLocal = CacheImpl.class.getDeclaredMethod("_evict", Fqn.class);
         evictVersionedNodeMethodLocal = CacheImpl.class.getDeclaredMethod("_evict", Fqn.class, DataVersion.class);

         // evictKeyValueMethodLocal=CacheImpl.class.getDeclaredMethod("_evict", new Class[]{Fqn.class, Object.class});
         prepareMethod = CacheImpl.class.getDeclaredMethod("prepare", GlobalTransaction.class, List.class, Address.class, boolean.class);
         commitMethod = CacheImpl.class.getDeclaredMethod("commit", GlobalTransaction.class);
         rollbackMethod = CacheImpl.class.getDeclaredMethod("rollback", GlobalTransaction.class);
         addChildMethodLocal = CacheImpl.class.getDeclaredMethod("_addChild", GlobalTransaction.class, Fqn.class, Object.class, Node.class, boolean.class);
         getKeyValueMethodLocal = CacheImpl.class.getDeclaredMethod("_get", Fqn.class, Object.class, boolean.class);
         getNodeMethodLocal = CacheImpl.class.getDeclaredMethod("_get", Fqn.class);
         getKeysMethodLocal = CacheImpl.class.getDeclaredMethod("_getKeys", Fqn.class);
         getChildrenNamesMethodLocal = CacheImpl.class.getDeclaredMethod("_getChildrenNames", Fqn.class);
         replicateMethod = CacheImpl.class.getDeclaredMethod("_replicate", MethodCall.class);
         replicateAllMethod = CacheImpl.class.getDeclaredMethod("_replicate", List.class);
         releaseAllLocksMethodLocal = CacheImpl.class.getDeclaredMethod("_releaseAllLocks", Fqn.class);
         printMethodLocal = CacheImpl.class.getDeclaredMethod("_print", Fqn.class);
         lockMethodLocal = CacheImpl.class.getDeclaredMethod("_lock", Fqn.class, NodeLock.LockType.class, boolean.class);

         optimisticPrepareMethod = CacheImpl.class.getDeclaredMethod("optimisticPrepare", GlobalTransaction.class, List.class, Map.class, Address.class, boolean.class);

         clusteredGetMethod = CacheImpl.class.getDeclaredMethod("_clusteredGet", MethodCall.class, Boolean.class);

         // ------------ buddy replication

         remoteAnnounceBuddyPoolNameMethod = CacheImpl.class.getDeclaredMethod("_remoteAnnounceBuddyPoolName", Address.class, String.class);
         remoteRemoveFromBuddyGroupMethod = CacheImpl.class.getDeclaredMethod("_remoteRemoveFromBuddyGroup", String.class);
         remoteAssignToBuddyGroupMethod = CacheImpl.class.getDeclaredMethod("_remoteAssignToBuddyGroup", BuddyGroup.class, Map.class);

         dataGravitationCleanupMethod = CacheImpl.class.getDeclaredMethod("_dataGravitationCleanup", GlobalTransaction.class, Fqn.class, Fqn.class);
         dataGravitationMethod = CacheImpl.class.getDeclaredMethod("gravitateData", Fqn.class, boolean.class);

         // ------------ move() api
         moveMethodLocal = CacheImpl.class.getDeclaredMethod("_move", Fqn.class, Fqn.class);

         // ------------ Channel BLOCK event 
         blockChannelLocal = CacheImpl.class.getDeclaredMethod("_block");
         unblockChannelLocal = CacheImpl.class.getDeclaredMethod("_unblock");

         // ------------ version-aware methods - see JBCACHE-843
         putDataVersionedMethodLocal = CacheImpl.class.getDeclaredMethod("_put", GlobalTransaction.class, Fqn.class, Map.class, boolean.class, DataVersion.class);
         putDataEraseVersionedMethodLocal = CacheImpl.class.getDeclaredMethod("_put", GlobalTransaction.class, Fqn.class, Map.class, boolean.class, boolean.class, DataVersion.class);
         putKeyValVersionedMethodLocal = CacheImpl.class.getDeclaredMethod("_put", GlobalTransaction.class, Fqn.class, Object.class, Object.class, boolean.class, DataVersion.class);
         putForExternalReadVersionedMethodLocal = CacheImpl.class.getDeclaredMethod("_putForExternalRead", GlobalTransaction.class, Fqn.class, Object.class, Object.class, DataVersion.class);
         removeNodeVersionedMethodLocal = CacheImpl.class.getDeclaredMethod("_remove", GlobalTransaction.class, Fqn.class, boolean.class, DataVersion.class);
         removeKeyVersionedMethodLocal = CacheImpl.class.getDeclaredMethod("_remove", GlobalTransaction.class, Fqn.class, Object.class, boolean.class, DataVersion.class);
         removeDataVersionedMethodLocal = CacheImpl.class.getDeclaredMethod("_removeData", GlobalTransaction.class, Fqn.class, boolean.class, DataVersion.class);

         invalidateMethodLocal = CacheImpl.class.getDeclaredMethod("invalidate", Fqn.class, DataVersion.class);

      }
      catch (NoSuchMethodException e)
      {
         throw new ExceptionInInitializerError(e);
      }

      methods.put(putDataMethodLocal_id, putDataMethodLocal);
      methods.put(putDataEraseMethodLocal_id, putDataEraseMethodLocal);
      methods.put(putKeyValMethodLocal_id, putKeyValMethodLocal);
      methods.put(removeNodeMethodLocal_id, removeNodeMethodLocal);
      methods.put(removeKeyMethodLocal_id, removeKeyMethodLocal);
      methods.put(removeDataMethodLocal_id, removeDataMethodLocal);
      methods.put(evictNodeMethodLocal_id, evictNodeMethodLocal);
      methods.put(evictVersionedNodeMethodLocal_id, evictVersionedNodeMethodLocal);
      methods.put(prepareMethod_id, prepareMethod);
      methods.put(commitMethod_id, commitMethod);
      methods.put(rollbackMethod_id, rollbackMethod);
      methods.put(replicateMethod_id, replicateMethod);
      methods.put(replicateAllMethod_id, replicateAllMethod);
      methods.put(addChildMethodLocal_id, addChildMethodLocal);
      methods.put(existsMethod_id, existsMethod);
      methods.put(releaseAllLocksMethodLocal_id, releaseAllLocksMethodLocal);
      methods.put(optimisticPrepareMethod_id, optimisticPrepareMethod);
      methods.put(clusteredGetMethod_id, clusteredGetMethod);
      methods.put(getChildrenNamesMethodLocal_id, getChildrenNamesMethodLocal);
      methods.put(getDataMapMethodLocal_id, getDataMapMethodLocal);
      methods.put(getKeysMethodLocal_id, getKeysMethodLocal);
      methods.put(getKeyValueMethodLocal_id, getKeyValueMethodLocal);
      methods.put(remoteAnnounceBuddyPoolNameMethod_id, remoteAnnounceBuddyPoolNameMethod);
      methods.put(remoteAssignToBuddyGroupMethod_id, remoteAssignToBuddyGroupMethod);
      methods.put(remoteRemoveFromBuddyGroupMethod_id, remoteRemoveFromBuddyGroupMethod);
      /* Mappings added as they did not exist before refactoring */
      methods.put(getNodeMethodLocal_id, getNodeMethodLocal);
      methods.put(printMethodLocal_id, printMethodLocal);
      methods.put(lockMethodLocal_id, lockMethodLocal);

      methods.put(dataGravitationCleanupMethod_id, dataGravitationCleanupMethod);
      methods.put(dataGravitationMethod_id, dataGravitationMethod);

      methods.put(moveMethodLocal_id, moveMethodLocal);
      methods.put(blockChannelMethodLocal_id, blockChannelLocal);
      methods.put(unblockChannelMethodLocal_id, unblockChannelLocal);

      methods.put(putDataVersionedMethodLocal_id, putDataVersionedMethodLocal);
      methods.put(putDataEraseVersionedMethodLocal_id, putDataEraseVersionedMethodLocal);
      methods.put(putKeyValVersionedMethodLocal_id, putKeyValVersionedMethodLocal);
      methods.put(removeDataVersionedMethodLocal_id, removeDataVersionedMethodLocal);
      methods.put(removeKeyVersionedMethodLocal_id, removeKeyVersionedMethodLocal);
      methods.put(removeNodeVersionedMethodLocal_id, removeNodeVersionedMethodLocal);

      methods.put(putForExternalReadVersionedMethodLocal_id, putForExternalReadVersionedMethodLocal);
      methods.put(putForExternalReadMethodLocal_id, putForExternalReadMethodLocal);

      methods.put(invalidateMethodLocal_id, invalidateMethodLocal);

      for (Integer id : methods.keySet())
      {
         methodIds.put(methods.get(id), id);
      }

      putMethodIds.add(putDataMethodLocal_id);
      putMethodIds.add(putDataEraseMethodLocal_id);
      putMethodIds.add(putKeyValMethodLocal_id);
      putMethodIds.add(putDataEraseVersionedMethodLocal_id);
      putMethodIds.add(putDataVersionedMethodLocal_id);
      putMethodIds.add(putKeyValVersionedMethodLocal_id);
      putMethodIds.add(putForExternalReadMethodLocal_id);
      putMethodIds.add(putForExternalReadVersionedMethodLocal_id);

      crudMethodIds.addAll(putMethodIds);
      crudMethodIds.add(removeNodeMethodLocal_id);
      crudMethodIds.add(removeKeyMethodLocal_id);
      crudMethodIds.add(removeDataMethodLocal_id);
      crudMethodIds.add(dataGravitationCleanupMethod_id);
      crudMethodIds.add(moveMethodLocal_id);
      crudMethodIds.add(putDataVersionedMethodLocal_id);
      crudMethodIds.add(putDataEraseVersionedMethodLocal_id);
      crudMethodIds.add(putKeyValVersionedMethodLocal_id);
      crudMethodIds.add(removeNodeVersionedMethodLocal_id);
      crudMethodIds.add(removeKeyVersionedMethodLocal_id);
      crudMethodIds.add(removeDataVersionedMethodLocal_id);


      transactionLifecycleMethodIds.add(commitMethod_id);
      transactionLifecycleMethodIds.add(rollbackMethod_id);
      transactionLifecycleMethodIds.add(prepareMethod_id);
      transactionLifecycleMethodIds.add(optimisticPrepareMethod_id);

      buddyGroupOrganisationMethodIds.add(remoteAnnounceBuddyPoolNameMethod_id);
      buddyGroupOrganisationMethodIds.add(remoteAssignToBuddyGroupMethod_id);
      buddyGroupOrganisationMethodIds.add(remoteRemoveFromBuddyGroupMethod_id);

      blockUnblockMethodIds.add(blockChannelMethodLocal_id);
      blockUnblockMethodIds.add(unblockChannelMethodLocal_id);

   }

   protected static int lookupMethodId(Method method)
   {
      Integer methodIdInteger = methodIds.get(method);
      int methodId = -1;

      if (methodIdInteger != null)
      {
         methodId = methodIdInteger;
      }
      else
      {
         if (log.isWarnEnabled())
         {
            log.warn("Method " + method + " is not registered with " + CacheMarshaller200.class);
         }
      }

      return methodId;
   }

   protected static Method lookupMethod(int id)
   {
      Method method = methods.get(id);
      if (method == null)
      {
         if (log.isErrorEnabled())
         {
            log.error("Method id " + id + " is not registered");
         }
         throw new CacheException("Method id " + id + " is not registered with " + CacheMarshaller200.class);
      }
      return method;
   }

   /**
    * Returns true if the method is a CRUD method.
    */
   public static boolean isCrudMethod(int id)
   {
      return crudMethodIds.contains(Integer.valueOf(id));
   }

   public static boolean isTransactionLifecycleMethod(int id)
   {
      return transactionLifecycleMethodIds.contains(id);
   }

   public static boolean isBuddyGroupOrganisationMethod(int id)
   {
      return buddyGroupOrganisationMethodIds.contains(id);
   }

   public static boolean isPutMethod(int id)
   {
      return putMethodIds.contains(id);
   }

   public static boolean isGetMethod(int methodId)
   {
      return methodId == getChildrenNamesMethodLocal_id || methodId == getDataMapMethodLocal_id || methodId == existsMethod_id
              || methodId == getKeysMethodLocal_id || methodId == getKeyValueMethodLocal_id || methodId == getNodeMethodLocal_id;
   }

   public static boolean isBlockUnblockMethod(int id)
   {
      return blockUnblockMethodIds.contains(id);
   }

   /**
    * Returns the versioned equivalent of a crud method.
    */
   public static Method getVersionedMethod(int methodId)
   {
      if (isCrudMethod(methodId))
      {
         switch (methodId)
         {
            case putForExternalReadMethodLocal_id:
               return putForExternalReadVersionedMethodLocal;
            case putDataEraseMethodLocal_id:
               return putDataEraseVersionedMethodLocal;
            case putDataMethodLocal_id:
               return putDataVersionedMethodLocal;
            case putKeyValMethodLocal_id:
               return putKeyValVersionedMethodLocal;
            case removeDataMethodLocal_id:
               return removeDataVersionedMethodLocal;
            case removeKeyMethodLocal_id:
               return removeKeyVersionedMethodLocal;
            case removeNodeMethodLocal_id:
               return removeNodeVersionedMethodLocal;
            case moveMethodLocal_id:
               return moveMethodLocal;
            default:
               throw new CacheException("Unrecognised method id " + methodId);
         }
      }
      else
      {
         throw new CacheException("Attempting to look up a versioned equivalent of a non-crud method");
      }
   }

   /**
    * Counterpart to {@link #getVersionedMethod(int)}
    */
   public static Method getUnversionedMethod(int methodId)
   {
      if (isCrudMethod(methodId))
      {
         switch (methodId)
         {
            case putForExternalReadVersionedMethodLocal_id:
               return putForExternalReadMethodLocal;
            case putDataEraseVersionedMethodLocal_id:
               return putDataEraseMethodLocal;
            case putDataVersionedMethodLocal_id:
               return putDataMethodLocal;
            case putKeyValVersionedMethodLocal_id:
               return putKeyValMethodLocal;
            case removeDataVersionedMethodLocal_id:
               return removeDataMethodLocal;
            case removeKeyVersionedMethodLocal_id:
               return removeKeyMethodLocal;
            case removeNodeVersionedMethodLocal_id:
               return removeNodeMethodLocal;
            case moveMethodLocal_id:
               return moveMethodLocal;
            default:
               throw new CacheException("Unrecognised method id " + methodId);
         }
      }
      else
      {
         throw new CacheException("Attempting to look up a versioned equivalent of a non-crud method");
      }
   }


   public static boolean isDataGravitationMethod(int methodId)
   {
      return methodId == MethodDeclarations.dataGravitationCleanupMethod_id || methodId == MethodDeclarations.dataGravitationMethod_id;
   }

   /**
    * Tests whether remote calls to this method should return the value to the caller or just return a null (more efficient if the caller won't use this value anyway)
    *
    * @param methodId
    * @return true if the caller expects the value to come back.
    */
   public static boolean returnValueForRemoteCall(int methodId)
   {
      return methodsThatNeedToReturnValuesToRemoteCallers.contains(methodId);
   }
}

