/*
 * JBoss, Home of Professional Open Source
 *
 * Distributable under LGPL license.
 * See terms of license at gnu.org.
 */
package org.jboss.cache.marshall;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.jboss.cache.Fqn;
import org.jboss.cache.RegionManager;
import org.jboss.cache.config.Configuration;

import java.io.ByteArrayOutputStream;
import java.io.InputStream;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.util.HashMap;
import java.util.Map;
import java.util.StringTokenizer;

/**
 * A delegate to various other marshallers like {@link org.jboss.cache.marshall.CacheMarshaller200}.
 * This delegating marshaller adds versioning information to the stream when marshalling objects and
 * is able to pick the appropriate marshaller to delegate to based on the versioning information when
 * unmarshalling objects.
 *
 * @author <a href="mailto:manik@jboss.org">Manik Surtani (manik@jboss.org)</a>
 * @author <a href="mailto:galder.zamarreno@jboss.com">Galder Zamarreno</a>
 */
public class VersionAwareMarshaller extends AbstractMarshaller
{
   
   private static final Log log = LogFactory.getLog(VersionAwareMarshaller.class);
   private static final int VERSION_200 = 20;
   
   private RegionManager manager;
   private boolean defaultInactive, useRegionBasedMarshalling;
   AbstractMarshaller defaultMarshaller;
   Map<Integer, Marshaller> marshallers = new HashMap<Integer, Marshaller>();
   private int versionInt;

   public VersionAwareMarshaller(RegionManager manager, boolean defaultInactive, boolean useRegionBasedMarshalling, String version)
   {
      this.manager = manager;
      this.defaultInactive = defaultInactive;
      this.useRegionBasedMarshalling = useRegionBasedMarshalling;

      // "Rounds down" the replication version passed in to the MINOR version.
      // E.g., 1.4.1.SP3 -> 1.4.0

      versionInt = toMinorVersionInt(version);

      switch (versionInt)
      {
         case VERSION_200:
         default:
            defaultMarshaller = new CacheMarshaller200(manager, defaultInactive, useRegionBasedMarshalling);
            marshallers.put(VERSION_200, defaultMarshaller);
            break;
      }

      if (log.isDebugEnabled())
      {
         log.debug("Initialised with version " + version + " and versionInt " + versionInt);
         log.debug("Using default marshaller " + defaultMarshaller.getClass());
      }
   }

   public VersionAwareMarshaller(RegionManager regionManager, Configuration configuration)
   {
      this(regionManager, configuration.isInactiveOnStartup(), configuration.isUseRegionBasedMarshalling(), configuration.getReplVersionString());
   }

   /**
    * Converts versions to known compatible version ids.
    * <p/>
    * Typical return values:
    * <p/>
    * < 1.4.0 = "1"
    * 1.4.x = "14"
    * 1.5.x = "15"
    * 2.0.x = "20"
    * 2.1.x = "21"
    * <p/>
    * etc.
    *
    * @param version
    * @return a version integer
    */
   private int toMinorVersionInt(String version)
   {
      try
      {
         StringTokenizer strtok = new StringTokenizer(version, ".");

         // major, minor, micro, patch
         String[] versionComponents = {null, null, null, null};
         int i = 0;
         while (strtok.hasMoreTokens())
         {
            versionComponents[i++] = strtok.nextToken();
         }

         int major = Integer.parseInt(versionComponents[0]);
         int minor = Integer.parseInt(versionComponents[1]);

         return (major > 1 || minor > 3) ? (10 * major) + minor : 1;
      }
      catch (Exception e)
      {
         throw new IllegalArgumentException("Unsupported replication version string " + version);
      }
   }

   public byte[] objectToByteBuffer(Object obj) throws Exception
   {
      ByteArrayOutputStream bos = new ByteArrayOutputStream();
      ObjectOutputStream out;

      // based on the default marshaller, construct an object output stream based on what's compatible.
      out = ObjectSerializationFactory.createObjectOutputStream(bos);
      out.writeShort(versionInt);
      if (log.isTraceEnabled()) log.trace("Wrote version " + versionInt);

      //now marshall the contents of the object
      defaultMarshaller.objectToObjectStream(obj, out);
      out.close();

      // and return bytes.
      return bos.toByteArray();
   }

   public Object objectFromByteBuffer(byte[] buf) throws Exception
   {
      Marshaller marshaller;
      int versionId;
      ObjectInputStream in;
      try
      {
         in = ObjectSerializationFactory.createObjectInputStream(buf);
         versionId = in.readShort();
         if (log.isTraceEnabled()) log.trace("Read version " + versionId);
      }
      catch (Exception e)
      {
         log.error("Unable to read version id from first two bytes of stream, barfing.");
         throw e;
      }

      marshaller = getMarshaller(versionId);

      return marshaller.objectFromObjectStream(in);
   }

   public Object objectFromStream(InputStream is) throws Exception
   {
      Marshaller marshaller;
      int versionId;
      ObjectInputStream in;
      try
      {
         in = ObjectSerializationFactory.createObjectInputStream(is);
         versionId = in.readShort();
         if (log.isTraceEnabled()) log.trace("Read version " + versionId);
      }
      catch (Exception e)
      {
         log.error("Unable to read version id from first two bytes of stream, barfing.");
         throw e;
      }

      marshaller = getMarshaller(versionId);

      return marshaller.objectFromObjectStream(in);
   }

   public void objectToObjectStream(Object obj, ObjectOutputStream out, Fqn region) throws Exception
   {
      out.writeShort(versionInt);
      if (log.isTraceEnabled()) log.trace("Wrote version " + versionInt);
      defaultMarshaller.objectToObjectStream(obj, out, region);
   }

   /**
    * Lazily instantiates and loads the relevant marshaller for a given version.
    *
    * @param versionId
    * @return appropriate marshaller for the version requested.
    */
   Marshaller getMarshaller(int versionId)
   {
      Marshaller marshaller;
      switch (versionId)
      {
         case VERSION_200:
         default:
            marshaller = marshallers.get(VERSION_200);
            if (marshaller == null)
            {
               marshaller = new CacheMarshaller200(manager, defaultInactive, useRegionBasedMarshalling);
               marshallers.put(VERSION_200, marshaller);
            }
            break;
      }
      return marshaller;
   }

   public void objectToObjectStream(Object obj, ObjectOutputStream out) throws Exception
   {
      out.writeShort(versionInt);
      if (log.isTraceEnabled()) log.trace("Wrote version " + versionInt);
      defaultMarshaller.objectToObjectStream(obj, out);
   }

   public Object objectFromObjectStream(ObjectInputStream in) throws Exception
   {
      Marshaller marshaller;
      int versionId;
      try
      {
         versionId = in.readShort();
         if (log.isTraceEnabled()) log.trace("Read version " + versionId);
      }
      catch (Exception e)
      {
         log.error("Unable to read version id from first two bytes of stream, barfing.");
         throw e;
      }

      marshaller = getMarshaller(versionId);

      return marshaller.objectFromObjectStream(in);
   }
}
