/*
 * JBoss, Home of Professional Open Source
 *
 * Distributable under LGPL license.
 * See terms of license at gnu.org.
 */
package org.jboss.cache.optimistic;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.jboss.cache.AbstractNode;
import org.jboss.cache.CacheSPI;
import org.jboss.cache.Fqn;
import org.jboss.cache.Node;
import org.jboss.cache.NodeFactory;
import org.jboss.cache.NodeSPI;
import org.jboss.cache.VersionedNode;
import org.jboss.cache.transaction.GlobalTransaction;

import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

/**
 * Wraps an ordinary {@link Node} and adds versioning and other meta data to it.
 *
 * @author Manik Surtani (<a href="mailto:manik@jboss.org">manik@jboss.org</a>)
 * @author Steve Woodcock (<a href="mailto:stevew@jofti.com">stevew@jofti.com</a>)
 */
public class WorkspaceNodeImpl<K, V> extends AbstractNode<K, V> implements WorkspaceNode<K, V>
{

   private static Log log = LogFactory.getLog(WorkspaceNodeImpl.class);

   private NodeSPI<K, V> node;
   private TransactionWorkspace workspace;
   private DataVersion version = DefaultDataVersion.ZERO;
   private boolean modified;
   private boolean created;
   private boolean childrenModified;
   private Map<Object, NodeSPI<K, V>> optimisticChildNodeMap;
   private Set<Fqn> childrenAdded = new HashSet<Fqn>();
   private Set<Fqn> childrenRemoved = new HashSet<Fqn>();
   private Map<K, V> optimisticDataMap;
   private boolean versioningImplicit = true; // default

   /**
    * Constructs with a node and workspace.
    */
   public WorkspaceNodeImpl(NodeSPI<K, V> node, TransactionWorkspace workspace)
   {
      if (!(node instanceof VersionedNode))
      {
         throw new IllegalArgumentException("node " + node + " not VersionedNode");
      }
      this.node = node;
      this.workspace = workspace;
      optimisticDataMap = new HashMap<K, V>(node.getDataDirect());
      this.version = node.getVersion();
      if (version == null)
      {
         throw new IllegalStateException("VersionedNode version null");
      }
   }

   public boolean isChildrenModified()
   {
      return childrenModified;
   }

   public boolean isChildrenLoaded()
   {
      return optimisticChildNodeMap != null;
   }

   /**
    * Returns true if this node is modified.
    */
   public boolean isModified()
   {
      return modified;
   }

   /**
    * A convenience method that returns whether a node is dirty, i.e., it has been created, deleted or modified.
    *
    * @return true if the node has been either created, deleted or modified.
    */
   public boolean isDirty()
   {
      return modified || created || deleted;
   }

   public Fqn getFqn()
   {
      return node.getFqn();
   }

   public void putAll(Map<K, V> data)
   {
      realPut(data, false);
      modified = true;
   }

   public void replaceAll(Map<K, V> data)
   {
      clearData();
      putAll(data);
   }

   public V put(K key, V value)
   {
      modified = true;
      return optimisticDataMap.put(key, value);

   }

   public V remove(K key)
   {
      modified = true;
      return optimisticDataMap.remove(key);

   }

   public V get(K key)
   {
      return optimisticDataMap.get(key);
   }

   public Set<K> getKeys()
   {
      return optimisticDataMap.keySet();
   }

   //not able to delete from this
   public Set<Object> getChildrenNames()
   {
      if (optimisticChildNodeMap == null)
      {
         initialiseChildMap();
      }

      Set<Object> names = new HashSet<Object>(optimisticChildNodeMap.keySet());
      
      // process deltas
      for (Fqn child : childrenAdded) names.add(child.getLastElement());
      for (Fqn child : childrenRemoved) names.remove(child.getLastElement());
      
      return names;
   }

   @SuppressWarnings("unchecked")
   private void initialiseChildMap()
   {
      Map<Object, Node<K, V>> childrenMap = node.getChildrenMapDirect();
      this.optimisticChildNodeMap = childrenMap == null ? new HashMap() : new HashMap(childrenMap);
   }

   private void realPut(Map<K, V> data, boolean eraseData)
   {
      realPut(data, eraseData, true);
   }

   private void realPut(Map<K, V> data, boolean eraseData, boolean forceDirtyFlag)
   {
      if (forceDirtyFlag) modified = true;
      if (eraseData)
      {
         optimisticDataMap.clear();
      }
      if (data != null) optimisticDataMap.putAll(data);
   }

   public Node<K, V> getParent()
   {
      return node.getParent();
   }

   @SuppressWarnings("unchecked")
   public NodeSPI<K, V> createChild(Object child_name, NodeSPI<K, V> parent, CacheSPI<K, V> cache, DataVersion version)
   {
      if (child_name == null)
      {
         return null;
      }

      //see if we already have it
//      NodeSPI<K, V> child = optimisticChildNodeMap.get(child_name);

      // if not we need to create it
//      if (child == null)
//      {
         NodeFactory<K, V> factory = cache.getConfiguration().getRuntimeConfig().getNodeFactory();
         NodeSPI<K, V> child = (NodeSPI<K, V>) factory.createNodeOfType(parent, child_name, parent, null);
//         optimisticChildNodeMap.put(child_name, child);
         childrenAdded.add(child.getFqn());
         childrenRemoved.remove(child.getFqn());
//      }
      childrenModified = true;
      return child;
   }

   public boolean isVersioningImplicit()
   {
      return versioningImplicit;
   }

   public void setVersioningImplicit(boolean b)
   {
      versioningImplicit = b;
   }

   public NodeSPI<K, V> getChild(Object childName)
   {
      //see if in the the transaction map
//      return optimisticChildNodeMap.get(childName);
      throw new UnsupportedOperationException("This call should be intercepted by the optimisticNodeInterceptor and the child node should be retrieved from the workspace.");
   }

   public NodeSPI<K, V> getNode()
   {
      return node;
   }

   public DataVersion getVersion()
   {
      return version;
   }

   public void setVersion(DataVersion version)
   {
      this.version = version;
   }

   public List<Set<Fqn>> getMergedChildren()
   {
      //return optimisticChildNodeMap;
      List<Set<Fqn>> l = new ArrayList<Set<Fqn>>(2);
      l.add(childrenAdded);
      l.add(childrenRemoved);
      return l;
   }

   public Map<K, V> getMergedData()
   {
      return optimisticDataMap;
   }

   public TransactionWorkspace getTransactionWorkspace()
   {
      return workspace;
   }

   public boolean isCreated()
   {
      return created;
   }

   public void markAsCreated()
   {
      created = true;
      // created != modified!!!
//      modified = true;
   }

   public Map<K, V> getData()
   {
      return Collections.unmodifiableMap(optimisticDataMap);
   }

   public String toString()
   {
      StringBuffer sb = new StringBuffer();
      if (deleted) sb.append("del ");
      if (modified) sb.append("modified ");
      if (created) sb.append("new ");
      return getClass().getSimpleName() + " [ fqn=" + getFqn() + " " + sb + "ver=" + version + " " + (versioningImplicit ? "implicit" : "explicit") + "]";
   }

   public Node<K, V> addChild(Fqn f)
   {
      CacheSPI cache = getCache();
      Node<K, V> newNode = this;
      GlobalTransaction gtx = cache.getInvocationContext().getGlobalTransaction();

      if (f.size() == 1)
      {
         newNode = createChild(f.get(0), node, getCache(), version);
      }
      else
      {
         // recursively create children
         Node<K,V> currentParent = this;
         for (Object o : f.peekElements())
         {
            if (currentParent instanceof WorkspaceNode)
            {
               newNode = ((WorkspaceNode<K,V>) currentParent).getNode().getOrCreateChild(o, gtx);
            }
            else
            {
               if (currentParent instanceof WorkspaceNode)
               {
                  newNode = ((WorkspaceNode<K,V>) currentParent).getNode().getOrCreateChild(o, gtx);
               }
               else
               {
                  newNode = ((NodeSPI<K,V>) currentParent).getOrCreateChild(o, gtx);
               }
            }
            currentParent = newNode;
         }
      }
      return newNode;
   }

   public void addChild(WorkspaceNode<K,V> child)
   {
//      optimisticChildNodeMap.put(child.getFqn().getLastElement(), child.getNode());
      childrenAdded.add(child.getFqn());
      childrenRemoved.remove(child.getFqn());
      if (log.isTraceEnabled()) log.trace("Adding child " + child.getFqn());
   }

   public void clearData()
   {
      modified = true;
      optimisticDataMap.clear();
   }

   public int dataSize()
   {
      return optimisticDataMap.size();
   }

   public boolean hasChild(Object o)
   {
      throw new UnsupportedOperationException();
   }

   public boolean isValid()
   {
      throw new UnsupportedOperationException();
   }

   public boolean isLockForChildInsertRemove()
   {
      throw new UnsupportedOperationException();
   }

   public void setLockForChildInsertRemove(boolean lockForChildInsertRemove)
   {
      throw new UnsupportedOperationException();
   }

   public NodeSPI<K, V> getChild(Fqn f)
   {
      if (f.size() > 1)
      {
         throw new UnsupportedOperationException("Workspace node does not support fetching indirect children");
      }
      return getChild(f.getLastElement());
   }

   public Set<Node<K, V>> getChildren()
   {
      throw new UnsupportedOperationException();
   }

   public boolean hasChild(Fqn f)
   {
      throw new UnsupportedOperationException();
   }

   public NodeSPI<K, V> getNodeSPI()
   {
      throw new UnsupportedOperationException("WorkspaceNode has no access to a NodeSPI");
   }

   public V putIfAbsent(K k, V v)
   {
      throw new UnsupportedOperationException();
   }

   public V replace(K key, V value)
   {
      throw new UnsupportedOperationException();
   }

   public boolean replace(K key, V oldValue, V newValue)
   {
      throw new UnsupportedOperationException();
   }

   public boolean removeChild(Fqn f)
   {
      if (f.size() > 1) throw new UnsupportedOperationException("Workspace nodes can only remove direct children!");
      Object key = f.getLastElement();
      return removeChild(key);
   }

   public boolean removeChild(Object childName)
   {
      //NodeSPI n = node.getChildDirect(childName);
      Fqn childFqn = new Fqn(getFqn(), childName);
      /*if (n != null)
      {*/
         childrenRemoved.add(childFqn);
         childrenAdded.remove(childFqn);
         childrenModified = true;
         return node.getChildDirect(childName) != null;
      /*}
      else
      {
         return false;
      }*/
   }

   protected CacheSPI<K, V> getCache()
   {
      return node.getCache();
   }
}
