/*
 * JBoss, the OpenSource J2EE webOS
 *
 * Distributable under LGPL license.
 * See terms of license at gnu.org.
 */
package org.jboss.cache.interceptors;

import static org.testng.AssertJUnit.assertEquals;
import static org.testng.AssertJUnit.assertFalse;
import static org.testng.AssertJUnit.assertNotNull;
import static org.testng.AssertJUnit.assertNull;
import static org.testng.AssertJUnit.assertTrue;

import java.util.HashMap;
import java.util.Map;

import org.jboss.cache.CacheImpl;
import org.jboss.cache.DefaultCacheFactory;
import org.jboss.cache.Fqn;
import org.jboss.cache.InvocationContext;
import org.jboss.cache.Node;
import org.jboss.cache.Region;
import org.jboss.cache.RegionManager;
import org.jboss.cache.config.EvictionPolicyConfig;
import org.jboss.cache.eviction.DummyEvictionConfiguration;
import org.jboss.cache.eviction.EvictedEventNode;
import org.jboss.cache.eviction.NodeEventType;
import org.jboss.cache.lock.IsolationLevel;
import org.jboss.cache.marshall.MethodCall;
import org.jboss.cache.marshall.MethodCallFactory;
import org.jboss.cache.marshall.MethodDeclarations;
import org.testng.annotations.AfterMethod;
import org.testng.annotations.BeforeMethod;
import org.testng.annotations.Test;

/**
 * @author Daniel Huang (dhuang@jboss.org)
 * @version $Revision: $
 */
@Test(groups = {"functional"})
public class EvictionInterceptorTest
{
   private static final String fqn1 = "/a/b/c";
   private static final String fqn2 = "/a/b";
   private static final String fqn3 = "/a/b/d";

   private static final String fqn4 = "/d/e/f";

   private CacheImpl<Object, Object> cache;
   private Interceptor interceptor;
   private RegionManager regionManager;

   @BeforeMethod(alwaysRun = true)
   public void setUp() throws Exception
   {
      regionManager = new RegionManager();
      EvictionPolicyConfig config = new DummyEvictionConfiguration();
      regionManager.setUsingEvictions(true);

      regionManager.getRegion(RegionManager.DEFAULT_REGION, true).setEvictionPolicy(config);

      regionManager.getRegion("/a/b/c", true).setEvictionPolicy(config);
      regionManager.getRegion("/a/b/c/d", true).setEvictionPolicy(config);
      regionManager.getRegion("/a/b", true).setEvictionPolicy(config);

      regionManager.getRegion("/d/e/f", true).setEvictionPolicy(config);
      regionManager.getRegion("/d/e/g", true).setEvictionPolicy(config);
      regionManager.getRegion("/d/e", true).setEvictionPolicy(config);

      cache = (CacheImpl<Object, Object>) DefaultCacheFactory.getInstance().createCache(false);
      cache.getConfiguration().setTransactionManagerLookupClass("org.jboss.cache.transaction.DummyTransactionManagerLookup");
      cache.getConfiguration().setIsolationLevel(IsolationLevel.SERIALIZABLE);

      // make the interceptor chain (separate from the CacheImpl object.

      interceptor = new CacheMgmtInterceptor();
      interceptor.setCache(cache);

      TxInterceptor ti = new TxInterceptor();
      ti.setCache(cache);
      interceptor.setNext(ti);

      UnlockInterceptor ui = new UnlockInterceptor();
      ui.setCache(cache);
      ti.setNext(ui);

      PessimisticLockInterceptor pli = new PessimisticLockInterceptor();
      pli.setCache(cache);
      ui.setNext(pli);

      EvictionInterceptor ei = new EvictionInterceptor();
      ei.setCache(cache);
      ei.setRegionManager(regionManager);
      pli.setNext(ei);

      CallInterceptor ci = new CallInterceptor();
      ci.setCache(cache);
      ci.setTreeCacheInstance(cache);
      ei.setNext(ci);

      cache.getConfiguration().setCacheMode("LOCAL");
      cache.start();
   }

   @AfterMethod(alwaysRun = true)
   public void tearDown() throws Exception
   {
      cache.remove("/");
      cache.stop();
   }

   @SuppressWarnings("unchecked")
   private Node<Object, Object> cast(Node node)
   {
      return node;
   }

   public void testVisitNode() throws Throwable
   {
      // make sure node that doesn't exist does not result in a node visit event.

      MethodCall mc = MethodCallFactory.create(MethodDeclarations.getNodeMethodLocal,
              Fqn.fromString(fqn1));
      interceptor.invoke(InvocationContext.fromMethodCall(mc));
      Region regionABC = regionManager.getRegion(fqn1, false);
      assertNull(regionABC.takeLastEventNode());

      cache.put(fqn1, "key", "value");
      assertEquals("value", cache.get(fqn1, "key"));
      Node<Object, Object> node = cast(cache.get(fqn1));
      assertNotNull(node);
      assertEquals("value", node.get("key"));

      cache.put(fqn3, "key", "value");
      assertEquals("value", cache.get(fqn3, "key"));
      node = cast(cache.get(fqn3));
      assertNotNull(node);
      assertEquals("value", node.get("key"));


      mc = MethodCallFactory.create(MethodDeclarations.getNodeMethodLocal,
              Fqn.fromString(fqn1));
      interceptor.invoke(InvocationContext.fromMethodCall(mc));

      regionABC = regionManager.getRegion(fqn1, false);
      EvictedEventNode event = regionABC.takeLastEventNode();
      assertEquals(NodeEventType.VISIT_NODE_EVENT, event.getEventType());
      assertEquals(fqn1, event.getFqn().toString());
      assertNull(regionABC.takeLastEventNode());

      mc = MethodCallFactory.create(MethodDeclarations.getNodeMethodLocal,
              Fqn.fromString(fqn2));
      interceptor.invoke(InvocationContext.fromMethodCall(mc));

      Region regionAB = regionManager.getRegion(fqn2, false);
      event = regionAB.takeLastEventNode();
      assertEquals(NodeEventType.VISIT_NODE_EVENT, event.getEventType());
      assertEquals(fqn2, event.getFqn().toString());
      assertNull(regionAB.takeLastEventNode());

      mc = MethodCallFactory.create(MethodDeclarations.getDataMapMethodLocal, Fqn.fromString(fqn3));
      interceptor.invoke(InvocationContext.fromMethodCall(mc));
      Region regionABD = regionManager.getRegion(fqn3, false);
      event = regionABD.takeLastEventNode();
      assertEquals(NodeEventType.VISIT_NODE_EVENT, event.getEventType());
      assertEquals(fqn3, event.getFqn().toString());
      assertNull(regionABD.takeLastEventNode());

      for (int i = 0; i < 10; i++)
      {
         Fqn fqn = Fqn.fromString(fqn3);
         mc = MethodCallFactory.create(MethodDeclarations.getNodeMethodLocal, fqn);
         interceptor.invoke(InvocationContext.fromMethodCall(mc));
      }

      for (int i = 0; i < 10; i++)
      {
         Region region = regionManager.getRegion(fqn3, false);
         event = region.takeLastEventNode();
         assertEquals(NodeEventType.VISIT_NODE_EVENT, event.getEventType());
         assertEquals(fqn3, event.getFqn().toString());
      }

      assertNull(regionManager.getRegion(fqn3, false).takeLastEventNode());

      // check null handling.
      mc = MethodCallFactory.create(MethodDeclarations.getDataMapMethodLocal, new Object[]{null});
      interceptor.invoke(InvocationContext.fromMethodCall(mc));

   }

   public void testVisitElement() throws Throwable
   {
      // make sure a get/visit on an empty node and empty element results in no cache events being added to event queue
      // aka MarshRegion.
      Fqn fqn = Fqn.fromString(fqn4);
      Object key = "key";
      MethodCall mc = MethodCallFactory.create(MethodDeclarations.getKeyValueMethodLocal, fqn, key, false);
      interceptor.invoke(InvocationContext.fromMethodCall(mc));
      Region region = regionManager.getRegion(fqn.toString(), false);
      assertNull(region.takeLastEventNode());

      // add the node but try to get on a null element should result in no cache events being added to MarshRegion.
      cache.put(fqn, "wrongkey", "");
      mc = MethodCallFactory.create(MethodDeclarations.getKeyValueMethodLocal, fqn, key, false);
      interceptor.invoke(InvocationContext.fromMethodCall(mc));
      assertNull(region.takeLastEventNode());

      // now make sure if we try to get on the node/key we just created in cache, that this DOES add a EvictedEventNode to
      // the MarshRegion.
      mc = MethodCallFactory.create(MethodDeclarations.getKeyValueMethodLocal, fqn, "wrongkey", false);
      interceptor.invoke(InvocationContext.fromMethodCall(mc));
      EvictedEventNode event = region.takeLastEventNode();
      assertEquals(fqn, event.getFqn());
      assertEquals(NodeEventType.VISIT_NODE_EVENT, event.getEventType());

      assertNull(region.takeLastEventNode());

      cache.put(fqn4, key, "value");
      // test on element granularity configured node.
      fqn = Fqn.fromString(fqn4);
      mc = MethodCallFactory.create(MethodDeclarations.getKeyValueMethodLocal, fqn, key, false);
      interceptor.invoke(InvocationContext.fromMethodCall(mc));

      region = regionManager.getRegion(fqn.toString(), false);
      event = region.takeLastEventNode();

      assertEquals(NodeEventType.VISIT_NODE_EVENT, event.getEventType());
      assertEquals(fqn, event.getFqn());

      assertNull(region.takeLastEventNode());

      fqn = Fqn.fromString("/d/e/g");
      for (int i = 0; i < 1000; i++)
      {
         key = i;

         cache.put("/d/e/g", key, "");

         mc = MethodCallFactory.create(MethodDeclarations.getKeyValueMethodLocal, fqn, key, true);
         interceptor.invoke(InvocationContext.fromMethodCall(mc));
      }

      region = regionManager.getRegion(fqn.toString(), false);
      for (int i = 0; i < 1000; i++)
      {
         event = region.takeLastEventNode();
         assertEquals(fqn, event.getFqn());
         assertEquals(NodeEventType.VISIT_NODE_EVENT, event.getEventType());
      }

      cache.put("/a/b/c", key, "");
      fqn = Fqn.fromString("/a/b/c");
      mc = MethodCallFactory.create(MethodDeclarations.getKeyValueMethodLocal, fqn, key, false);
      interceptor.invoke(InvocationContext.fromMethodCall(mc));

      region = regionManager.getRegion(fqn.toString(), false);
      event = region.takeLastEventNode();

      assertNull(region.takeLastEventNode());
      assertEquals(NodeEventType.VISIT_NODE_EVENT, event.getEventType());
      assertEquals(fqn, event.getFqn());

      for (int i = 0; i < 1000; i++)
      {
         key = i;

         cache.put(fqn.toString(), key, "");

         mc = MethodCallFactory.create(MethodDeclarations.getKeyValueMethodLocal, fqn, key, true);
         interceptor.invoke(InvocationContext.fromMethodCall(mc));
      }

      for (int i = 0; i < 1000; i++)
      {
         event = region.takeLastEventNode();
         assertEquals(NodeEventType.VISIT_NODE_EVENT, event.getEventType());
         assertEquals(fqn, event.getFqn());
      }

      assertNull(region.takeLastEventNode());

      // check null handling
      mc = MethodCallFactory.create(MethodDeclarations.getKeyValueMethodLocal, null, key, true);
      interceptor.invoke(InvocationContext.fromMethodCall(mc));
      assertNull(region.takeLastEventNode());

      mc = MethodCallFactory.create(MethodDeclarations.getKeyValueMethodLocal, null, null, true);
      interceptor.invoke(InvocationContext.fromMethodCall(mc));
      assertNull(region.takeLastEventNode());

      mc = MethodCallFactory.create(MethodDeclarations.getKeyValueMethodLocal, fqn, null, true);
      interceptor.invoke(InvocationContext.fromMethodCall(mc));
      assertNull(region.takeLastEventNode());
   }

   public void testCreateNode() throws Throwable
   {
      Map<Object, Object> data = new HashMap<Object, Object>();
      for (int i = 0; i < 100; i++)
      {
         data.put(i, i);
      }

      // this region is node granularity
      Fqn fqn = Fqn.fromString("/a/b/c");

      MethodCall mc = MethodCallFactory.create(MethodDeclarations.putDataMethodLocal, null, fqn, data, false);
      interceptor.invoke(InvocationContext.fromMethodCall(mc));

      Region region = regionManager.getRegion(fqn.toString(), false);
      EvictedEventNode event = region.takeLastEventNode();

      assertEquals(NodeEventType.ADD_NODE_EVENT, event.getEventType());
      assertEquals(fqn, event.getFqn());
      assertEquals(100, event.getElementDifference());

      Node<Object, Object> node = cast(cache.get(fqn.toString()));
      assertNotNull(node);

      for (int i = 0; i < 100; i++)
      {
         assertTrue(node.getData().containsKey(i));
         assertEquals(i, node.get(i));
      }

      for (int i = 0; i < 100; i++)
      {
         mc = MethodCallFactory.create(MethodDeclarations.putKeyValMethodLocal, null, fqn, i,
                 "value", false);
         interceptor.invoke(InvocationContext.fromMethodCall(mc));

         assertEquals("value", cache.get(fqn.toString(), i));
      }

      for (int i = 0; i < 100; i++)
      {
         event = region.takeLastEventNode();
         assertNotNull(event);
         assertEquals(fqn, event.getFqn());
         assertEquals(NodeEventType.ADD_ELEMENT_EVENT, event.getEventType());
      }

      assertNull(region.takeLastEventNode());

      fqn = Fqn.fromString("/a/b");
      mc = MethodCallFactory.create(MethodDeclarations.putDataEraseMethodLocal, null, fqn, data, false, false);
      interceptor.invoke(InvocationContext.fromMethodCall(mc));
      event = regionManager.getRegion(fqn.toString(), false).takeLastEventNode();
      assertFalse(event.isResetElementCount());
      assertEquals(NodeEventType.ADD_NODE_EVENT, event.getEventType());
      assertEquals(fqn, event.getFqn());
      assertEquals(100, event.getElementDifference());
      assertNull(regionManager.getRegion(fqn.toString(), false).takeLastEventNode());
      node = cast(cache.get(fqn.toString()));
      assertEquals(100, node.getData().size());

      assertNotNull(node);

      for (int i = 0; i < 100; i++)
      {
         assertTrue(node.getData().containsKey(i));
         assertEquals(i, node.get(i));
      }

      mc = MethodCallFactory.create(MethodDeclarations.putDataEraseMethodLocal, null, fqn, data, false, true);
      interceptor.invoke(InvocationContext.fromMethodCall(mc));
      event = regionManager.getRegion(fqn.toString(), false).takeLastEventNode();
      assertEquals(NodeEventType.ADD_NODE_EVENT, event.getEventType());
      assertEquals(fqn, event.getFqn());
      assertEquals(100, event.getElementDifference());
      assertTrue(event.isResetElementCount());
      assertNull(regionManager.getRegion(fqn.toString(), false).takeLastEventNode());


      node = cast(cache.get(fqn.toString()));
      assertEquals(100, node.getData().size());
      assertNotNull(node);

      for (int i = 0; i < 100; i++)
      {
         assertTrue(node.getData().containsKey(i));
         assertEquals(i, node.get(i));
      }

   }

   public void testCreateElement() throws Throwable
   {
      Fqn fqn = Fqn.fromString("/d/e/f");
      Region region = regionManager.getRegion(fqn.toString(), false);
      Object key = "key";
      Object value = "value";

      MethodCall mc = MethodCallFactory.create(MethodDeclarations.putKeyValMethodLocal,
              null, fqn, key, value, false);
      interceptor.invoke(InvocationContext.fromMethodCall(mc));
      assertEquals("value", cache.get(fqn, key));
      EvictedEventNode event = region.takeLastEventNode();
      assertEquals(NodeEventType.ADD_ELEMENT_EVENT, event.getEventType());
      assertEquals(1, event.getElementDifference());
      assertEquals(fqn, event.getFqn());
      assertEquals("value", cache.get(fqn, key));
      assertNull(region.takeLastEventNode());

      mc = MethodCallFactory.create(MethodDeclarations.putKeyValMethodLocal,
              null, fqn, key, value, false);
      interceptor.invoke(InvocationContext.fromMethodCall(mc));
      assertEquals("value", cache.get(fqn, key));
      event = region.takeLastEventNode();
      assertEquals(NodeEventType.ADD_ELEMENT_EVENT, event.getEventType());
      assertEquals(1, event.getElementDifference());
      assertEquals(fqn, event.getFqn());
      assertEquals("value", cache.get(fqn, key));
      assertNull(region.takeLastEventNode());

   }

   public void testRemoveNode() throws Throwable
   {
      Fqn fqn = Fqn.fromString("/a/b/c");
      cache.put(fqn, "a", "b");
      cache.put(fqn, "b", "c");

      MethodCall mc = MethodCallFactory.create(MethodDeclarations.removeDataMethodLocal, null, fqn, false);
      interceptor.invoke(InvocationContext.fromMethodCall(mc));

      assertEquals(0, cache.get(fqn).getData().size());
      Region region = regionManager.getRegion(fqn.toString(), false);
      EvictedEventNode event = region.takeLastEventNode();
      assertEquals(NodeEventType.REMOVE_NODE_EVENT, event.getEventType());
      assertEquals(fqn, event.getFqn());
      assertNull(region.takeLastEventNode());

      mc = MethodCallFactory.create(MethodDeclarations.removeNodeMethodLocal, null, fqn, false);
      interceptor.invoke(InvocationContext.fromMethodCall(mc));

      assertNull(cache.get(fqn));
      event = region.takeLastEventNode();
      assertEquals(NodeEventType.REMOVE_NODE_EVENT, event.getEventType());
      assertEquals(fqn, event.getFqn());
      assertNull(region.takeLastEventNode());
   }

   public void testRemoveElement() throws Throwable
   {
      Fqn fqn = Fqn.fromString("/a/b/c");
      cache.put(fqn, "a", "b");
      cache.put(fqn, "b", "c");

      MethodCall mc = MethodCallFactory.create(MethodDeclarations.removeKeyMethodLocal,
              null, fqn, "a", false);
      interceptor.invoke(InvocationContext.fromMethodCall(mc));

      assertNull(cache.get(fqn, "a"));
      Region region = regionManager.getRegion(fqn.toString(), false);
      EvictedEventNode event = region.takeLastEventNode();
      assertEquals(NodeEventType.REMOVE_ELEMENT_EVENT, event.getEventType());
      assertEquals(fqn, event.getFqn());
      assertEquals(1, event.getElementDifference());
      assertNull(region.takeLastEventNode());

      mc = MethodCallFactory.create(MethodDeclarations.removeKeyMethodLocal,
              null, fqn, "b", false);
      interceptor.invoke(InvocationContext.fromMethodCall(mc));

      assertNull(cache.get(fqn, "b"));
      event = region.takeLastEventNode();
      assertEquals(NodeEventType.REMOVE_ELEMENT_EVENT, event.getEventType());
      assertEquals(fqn, event.getFqn());
      assertEquals(1, event.getElementDifference());
      assertNull(region.takeLastEventNode());

      mc = MethodCallFactory.create(MethodDeclarations.removeKeyMethodLocal,
              null, fqn, "b", false);
      interceptor.invoke(InvocationContext.fromMethodCall(mc));

      event = region.takeLastEventNode();
      assertNull(event);
   }

   public void testMixedEvent() throws Throwable
   {
      Map<Object, Object> data = new HashMap<Object, Object>();
      for (int i = 0; i < 100; i++)
      {
         data.put(i, i);
      }

      // this region is node granularity
      Fqn fqn = Fqn.fromString("/a/b/c");

      MethodCall mc = MethodCallFactory.create(MethodDeclarations.putDataMethodLocal, null, fqn, data, false);
      interceptor.invoke(InvocationContext.fromMethodCall(mc));

      Region region = regionManager.getRegion(fqn.toString(), false);
      EvictedEventNode event = region.takeLastEventNode();

      assertEquals(NodeEventType.ADD_NODE_EVENT, event.getEventType());
      assertEquals(fqn, event.getFqn());
      assertEquals(100, event.getElementDifference());
      assertNull(region.takeLastEventNode());

      mc = MethodCallFactory.create(MethodDeclarations.getNodeMethodLocal,
              fqn);
      interceptor.invoke(InvocationContext.fromMethodCall(mc));
      event = region.takeLastEventNode();
      assertEquals(NodeEventType.VISIT_NODE_EVENT, event.getEventType());
      assertEquals(fqn, event.getFqn());
      assertNull(region.takeLastEventNode());

      mc = MethodCallFactory.create(MethodDeclarations.removeNodeMethodLocal, null, fqn, false);
      interceptor.invoke(InvocationContext.fromMethodCall(mc));
      assertNull(cache.get(fqn));
      event = region.takeLastEventNode();
      assertEquals(NodeEventType.REMOVE_NODE_EVENT, event.getEventType());
      assertEquals(fqn, event.getFqn());
      assertNull(region.takeLastEventNode());

      Object key = "key";
      Object value = "value";
      mc = MethodCallFactory.create(MethodDeclarations.putKeyValMethodLocal,
              null, fqn, key, value, false);
      interceptor.invoke(InvocationContext.fromMethodCall(mc));
      assertEquals("value", cache.get(fqn, key));
      event = region.takeLastEventNode();
      assertEquals(NodeEventType.ADD_ELEMENT_EVENT, event.getEventType());
      assertEquals(1, event.getElementDifference());
      assertEquals(fqn, event.getFqn());
      assertEquals("value", cache.get(fqn, key));
      assertNull(region.takeLastEventNode());

      mc = MethodCallFactory.create(MethodDeclarations.getKeyValueMethodLocal, fqn, key, false);
      interceptor.invoke(InvocationContext.fromMethodCall(mc));
      region = regionManager.getRegion(fqn.toString(), false);
      event = region.takeLastEventNode();
      assertEquals(NodeEventType.VISIT_NODE_EVENT, event.getEventType());
      assertEquals(fqn, event.getFqn());
      assertNull(region.takeLastEventNode());

      mc = MethodCallFactory.create(MethodDeclarations.removeKeyMethodLocal,
              null, fqn, key, false);
      interceptor.invoke(InvocationContext.fromMethodCall(mc));

      assertNull(cache.get(fqn, key));
      event = region.takeLastEventNode();
      assertEquals(NodeEventType.REMOVE_ELEMENT_EVENT, event.getEventType());
      assertEquals(fqn, event.getFqn());
      assertEquals(1, event.getElementDifference());
      assertNull(region.takeLastEventNode());
   }
}