package org.jboss.cache.invocationcontext;

import static org.testng.AssertJUnit.assertEquals;
import static org.testng.AssertJUnit.assertNotNull;
import static org.testng.AssertJUnit.assertNull;

import java.util.Map;

import javax.transaction.TransactionManager;

import org.jboss.cache.CacheSPI;
import org.jboss.cache.DefaultCacheFactory;
import org.jboss.cache.Fqn;
import org.testng.annotations.AfterMethod;
import org.testng.annotations.BeforeMethod;
import org.testng.annotations.Test;

/**
 * A test to ensure the transactional context is properly set up in the IC
 *
 * @author <a href="mailto:manik@jboss.org">Manik Surtani</a>
 */
@Test(groups = {"functional", "transaction"})
public class TransactionTest
{
   private CacheSPI<Object, Object> cache;
   private TransactionManager tm;

   @BeforeMethod(alwaysRun = true)
   public void setUp()
   {
      cache = (CacheSPI<Object, Object>) DefaultCacheFactory.getInstance().createCache("META-INF/local-tx-service.xml");
      tm = cache.getTransactionManager();
   }

   @AfterMethod(alwaysRun = true)
   public void tearDown()
   {
      if (cache != null)
      {
         cache.stop();
         cache = null;
      }
   }

   public void testTxExistenceAfterWrite() throws Exception
   {
      // make sure we have a running transaction.
      tm.begin();

      assertNull("Tx should not have been set up yet", cache.getInvocationContext().getTransaction());
      assertNull("Gtx should not have been set up yet", cache.getInvocationContext().getGlobalTransaction());

      // now make a WRITE call into the cache (should start a tx)
      cache.getRoot().put("k", "v");
      Map data = cache.getRoot().getData();
      assertEquals("Data map should not empty", 1, data.size());

      // but now we should have a local tx registered in the invocation context
      assertNotNull("Tx should have been set up by now", cache.getInvocationContext().getTransaction());
      assertEquals("The same current transaction should be associated with this invocation ctx.", tm.getTransaction(), cache.getInvocationContext().getTransaction());
      assertNotNull("Gtx should have been set up by now", cache.getInvocationContext().getGlobalTransaction());

      tm.commit();
   }

   public void testTxExistenceAfterRead() throws Exception
   {
      // make sure we have a running transaction.
      tm.begin();

      assertNull("Tx should not have been set up yet", cache.getInvocationContext().getTransaction());
      assertNull("Gtx should not have been set up yet", cache.getInvocationContext().getGlobalTransaction());

      // now make a WRITE call into the cache (should start a tx)
      Object value = cache.get(Fqn.ROOT, "k");
      assertNull("Value should be null", value);

      // but now we should have a local tx registered in the invocation context
      assertNotNull("Tx should have been set up by now", cache.getInvocationContext().getTransaction());
      assertEquals("The same current transaction should be associated with this invocation ctx.", tm.getTransaction(), cache.getInvocationContext().getTransaction());
      assertNotNull("Gtx should have been set up by now", cache.getInvocationContext().getGlobalTransaction());

      tm.commit();
   }

   public void testScrubbingAfterCommit() throws Exception
   {
      doScrubbingTest(true);
   }

   public void testScrubbingAfterRollback() throws Exception
   {
      doScrubbingTest(false);
   }

   private void doScrubbingTest(boolean commit) throws Exception
   {
      tm.begin();
      cache.getRoot().put("key", "value");
      assertNotNull("Tx should have been set up by now", cache.getInvocationContext().getTransaction());
      assertEquals("The same current transaction should be associated with this invocation ctx.", tm.getTransaction(), cache.getInvocationContext().getTransaction());
      assertNotNull("Gtx should have been set up by now", cache.getInvocationContext().getGlobalTransaction());

      if (commit)
      {
         tm.commit();
      }
      else
      {
         tm.rollback();
      }

      assertNull("Tx should have been scrubbed", cache.getInvocationContext().getTransaction());
      assertNull("Gtx should have been scrubbed", cache.getInvocationContext().getGlobalTransaction());
   }
}
