/*
 * JBoss, the OpenSource J2EE webOS
 *
 * Distributable under LGPL license.
 * See terms of license at gnu.org.
 */

package org.jboss.cache.marshall;

import static org.testng.AssertJUnit.assertEquals;
import static org.testng.AssertJUnit.assertFalse;
import static org.testng.AssertJUnit.assertTrue;
import static org.testng.AssertJUnit.fail;

import java.lang.reflect.Method;

import org.jboss.cache.CacheImpl;
import org.jboss.cache.DefaultCacheFactory;
import org.jboss.cache.Fqn;
import org.jboss.cache.Region;
import org.jboss.cache.RegionManager;
import org.jboss.cache.Version;
import org.testng.annotations.AfterMethod;
import org.testng.annotations.BeforeMethod;
import org.testng.annotations.Test;

/**
 * Tests the "activate/deactivate" functionality of LegacyTreeCacheMarshaller.
 *
 * @author <a href="mailto://brian.stansberry@jboss.com">Brian Stansberry</a>
 * @version $Revision$
 */
@Test(groups = {"functional"})
public class ActiveInactiveTest
{
   RegionManager rman;
   CacheImpl c;
   Fqn<String> A = Fqn.fromString("/a");
   Fqn<String> I = Fqn.fromString("/i");
   Fqn<String> A_B = new Fqn<String>(A, "b");

   @BeforeMethod(alwaysRun = true)
   public void setUp() throws Exception
   {
      c = (CacheImpl) DefaultCacheFactory.getInstance().createCache(false);
      c.getConfiguration().setUseRegionBasedMarshalling(true);
      c.getConfiguration().setFetchInMemoryState(false);
      c.start();
      rman = c.getRegionManager();
   }

   @AfterMethod(alwaysRun = true)
   public void tearDown() throws Exception
   {
      c.stop();
      c = null;
      rman = null;
   }

   public void testDefaultActive() throws Exception
   {
      rman.setDefaultInactive(false);
      assertFalse("Root is not active", rman.isInactive(Fqn.ROOT));

      rman.deactivate(A);
      assertFalse("Root is not active after inactivating subtree",
                  rman.isInactive(Fqn.ROOT));

      rman.activate(A);
      assertFalse("Root is not active after activating subtree",
                  rman.isInactive(Fqn.ROOT));

      rman.activate(A_B);

      rman.deactivate(Fqn.ROOT);
      assertTrue("Root is active", rman.isInactive(Fqn.ROOT));
   }

   public void testDefaultInactive() throws Exception
   {
      rman.setDefaultInactive(true);

      assertTrue("Root is not inactive", rman.isInactive(Fqn.ROOT));

      rman.activate(A);
      assertTrue("Root is not inactive after activating subtree",
                 rman.isInactive(Fqn.ROOT));

      rman.deactivate(A);
      assertTrue("Root is not inactive after inactivating subtree",
                 rman.isInactive(Fqn.ROOT));

      rman.deactivate(A_B);

      rman.activate(Fqn.ROOT);
      assertFalse("Root is not active", rman.isInactive(Fqn.ROOT));
   }

   public void testActivate() throws Exception
   {
      rman.setDefaultInactive(false);
      rman.activate(A);
      assertFalse("/a is not active after activating",
                  rman.isInactive(A));

      rman.deactivate(A);
      rman.activate(A);
      assertFalse("/a is not active after reactivating",
                  rman.isInactive(A));

      rman.reset();
      rman.setDefaultInactive(true);

      rman.activate(I);
      assertFalse("/i is not active after activating",
                  rman.isInactive(I));
      assertFalse("/i/k is not active after activating /i",
                  rman.isInactive(Fqn.fromString("/i/k")));

      rman.deactivate(I);
      rman.activate(I);
      assertFalse("/i is not active after reactivating",
                  rman.isInactive(I));
   }

   public void testInactivate() throws Exception
   {
      rman.setDefaultInactive(true);

      rman.deactivate(I);
      assertTrue("/i is not inactive after inactivating",
                 rman.isInactive(I));

      rman.activate(I);
      rman.deactivate(I);
      assertTrue("/i is not inactive after re-inactivating",
                 rman.isInactive(I));

      rman.reset();
      rman.setDefaultInactive(false);

      rman.deactivate(A);
      assertTrue("/a is not inactive after inactivating",
                 rman.isInactive(A));
      assertTrue("/a/b is not inactive after inactivating /a",
                 rman.isInactive(A_B));

      rman.activate(A);
      rman.deactivate(A);
      assertTrue("/a is not inactive after re-inactivating",
                 rman.isInactive(A));
   }

   public void testObjectFromByteBuffer() throws Exception
   {
      MethodCall put = MethodCallFactory.create(MethodDeclarations.putKeyValMethodLocal,
                                                null, A_B, "name", "Joe", false);

      MethodCall replicate = MethodCallFactory.create(MethodDeclarations.replicateMethod, put);

      rman.setDefaultInactive(true);
      // register A as an inactive marshalling region
      Region region_A = rman.getRegion(A, true);
      region_A.registerContextClassLoader(this.getClass().getClassLoader());
      assertFalse("New regions created should be inactive by default", region_A.isActive());
      VersionAwareMarshaller testee = new VersionAwareMarshaller(rman, true, true, Version.getVersionString(Version.getVersionShort()));
      byte[] callBytes = testee.objectToByteBuffer(replicate);

      try
      {
         testee.objectFromByteBuffer(callBytes);
         fail("Expected to fail since region is inactive");
      }
      catch (Exception e)
      {
         // expected to fail since region is inactive
      }

      rman.activate(A);
      assertTrue(rman.hasRegion(A, Region.Type.MARSHALLING));

      MethodCall result = (MethodCall) testee.objectFromByteBuffer(callBytes);
      Method method = result.getMethod();
      assertEquals("Did not get replicate method when passing" +
                   " call for active node", MethodDeclarations.replicateMethod, method);
   }

}
