/*
 *
 * JBoss, the OpenSource J2EE webOS
 *
 * Distributable under LGPL license.
 * See terms of license at gnu.org.
 */

package org.jboss.cache.marshall;


import static org.testng.AssertJUnit.assertEquals;
import static org.testng.AssertJUnit.assertFalse;
import static org.testng.AssertJUnit.assertNotNull;

import java.lang.reflect.Method;

import javax.transaction.HeuristicMixedException;
import javax.transaction.HeuristicRollbackException;
import javax.transaction.NotSupportedException;
import javax.transaction.RollbackException;
import javax.transaction.SystemException;
import javax.transaction.Transaction;
import javax.transaction.TransactionManager;

import org.jboss.cache.CacheSPI;
import org.jboss.cache.DefaultCacheFactory;
import org.jboss.cache.Fqn;
import org.jboss.cache.Region;
import org.jboss.cache.config.Configuration.CacheMode;
import org.jboss.cache.factories.UnitTestCacheConfigurationFactory;
import org.jboss.cache.marshall.data.Address;
import org.jboss.cache.marshall.data.Person;
import org.jboss.cache.misc.TestingUtil;
import org.testng.annotations.AfterMethod;
import org.testng.annotations.BeforeMethod;
import org.testng.annotations.Test;
/**
 * Test marshalling for async mode.
 *
 * @author Ben Wang
 * @version $Revision: 4536 $
 */
@Test(groups = {"functional", "jgroups", "transaction"})
public class AsyncReplTest extends RegionBasedMarshallingTestBase
{
   CacheSPI<Object, Object> cache1, cache2;
   String props = null;
   Person ben_;
   Address addr_;
   Throwable ex_;
   private Fqn<String> aop = Fqn.fromString("/aop");

   @BeforeMethod(alwaysRun = true)
   public void setUp() throws Exception
   {
      log("creating cache1");
      cache1 = createCache("TestCache");

      log("creating cache2");

      cache2 = createCache("TestCache");

      addr_ = new Address();
      addr_.setCity("San Jose");
      ben_ = new Person();
      ben_.setName("Ben");
      ben_.setAddress(addr_);

      // Pause to give caches time to see each other
      TestingUtil.blockUntilViewsReceived(new CacheSPI[]{cache1, cache2}, 60000);
   }

   private CacheSPI<Object, Object> createCache(String name) throws Exception
   {
      CacheSPI<Object, Object> cache = (CacheSPI<Object, Object>) DefaultCacheFactory.getInstance().createCache(UnitTestCacheConfigurationFactory.createConfiguration(CacheMode.REPL_ASYNC), false);
      cache.getConfiguration().setClusterName(name);
      // Use marshaller
      cache.getConfiguration().setUseRegionBasedMarshalling(true);
      cache.create();
      cache.start();
      return cache;
   }

   @AfterMethod(alwaysRun = true)
   public void tearDown() throws Exception
   {
      cache1.removeNode(Fqn.ROOT);
      if (cache1 != null)
      {
         log("stopping cache1");
         cache1.stop();
      }

      if (cache2 != null)
      {
         log("stopping cache2");
         cache2.stop();
      }
   }

   /**
    * Test replication with classloaders.
    *
    * @throws Exception
    */
   public void testCLSet2() throws Exception
   {
      ClassLoader cla = getClassLoader();

      Region existing = cache1.getRegion(aop, false);
      if (existing == null)
      {
         existing = cache1.getRegion(aop, true);
      }
      existing.registerContextClassLoader(cla);


      ClassLoader clb = getClassLoader();
      existing = cache2.getRegion(aop, false);
      if (existing == null)
      {
         existing = cache2.getRegion(aop, true);
      }
      existing.registerContextClassLoader(clb);

      cache1.put(aop, "person", ben_);
      cache1.put(new Fqn<String>("alias"), "person", ben_);

      TestingUtil.sleepThread(1000);
      Object ben2 = null;
      // Can't cast it to Person. CCE will resutl.
      ben2 = cache2.get(aop, "person");
      assertNotNull(ben2);
      assertEquals(ben_.toString(), ben2.toString());

      Class<?> claz = clb.loadClass(ADDRESS_CLASSNAME);
      Object add = claz.newInstance();
      Class<?>[] types = {String.class};
      Method setValue = claz.getMethod("setCity", types);
      Object[] margs = {"Sunnyvale"};
      setValue.invoke(add, margs);
      Class<?> clasz1 = clb.loadClass(PERSON_CLASSNAME);
      types = new Class[]{claz};
      setValue = clasz1.getMethod("setAddress", types);
      margs = new Object[]{add};
      setValue.invoke(ben2, margs);

      // Set it back to the cache
      // Can't cast it to Person. CCE will resutl.
      cache2.put(aop, "person", ben2);
      TestingUtil.sleepThread(1000);
      Object ben3 = cache1.get(aop, "person");
      assertEquals(ben2.toString(), ben3.toString());
   }

   public void testPuts() throws Exception
   {
      ClassLoader cl = getClassLoader();
      Region r1 = cache1.getRegion(aop, false) == null ? cache1.getRegion(aop, true) : cache1.getRegion(aop, false);
      r1.registerContextClassLoader(cl);
      // Create an empty Person loaded by this classloader
      Object scopedBen1 = getPersonFromClassloader(cl);

      cl = getClassLoader();
      Region r2 = cache2.getRegion(aop, false) == null ? cache2.getRegion(aop, true) : cache2.getRegion(aop, false);
      r2.registerContextClassLoader(cl);
      // Create another empty Person loaded by this classloader
      Object scopedBen2 = getPersonFromClassloader(cl);

      cache1.put(Fqn.fromString("/aop/1"), "person", ben_);
      cache1.put(Fqn.fromString("/aop/2"), "person", scopedBen1);
      TestingUtil.sleepThread(1000);

      Object ben2 = null;
      // Can't cast it to Person. CCE will resutl.
      ben2 = cache2.get(Fqn.fromString("/aop/1"), "person");
      assertEquals(ben_.toString(), ben2.toString());

      ben2 = cache2.get(Fqn.fromString("/aop/2"), "person");
      assertFalse("cache2 deserialized with scoped classloader", ben2 instanceof Person);
      assertFalse("cache2 deserialized with cache2 classloader", scopedBen1.equals(ben2));
      assertEquals("scopedBen deserialized properly", scopedBen2, ben2);
   }

   public void testTxPut() throws Exception
   {
      beginTransaction();
      cache1.put(aop, "person", ben_);
      cache1.put(aop, "person1", ben_);
      commit();
      TestingUtil.sleepThread(1000);
      Person ben2 = (Person) cache2.get(aop, "person");
      assertNotNull("Person from 2nd cache should not be null ", ben2);
      assertEquals(ben_.toString(), ben2.toString());
   }

   public void testTxCLSet2() throws Exception
   {
      ClassLoader cla = getClassLoader();
      Region r1 = cache1.getRegion(aop, false) == null ? cache1.getRegion(aop, true) : cache1.getRegion(aop, false);
      r1.registerContextClassLoader(cla);
      ClassLoader clb = getClassLoader();
      Region r2 = cache2.getRegion(aop, false) == null ? cache2.getRegion(aop, true) : cache2.getRegion(aop, false);
      r2.registerContextClassLoader(clb);

      beginTransaction();
      cache1.put(aop, "person", ben_);
      commit();
      TestingUtil.sleepThread(1000);

      Object ben2 = null;
      // Can't cast it to Person. CCE will resutl.
      ben2 = cache2.get(aop, "person");
      assertEquals(ben_.toString(), ben2.toString());

      Class<?> claz = clb.loadClass(ADDRESS_CLASSNAME);
      Object add = claz.newInstance();
      Class<?>[] types = {String.class};
      Method setValue = claz.getMethod("setCity", types);
      Object[] margs = {"Sunnyvale"};
      setValue.invoke(add, margs);
      Class<?> clasz1 = clb.loadClass(PERSON_CLASSNAME);
      types = new Class[]{claz};
      setValue = clasz1.getMethod("setAddress", types);
      margs = new Object[]{add};
      setValue.invoke(ben2, margs);

      // Set it back to the cache
      // Can't cast it to Person. CCE will resutl.
      cache2.put(aop, "person", ben2);
      TestingUtil.sleepThread(1000);
      Object ben3 = cache1.get(aop, "person");
      assertEquals(ben2.toString(), ben3.toString());
   }

   public void testCustomFqn() throws Exception
   {
      FooClassLoader cl1 = new FooClassLoader(Thread.currentThread().getContextClassLoader());
      Region r1 = cache1.getRegion(aop, false) == null ? cache1.getRegion(aop, true) : cache1.getRegion(aop, false);
      r1.registerContextClassLoader(cl1);

      FooClassLoader cl2 = new FooClassLoader(Thread.currentThread().getContextClassLoader());
      Region r2 = cache2.getRegion(aop, false) == null ? cache2.getRegion(aop, true) : cache2.getRegion(aop, false);
      r2.registerContextClassLoader(cl2);

      Class<?> clazz = cl1.loadFoo();
      Object custom1 = clazz.newInstance();

      clazz = cl2.loadFoo();
      Object custom2 = clazz.newInstance();

      Fqn<Object> base = new Fqn<Object>("aop");
      Fqn<Object> fqn = new Fqn<Object>(base, custom1);
      cache1.put(fqn, "key", "value");
      TestingUtil.sleepThread(1000);

      Fqn<Object> fqn2 = new Fqn<Object>(base, custom2);
      Object val = cache2.get(fqn2, "key");
      assertEquals("value", val);
   }

   private Transaction beginTransaction() throws SystemException, NotSupportedException
   {
      TransactionManager mgr = cache1.getConfiguration().getRuntimeConfig().getTransactionManager();
      mgr.begin();
      return mgr.getTransaction();
   }

   private void commit() throws SecurityException, IllegalStateException, RollbackException, HeuristicMixedException, HeuristicRollbackException, SystemException
   {
      cache1.getConfiguration().getRuntimeConfig().getTransactionManager().commit();
   }

   protected Object getPersonFromClassloader(ClassLoader cl) throws Exception
   {
      Class<?> clazz = cl.loadClass(PERSON_CLASSNAME);
      return clazz.newInstance();
   }

   private void log(String msg)
   {
      System.out.println("-- [" + Thread.currentThread() + "]: " + msg);
   }
}
