/**
 *
 */
package org.jboss.cache.optimistic;

import java.io.DataInputStream;
import java.io.DataOutputStream;
import java.io.ObjectInput;
import java.io.ObjectOutput;
import java.util.LinkedList;
import java.util.List;
import java.util.Random;

import javax.transaction.SystemException;
import javax.transaction.TransactionManager;

import org.jboss.cache.Cache;
import org.jboss.cache.CacheImpl;
import org.jboss.cache.CacheSPI;
import org.jboss.cache.DefaultCacheFactory;
import org.jboss.cache.Fqn;
import org.jboss.cache.config.CacheLoaderConfig;
import org.jboss.cache.config.Configuration;
import org.jboss.cache.factories.XmlConfigurationParser;
import org.jboss.cache.interceptors.Interceptor;
import org.jboss.cache.interceptors.InvocationContextInterceptor;
import org.jboss.cache.interceptors.OptimisticCreateIfNotExistsInterceptor;
import org.jboss.cache.interceptors.OptimisticNodeInterceptor;
import org.jboss.cache.interceptors.OptimisticReplicationInterceptor;
import org.jboss.cache.interceptors.TxInterceptor;
import org.jboss.cache.loader.DummyInMemoryCacheLoader;
import org.jboss.cache.loader.DummySharedInMemoryCacheLoader;
import org.jboss.cache.lock.IsolationLevel;
import org.jboss.cache.marshall.MethodCall;
import org.jboss.cache.marshall.MethodCallFactory;
import org.jboss.cache.marshall.MethodDeclarations;
import org.jboss.cache.misc.TestingUtil;
import org.jboss.cache.transaction.TransactionSetup;
import org.jboss.cache.xml.XmlHelper;
import org.jgroups.Address;
import org.testng.annotations.AfterMethod;
import org.testng.annotations.Test;
import org.w3c.dom.Element;

/**
 * @author manik
 */
@Test(groups = "functional")
public abstract class AbstractOptimisticTestCase
{
   // some test data shared among all the test cases
   protected Fqn<String> fqn = Fqn.fromString("/blah");
   protected String key = "myKey", value = "myValue";

   protected CacheImpl<Object, Object> createCacheUnstarted() throws Exception
   {
      return createCacheUnstarted(true);
   }

   protected CacheImpl<Object, Object> createCacheUnstarted(boolean optimistic) throws Exception
   {
      CacheImpl<Object, Object> cache = (CacheImpl<Object, Object>) DefaultCacheFactory.getInstance().createCache(false);
      Configuration c = new Configuration();
      cache.setConfiguration(c);
      if (optimistic) c.setNodeLockingScheme("OPTIMISTIC");

      c.setTransactionManagerLookupClass("org.jboss.cache.transaction.DummyTransactionManagerLookup");
      c.setCacheMode("LOCAL");
      return cache;
   }

   protected CacheImpl<Object, Object> createCacheWithListener() throws Exception
   {
      return createCacheWithListener(new TestListener());
   }

   protected CacheImpl<Object, Object> createCacheWithListener(Object listener) throws Exception
   {
      CacheImpl<Object, Object> cache = createCacheUnstarted();
      cache.create();
      cache.start();
      cache.getNotifier().addCacheListener(listener);
      return cache;
   }

   /**
    * Returns a tree cache with passivation disabled in the loader.
    *
    * @return
    * @throws Exception
    */
   protected CacheImpl<Object, Object> createCacheWithLoader() throws Exception
   {
      return createCacheWithLoader(false);
   }

   protected CacheLoaderConfig getCacheLoaderConfig(boolean shared, boolean passivation) throws Exception
   {
      String xml = "            <config>\n" +
              "                <passivation>" + passivation + "</passivation>\n" +
              "                <preload></preload>\n" +
              "                <shared>" + shared + "</shared>\n" +
              "                <cacheloader>\n" +
              "                    <class>" + (shared ? DummySharedInMemoryCacheLoader.class.getName() : DummyInMemoryCacheLoader.class.getName()) + "</class>\n" +
              "                    <properties>\n" +
              "                    </properties>\n" +
              "                    <async>false</async>\n" +
              "                    <fetchPersistentState>" + (!shared) + "</fetchPersistentState>\n" +
              "                    <ignoreModifications>false</ignoreModifications>\n" +
              "                </cacheloader>\n" +
              "            </config>";
      Element element = XmlHelper.stringToElement(xml);
      return XmlConfigurationParser.parseCacheLoaderConfig(element);
   }

   protected CacheImpl<Object, Object> createCacheWithLoader(boolean passivationEnabled) throws Exception
   {
      CacheImpl<Object, Object> cache = createCacheUnstarted();
      Configuration c = cache.getConfiguration();
      cache.setConfiguration(c);
      c.setCacheLoaderConfig(getCacheLoaderConfig(true, passivationEnabled));
      cache.create();
      cache.start();
      return cache;
   }


   protected CacheImpl<Object, Object> createCache() throws Exception
   {
      CacheImpl<Object, Object> cache = createCacheUnstarted();
      cache.create();
      cache.start();
      return cache;
   }

   protected void destroyCache(Cache<Object, Object> c)
   {
      c.stop();
      c.destroy();
   }


   protected CacheImpl<Object, Object> createPessimisticCache() throws Exception
   {
      CacheImpl<Object, Object> cache = (CacheImpl<Object, Object>) DefaultCacheFactory.getInstance().createCache(false);
      Configuration c = new Configuration();
      cache.setConfiguration(c);

      c.setClusterName("name");
      c.setStateRetrievalTimeout(5000);
      c.setClusterConfig(getDefaultProperties());
      c.setCacheMode(Configuration.CacheMode.REPL_SYNC);
      c.setIsolationLevel(IsolationLevel.REPEATABLE_READ);
      c.setTransactionManagerLookupClass("org.jboss.cache.transaction.DummyTransactionManagerLookup");
      cache.create();
      cache.start();


      return cache;
   }

   protected CacheImpl<Object, Object> createPessimisticCacheLocal() throws Exception
   {
      CacheImpl<Object, Object> cache = (CacheImpl<Object, Object>) DefaultCacheFactory.getInstance().createCache(false);
      Configuration c = new Configuration();
      cache.setConfiguration(c);

      c.setClusterName("name");
      c.setStateRetrievalTimeout(5000);
      c.setClusterConfig(getDefaultProperties());

      c.setCacheMode(Configuration.CacheMode.LOCAL);
      c.setIsolationLevel(IsolationLevel.REPEATABLE_READ);
      c.setTransactionManagerLookupClass("org.jboss.cache.transaction.DummyTransactionManagerLookup");
      cache.create();
      cache.start();


      return cache;
   }

   protected String getDefaultProperties()
   {
      return "UDP(mcast_addr=228.1.2.3;mcast_port=48866;ip_ttl=32;" +
              "mcast_send_buf_size=150000;mcast_recv_buf_size=80000;loopback=true;ip_mcast=true;bind_addr=127.0.0.1):" +
              "PING(timeout=1000;num_initial_members=2):" +
              "MERGE2(min_interval=5000;max_interval=10000):" +
              "FD_SOCK:" +
              "VERIFY_SUSPECT(timeout=1500):" +
              "pbcast.NAKACK(gc_lag=50;max_xmit_size=8192;retransmit_timeout=600,1200,2400,4800):" +
              "UNICAST(timeout=600,1200,2400,4800):" +
              "pbcast.STABLE(desired_avg_gossip=20000):" +
              "FRAG(frag_size=8192;down_thread=false;up_thread=false):" +
              "pbcast.GMS(join_timeout=5000;join_retry_timeout=2000;" +
              "shun=false;print_local_addr=true):" +
              "pbcast.STATE_TRANSFER";
   }

   protected CacheImpl<Object, Object> createReplicatedCache(Configuration.CacheMode mode) throws Exception
   {
      return createReplicatedCache("test", mode);
   }

   protected CacheImpl<Object, Object> createReplicatedCache(String name, Configuration.CacheMode mode) throws Exception
   {
      return createReplicatedCache(name, mode, true);
   }
   
   protected CacheImpl<Object, Object> createReplicatedCache(String name, Configuration.CacheMode mode, boolean start) throws Exception
   {  
      CacheImpl<Object, Object> cache = (CacheImpl<Object, Object>) DefaultCacheFactory.getInstance().createCache(false);
      Configuration c = new Configuration();
      cache.setConfiguration(c);

      c.setClusterName(name);
      c.setStateRetrievalTimeout(5000);
      c.setClusterConfig(getDefaultProperties());
      c.setCacheMode(mode);
      if (mode == Configuration.CacheMode.REPL_SYNC)
      {
         // make sure commits and rollbacks are sync as well
         c.setSyncCommitPhase(true);
         c.setSyncRollbackPhase(true);
      }
      c.setNodeLockingScheme("OPTIMISTIC");
      c.setTransactionManagerLookupClass(TransactionSetup.getManagerLookup());
      
      if (start)
      {
         cache.create();
         cache.start();
      }

      return cache;
   }

   protected CacheImpl<Object, Object> createReplicatedCacheWithLoader(boolean shared, Configuration.CacheMode cacheMode) throws Exception
   {
      return createReplicatedCacheWithLoader("temp-loader", shared, cacheMode);
   }

   protected CacheImpl<Object, Object> createReplicatedCacheWithLoader(boolean shared) throws Exception
   {
      return createReplicatedCacheWithLoader("temp-loader", shared, Configuration.CacheMode.REPL_SYNC);
   }

   protected CacheImpl<Object, Object> createReplicatedCacheWithLoader(String name, boolean shared) throws Exception
   {
      return createReplicatedCacheWithLoader(name, shared, Configuration.CacheMode.REPL_SYNC);
   }

   protected CacheImpl<Object, Object> createReplicatedCacheWithLoader(String name, boolean shared, Configuration.CacheMode cacheMode) throws Exception
   {
      CacheImpl<Object, Object> cache = (CacheImpl<Object, Object>) DefaultCacheFactory.getInstance().createCache(false);
      Configuration c = new Configuration();
      cache.setConfiguration(c);
      c.setClusterName(name);
      c.setStateRetrievalTimeout(5000);
      c.setClusterConfig(getDefaultProperties());
      c.setCacheMode(cacheMode);
      c.setSyncCommitPhase(true);
      c.setSyncRollbackPhase(true);
      c.setNodeLockingScheme("OPTIMISTIC");
      c.setTransactionManagerLookupClass("org.jboss.cache.transaction.DummyTransactionManagerLookup");
      c.setCacheLoaderConfig(getCacheLoaderConfig(shared, false));

      cache.create();
      cache.start();
      return cache;
   }

   protected Random random;

   protected void randomSleep(int min, int max)
   {
      if (random == null) random = new Random();
      long l = -1;
      while (l < min) l = random.nextInt(max);
      TestingUtil.sleepThread(l);
   }

   @AfterMethod(alwaysRun = true)
   public void tearDown()
   {
      TransactionManager mgr = TransactionSetup.getManager();
      try
      {
         if (mgr.getTransaction() != null)
         {
            mgr.rollback();
         }
      }
      catch (SystemException e)
      {
         // do nothing
      }
   }

   protected Interceptor getAlteredInterceptorChain(Interceptor newLast, CacheSPI<Object, Object> spi, boolean replicated)
   {
      Interceptor ici = new InvocationContextInterceptor();
      ici.setCache(spi);

      Interceptor txInterceptor = new TxInterceptor();
      txInterceptor.setCache(spi);

      Interceptor replicationInterceptor = new OptimisticReplicationInterceptor();
      replicationInterceptor.setCache(spi);

      Interceptor createInterceptor = new OptimisticCreateIfNotExistsInterceptor();
      createInterceptor.setCache(spi);

      Interceptor nodeInterceptor = new OptimisticNodeInterceptor();
      nodeInterceptor.setCache(spi);

      ici.setNext(txInterceptor);
      if (replicated)
      {
         txInterceptor.setNext(replicationInterceptor);
         replicationInterceptor.setNext(createInterceptor);
      }
      else
      {
         txInterceptor.setNext(createInterceptor);
      }
      createInterceptor.setNext(nodeInterceptor);
      nodeInterceptor.setNext(newLast);

      return ici;
   }

   public abstract class ExceptionThread extends Thread
   {
      protected Exception exception;

      public void setException(Exception e)
      {
         exception = e;
      }

      public Exception getException()
      {
         return exception;
      }
   }

   protected List<MethodCall> injectDataVersion(List<MethodCall> modifications)
   {
      List<MethodCall> newList = new LinkedList<MethodCall>();
      for (MethodCall c : modifications)
      {
         Object[] oa = c.getArgs();
         Object[] na = new Object[oa.length + 1];
         System.out.println("*** " + oa.length);
         System.arraycopy(oa, 0, na, 0, oa.length);
         na[oa.length] = new DefaultDataVersion();
         newList.add(MethodCallFactory.create(MethodDeclarations.getVersionedMethod(c.getMethodId()), na));
      }
      return newList;
   }

   protected class DummyAddress implements Address
   {
      private static final long serialVersionUID = -2628268587640985944L;

      public int compareTo(Object arg0)
      {
         return 0;
      }

      public void readFrom(DataInputStream
              arg0)
      {
      }

      public void writeTo(DataOutputStream
              arg0)
      {
      }

      public void readExternal(ObjectInput
              arg0)
      {
      }

      public void writeExternal(ObjectOutput
              arg0)
      {
      }

      public int size()
      {
         return 0;
      }

      public boolean isMulticastAddress()
      {
         return false;
      }

   }

}
