package org.jboss.cache.optimistic;

import org.jboss.cache.CacheImpl;
import org.jboss.cache.Fqn;
import org.jboss.cache.interceptors.Interceptor;
import org.jboss.cache.interceptors.OptimisticCreateIfNotExistsInterceptor;
import org.jboss.cache.interceptors.OptimisticNodeInterceptor;
import org.jboss.cache.loader.SamplePojo;
import org.jboss.cache.transaction.GlobalTransaction;
import org.jboss.cache.transaction.OptimisticTransactionEntry;
import org.jboss.cache.transaction.TransactionTable;
import static org.testng.AssertJUnit.*;
import org.testng.annotations.Test;

import javax.transaction.Transaction;
import javax.transaction.TransactionManager;
import java.util.Iterator;

/**
 * @author xenephon
 */
@Test (groups = {"functional"})
public class NodeInterceptorGetChildrenNamesTest extends AbstractOptimisticTestCase
{
   public void testTransactionGetNamesMethod() throws Exception
   {

      TestListener listener = new TestListener();
      final CacheImpl<Object, Object> cache = createCacheWithListener(listener);

      Interceptor interceptor = new OptimisticCreateIfNotExistsInterceptor();
      interceptor.setCache(cache);
      Interceptor nodeInterceptor = new OptimisticNodeInterceptor();
      nodeInterceptor.setCache(cache);
      MockInterceptor dummy = new MockInterceptor();
      dummy.setCache(cache);

      interceptor.setNext(nodeInterceptor);
      nodeInterceptor.setNext(dummy);

      cache.setInterceptorChain(interceptor);

//		 first set up a node with a pojo
      TransactionManager mgr = cache.getConfiguration().getRuntimeConfig().getTransactionManager();
      mgr.begin();
      Transaction tx = mgr.getTransaction();

      // inject InvocationContext
      cache.getInvocationContext().setTransaction(tx);
      cache.getInvocationContext().setGlobalTransaction(cache.getCurrentTransaction(tx));


      SamplePojo pojo = new SamplePojo(21, "test");

      cache.put("/one/two", "key1", pojo);

      assertEquals(null, dummy.getCalled());
      TransactionTable table = cache.getTransactionTable();

      GlobalTransaction gtx = table.get(tx);

      OptimisticTransactionEntry entry = (OptimisticTransactionEntry) table.get(gtx);

      @SuppressWarnings("unchecked")
      TransactionWorkspace<Object, Object> workspace = entry.getTransactionWorkSpace();

      //assert we can see this with a key value get in the transaction
      assertEquals(1, cache.getChildrenNames("/one").size());
      mgr.commit();

      //assert what should be the results of our call
      assertEquals(3, workspace.getNodes().size());
      assertNotNull(workspace.getNode(Fqn.fromString("/one/two")));
      assertEquals(pojo, workspace.getNode(Fqn.fromString("/one/two")).get("key1"));
      assertTrue(entry.getLocks().isEmpty());
      assertEquals(1, entry.getModifications().size());
      assertTrue(!cache.exists("/one/two"));
      assertEquals(null, dummy.getCalled());

      //assert that we cannot see the change if we have not put it into the cache
      // we need to do this as we do not have the tx interceptor in this stack
      mgr.begin();

      Transaction tx2 = mgr.getTransaction();

      // inject InvocationContext
      cache.getInvocationContext().setTransaction(tx2);
      cache.getInvocationContext().setGlobalTransaction(cache.getCurrentTransaction(tx2));

      assertEquals(0, cache.getChildrenNames("/").size());
      mgr.commit();
      cache.stop();
   }


   public void testTransactionGetNoNamesMethod() throws Exception
   {

      TestListener listener = new TestListener();
      final CacheImpl<Object, Object> cache = createCacheWithListener(listener);

      Interceptor interceptor = new OptimisticCreateIfNotExistsInterceptor();
      interceptor.setCache(cache);
      Interceptor nodeInterceptor = new OptimisticNodeInterceptor();
      nodeInterceptor.setCache(cache);
      MockInterceptor dummy = new MockInterceptor();
      dummy.setCache(cache);

      interceptor.setNext(nodeInterceptor);
      nodeInterceptor.setNext(dummy);

      cache.setInterceptorChain(interceptor);

//		 first set up a node with a pojo
      TransactionManager mgr = cache.getConfiguration().getRuntimeConfig().getTransactionManager();
      mgr.begin();
      Transaction tx = mgr.getTransaction();

      // inject InvocationContext
      cache.getInvocationContext().setTransaction(tx);
      cache.getInvocationContext().setGlobalTransaction(cache.getCurrentTransaction(tx));


      assertEquals(null, dummy.getCalled());
      TransactionTable table = cache.getTransactionTable();

      GlobalTransaction gtx = table.get(tx);

      OptimisticTransactionEntry entry = (OptimisticTransactionEntry) table.get(gtx);

      //assert we can see this with a key value get in the transaction
      assertEquals(0, cache.getChildrenNames("/").size());
      mgr.commit();


      assertTrue(entry.getLocks().isEmpty());
      assertEquals(0, entry.getModifications().size());
      assertTrue(!cache.exists("/one/two"));
      assertEquals(null, dummy.getCalled());


      cache.stop();
   }


   public void testTransactionGetNamesIteratorMethod() throws Exception
   {

      TestListener listener = new TestListener();
      final CacheImpl<Object, Object> cache = createCacheWithListener(listener);

      Interceptor interceptor = new OptimisticCreateIfNotExistsInterceptor();
      interceptor.setCache(cache);
      Interceptor nodeInterceptor = new OptimisticNodeInterceptor();
      nodeInterceptor.setCache(cache);
      MockInterceptor dummy = new MockInterceptor();
      dummy.setCache(cache);

      interceptor.setNext(nodeInterceptor);
      nodeInterceptor.setNext(dummy);

      cache.setInterceptorChain(interceptor);

//		 first set up a node with a pojo
      TransactionManager mgr = cache.getConfiguration().getRuntimeConfig().getTransactionManager();
      mgr.begin();
      Transaction tx = mgr.getTransaction();

      // inject InvocationContext
      cache.getInvocationContext().setTransaction(tx);
      cache.getInvocationContext().setGlobalTransaction(cache.getCurrentTransaction(tx));

      SamplePojo pojo = new SamplePojo(21, "test");

      cache.put("/one/two", "key1", pojo);

      assertEquals(null, dummy.getCalled());
      TransactionTable table = cache.getTransactionTable();

      GlobalTransaction gtx = table.get(tx);

      OptimisticTransactionEntry entry = (OptimisticTransactionEntry) table.get(gtx);

      @SuppressWarnings("unchecked")
      TransactionWorkspace<Object, Object> workspace = entry.getTransactionWorkSpace();

      //assert we can see this
      assertEquals(1, cache.getChildrenNames("/one").size());

      try
      {
         for (Iterator<?> it = cache.getChildrenNames("/one").iterator(); it.hasNext();)
         {
            it.next();
            it.remove();
         }
         fail("Should not be allowed to modify elements in the set returned by getChildrenNames()");
      }
      catch (UnsupportedOperationException uoe)
      {
         // the returned set should be unmodifiable and a remove on the iterator should fail.
      }

      //assert the removal has had no effect
      assertEquals(1, cache.getChildrenNames("/one").size());
      assertNotNull(workspace.getNode(Fqn.fromString("/one/two")));

      mgr.commit();


      assertTrue(entry.getLocks().isEmpty());
      assertEquals(1, entry.getModifications().size());
      assertTrue(!cache.exists("/one/two"));
      assertEquals(null, dummy.getCalled());


      cache.stop();
   }
}
