/*
 * Created on 17-Feb-2005
 *
 *
 *
 */
package org.jboss.cache.optimistic;

import org.jboss.cache.CacheImpl;
import org.jboss.cache.Fqn;
import org.jboss.cache.Node;
import org.jboss.cache.factories.InterceptorChainFactory;
import org.jboss.cache.interceptors.CallInterceptor;
import org.jboss.cache.interceptors.Interceptor;
import org.jboss.cache.loader.SamplePojo;
import org.jboss.cache.transaction.GlobalTransaction;
import org.jboss.cache.transaction.OptimisticTransactionEntry;
import org.jboss.cache.transaction.TransactionTable;
import static org.testng.AssertJUnit.*;
import org.testng.annotations.AfterMethod;
import org.testng.annotations.BeforeMethod;
import org.testng.annotations.Test;

import javax.transaction.Transaction;
import javax.transaction.TransactionManager;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Set;

/**
 * @author xenephon
 */
@Test (groups = {"functional"})
@SuppressWarnings("unchecked")
public class NodeInterceptorRemoveNodeTest extends AbstractOptimisticTestCase
{
   private CacheImpl<Object, Object> cache;
   private TestListener listener;
   private MockInterceptor dummy;
   private TransactionManager mgr;

   @BeforeMethod(alwaysRun = true) public void setUp() throws Exception
   {
      listener = new TestListener();
      cache = createCacheWithListener(listener);

      dummy = new MockInterceptor();
      dummy.setCache(cache);

      Interceptor interceptor = cache.getInterceptorChain().get(0); // get the first interceptor
      // replace the last (call) interceptor with the mock interceptor.
      Interceptor next = interceptor, prev = interceptor;
      while (!(next instanceof CallInterceptor))
      {
         prev = next;
         next = next.getNext();
      }

      prev.setNext(dummy);

      InterceptorChainFactory.getInstance().setLastInterceptorPointer(interceptor, dummy);

      cache.setInterceptorChain(interceptor);
      mgr = cache.getTransactionManager();
   }

   @AfterMethod(alwaysRun = true) public void tearDown()
   {
      super.tearDown();
      cache.stop();
   }

   public void testTransactionRemoveNotExistsNodeMethod() throws Exception
   {
      mgr.begin();
      Transaction tx = mgr.getTransaction();

      // inject InvocationContext
      cache.getInvocationContext().setTransaction(tx);
      cache.getInvocationContext().setGlobalTransaction(cache.getCurrentTransaction(tx));


      cache.remove("/one/two");

      TransactionTable table = cache.getTransactionTable();
      GlobalTransaction gtx = table.get(tx);
      OptimisticTransactionEntry entry = (OptimisticTransactionEntry) table.get(gtx);
      TransactionWorkspace workspace = entry.getTransactionWorkSpace();

      mgr.commit();
      assertEquals(null, dummy.getCalled());

      //assert what should be the results of our call
      assertEquals(0, workspace.getNodes().size());

      assertTrue(entry.getLocks().isEmpty());
      assertEquals(1, entry.getModifications().size());
      assertTrue(!cache.exists("/one/two"));
      assertEquals(null, dummy.getCalled());
      assertEquals(0, listener.getNodesAdded());
   }

   public void testTransactionRemoveNodeMethod() throws Exception
   {
      mgr.begin();
      Transaction tx = mgr.getTransaction();

      // inject InvocationContext
      cache.getInvocationContext().setTransaction(tx);
      cache.getInvocationContext().setGlobalTransaction(cache.getCurrentTransaction(tx));

      SamplePojo pojo = new SamplePojo(21, "test");
      Map temp = new HashMap();
      temp.put("key1", pojo);
      cache.put("/one/two", temp);

      cache.remove("/one/two");
      assertEquals(null, dummy.getCalled());
      TransactionTable table = cache.getTransactionTable();

      GlobalTransaction gtx = table.get(tx);

      OptimisticTransactionEntry entry = (OptimisticTransactionEntry) table.get(gtx);

      TransactionWorkspace<?, ?> workspace = entry.getTransactionWorkSpace();

      mgr.commit();

      //assert what should be the results of our call
      assertEquals(3, workspace.getNodes().size());

      assertNotNull(workspace.getNode(Fqn.fromString("/one/two")));
      assertEquals(true, workspace.getNode(Fqn.fromString("/one/two")).isDeleted());
      assertNotNull(workspace.getNode(Fqn.fromString("/one")));
      assertEquals(false, workspace.getNode(Fqn.fromString("/one")).isDeleted());
      List<Set<Fqn>> mergedChildren = workspace.getNode(Fqn.fromString("/one")).getMergedChildren();
      assertEquals(1, mergedChildren.get(1).size());
      System.out.println(entry.getModifications());
      assertEquals(2, entry.getModifications().size());
      assertTrue(!cache.exists("/one/two"));
      assertEquals(null, dummy.getCalled());
   }

   public void testTransactionRemoveIntermediateNodeMethod() throws Exception
   {
      mgr.begin();
      Transaction tx = mgr.getTransaction();

      // inject InvocationContext
      cache.getInvocationContext().setTransaction(tx);
      cache.getInvocationContext().setGlobalTransaction(cache.getCurrentTransaction(tx));

      SamplePojo pojo = new SamplePojo(21, "test");
      Map temp = new HashMap();
      temp.put("key1", pojo);
      cache.put("/one/two", temp);

      cache.remove("/one");
      assertEquals(null, dummy.getCalled());
      TransactionTable table = cache.getTransactionTable();

      GlobalTransaction gtx = table.get(tx);

      OptimisticTransactionEntry entry = (OptimisticTransactionEntry) table.get(gtx);

      TransactionWorkspace<?, ?> workspace = entry.getTransactionWorkSpace();

      mgr.commit();

      //assert what should be the results of our call
      assertEquals(3, workspace.getNodes().size());

      assertNotNull(workspace.getNode(Fqn.fromString("/one/two")));
      assertEquals(true, workspace.getNode(Fqn.fromString("/one/two")).isDeleted());
      assertNotNull(workspace.getNode(Fqn.fromString("/one")));
      assertEquals(true, workspace.getNode(Fqn.fromString("/one")).isDeleted());
      assertEquals(1, workspace.getNode(Fqn.fromString("/one")).getMergedChildren().get(0).size());
      assertEquals(2, entry.getModifications().size());
      assertTrue(!cache.exists("/one/two"));
      assertEquals(null, dummy.getCalled());
   }

   public void testTransactionRemoveTwiceMethod() throws Exception
   {
      mgr.begin();
      Transaction tx = mgr.getTransaction();

      // inject InvocationContext
      cache.getInvocationContext().setTransaction(tx);
      cache.getInvocationContext().setGlobalTransaction(cache.getCurrentTransaction(tx));

      SamplePojo pojo = new SamplePojo(21, "test");
      Map temp = new HashMap();
      temp.put("key1", pojo);
      cache.put("/one/two", temp);

      // get the transaction stuff
      TransactionTable table = cache.getTransactionTable();
      GlobalTransaction gtx = table.get(tx);

      OptimisticTransactionEntry entry = (OptimisticTransactionEntry) table.get(gtx);

      TransactionWorkspace workspace = entry.getTransactionWorkSpace();

      workspace.getNode(Fqn.fromString("/one"));
      workspace.getNode(Fqn.fromString("/one/two"));


      cache.remove("/one");

      assertNotNull(workspace.getNode(Fqn.fromString("/one/two")));
      assertEquals(true, workspace.getNode(Fqn.fromString("/one/two")).isDeleted());
      assertNotNull(workspace.getNode(Fqn.fromString("/one")));
      assertEquals(true, workspace.getNode(Fqn.fromString("/one")).isDeleted());

      //now put /one/two back in
      cache.remove("/one");


      assertNotNull(workspace.getNode(Fqn.fromString("/one/two")));
      assertEquals(true, workspace.getNode(Fqn.fromString("/one/two")).isDeleted());
      assertNotNull(workspace.getNode(Fqn.fromString("/one")));
      assertEquals(true, workspace.getNode(Fqn.fromString("/one")).isDeleted());

      assertEquals(null, dummy.getCalled());

      mgr.commit();

      //assert what should be the results of our call
      assertEquals(3, workspace.getNodes().size());


      assertEquals(3, entry.getModifications().size());
      assertTrue(!cache.exists("/one/two"));
      assertEquals(null, dummy.getCalled());
   }


   public void testTransactionRemovePutNodeMethod() throws Exception
   {
      mgr.begin();
      Transaction tx = mgr.getTransaction();

      // inject InvocationContext
      cache.getInvocationContext().setTransaction(tx);
      cache.getInvocationContext().setGlobalTransaction(cache.getCurrentTransaction(tx));

      SamplePojo pojo = new SamplePojo(21, "test");
      Map temp = new HashMap();
      temp.put("key1", pojo);
      cache.put("/one/two", temp);

      // get the transaction stuff
      TransactionTable table = cache.getTransactionTable();
      GlobalTransaction gtx = table.get(tx);

      OptimisticTransactionEntry entry = (OptimisticTransactionEntry) table.get(gtx);

      TransactionWorkspace workspace = entry.getTransactionWorkSpace();

      Node one = workspace.getNode(Fqn.fromString("/one"));
      Node two = workspace.getNode(Fqn.fromString("/one/two"));


      cache.remove("/one");

      assertNotNull(workspace.getNode(Fqn.fromString("/one/two")));
      assertEquals(true, workspace.getNode(Fqn.fromString("/one/two")).isDeleted());
      assertNotNull(workspace.getNode(Fqn.fromString("/one")));
      assertEquals(true, workspace.getNode(Fqn.fromString("/one")).isDeleted());

      //now put /one/two back in
      cache.put("/one/two", temp);

      WorkspaceNode oneAfter = workspace.getNode(Fqn.fromString("/one"));
      WorkspaceNode twoAfter = workspace.getNode(Fqn.fromString("/one/two"));

      assertSame(one, oneAfter);
      assertEquals(false, oneAfter.isDeleted());
      assertSame(two, twoAfter);
      assertEquals(false, twoAfter.isDeleted());

      assertEquals(null, dummy.getCalled());


      mgr.commit();

      //assert what should be the results of our call
      assertEquals(3, workspace.getNodes().size());


      assertEquals(3, entry.getModifications().size());
      assertEquals(null, dummy.getCalled());
      assertEquals(2, listener.getNodesAdded());
   }


   public void testTransactionRemovePutkeyValMethod() throws Exception
   {
      mgr.begin();
      Transaction tx = mgr.getTransaction();

      // inject InvocationContext
      cache.getInvocationContext().setTransaction(tx);
      cache.getInvocationContext().setGlobalTransaction(cache.getCurrentTransaction(tx));

      SamplePojo pojo = new SamplePojo(21, "test");
      Map temp = new HashMap();
      temp.put("key1", pojo);
      cache.put("/one/two", temp);

      // get the transaction stuff
      TransactionTable table = cache.getTransactionTable();
      GlobalTransaction gtx = table.get(tx);

      OptimisticTransactionEntry entry = (OptimisticTransactionEntry) table.get(gtx);

      TransactionWorkspace workspace = entry.getTransactionWorkSpace();

      Node one = workspace.getNode(Fqn.fromString("/one"));
      Node two = workspace.getNode(Fqn.fromString("/one/two"));


      cache.remove("/one");

      assertNotNull(workspace.getNode(Fqn.fromString("/one/two")));
      assertEquals(true, workspace.getNode(Fqn.fromString("/one/two")).isDeleted());
      assertNotNull(workspace.getNode(Fqn.fromString("/one")));
      assertEquals(true, workspace.getNode(Fqn.fromString("/one")).isDeleted());

      //now put /one back in
      SamplePojo pojo2 = new SamplePojo(21, "test");
      cache.put(new Fqn("one"), "key1", pojo2);

      WorkspaceNode oneAfter = workspace.getNode(Fqn.fromString("/one"));
      WorkspaceNode twoAfter = workspace.getNode(Fqn.fromString("/one/two"));

      assertSame(one, oneAfter);
      assertEquals(false, oneAfter.isDeleted());
      assertEquals(two, twoAfter);
      assertEquals(true, twoAfter.isDeleted());

      assertEquals(null, dummy.getCalled());

      mgr.commit();

      assertEquals(pojo2, workspace.getNode(Fqn.fromString("/one")).get("key1"));
      //assert what should be the results of our call
      assertEquals(3, workspace.getNodes().size());


      assertEquals(3, entry.getModifications().size());
      assertTrue(!cache.exists("/one/two"));
      assertEquals(null, dummy.getCalled());
      assertEquals(2, listener.getNodesAdded());
   }

   public void testTransactionRemoveSubNodeMethod() throws Exception
   {
      mgr.begin();
      Transaction tx = mgr.getTransaction();

      // inject InvocationContext
      cache.getInvocationContext().setTransaction(tx);
      cache.getInvocationContext().setGlobalTransaction(cache.getCurrentTransaction(tx));

      SamplePojo pojo = new SamplePojo(21, "test");
      Map temp = new HashMap();
      temp.put("key1", pojo);
      cache.put("/one/two", temp);

      // get the transaction stuff
      TransactionTable table = cache.getTransactionTable();
      GlobalTransaction gtx = table.get(tx);

      OptimisticTransactionEntry entry = (OptimisticTransactionEntry) table.get(gtx);

      TransactionWorkspace<?, ?> workspace = entry.getTransactionWorkSpace();

      WorkspaceNode<?, ?> one = workspace.getNode(Fqn.fromString("/one"));

      assertEquals(1, one.getMergedChildren().get(0).size());

      cache.remove("/one/two");

      assertNotNull(workspace.getNode(Fqn.fromString("/one/two")));
      assertEquals(true, workspace.getNode(Fqn.fromString("/one/two")).isDeleted());
      assertNotNull(workspace.getNode(Fqn.fromString("/one")));
      assertEquals(false, workspace.getNode(Fqn.fromString("/one")).isDeleted());

      assertEquals(null, dummy.getCalled());

      mgr.commit();

      assertEquals(1, workspace.getNode(Fqn.fromString("/one")).getMergedChildren().get(1).size());
      //assert what should be the results of our call
      assertEquals(3, workspace.getNodes().size());


      assertEquals(2, entry.getModifications().size());
      assertTrue(!cache.exists("/one/two"));
      assertEquals(null, dummy.getCalled());
      assertEquals(2, listener.getNodesAdded());
   }
}
