/*
 * JBoss, Home of Professional Open Source
 *
 * Distributable under LGPL license.
 * See terms of license at gnu.org.
 */
package org.jboss.cache.options.cachemodelocal;

import static org.testng.AssertJUnit.assertEquals;
import static org.testng.AssertJUnit.assertNull;

import java.util.HashMap;
import java.util.Map;

import javax.transaction.SystemException;
import javax.transaction.TransactionManager;

import junit.framework.Assert;

import org.jboss.cache.CacheFactory;
import org.jboss.cache.CacheImpl;
import org.jboss.cache.DefaultCacheFactory;
import org.jboss.cache.Fqn;
import org.jboss.cache.config.Configuration;
import org.testng.annotations.AfterMethod;
import org.testng.annotations.BeforeMethod;
import org.testng.annotations.Test;

/**
 * Tests the cache mode local override in various scenarios.  To be subclassed to test REPL_SYNC, REPL_ASYNC, INVALIDATION_SYNC, INVALIDATION_ASYNC for Opt and Pess locking.
 * <p/>
 * Option.setCacheModeLocal() only applies to put() and remove() methods.
 *
 * @author <a href="mailto:manik@jboss.org">Manik Surtani (manik@jboss.org)</a>
 */
@Test(groups = { "functional" })
public abstract class CacheModeLocalTestBase
{
   // to be subclassed.
   protected Configuration.CacheMode cacheMode;
   protected String nodeLockingScheme;
   /**
    * set this to true if the implementing class plans to use an invalidating cache mode *
    */
   protected boolean isInvalidation;

   private CacheImpl<String, String> cache1;
   private CacheImpl<String, String> cache2;

   private Fqn fqn = Fqn.fromString("/a");
   private String key = "key";

   @BeforeMethod(alwaysRun = true)
   public void setUp() throws Exception
   {
      // force a tear down if the test runner didn't run one before (happens in IDEA)
      if (cache1 != null || cache2 != null)
         tearDown();

      CacheFactory<String, String> instance = DefaultCacheFactory.getInstance();
      cache1 = (CacheImpl<String, String>)instance.createCache(false);
      cache1.getConfiguration().setClusterName("test");
      cache1.getConfiguration().setStateRetrievalTimeout(1000);
      cache1.getConfiguration().setTransactionManagerLookupClass("org.jboss.cache.transaction.DummyTransactionManagerLookup");
      cache1.getConfiguration().setNodeLockingScheme(nodeLockingScheme);
      cache1.getConfiguration().setCacheMode(cacheMode);
      cache1.start();

      cache2 = (CacheImpl<String, String>)instance.createCache(false);
      cache2.getConfiguration().setClusterName("test");
      cache2.getConfiguration().setStateRetrievalTimeout(1000);
      cache2.getConfiguration().setTransactionManagerLookupClass("org.jboss.cache.transaction.DummyTransactionManagerLookup");
      cache2.getConfiguration().setNodeLockingScheme(nodeLockingScheme);
      cache2.getConfiguration().setCacheMode(cacheMode);
      cache2.start();
   }

   @AfterMethod(alwaysRun = true)
   public void tearDown()
   {
      if (cache1 != null)
      {
         cache1.stop();
         flushTxs(cache1.getTransactionManager());
         cache1 = null;
      }

      if (cache2 != null)
      {
         cache2.stop();
         flushTxs(cache2.getTransactionManager());
         cache2 = null;
      }
   }

   private void flushTxs(TransactionManager mgr)
   {
      if (mgr != null)
      {
         try
         {
            if (mgr.getTransaction() != null)
            {
               mgr.rollback();
            }
         }
         catch (SystemException e)
         {
            // do nothing
         }
      }
   }

   public void testPutKeyValue() throws Exception
   {
      cache1.getInvocationContext().getOptionOverrides().setCacheModeLocal(true);
      cache1.put(fqn, key, "value");
      delay();
      // cache1 should still have this
      Assert.assertEquals("value", cache1.get(fqn, key));

      // cache 2 should not
      Assert.assertNull("Should be null", cache2.get(fqn, key));

      // now try again with passing the default options
      cache1.getInvocationContext().getOptionOverrides().reset();
      cache1.put(fqn, key, "value");
      delay();
      // cache1 should still have this
      Assert.assertEquals("value", cache1.get(fqn, key));

      // cache 2 should as well
      if (!isInvalidation)
      {
         Assert.assertEquals("value", cache2.get(fqn, key));
      }
      else
      {
         Assert.assertNull("should be invalidated", cache2.get(fqn, key));
      }

      // now cache2
      cache2.getInvocationContext().getOptionOverrides().setCacheModeLocal(true);
      cache2.put(fqn, key, "value2");
      delay();
      Assert.assertEquals("value2", cache2.get(fqn, key));
      Assert.assertEquals("value", cache1.get(fqn, key));

      cache2.getInvocationContext().getOptionOverrides().reset();
      cache2.put(fqn, key, "value2");
      delay();
      Assert.assertEquals("value2", cache2.get(fqn, key));
      if (!isInvalidation)
      {
         Assert.assertEquals("value2", cache1.get(fqn, key));
      }
      else
      {
         Assert.assertNull("should be invalidated", cache1.get(fqn, key));
      }
   }

   public void testPutData() throws Exception
   {
      Map<String, String> map = new HashMap<String, String>();
      map.put(key, "value");

      cache1.getInvocationContext().getOptionOverrides().setCacheModeLocal(true);
      cache1.put(fqn, map);
      delay();
      // cache1 should still have this
      Assert.assertEquals("value", cache1.get(fqn, key));
      // cache 2 should not
      Assert.assertNull("Should be null", cache2.get(fqn, key));

      // now try again with passing the default options
      cache1.getInvocationContext().getOptionOverrides().setCacheModeLocal(false);
      cache1.put(fqn, map);
      delay();
      // cache1 should still have this
      Assert.assertEquals("value", cache1.get(fqn, key));
      // cache 2 should as well
      if (!isInvalidation)
      {
         Assert.assertEquals("value", cache2.get(fqn, key));
      }
      else
      {
         Assert.assertNull("should be invalidated", cache2.get(fqn, key));
      }

      // now cache2
      map.put(key, "value2");
      cache2.getInvocationContext().getOptionOverrides().setCacheModeLocal(true);
      cache2.put(fqn, map);
      delay();
      Assert.assertEquals("value2", cache2.get(fqn, key));
      Assert.assertEquals("value", cache1.get(fqn, key));

      cache2.getInvocationContext().getOptionOverrides().reset();
      cache2.put(fqn, key, "value2");
      delay();
      Assert.assertEquals("value2", cache2.get(fqn, key));
      if (!isInvalidation)
      {
         Assert.assertEquals("value2", cache1.get(fqn, key));
      }
      else
      {
         Assert.assertNull("should be invalidated", cache1.get(fqn, key));
      }
   }

   public void testRemoveNode() throws Exception
   {
      // put some stuff in the cache first
      // make sure we cleanup thread local vars.
      cache1.getInvocationContext().setOptionOverrides(null);
      cache1.put(fqn, key, "value");
      delay();
      Assert.assertEquals("value", cache1.get(fqn, key));
      if (isInvalidation)
      {
         Assert.assertNull("Should be null", cache2.get(fqn, key));
      }
      else
      {
         Assert.assertEquals("value", cache2.get(fqn, key));
      }

      cache1.getInvocationContext().getOptionOverrides().setCacheModeLocal(true);
      cache1.removeNode(fqn);
      delay();

      // should be removed in cache1
      Assert.assertNull("should be null", cache1.get(fqn, key));
      // Not in cache2
      if (isInvalidation)
      {
         Assert.assertNull("Should be null", cache2.get(fqn, key));
      }
      else
      {
         Assert.assertEquals("value", cache2.get(fqn, key));
      }

      // replace cache entries
      cache1.put(fqn, key, "value");
      delay();
      Assert.assertEquals("value", cache1.get(fqn, key));
      if (isInvalidation)
      {
         Assert.assertNull("Should be null", cache2.get(fqn, key));
      }
      else
      {
         Assert.assertEquals("value", cache2.get(fqn, key));
      }

      // now try again with passing the default options
      cache1.getInvocationContext().getOptionOverrides().setCacheModeLocal(false);
      cache1.remove(fqn);
      delay();

      // both should be null
      Assert.assertNull("should be null", cache1.get(fqn, key));
      Assert.assertNull("should be null", cache2.get(fqn, key));
   }

   public void testRemoveKey() throws Exception
   {
      // put some stuff in the cache first
      cache1.getInvocationContext().setOptionOverrides(null);
      cache1.put(fqn, key, "value");
      delay();
      Assert.assertEquals("value", cache1.get(fqn, key));
      if (isInvalidation)
      {
         Assert.assertNull("Should be null", cache2.get(fqn, key));
      }
      else
      {
         Assert.assertEquals("value", cache2.get(fqn, key));
      }

      cache1.getInvocationContext().getOptionOverrides().setCacheModeLocal(true);
      cache1.remove(fqn, key);
      delay();

      // should be removed in cache1
      Assert.assertNull("should be null", cache1.get(fqn, key));
      // Not in cache2
      if (isInvalidation)
      {
         Assert.assertNull("Should be null", cache2.get(fqn, key));
      }
      else
      {
         Assert.assertEquals("value", cache2.get(fqn, key));
      }

      // replace cache entries
      cache1.put(fqn, key, "value");
      delay();
      Assert.assertEquals("value", cache1.get(fqn, key));
      if (isInvalidation)
      {
         Assert.assertNull("Should be null", cache2.get(fqn, key));
      }
      else
      {
         Assert.assertEquals("value", cache2.get(fqn, key));
      }

      // now try again with passing the default options
      cache1.getInvocationContext().getOptionOverrides().reset();
      cache1.remove(fqn, key);
      delay();

      // both should be null
      Assert.assertNull("should be null", cache1.get(fqn, key));
      Assert.assertNull("should be null", cache2.get(fqn, key));
   }

   public void testTransactionalBehaviour() throws Exception
   {
      TransactionManager mgr = cache1.getTransactionManager();
      mgr.begin();
      cache1.getInvocationContext().getOptionOverrides().reset();
      cache1.put(fqn, key, "value1");
      cache1.getInvocationContext().getOptionOverrides().setCacheModeLocal(true);
      cache1.put(fqn, key, "value2");
      mgr.commit();
      delay();
      // cache1 should still have this
      Assert.assertEquals("value2", cache1.get(fqn, key));

      if (!isInvalidation)
      {
         Assert.assertEquals("value1", cache2.get(fqn, key));
      }
      else
      {
         assertNull(cache2.get(fqn, key));
      }

      // now try again with passing the default options
      mgr.begin();
      cache1.getInvocationContext().getOptionOverrides().setCacheModeLocal(true);
      cache1.put(fqn, key, "value3");
      cache1.getInvocationContext().getOptionOverrides().reset();
      cache1.put(fqn, key, "value");
      mgr.commit();
      delay();
      // cache1 should still have this
      Assert.assertEquals("value", cache1.get(fqn, key));

      // cache 2 should as well
      if (!isInvalidation)
      {
         Assert.assertEquals("value", cache2.get(fqn, key));
      }
      else
      {
         Assert.assertNull("should be invalidated", cache2.get(fqn, key));
      }

      // now cache2
      mgr = cache2.getTransactionManager();
      mgr.begin();
      cache2.getInvocationContext().getOptionOverrides().setCacheModeLocal(false);
      cache2.put(fqn, key, "value3");
      cache2.getInvocationContext().getOptionOverrides().setCacheModeLocal(true);
      cache2.put(fqn, key, "value2");
      mgr.commit();
      delay();

      Assert.assertEquals("value2", cache2.get(fqn, key));

      if (!isInvalidation)
      {
         assertEquals("value3", cache1.get(fqn, key));
      }
      else
      {
         assertNull(cache1.get(fqn, key));
      }

      mgr.begin();
      cache2.getInvocationContext().getOptionOverrides().setCacheModeLocal(true);
      cache2.put(fqn, key, "value2");
      cache2.getInvocationContext().getOptionOverrides().reset();
      cache2.put(fqn, key, "value2");
      mgr.commit();
      delay();
      Assert.assertEquals("value2", cache2.get(fqn, key));
      if (!isInvalidation)
      {
         Assert.assertEquals("value2", cache1.get(fqn, key));
      }
      else
      {
         Assert.assertNull("should be invalidated", cache1.get(fqn, key));
      }

   }

   protected abstract void delay();

}