/*
 * JBoss, Home of Professional Open Source
 *
 * Distributable under LGPL license.
 * See terms of license at gnu.org.
 */

package org.jboss.cache.passivation;

import static org.testng.AssertJUnit.assertEquals;
import static org.testng.AssertJUnit.assertFalse;
import static org.testng.AssertJUnit.assertNotNull;
import static org.testng.AssertJUnit.fail;

import org.jboss.cache.CacheFactory;
import org.jboss.cache.CacheImpl;
import org.jboss.cache.DefaultCacheFactory;
import org.jboss.cache.Fqn;
import org.jboss.cache.factories.XmlConfigurationParser;
import org.jboss.cache.loader.DummyInMemoryCacheLoader;
import org.jboss.cache.misc.TestingUtil;
import org.jboss.cache.notifications.annotation.CacheListener;
import org.jboss.cache.notifications.annotation.NodeActivated;
import org.jboss.cache.notifications.annotation.NodePassivated;
import org.jboss.cache.notifications.event.Event;
import org.jboss.cache.notifications.event.NodeEvent;
import org.testng.annotations.AfterMethod;
import org.testng.annotations.BeforeMethod;
import org.testng.annotations.Test;

/**
 * @author Ben Wang
 * @version $Revision: 4444 $
 */
@Test(groups = { "functional" })
public class BasicPassivationTest
{
   CacheImpl<String, String> cache_;
   int wakeupIntervalMillis_ = 0;
   final String ROOT_STR = "/test";
   Throwable t1_ex, t2_ex;
   final long DURATION = 10000;
   boolean isTrue;
   final String FQNSTR = "/org/jboss/3";
   int activationCount = 0;
   int passivationCount = 0;

   @BeforeMethod(alwaysRun = true)
   public void setUp() throws Exception
   {
      initCaches();
      wakeupIntervalMillis_ = cache_.getConfiguration().getEvictionConfig().getWakeupIntervalSeconds() * 1000;
      log("wakeupInterval is " + wakeupIntervalMillis_);
      if (wakeupIntervalMillis_ < 0)
      {
         fail("testEviction(): eviction thread wake up interval is illegal " + wakeupIntervalMillis_);
      }

      t1_ex = t2_ex = null;
      isTrue = true;
   }

   private void initCaches() throws Exception
   {
      CacheFactory<String, String> instance = DefaultCacheFactory.getInstance();
      cache_ = (CacheImpl<String, String>)instance.createCache(false);
      cache_.setConfiguration(new XmlConfigurationParser().parseFile("META-INF/local-passivation-service.xml"));// read in generic local xml
      cache_.getConfiguration().setTransactionManagerLookupClass("org.jboss.cache.transaction.DummyTransactionManagerLookup");
      Object listener = new TestCacheListener();
      cache_.getConfiguration().getCacheLoaderConfig().getFirstCacheLoaderConfig().setClassName(DummyInMemoryCacheLoader.class.getName());
      cache_.start();
      cache_.getNotifier().addCacheListener(listener);
   }

   @AfterMethod(alwaysRun = true)
   public void tearDown() throws Exception
   {
      cache_.stop();
   }

   public void testBasic()
   {
      activationCount = 0;
      passivationCount = 0;
      Fqn fqn = Fqn.fromString(FQNSTR);
      try
      {
         cache_.put(fqn, FQNSTR, FQNSTR);
      }
      catch (Exception e)
      {
         fail("Failed to insert data" + e);
         e.printStackTrace();
      }
      System.out.println(cache_.toString());
      TestingUtil.sleepThread(21000);
      System.out.println(cache_.toString());
      try
      {
         assertFalse(cache_.exists(FQNSTR, FQNSTR));
         String val = cache_.get(FQNSTR, FQNSTR);
         assertNotNull("DataNode should not be empty ", val);
      }
      catch (Exception e)
      {
         e.printStackTrace();
         fail("Failed to get" + e);
      }
      assertEquals("activation count:", 1, activationCount);
      assertEquals("passivation count:", 1, passivationCount);
   }

   public void testDualPassivation() throws Exception
   {
      Fqn fqn = Fqn.fromString(FQNSTR);
      cache_.put(fqn, "key", "value");
      cache_.evict(fqn);
      cache_.evict(fqn);
      assertEquals("Proper value after 2 passivations", "value", cache_.get(fqn, "key"));
   }

   public void testIntermingledPassivation() throws Exception
   {
      Fqn fqn = Fqn.fromString(FQNSTR);
      cache_.put(fqn, "key1", "value");
      cache_.evict(fqn);
      cache_.put(fqn, "key2", "value");
      cache_.evict(fqn);
      assertEquals("Proper value after 2 passivations", "value", cache_.get(fqn, "key1"));

   }

   private void log(String msg)
   {
      System.out.println("-- " + msg);
   }

   @CacheListener
   public class TestCacheListener
   {
      @SuppressWarnings("unchecked")
      @NodeActivated
      @NodePassivated
      public void callback(NodeEvent ne)
      {
         if (ne.isPre())
            return;// we are not interested in postActivate event
         if (!ne.getFqn().isChildOrEquals(Fqn.fromString(FQNSTR)))
            return;// don't care about fqn that doesn't belong to me.

         log("got event " + ne);
         if (ne.getType() == Event.Type.NODE_ACTIVATED)
            activationCount++;
         else if (ne.getType() == Event.Type.NODE_PASSIVATED)
            passivationCount++;
      }
   }
}
