/*
 * JBoss, Home of Professional Open Source.
 * Copyright 2006, Red Hat Middleware LLC, and individual contributors
 * as indicated by the @author tags. See the copyright.txt file in the
 * distribution for a full listing of individual contributors.
 *
 * This is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as
 * published by the Free Software Foundation; either version 2.1 of
 * the License, or (at your option) any later version.
 *
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this software; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA, or see the FSF site: http://www.fsf.org.
 */

package org.jboss.cache.passivation;

import static org.testng.AssertJUnit.assertFalse;
import static org.testng.AssertJUnit.assertNotNull;
import static org.testng.AssertJUnit.fail;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.jboss.cache.CacheFactory;
import org.jboss.cache.CacheImpl;
import org.jboss.cache.DefaultCacheFactory;
import org.jboss.cache.factories.XmlConfigurationParser;
import org.jboss.cache.loader.DummyInMemoryCacheLoader;
import org.jboss.cache.misc.TestingUtil;
import org.jboss.cache.notifications.annotation.CacheListener;
import org.jboss.cache.notifications.annotation.NodeActivated;
import org.jboss.cache.notifications.annotation.NodeLoaded;
import org.jboss.cache.notifications.annotation.NodePassivated;
import org.jboss.cache.notifications.event.NodeEvent;
import org.testng.annotations.AfterMethod;
import org.testng.annotations.BeforeMethod;
import org.testng.annotations.Test;

/**
 * @author Ben Wang, Feb 11, 2004
 */
@Test(groups = { "functional", "jgroups" })
public class ReplicatedPassivationIntegrationTest
{
   private CacheImpl<String, String> cache_;
   private CacheImpl<String, String> cache1_;
   protected final static Log log = LogFactory.getLog(ReplicatedPassivationIntegrationTest.class);
   int wakeupIntervalMillis_ = 0;
   PassivationListener listener_;

   public ReplicatedPassivationIntegrationTest()
   {
      listener_ = new ReplicatedPassivationIntegrationTest.PassivationListener();
   }

   @BeforeMethod(alwaysRun = true)
   public void setUp() throws Exception
   {
      CacheFactory<String, String> instance = DefaultCacheFactory.getInstance();
      cache_ = (CacheImpl<String, String>)instance.createCache(false);
      initCaches(cache_);
      cache_.getConfiguration().setUseRegionBasedMarshalling(true);
      cache_.start();

      cache1_ = (CacheImpl<String, String>)instance.createCache(false);
      initCaches(cache1_);
      cache1_.getConfiguration().setUseRegionBasedMarshalling(true);

      cache1_.start();
      cache1_.getNotifier().addCacheListener(listener_);
      listener_.resetCounter();

      wakeupIntervalMillis_ = cache1_.getConfiguration().getEvictionConfig().getWakeupIntervalSeconds() * 1000;
      log("wakeupInterval is " + wakeupIntervalMillis_);
      if (wakeupIntervalMillis_ <= 0)
      {
         fail("testEviction(): eviction thread wake up interval is illegal " + wakeupIntervalMillis_);
      }
   }

   void initCaches(CacheImpl<String, String> cache) throws Exception
   {
      cache.setConfiguration(new XmlConfigurationParser().parseFile("META-INF/replSync-passivation-service.xml"));// read in generic local xml
      cache.getConfiguration().setTransactionManagerLookupClass("org.jboss.cache.transaction.DummyTransactionManagerLookup");
      cache.getConfiguration().getCacheLoaderConfig().getFirstCacheLoaderConfig().setClassName(DummyInMemoryCacheLoader.class.getName());
   }

   @AfterMethod(alwaysRun = true)
   public void tearDown() throws Exception
   {
      cache_.stop();
      cache1_.stop();
   }

   /**
    */
   public void testActivationEvent() throws Exception
   {
      String rootStr = "/org/jboss/test/data/";
      String rootStr1 = "/__JBossInternal__/5c4o12-pzhlhj-esnuy3sg-1-esnuy3sg-2";
      String str = rootStr + "0";
      cache_.remove("/");

      cache_.put(str, str, str);
      cache_.put(rootStr1, str, str);

      TestingUtil.sleepThread(11000);
      assertFalse("UnversionedNode should not exist", cache1_.exists(str, str));
      String val;
      val = cache1_.get(str, str);
      val = cache1_.get(rootStr1, str);
      assertNotNull("DataNode should be activated ", val);
   }

   void log(String msg)
   {
      System.out.println("-- " + msg);
   }

   @CacheListener
   public class PassivationListener
   {
      int counter = 0;
      int loadedCounter = 0;

      public int getCounter()
      {
         return counter;
      }

      public void resetCounter()
      {
         counter = 0;
         loadedCounter = 0;
      }

      @NodeActivated
      public void nodeActivated(NodeEvent ne)
      {
         if (!ne.isPre())
         {
            counter++;
            System.out.println("nodeActivate(): counter: " + counter);
            System.out.println("nodeActivate(): " + ne.getFqn());
         }
      }

      @NodePassivated
      public void nodePassivated(NodeEvent ne)
      {
         if (ne.isPre())
         {
            System.out.println("nodePassivate(): " + ne.getFqn());
         }
      }

      @NodeLoaded
      public void nodeLoaded(NodeEvent ne)
      {
         if (!ne.isPre())
         {
            loadedCounter++;
         }
      }

   }
}
