/*
 * JBoss, Home of Professional Open Source
 *
 * Distributable under LGPL license.
 * See terms of license at gnu.org.
 */
package org.jboss.cache.transaction;

import static org.testng.AssertJUnit.assertEquals;
import static org.testng.AssertJUnit.assertNotSame;
import static org.testng.AssertJUnit.assertSame;
import static org.testng.AssertJUnit.assertTrue;

import javax.transaction.RollbackException;
import javax.transaction.Synchronization;
import javax.transaction.SystemException;
import javax.transaction.Transaction;
import javax.transaction.TransactionManager;

import org.jboss.cache.CacheException;
import org.jboss.cache.CacheImpl;
import org.jboss.cache.interceptors.OrderedSynchronizationHandler;
import org.jboss.cache.misc.TestingUtil;
import org.jboss.cache.transaction.NotifyingTransactionManager.Notification;
import org.jgroups.Channel;
import org.jgroups.JChannel;
import org.testng.annotations.AfterMethod;
import org.testng.annotations.BeforeMethod;
import org.testng.annotations.Test;

/**
 * @author <a href="mailto:manik@jboss.org">Manik Surtani (manik@jboss.org)</a>
 */
@Test(groups={"functional"})
public class AbortionTest
{
   private MyTC cache1, cache2, cache3;

   @BeforeMethod(alwaysRun = true)
   public void setUp() throws Exception
   {
      System.out.println("********* START: SET UP *************");
      cache1 = initCache(false);
      TestingUtil.sleepThread(1500); // to ensure cache1 is the coordinator
      cache2 = initCache(false);
      cache3 = initCache(true);
      System.out.println("********* END: SET UP *************");
   }

   @AfterMethod(alwaysRun = true)
   public void tearDown() throws Exception
   {
      System.out.println("********* START: TEAR DOWN *************");
      destroyCache(cache3);
      destroyCache(cache2);
      destroyCache(cache1);
      cache1 = null;
      cache2 = null;
      cache3 = null;
      System.out.println("********* END: TEAR DOWN *************");
   }

   private MyTC initCache(boolean notifying) throws Exception
   {
      MyTC c = new MyTC();
      c.getConfiguration().setCacheMode("REPL_SYNC");
      c.getConfiguration().setClusterConfig(getJGroupsStack());
      c.getConfiguration().setFetchInMemoryState(false);
      if (!notifying)
      {
         c.getConfiguration().setTransactionManagerLookupClass("org.jboss.cache.transaction.DummyTransactionManagerLookup");
      }
      else
      {
         c.getConfiguration().setTransactionManagerLookupClass("org.jboss.cache.transaction.NotifyingTransactionManager");
      }
      c.start();
      return c;
   }

   // we need a 'special' stack that does not attempt redelivery since we kill a channel midway during a tx in this test.
   private String getJGroupsStack()
   {
      //       return "UDP(mcast_addr=224.0.0.36;mcast_port=55566;ip_ttl=32;" +
      //          "mcast_send_buf_size=150000;mcast_recv_buf_size=80000):" +
      //          "PING(timeout=10;num_initial_members=1):" +
      //          "pbcast.NAKACK(gc_lag=50;max_xmit_size=8192;retransmit_timeout=10):" +
      //          "UNICAST(timeout=600):" +
      //          "FRAG(frag_size=8192;down_thread=false;up_thread=false):" +
      //          "pbcast.GMS(join_timeout=5000;join_retry_timeout=2000;" +
      //          "shun=false;print_local_addr=true):" +
      //          "pbcast.STATE_TRANSFER";
      return JChannel.DEFAULT_PROTOCOL_STACK;
   }

   private void destroyCache(MyTC c)
   {
      if (c != null)
      {
         c.stop();
         c.destroy();
      }
   }

   public void testSyncCaches() throws Exception
   {
      performTest(false, false);
   }

   public void testSyncCachesSyncCommitRollback() throws Exception
   {
      performTest(true, false);
   }

   /**
    * Note that this tests a *remote* beforeCompletion abort - which is a part of the calling instance's afterCompletion.
    *
    * @throws Exception
    */
   public void testAbortBeforeCompletion() throws Exception
   {
      performTest(true, true);
   }

   private void performTest(boolean syncCommitRollback, boolean abortBeforeCompletion) throws Exception
   {
      cache1.getConfiguration().setSyncCommitPhase(syncCommitRollback);
      cache1.getConfiguration().setSyncRollbackPhase(syncCommitRollback);
      cache2.getConfiguration().setSyncCommitPhase(syncCommitRollback);
      cache2.getConfiguration().setSyncRollbackPhase(syncCommitRollback);
      cache3.getConfiguration().setSyncCommitPhase(syncCommitRollback);
      cache3.getConfiguration().setSyncRollbackPhase(syncCommitRollback);

      TransactionManager mgr1 = cache1.getTransactionManager();
      TransactionManager mgr2 = cache2.getTransactionManager();
      assertTrue(cache3.getTransactionManager() instanceof NotifyingTransactionManager);
      NotifyingTransactionManager mgr3 = (NotifyingTransactionManager)cache3.getTransactionManager();

      assertSame(mgr1, mgr2);
      assertNotSame(mgr1, mgr3);
      assertNotSame(mgr2, mgr3);

      assertTrue(mgr1 instanceof DummyTransactionManager);
      assertTrue(mgr2 instanceof DummyTransactionManager);

      cache1.put("/test", "key", "value");

      assertEquals("value", cache1.get("/test", "key"));
      assertEquals("value", cache2.get("/test", "key"));
      assertEquals("value", cache3.get("/test", "key"));

      mgr3.setNotification(new TestNotification(abortBeforeCompletion));

      mgr1.begin();
      cache1.put("/test", "key", "value2");
      mgr1.commit();

      TestingUtil.sleepThread(5000);

      // only test cache1 and cache2.  Assume cache3 has crashed out.
      assertEquals(0, cache1.getNumberOfLocksHeld());
      assertEquals(0, cache2.getNumberOfLocksHeld());
      assertEquals("put in transaction should NOT have been rolled back", "value2", cache1.get("/test", "key"));
      assertEquals("put in transaction should NOT have been rolled back", "value2", cache2.get("/test", "key"));

   }

   class TestNotification implements Notification
   {
      boolean abortBeforeCompletion;

      public TestNotification(boolean abortBeforeCompletion)
      {
         this.abortBeforeCompletion = abortBeforeCompletion;
      }

      public void notify(Transaction tx) throws SystemException, RollbackException
      {
         final Transaction finalTx = tx;
         System.out.println("Notify called.");
         // add an aborting sync handler.
         Synchronization abort = new Synchronization() {

            public void beforeCompletion()
            {
               if (abortBeforeCompletion)
               {
                  cache3.myChannel.close();
                  System.out.println("Returning from abort.beforeCompletion");
                  try
                  {
                     finalTx.setRollbackOnly();
                  }
                  catch (SystemException e)
                  {
                     throw new RuntimeException("Unable to set rollback", e);
                  }
                  throw new RuntimeException("Dummy exception");
               }
            }

            public void afterCompletion(int i)
            {
               if (!abortBeforeCompletion)
               {
                  cache3.myChannel.close();
                  System.out.println("Returning from abort.afterCompletion");
                  throw new RuntimeException("Dummy exception");
               }
            }
         };

         OrderedSynchronizationHandler osh = OrderedSynchronizationHandler.getInstance(tx);
         osh.registerAtHead(abort);
         System.out.println("Added sync handler.");
      }

   }

   ;

   public static class MyTC extends CacheImpl<Object, Object>
   {
      Channel myChannel;

      public MyTC() throws Exception
      {
         super();
      }

      public void start() throws CacheException
      {
         super.start();
         myChannel = channel;
      }
   }

}
